Package: tutorial.helpers
Title: Helper Functions for Creating Tutorials
Version: 0.7.0
Authors@R: 
    person(given = "David",
           family = "Kane",
           role = c("aut", "cre", "cph"),
           email = "dave.kane@gmail.com",
           comment = c(ORCID = "0000-0002-6660-3934"))
Description: Helper functions for creating, editing, and testing tutorials
    created with the 'learnr' package. Provides a simple method for allowing
    students to download their answers to tutorial questions. For examples
    of its use, see the 'r4ds.tutorials' package.
Depends: R (>= 4.1.0)
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: quarto
Suggests: googledrive, quarto, testthat (>= 3.0.0)
Config/testthat/edition: 3
Imports: knitr, learnr, rmarkdown, rstudioapi, rvest, shiny, tibble,
        utils
BugReports: https://github.com/PPBDS/tutorial.helpers/issues
URL: https://ppbds.github.io/tutorial.helpers/,
        https://github.com/PPBDS/tutorial.helpers
NeedsCompilation: no
Packaged: 2026-07-27 01:28:28 UTC; dkane
Author: David Kane [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-6660-3934>)
Maintainer: David Kane <dave.kane@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-27 06:50:09 UTC

# Generated by roxygen2: do not edit by hand

export(check_current_tutorial)
export(check_key_vars)
export(check_membership)
export(check_tutorial_defaults)
export(determine_code_chunk_name)
export(determine_exercise_number)
export(download_google_drive)
export(format_tutorial)
export(gather_submissions)
export(get_submissions_from_learnr_session)
export(knit_tutorials)
export(make_exercise)
export(match_questions)
export(open_github_pages)
export(return_tutorial_paths)
export(show_file)
export(submission_server)
export(submission_ui)
export(submissions_answers)
export(submissions_summary)
export(write_answers)

<!-- README.md is generated from README.Rmd. Edit ONLY this file if you need to make a change in README.md. But, after you edit it, you need to run devtools::build_readme() in order to create the new README.md, which is the thing which is actually used. -->

# Tutorial Helper Functions

<!-- badges: start -->

[![R build
status](https://github.com/PPBDS/tutorial.helpers/workflows/R-CMD-check/badge.svg)](https://github.com/PPBDS/tutorial.helpers/actions)
<!-- badges: end -->

## About this package

**tutorial.helpers** provides functions to help you write R tutorials,
especially if you follow [my
advice](https://ppbds.github.io/tutorial.helpers/articles/ai.html) about
the best way to do so.

## Installation

To install the package from CRAN:

``` r
install.packages("tutorial.helpers")
```

You can install the development version from
[GitHub](https://github.com/) with:

``` r
remotes::install_github("PPBDS/tutorial.helpers")
```

## Useful tools

#### Creating a tutorial

To create a new tutorial, you first need a new directory, located in the
`inst/tutorials` directory of your package. Create a `tutorial.Rmd` file
within it. The [AI
vignette](https://ppbds.github.io/tutorial.helpers/articles/ai.html)
provides a highly opinionated framework for teaching students how to use
AI to do data science, including the required `info_section` and
`download-answers` chunks which every tutorial should contain.

#### Testing

The package includes two important functions: `make_exercise()` and
`check_current_tutorial()`. Use `make_exercise()` to add a new exercise
to the current tutorial. It will number the exercise, and the code chunk
labels, automatically. Use `check_current_tutorial()` to renumber all
the exercises and relabel all the code chunks. This is especially useful
if you add or delete an exercise in the middle of a section.

There are three exported functions for checking the tutorials in your
package. See the [testing
vignette](https://ppbds.github.io/tutorial.helpers/articles/testing.html)
for details on their use.

#### Tutorials

The package includes the “Getting Started” tutorial, which uses VS Code
on GitHub Codespaces. You should require your students to complete it if
you are creating your own tutorials with the **tutorial.helpers**
package. Have students run:

    learnr::run_tutorial("getting-started", package = "tutorial.helpers")

This will teach them enough about how tutorials work to be able to
complete the tutorials you write.

#### Download answers

The most useful tool is the [download answers
trick](https://ppbds.github.io/tutorial.helpers/articles/downloads.html).
In your tutorial, simply include `library(tutorial.helpers)` in the
setup R code chunk and then insert this empty code chunk at the end.

```` default
```{r download-answers, child = system.file("child_documents/download_answers.Rmd", package = "tutorial.helpers")}
```
````

This will ask the student to provide an estimate of how long the
tutorial took to complete. It will then provide the ability to download
the student’s answers in html format. Students submit these files to
their instructors, who can then confirm that the work was completed and
look for any patterns in student (mis)understandings.

#### Recommended components

We recommend including this child document at the start of each
tutorial:

```` default
```{r info-section, child = system.file("child_documents/info_section.Rmd", package = "tutorial.helpers")}
```
````

This will insert (optional) questions asking for the student’s name,
email and id.

We recommend ending the tutorial with the download-answers child
document, as discussed above.

```` default
```{r download-answers, child = system.file("child_documents/download_answers.Rmd", package = "tutorial.helpers")}
```
````

This folder contains actual submissions from students who took Kane's Data Science Bootcamp (https://bootcamp.davidkane.info/) during the summer of 2025. All students answers were submitted to this Google Drive:

https://drive.google.com/drive/folders/1RgxfiwPmcKrhYLOqitLQt4KeODYZEu9ZMVvh029qBfy37V0wEbxIg95gwUE0JWL_hXEBuAlT?usp=sharing

This folder contains all the answers to the regular expressions *data-import* and *stops*. The former corresponds to the Data Import tutorial from the r4ds.tutorials package and the later corresponds to the Stops tutorial from the primer.tutorials package.

## ## Interacting with sites with `GET()`
## ###
## 
## In order to get data from an API, we use the **httr** package.
## The package is designed to imitate standard HTTP in R.
## Read more about HTTP [here](https://www.jmarshall.com/easy/http/).
## 
## ### Exercise 1

## ### Exercise 2
## 
## Start your code with `cces`. Use the pipe operator `|>` to add the
## function `filter()`, selecting the  rows where `state` is equal to
## "Massachusetts". To set something equal to a value in `filter()`
## use two equal signs: `==`.
## 
## ```{r filter-2, exercise = TRUE}
## ```
## 
## ```{r filter-2-hint-1, eval = FALSE}
## cces |>
##   filter(state == "...")
## ```
## 
## ```{r filter-2-test, include = FALSE}
## cces |>
##   filter(state == "Massachusetts")
## ```
## ###
## 
## `==` is used because it is **checking** whether the value of the
## variable on the left is equal to the value on the left. See
## [here](https://stat.ethz.ch/R-manual/R-devel/library/base/html/Comparison.html)
## for discussion of other relational operators in R.
## A single equation symbol, `=`, is used to set something equal to
## something else.

## <button onclick = "transfer_code(this)">Copy previous code</button>

## ```{r ex-1-hint-1, eval = FALSE}
## This is an example hint. Normally sentences like these
## would cause an error in R because it is not proper code.
## However, since we include eval = FALSE in the r-chunk,
## R ignores all errors!
## ```

## ```{r ex-1-hint-2, eval = FALSE}
## ... |>
##   filter(year = ...) |>
##   ...(flights)
## ```

## ### Exercise 6
## 
## Explain potential outcomes in about two sentences.
## 
## ```{r definitions-6}
## question_text(NULL,
##     message = "This is where we place the correct answer. It will appear only after
##     students have submitted their own answers. Note that we do not need to wrap the
##     answer text by hand.",
##     answer(NULL,
##            correct = TRUE),
##     allow_retry = FALSE,
##     incorrect = NULL,
##     rows = 6)
## ```

## ### Exercise 7
## 
## From the Console, run `list.files()`. CP/CR.
## 
## ```{r file-creation-7}
##   question_text(NULL,
##     answer(NULL, correct = TRUE),
##     allow_retry = TRUE,
##     try_again_button = "Edit Answer",
##     incorrect = NULL,
##     rows = 3)
## ```

## In the Console, run:
## 
## ```
## tutorial.helpers::show_file("_quarto.yml")
## ```
## 
## CP/CR.

## ----eval = FALSE-------------------------------------------------------------
# x <- read_csv("data/myfile.csv")


## > getwd()
## [1] "/Users/davidkane/Desktop/projects/r4ds.tutorials"
## > setwd("inst/tutorials/031-data-files/")
## > getwd()
## [1] "/Users/davidkane/Desktop/projects/r4ds.tutorials/inst/tutorials/031-data-files"
## >

## ```{r}
## include_graphics("images/example.png")
## ```

## <pre><code>```{r}
## 1 + 1
## ```</code></pre>

## <pre><code>
## > tutorial.helpers::show_file("quarto-1.qmd")
## ---
## title: "Quarto 1"
## author: David Kane
## format: html
## execute:
##   echo: false
## ---
## <pre><code>```{r}
## suppressPackageStartupMessages(library(tidyverse))
## library(palmerpenguins)
## ```</code></pre>
## >
## </code></pre>

## ----eval = FALSE-------------------------------------------------------------
# rmarkdown::render("file_path")

## ### Exercise 1
## 
## Create a Github repo called `XX`. Make sure to click the "Add a README file" check box.
## 
## Connect the repo to a project on your computer using `File -> New Folder from Git ...`.  Make sure to select the "Open in a new window" box.
## 
## You need two Positon windows: this one for running the tutorial and the one you just created for writing your code and interacting with the Console.
## 
## In the new window, select `File -> New File -> Quarto Document ...`. Provide a title -- `"XX"` -- and an author (you). Render the document and save it as `analysis.qmd`.
## 
## Create a `.gitignore` file with `analysis_files` on the first line and then a blank line. Save and push.
## 
## In the Console, run:
## 
## ```
## show_file(".gitignore")
## ```
## 
## If that fails, it is probably because you have not yet loaded `library(tutorial.helpers)` in the Console.
## 
## CP/CR.
## 
## ```{r introduction-1}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 3)
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 2
## 
## In your QMD, put `library(tidyverse)` in a new code chunk. Render the file using `Cmd/Ctrl + Shift + K`.
## 
## Notice that the file does not look good because the code is visible and there are annoying messages. To take care of this, add `#| message: false` to remove all the messages in this `setup` chunk. Also, add the following to the YAML header to remove all code echos from the HTML:
## 
## ```
## execute:
##   echo: false
## ```
## 
## In the Console, run:
## 
## ```
## show_file("analysis.qmd", chunk = "Last")
## ```
## 
## CP/CR.
## 
## ```{r introduction-2}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 6)
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 3
## 
## Place your cursor in the QMD file on the `library(tidyverse)` line. Use `Cmd/Ctrl + Enter` to execute that line.
## 
## Note that this causes `library(tidyverse)` to be copied down to the Console and then executed.
## 
## CP/CR.
## 
## ```{r introduction-3}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 8)
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 12
## 
## Create a new code chunk in your QMD. Add a code chunk option: `#| cache: true`. Copy/paste the R code for the final model into the code chunk, assigning the result to `fit_XX`.
## 
## `Cmd/Ctrl + Shift + K`. It may take some time to render your QMD, depending on how complex your model is. But, by including `#| cache: true` you cause Quarto to cache the results of the chunk. The next time you render your QMD, as long as you have not changed the code, Quarto will just load up the saved object.
## 
## At the Console, run:
## 
## ```
## tutorial.helpers::show_file("XX.qmd", chunk = "Last")
## ```
## 
## CP/CR.
## 
## ```{r courage-12}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 8)
## ```
## 
## ###
## 
## To confirm, `Cmd/Ctrl + Shift + K` again. It should be quick.

## ### Exercise 13
## 
## Place your cursor in the QMD file on the `fit_XX` line. Use `Cmd/Ctrl + Enter` to execute that line.
## 
## At the Console, run `ls()`. CP/CR.
## 
## ```{r courage-13}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 8)
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 6
## 
## Using your favorite AI, prompt it to generate R code that ... Add the code to your QMD in a new chunk. Place your cursor on the first line of the code and run `Cmd/Ctrl + Enter`.
## 
## In the Console, run:
## 
## ```
## show_file("analysis.qmd", chunk = "Last")
## ```
## 
## CP/CR.
## 
## ```{r something-1}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 4)
## ```
## 
## ###
## 
## ```{r, echo = TRUE}
## 
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ```{r, echo = TRUE, eval = FALSE}
## 
## ```

## ### Exercise 8
## 
## Before creating a plot, we need to ensure that your data matches our data. In the QMD, replace your code from the previous exercise with our code.
## 
## In the Console, run:
## 
## ```
## show_file("analysis.qmd", chunk = "Last")
## ```
## 
## CP/CR.
## 
## ```{r ai-usage-8}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 6)
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 9
## 
## Within the latest code chunk, add the option: `#| cache: true`. Assign the result of the pipe to `x`.
## 
## `Cmd/Ctrl + Shift + K`. By including `#| cache: true` you cause Quarto to cache the results of the chunk. The next time you render your QMD, as long as you have not changed the code, Quarto will just load up the saved object.
## 
## If you have not done so already, you should add `analysis_cache` to the `.gitginore`. The content of the cache file does not belong on GitHub.
## 
## Place your cursor on the line where the pipe is assigned to `x`, run `Cmd/Ctrl + Enter`, thus ensuring that the workspace also includes a copy of `x`.
## 
## CP/CR.
## 
## ```{r ai-usage-9}
## question_text(NULL,
##     answer(NULL, correct = TRUE),
##     allow_retry = TRUE,
##     try_again_button = "Edit Answer",
##     incorrect = NULL,
##     rows = 8)
## ```
## 
## ###
## 
## Our code:
## 
## ```{r, echo = TRUE}
## x <- ...
## ```
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 10
## 
## Within the Console, type `x`, which we previously assigned to a pipe and ran in the Console. Hit `Enter`.
## 
## CP/CR.
## 
## ```{r ai-usage-10}
## question_text(NULL,
##     answer(NULL, correct = TRUE),
##     allow_retry = TRUE,
##     try_again_button = "Edit Answer",
##     incorrect = NULL,
##     rows = 8)
## ```
## 
## ###
## 
## Our code:
## 
## ```{r, echo=TRUE}
## x
## ...
## ```
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 11
## 
## Ask AI to generate R code that uses `x` to plot a basic graph showing XX... Mention you want to use the data from `x` and copy/paste the `x` you ran in the Console with the resulting tibble. You only need the top 3 lines, mainly to include column names.
## 
## Within `labs()`, edit or add a proper title, subtitle, and caption. If axis labels would be useful, add them, but if unnecessary, don't bother. Don't assign the code for the plot to any variable. Put the plot code in a new code chunk. Run `Cmd/Ctrl + Shift + K` to ensure that everything works. Make your plot look nice.
## 
## In the Console, run:
## 
## ```
## show_file("analysis.qmd", chunk = "Last")
## ```
## 
## CP/CR.
## 
## ```{r ai-usage-11}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 12)
## ```
## 
## ###
## 
## Our code:
## 
## ```{r, echo=TRUE}
##   ...
## ```
## 
## ###
## 
## <!-- XX: Make sure your plotting code is good! This will take some time. You had better have a subtitle which provides the take-away message of the plot. AI sometimes gives you too much code, lots of `theme()` stuff and so on. This is no good! In most cases, we are happy with concise, straightforward code. Insert a knowledge drop related to this project. -->

## ### Exercise 7
## 
## 
## ```{r plotting-7, exercise = TRUE}
## 
## ```
## 
## <button onclick = "transfer_code(this)">Copy previous code</button>
## 
## ```{r plotting-7-hint, eval = FALSE}
## 
## ```
## 
## ```{r plotting-7-test, include = FALSE}
## 
## ```
## 
## ###

## ### Exercise 9
## 
## ```{r plotting-9}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 3)
## ```
## 
## ###

## ### Exercise 8
## 
## ```{r plotting-8}
## question_text(NULL,
## 	message = "Place correct answer here.",
## 	answer(NULL, correct = TRUE),
## 	allow_retry = FALSE,
## 	incorrect = NULL,
## 	rows = 6)
## ```
## 
## ###

#' @title Tutorial submission functions
#' @rdname submission_functions
#'
#' @description 
#'   Provides the core Shiny server and UI hooks for collecting and downloading student answers
#'   from a learnr tutorial. `submission_server()` should be called in an Rmd code chunk with
#'   `context="server"`. 
#'   
#'   This function was modified from Colin Rundel's learnrhash package
#'   (https://github.com/rundel/learnrhash).
#'
#' @details
#'   The server function uses a Shiny downloadHandler to let students download their answers.
#'   All main logic must be wrapped in `local()` with `parent.frame()` to ensure access to the live
#'   learnr session and objects created in the parent environment.
#'
#'   The `session` object (created by Shiny) is only available inside the `downloadHandler$content`
#'   function, so any test-case extraction or answer writing must happen there.
#'
#'   For reference: the `file` argument in `content` is a temporary file path created by Shiny,
#'   and your handler's job is to write the downloadable file there.
#'
#'   If you want to generate test fixtures, insert `browser()` inside the `content` function,
#'   then use functions like `get_submissions_from_learnr_session(session)` at the prompt.
#'
#'   See also: https://mastering-shiny.org/action-transfer.html#downloading-reports
#'
#' @examples
#' if(interactive()){
#'   submission_server()
#' }
#'
#' @export
#'
#' @returns No return value; called for side effects in a Shiny/learnr session.

submission_server <- function() {
  p <- parent.frame()



  # All main logic is run in the parent frame to access live session/reactive values.
  local({
    output$downloadHtml <- shiny::downloadHandler(
      filename = paste0(learnr::get_tutorial_info()$tutorial_id, "_answers.html"),
      content = function(file) {
        write_answers(file, session)
      }
    )
  }, envir = p)
  
  NULL
}

#' @rdname submission_functions
#'
#' @description UI block to include a download button and simple instructions for students.
#'
#' @examples
#' if(interactive()){
#'   submission_ui
#' }
#'
#' @export
#'
#' @returns An object of class shiny.tag

submission_ui <- shiny::div(
  "When you have completed this tutorial, follow these steps:",
  shiny::tags$br(),
  shiny::tags$ol(
    shiny::tags$li("Click the button to download a file containing your answers."),
    shiny::tags$li("Save the file onto your computer in a convenient location.")
  ),
  shiny::fluidPage(
    shiny::mainPanel(
      shiny::div(id = "form",
                 shiny::downloadButton(outputId = "downloadHtml", label = "Download HTML"))
    )
  ),
  shiny::div(
    shiny::tags$br(),
    "(If no file seems to download, try right-clicking the download button and choose \"Save link as...\")",
    shiny::tags$br()
  )
)

#' Open Multiple GitHub Pages in Browser Tabs
#'
#' This function opens multiple GitHub.io pages in browser tabs, making it easy
#' for teaching fellows to quickly review student webpages. It can accept either
#' a vector of URLs or a tibble/data.frame containing URLs.
#'
#' @param urls Either a character vector of URLs to open, OR a tibble/data.frame 
#'        containing submission data with a URL column
#' @param url_var Character string specifying the column name containing URLs.
#'        Required when urls is a tibble/data.frame. Ignored when urls is a character vector.
#' @param label_var Character string specifying the column name to use for 
#'        identifying each submission in verbose output (e.g., "name", "email").
#'        Only used when urls is a tibble/data.frame.
#' @param delay_seconds Numeric value specifying delay between opening each URL
#'        (default is 0.5 seconds to allow browser to process each request)
#' @param browser Character string specifying which browser to use. Options are
#'        "default" (system default), "chrome", "firefox", "safari", or "edge".
#'        On Windows, also supports "msedge". Default is "default".
#' @param verbose Logical value (TRUE or FALSE) specifying verbosity level.
#'        If TRUE, reports each URL as it's being opened.
#'
#' @return Invisible NULL. Function is called for its side effect of opening browser tabs.
#'
#' @details
#' The function uses the system's default method to open URLs, which typically
#' opens them in the default browser. Most modern browsers will open multiple
#' URLs as tabs in the same window when called in quick succession.
#'
#' The delay between opening URLs helps ensure the browser has time to process
#' each request properly. You may need to adjust this delay based on your
#' system performance and browser behavior.
#'
#' @examples
#' \dontrun{
#' # Open multiple GitHub Pages from vector
#' student_sites <- c("https://github.com/Abdul-Hannan96/stops.git")
#' 
#' open_github_pages(student_sites, verbose = TRUE)
#'
#' # Open from tibble/data.frame
#' path <- system.file("extdata", "answers_html", package = "tutorial.helpers")
#'
#' result <- submissions_answers(
#'   path = path,
#'   title = c("stop"), 
#'   key_var = "email",
#'   membership = c("bluebird.jack.xu@gmail.com", "abdul.hannan20008@gmail.com"),
#'   vars = c("name","email","temperance-15"),
#'   verbose = TRUE
#' )
#'
#' open_github_pages(result, 
#'                   url_var = "temperance-15",
#'                   verbose = TRUE)
#' }
#'
#' @export
open_github_pages <- function(urls, 
                             url_var = NULL,
                             label_var = NULL,
                             delay_seconds = 0.5, 
                             browser = "default", 
                             verbose = FALSE) {
  
  # Input validation
  if (missing(urls) || is.null(urls)) {
    stop("'urls' must be provided.")
  }
  
  if (!is.numeric(delay_seconds) || length(delay_seconds) != 1 || delay_seconds < 0) {
    stop("'delay_seconds' must be a single non-negative numeric value.")
  }
  
  if (!is.logical(verbose) || length(verbose) != 1) {
    stop("'verbose' must be a single logical value (TRUE or FALSE).")
  }
  
  # Validate browser parameter
  valid_browsers <- c("default", "chrome", "firefox", "safari", "edge", "msedge")
  if (!browser %in% valid_browsers) {
    stop("'browser' must be one of: ", paste(valid_browsers, collapse = ", "))
  }
  
  # Determine input type and extract URLs accordingly
  if (is.data.frame(urls)) {
    # Handle tibble/data.frame input
    submission_data <- urls
    
    if (is.null(url_var)) {
      stop("'url_var' must be provided when 'urls' is a tibble/data.frame.")
    }
    
    if (nrow(submission_data) == 0) {
      stop("'submission_data' is empty.")
    }
    
    if (!url_var %in% colnames(submission_data)) {
      stop("Column '", url_var, "' not found in submission_data. Available columns: ", 
           paste(colnames(submission_data), collapse = ", "))
    }
    
    if (!is.null(label_var) && !label_var %in% colnames(submission_data)) {
      warning("Column '", label_var, "' not found in submission_data. Labels will not be used.")
      label_var <- NULL
    }
    
    # Extract URLs and labels
    url_vector <- submission_data[[url_var]]
    labels <- if (!is.null(label_var)) submission_data[[label_var]] else NULL
    
  } else if (is.character(urls)) {
    # Handle character vector input
    url_vector <- urls
    labels <- NULL
    
    if (length(url_vector) == 0) {
      stop("'urls' vector is empty.")
    }
    
  } else {
    stop("'urls' must be either a character vector or a tibble/data.frame.")
  }
  
  # Remove any empty or NA URLs
  valid_indices <- !is.na(url_vector) & nzchar(trimws(url_vector))
  clean_urls <- url_vector[valid_indices]
  
  if (!is.null(labels)) {
    clean_labels <- labels[valid_indices]
  } else {
    clean_labels <- NULL
  }
  
  if (length(clean_urls) == 0) {
    stop("No valid URLs found after cleaning.")
  }
  
  if (verbose) {
    message("Opening ", length(clean_urls), " URL(s) in browser...")
  }
  
  # Determine browser command based on operating system and browser choice
  get_browser_command <- function(browser_choice, url) {
    os_type <- Sys.info()["sysname"]
    
    if (browser_choice == "default") {
      # Use system default browser
      if (os_type == "Windows") {
        return(paste("start", shQuote(url)))
      } else if (os_type == "Darwin") {  # macOS
        return(paste("open", shQuote(url)))
      } else {  # Linux/Unix
        return(paste("xdg-open", shQuote(url)))
      }
    } else {
      # Use specific browser
      if (os_type == "Windows") {
        browser_paths <- list(
          chrome = "start chrome",
          firefox = "start firefox",
          edge = "start msedge",
          msedge = "start msedge"
        )
        if (browser_choice %in% names(browser_paths)) {
          return(paste(browser_paths[[browser_choice]], shQuote(url)))
        }
      } else if (os_type == "Darwin") {  # macOS
        browser_apps <- list(
          chrome = "open -a 'Google Chrome'",
          firefox = "open -a 'Firefox'",
          safari = "open -a 'Safari'",
          edge = "open -a 'Microsoft Edge'"
        )
        if (browser_choice %in% names(browser_apps)) {
          return(paste(browser_apps[[browser_choice]], shQuote(url)))
        }
      } else {  # Linux/Unix
        browser_commands <- list(
          chrome = "google-chrome",
          firefox = "firefox",
          edge = "microsoft-edge"
        )
        if (browser_choice %in% names(browser_commands)) {
          return(paste(browser_commands[[browser_choice]], shQuote(url)))
        }
      }
      
      # Fallback to default if specific browser not found
      warning("Specified browser '", browser_choice, "' not found. Using default browser.")
      return(get_browser_command("default", url))
    }
  }
  
  # Open each URL
  for (i in seq_along(clean_urls)) {
    url <- clean_urls[i]
    
    # Create label for verbose output
    label <- if (!is.null(clean_labels) && !is.na(clean_labels[i]) && nzchar(trimws(clean_labels[i]))) {
      paste0(" (", clean_labels[i], ")")
    } else {
      ""
    }
    
    if (verbose) {
      message("Opening (", i, "/", length(clean_urls), ")", label, ": ", url)
    }
    
    # Get the appropriate command for this URL
    command <- get_browser_command(browser, url)
    
    # Execute the command. show.output.on.console is Windows-only; passing
    # it on other platforms triggers a warning per call.
    tryCatch({
      if (.Platform$OS.type == "windows") {
        system(command, wait = FALSE, show.output.on.console = FALSE)
      } else {
        system(command, wait = FALSE)
      }
    }, error = function(e) {
      warning("Failed to open URL", label, ": ", url, ". Error: ", e$message)
    })
    
    # Add delay between URLs (except for the last one)
    if (i < length(clean_urls) && delay_seconds > 0) {
      Sys.sleep(delay_seconds)
    }
  }
  
  if (verbose) {
    message("Finished opening all URLs.")
  }
  
  invisible(NULL)
}

# Tests for open_github_pages(). We exercise all the input validation and
# URL-extraction logic. We deliberately stop short of asserting on the final
# system() calls that actually open browser tabs, since those have side effects
# and depend on the OS/browser. Those calls are reached only with valid input;
# the branches below all error or warn before any browser is opened.

test_that("open_github_pages validates its arguments", {
  expect_error(open_github_pages(), "'urls' must be provided")
  expect_error(open_github_pages(NULL), "'urls' must be provided")

  expect_error(open_github_pages("https://example.com", delay_seconds = -1),
               "'delay_seconds' must be a single non-negative numeric value")
  expect_error(
    open_github_pages("https://example.com", delay_seconds = c(1, 2)),
    "'delay_seconds' must be a single non-negative numeric value")

  expect_error(open_github_pages("https://example.com", verbose = "yes"),
               "'verbose' must be a single logical value")

  expect_error(open_github_pages("https://example.com", browser = "netscape"),
               "'browser' must be one of")
})

test_that("open_github_pages rejects bad data.frame input", {
  df <- data.frame(url = "https://example.com", stringsAsFactors = FALSE)

  # url_var required for data.frame input
  expect_error(open_github_pages(df),
               "'url_var' must be provided when 'urls' is a tibble/data.frame")

  # empty data.frame
  empty_df <- data.frame(url = character(0), stringsAsFactors = FALSE)
  expect_error(open_github_pages(empty_df, url_var = "url"),
               "'submission_data' is empty")

  # missing url column
  expect_error(open_github_pages(df, url_var = "nope"),
               "Column 'nope' not found")
})

test_that("open_github_pages errors on unusable url collections", {
  # not a character vector or data.frame
  expect_error(
    open_github_pages(list("https://example.com")),
    "'urls' must be either a character vector or a tibble/data.frame")

  # empty character vector
  expect_error(open_github_pages(character(0)),
               "'urls' vector is empty")

  # all URLs are NA or blank -> nothing left after cleaning
  expect_error(open_github_pages(c(NA, "", "   ")),
               "No valid URLs found after cleaning")
})

#' Remove embedded pagedtable HTML artifacts from a vector of lines
#'
#' VS Code's interactive chunk execution can cache a chunk's rendered output
#' directly back into the .qmd/.Rmd source file. When the chunk's output is a
#' tibble, that cached output is a pagedtable HTML widget (a `<div
#' data-pagedtable="false">` block wrapping a `<script
#' data-pagedtable-source>` JSON blob). Because this cached output can end up
#' inside the same fenced region as the code itself, show_file()'s chunk
#' extraction can otherwise return this HTML noise instead of clean source
#' code. This helper strips any such blocks out of a character vector of
#' lines before they are printed.
#'
#' @param lines A character vector of lines (as returned by readLines()).
#' @return A character vector with any pagedtable HTML blocks removed.
#' @keywords internal
strip_pagedtable_html <- function(lines) {
  if (length(lines) == 0) {
    return(lines)
  }

  keep <- rep(TRUE, length(lines))
  depth <- 0
  in_block <- FALSE

  for (i in seq_along(lines)) {
    line <- lines[i]

    # A pagedtable block always starts with a <div data-pagedtable...> tag.
    if (!in_block && grepl("^\\s*<div\\s+data-pagedtable", line)) {
      in_block <- TRUE
      depth <- 0
    }

    if (in_block) {
      keep[i] <- FALSE
      # Track nested <div ...> / </div> tags so we find the *matching*
      # closing tag rather than the first </div> we see.
      depth <- depth + lengths(regmatches(line, gregexpr("<div\\b", line)))
      depth <- depth - lengths(regmatches(line, gregexpr("</div>", line)))
      if (depth <= 0) {
        in_block <- FALSE
      }
    }
  }

  lines[keep]
}

#' Display the contents of a text file that match a pattern
#'
#' This function reads the contents of a text file and either prints the specified range of rows
#' that match a given regular expression pattern, prints the code lines within code chunks,
#' or extracts the YAML header. If start is a negative number, it prints the last abs(start) lines,
#' ignoring missing lines at the end of the file. If start is 0, it prints the entire file.
#'
#' The arguments are resolved in a fixed order of precedence: `chunk = "YAML"`
#' is handled first, then `start == 0` (whole file), then `start < 0` (last
#' abs(start) lines), then `chunk %in% c("All", "Last")` (code chunks), and
#' finally the `start`/`end` row range. The `pattern` filter is applied within
#' the whole-file, last-lines, and row-range cases, but is ignored when `chunk`
#' selects code chunks or the YAML header.
#'
#' @param path A character vector representing the path to the text file.
#' @param start An integer specifying the starting row number (inclusive) to consider. Default is 1.
#'              If negative, it represents the number of lines to print from the end of the file.
#'              If 0, prints the entire file.
#' @param end An integer specifying the ending row number (inclusive) to consider. Default is the last row.
#' @param pattern A regular expression pattern to match against each row. Default is NULL (no pattern matching).
#'              Applied to the whole-file (`start == 0`), last-lines (`start < 0`), and row-range cases;
#'              ignored when `chunk` is "All", "Last", or "YAML".
#' @param chunk A character string indicating what content to extract.
#'              Possible values are "None" (default - no chunk processing),
#'              "All" (print all code chunks), "Last" (print only the last code chunk),
#'              or "YAML" (extract the YAML header without delimiters). Code chunks of
#'              any language are recognized (e.g. `r`, `python`, `bash`), not just R.
#' @return The function prints the contents of the specified range of rows that match the pattern (if provided),
#'         the code lines within R code chunks (if chunk is "All" or "Last"),
#'         or the YAML header content (if chunk is "YAML") to the console. If no rows match the pattern,
#'         nothing is printed. If start is negative, the function prints the last abs(start) lines, ignoring
#'         missing lines at the end of the file. If start is 0, the function prints the entire file.
#'
#' @examples
#' \dontrun{
#' # Display all rows of a text file
#' show_file("path/to/your/file.txt")
#'
#' # Display the entire file
#' show_file("path/to/your/file.txt", start = 0)
#'
#' # Display rows 5 to 10 of a text file
#' show_file("path/to/your/file.txt", start = 5, end = 10)
#'
#' # Display all rows of a text file that contain the word "example"
#' show_file("path/to/your/file.txt", pattern = "example")
#'
#' # Print all code lines within R code chunks
#' show_file("path/to/your/file.txt", chunk = "All")
#'
#' # Print only the last R code chunk
#' show_file("path/to/your/file.txt", chunk = "Last")
#'
#' # Extract the YAML header
#' show_file("path/to/your/file.Rmd", chunk = "YAML")
#'
#' # Display the last 5 lines of a text file, ignoring missing lines at the end
#' show_file("path/to/your/file.txt", start = -5)
#' }
#'
#' @export

show_file <- function(path, start = 1, end = NULL, pattern = NULL, chunk = "None") {
  # Check if the file exists
  if (!file.exists(path)) {
    stop("File does not exist.")
  }
  
  # Validate chunk parameter
  allowed_chunks <- c("None", "All", "Last", "YAML")
  if (!chunk %in% allowed_chunks) {
    stop("chunk must be one of: ", paste(allowed_chunks, collapse = ", "))
  }
  
  # Read the contents of the file
  contents <- readLines(path)

  # VS Code's interactive chunk execution can cache a chunk's rendered
  # output directly back into the .qmd/.Rmd source file. When the chunk's
  # output is a tibble, that cached output is a pagedtable HTML widget (a
  # `<div data-pagedtable="false">` block wrapping a `<script
  # data-pagedtable-source>` JSON blob), and it can end up inside the same
  # fenced region as the code itself. Strip any such blocks here, up front,
  # so every code path below (YAML, whole-file, tail, chunk extraction, and
  # row-range) sees clean source text rather than rendered-output noise.
  contents <- strip_pagedtable_html(contents)
  
  # Remove trailing empty lines from the contents
  while (length(contents) > 0 && contents[length(contents)] == "") {
    contents <- contents[-length(contents)]
  }

  if (length(contents) == 0) {
    cat("File is empty.\n")
    return(invisible(NULL))
  }

  # If chunk is "YAML", extract and return the YAML header
  if (chunk == "YAML") {
    if (length(contents) == 0 || !grepl("^---\\s*$", contents[1])) {
      stop("No YAML header found.")
    }
    
    # Find the closing --- for the YAML header. seq_len(length(contents))[-1]
    # gives the indices from 2 to the end, and is empty (rather than c(2, 1))
    # when the file has only a single line.
    yaml_end <- NULL
    for (i in seq_len(length(contents))[-1]) {
      if (grepl("^---\\s*$", contents[i])) {
        yaml_end <- i
        break
      }
    }
    
    if (is.null(yaml_end)) {
      stop("No YAML header found.")
    }
    
    # Extract YAML content (excluding the --- delimiters)
    if (yaml_end > 2) {
      yaml_content <- contents[2:(yaml_end - 1)]
      cat(yaml_content, sep = "\n")
    }
    return(invisible(NULL))
  }
  
  # If start is 0, print the entire file
  if (start == 0) {
    selected_contents <- contents
    if (!is.null(pattern)) {
      selected_contents <- selected_contents[grepl(pattern, selected_contents)]
    }
    if (length(selected_contents) > 0) {
      cat(selected_contents, sep = "\n")
    }
    return(invisible(NULL))
  }

  # If start is negative, print the last abs(start) lines
  if (start < 0) {
    selected_contents <- utils::tail(contents, abs(start))
    if (!is.null(pattern)) {
      selected_contents <- selected_contents[grepl(pattern, selected_contents)]
    }
    if (length(selected_contents) > 0) {
      cat(selected_contents, sep = "\n")
    }
    return(invisible(NULL))
  }
  
  # If chunk is "All" or "Last", print code lines within code chunks. We match
  # the opening fence of any language (```{r}, ```{python}, ```{bash}, ...),
  # not just R, so that "Last"/"All" work for every kind of tutorial chunk.
  if (chunk %in% c("All", "Last")) {
    in_chunk <- FALSE
    code_chunks <- list()
    current_chunk <- character()

    for (line in contents) {
      if (grepl("^```\\{", line)) {
        in_chunk <- TRUE
        if (length(current_chunk) > 0) {
          code_chunks <- c(code_chunks, list(current_chunk))
          current_chunk <- character()
        }
      } else if (grepl("^```\\s*$", line)) {
        in_chunk <- FALSE
        if (length(current_chunk) > 0) {
          code_chunks <- c(code_chunks, list(current_chunk))
          current_chunk <- character()
        }
      } else if (in_chunk) {
        current_chunk <- c(current_chunk, line)
      }
    }
    
    if (chunk == "All") {
      for (i in seq_along(code_chunks)) {
        cat(code_chunks[[i]], sep = "\n")
        if (i < length(code_chunks)) {
          cat("\n")
        }
      }
    } else if (chunk == "Last") {
      if (length(code_chunks) > 0) {
        cat(code_chunks[[length(code_chunks)]], sep = "\n")
      }
    }
    return(invisible(NULL))
  }
  
  # Get the total number of rows
  total_rows <- length(contents)
  
  # Set default value for end if not provided
  if (is.null(end)) {
    end <- total_rows
  }
  
  # Check if start and end are within the valid range
  if (start < 1 || start > total_rows || end < 1 || end > total_rows) {
    stop("start and end must be within the valid range of rows.")
  }
  
  # Check if start is smaller or equal to end
  if (start > end) {
    stop("start must be smaller or equal to end.")
  }
  
  # Extract the specified range of rows
  selected_contents <- contents[start:end]
  
  # Filter the selected rows based on the pattern (if provided)
  if (!is.null(pattern)) {
    selected_contents <- selected_contents[grepl(pattern, selected_contents)]
  }
  
  # Print the selected contents to the console if there are matching rows
  if (length(selected_contents) > 0) {
    cat(selected_contents, sep = "\n")
  }
}

# Test file path using testthat::test_path()
# This automatically finds the file relative to the test directory
test_file <- test_path("fixtures", "show_file_test.txt")
test_file_yaml <- test_path("fixtures", "show_file_yaml_test.qmd")
test_file_python <- test_path("fixtures", "show_file_python_test.qmd")

# Test cases
test_that("show_file function works correctly", {
  # Test case 1: Display all rows
  expect_equal(paste(capture.output(show_file(test_file)), collapse = "\n"),
               paste(c(
                 "This is line 1.",
                 "This is line 2.",
                 "This is line 3 with the word example.",
                 "```{r}",
                 "# This is a code chunk",
                 "",
                 "x <- 1:10",
                 "print(x)",
                 "```",
                 "This is line 4.",
                 "This is line 5 with another example.",
                 "```{r}",
                 "# Another code chunk",
                 "y <- 20:30",
                 "mean(y)",
                 "```",
                 "This is line 6.",
                 "This is line 7.",
                 "This is line 8.",
                 "This is line 9.",
                 "This is line 10.",
                 "",
                 "This is line 11 with no matching pattern."
               ), collapse = "\n"))
  
  # Test case 2: Display entire file with start = 0
  expect_equal(paste(capture.output(show_file(test_file, start = 0)), collapse = "\n"),
               paste(c(
                 "This is line 1.",
                 "This is line 2.",
                 "This is line 3 with the word example.",
                 "```{r}",
                 "# This is a code chunk",
                 "",
                 "x <- 1:10",
                 "print(x)",
                 "```",
                 "This is line 4.",
                 "This is line 5 with another example.",
                 "```{r}",
                 "# Another code chunk",
                 "y <- 20:30",
                 "mean(y)",
                 "```",
                 "This is line 6.",
                 "This is line 7.",
                 "This is line 8.",
                 "This is line 9.",
                 "This is line 10.",
                 "",
                 "This is line 11 with no matching pattern."
               ), collapse = "\n"))
  
  # Test case 3: Display rows 3 to 7
  expect_equal(paste(capture.output(show_file(test_file, start = 3, end = 7)), collapse = "\n"),
               paste(c(
                 "This is line 3 with the word example.",
                 "```{r}",
                 "# This is a code chunk",
                 "",
                 "x <- 1:10"
               ), collapse = "\n"))
  
  # Test case 4: Display rows matching the pattern "example"
  expect_equal(paste(capture.output(show_file(test_file, pattern = "example")), collapse = "\n"),
               paste(c(
                 "This is line 3 with the word example.",
                 "This is line 5 with another example."
               ), collapse = "\n"))
  
  # Test case 5: Print all code chunks
  expect_equal(paste(capture.output(show_file(test_file, chunk = "All")), collapse = "\n"),
               paste(c(
                 "# This is a code chunk",
                 "",
                 "x <- 1:10",
                 "print(x)",
                 "",
                 "# Another code chunk",
                 "y <- 20:30",
                 "mean(y)"
               ), collapse = "\n"))
  
  # Test case 6: Print the last code chunk
  expect_equal(paste(capture.output(show_file(test_file, chunk = "Last")), collapse = "\n"),
               paste(c(
                 "# Another code chunk",
                 "y <- 20:30",
                 "mean(y)"
               ), collapse = "\n"))
  
  # Test case 7: Invalid chunk value (numeric)
  expect_error(show_file(test_file, chunk = 1), 
               "chunk must be one of: None, All, Last, YAML")
  
  # Test case 8: Invalid chunk value (incorrect string)
  expect_error(show_file(test_file, chunk = "invalid"), 
               "chunk must be one of: None, All, Last, YAML")
  
  # Test case 9: Extract YAML header (assuming test_file_yaml has proper YAML)

  expect_equal(paste(capture.output(show_file(test_file_yaml, chunk = "YAML")), collapse = "\n"),
               paste(c(
                 'title: "Test Document"',
                 'author: "Test Author"',
                 'date: "2024-01-01"'
               ), collapse = "\n"))
  
  # Test case 10: No YAML header found
  expect_error(show_file(test_file, chunk = "YAML"), 
               "No YAML header found.")
  
  # Test case 11: File does not exist
  expect_error(show_file("nonexistent_file.txt"), "File does not exist.")
  
  # Test case 12: Start is greater than end
  expect_error(show_file(test_file, start = 5, end = 3), "start must be smaller or equal to end.")
  
  # Test case 13: End is out of range
  expect_error(show_file(test_file, end = 30), "start and end must be within the valid range of rows.")
  
  # Test case 14: Print the last 3 lines of the file
  expect_equal(paste(capture.output(show_file(test_file, start = -3)), collapse = "\n"),
               paste(c(
                 "This is line 10.",
                 "",
                 "This is line 11 with no matching pattern."
               ), collapse = "\n"))
  
  # Test case 15: No rows matching the pattern
  expect_equal(paste(capture.output(show_file(test_file, pattern = "nomatch")), collapse = "\n"), "")

  # Test case 16: Empty file prints "File is empty."
  empty_file <- tempfile()
  file.create(empty_file)
  on.exit(unlink(empty_file), add = TRUE)
  expect_equal(paste(capture.output(show_file(empty_file)), collapse = "\n"),
               "File is empty.")

  # Test case 17: A file containing only blank lines prints "File is empty."
  blank_file <- tempfile()
  writeLines(c("", "", ""), blank_file)
  on.exit(unlink(blank_file), add = TRUE)
  expect_equal(paste(capture.output(show_file(blank_file)), collapse = "\n"),
               "File is empty.")

  # Test case 18: chunk = "None" (the default) shows the whole file unchanged
  expect_equal(paste(capture.output(show_file(test_file, chunk = "None")), collapse = "\n"),
               paste(capture.output(show_file(test_file)), collapse = "\n"))

  # Test case 19: pattern combined with a row range
  expect_equal(paste(capture.output(show_file(test_file, start = 1, end = 5, pattern = "example")), collapse = "\n"),
               paste(c(
                 "This is line 3 with the word example."
               ), collapse = "\n"))

  # Test case 20: pattern combined with start = 0 (whole file) -- Bug #1
  expect_equal(paste(capture.output(show_file(test_file, start = 0, pattern = "example")), collapse = "\n"),
               paste(c(
                 "This is line 3 with the word example.",
                 "This is line 5 with another example."
               ), collapse = "\n"))

  # Test case 21: pattern combined with a negative start -- Bug #1
  expect_equal(paste(capture.output(show_file(test_file, start = -5, pattern = "matching")), collapse = "\n"),
               paste(c(
                 "This is line 11 with no matching pattern."
               ), collapse = "\n"))

  # Test case 22: negative start larger than the file returns all lines
  expect_equal(paste(capture.output(show_file(test_file, start = -100)), collapse = "\n"),
               paste(capture.output(show_file(test_file, start = 0)), collapse = "\n"))

  # Test case 23: chunk = "Last" on a qmd with named R chunks
  expect_equal(paste(capture.output(show_file(test_file_yaml, chunk = "Last")), collapse = "\n"),
               paste(c(
                 "x <- 1:10",
                 "mean(x)"
               ), collapse = "\n"))

  # Test case 24: chunk = "All" on a qmd with named R chunks
  expect_equal(paste(capture.output(show_file(test_file_yaml, chunk = "All")), collapse = "\n"),
               paste(c(
                 "library(tidyverse)",
                 "",
                 "x <- 1:10",
                 "mean(x)"
               ), collapse = "\n"))

  # Test case 25: chunk = "Last" works on Python chunks
  expect_equal(paste(capture.output(show_file(test_file_python, chunk = "Last")), collapse = "\n"),
               paste(c(
                 'df = pd.DataFrame({"x": [1, 2, 3]})',
                 "df.head()"
               ), collapse = "\n"))

  # Test case 26: chunk = "All" captures all Python chunks
  expect_equal(paste(capture.output(show_file(test_file_python, chunk = "All")), collapse = "\n"),
               paste(c(
                 "import pandas as pd",
                 "",
                 'df = pd.DataFrame({"x": [1, 2, 3]})',
                 "df.head()"
               ), collapse = "\n"))

  # Test case 27: single-line file with no YAML errors cleanly -- Bug #2
  one_line_file <- tempfile(fileext = ".txt")
  writeLines("only one line", one_line_file)
  on.exit(unlink(one_line_file), add = TRUE)
  expect_error(show_file(one_line_file, chunk = "YAML"), "No YAML header found.")

  # Test case 28: chunk = "Last" strips embedded pagedtable HTML that VS
  # Code's interactive chunk execution can cache back into a .qmd file
  test_file_pagedtable <- test_path("fixtures", "show_file_pagedtable_test.qmd")
  expect_equal(paste(capture.output(show_file(test_file_pagedtable, chunk = "Last")), collapse = "\n"),
               "penguins")

  # Test case 29: chunk = "All" strips pagedtable HTML from every chunk
  expect_equal(paste(capture.output(show_file(test_file_pagedtable, chunk = "All")), collapse = "\n"),
               paste(c(
                 "library(tidyverse)",
                 "",
                 "penguins"
               ), collapse = "\n"))

  # Test case 30: pagedtable HTML is stripped even outside chunk mode
  expect_false(grepl("pagedtable",
                      paste(capture.output(show_file(test_file_pagedtable, start = 0)), collapse = "\n")))

  # Test case 31: nested <div> tags inside the pagedtable JSON payload do
  # not confuse the stripping logic
  nested_div_file <- tempfile(fileext = ".qmd")
  writeLines(c(
    "```{r}",
    "df",
    '<div data-pagedtable="false">',
    '  <script data-pagedtable-source type="application/json">',
    '{"data":[["<div>nested</div>"]]}',
    "  </script>",
    "</div>",
    "```"
  ), nested_div_file)
  on.exit(unlink(nested_div_file), add = TRUE)
  expect_equal(paste(capture.output(show_file(nested_div_file, chunk = "Last")), collapse = "\n"),
               "df")

  # Test case 32: strip_pagedtable_html() is a no-op on ordinary lines
  expect_equal(strip_pagedtable_html(c("a <- 1", "b <- 2")), c("a <- 1", "b <- 2"))

  # Test case 33: strip_pagedtable_html() handles an empty vector
  expect_equal(strip_pagedtable_html(character(0)), character(0))
})

# Regression test: a closing fence with trailing whitespace used to go
# unrecognized, so following prose leaked into the chunk output.

test_that("closing fences with trailing whitespace are recognized", {
  f <- tempfile(fileext = ".Rmd")
  on.exit(unlink(f))
  writeLines(c("```{r}", "x <- 1", "``` ", "Some prose."), f)
  out <- capture.output(show_file(f, chunk = "Last"))
  expect_equal(out, "x <- 1")
})

#' Download Files from a Google Drive Folder
#'
#' Downloads all files or filtered files from a public Google Drive folder to a local directory.
#'
#' @param url Character string. The Google Drive folder URL or ID. Supports standard folder URLs
#'   (e.g., \code{"https://drive.google.com/drive/folders/FOLDER_ID"}) or direct folder IDs.
#' @param path Character string or NULL. The local directory path for downloads.
#'   If NULL (default), uses the current working directory. If the directory doesn't exist, it will be created.
#' @param title Character vector or NULL. Patterns to match against file names for filtering.
#'   If provided, only files whose names contain any of these patterns (case-insensitive) are downloaded.
#'   If NULL (default), all files are downloaded.
#' @param overwrite Logical. If TRUE (default), overwrites existing files. If FALSE, skips files that already exist.
#' @param verbose Logical. If TRUE (default), prints detailed progress messages.
#'
#' @return Character string. The path to the directory where files were downloaded.
#'
#' @examples
#' \dontrun{
#' # Download all files to current directory
#' download_google_drive("https://drive.google.com/drive/folders/1Rgxfiw")
#'
#' # Download to a specific directory with filtering
#' download_google_drive(
#'   url = "https://drive.google.com/drive/folders/1Rgxfiw",
#'   path = "./my_data",
#'   title = c("report", ".csv"),
#'   overwrite = FALSE
#' )
#' }
#'
#' @export

download_google_drive <- function(url, path = NULL, title = NULL, overwrite = TRUE, verbose = TRUE) {
  # Extract folder ID from URL or use directly if ID is provided
  folder_id <- if (grepl("drive\\.google\\.com", url)) {
    # Extract the ID between /folders/ and the next / or ? or end of string
    # This handles various URL formats including those with additional parameters
    if (grepl("/folders/", url)) {
      # Extract everything after /folders/ until we hit ? or end
      id <- sub(".*/folders/([^/?]+).*", "\\1", url)
    } else if (grepl("id=", url)) {
      # Handle old-style URLs with id= parameter
      id <- sub(".*id=([^&]+).*", "\\1", url)
    } else {
      stop("Could not extract folder ID from URL. Please check the URL format.")
    }
    
    # Basic validation - folder IDs should be alphanumeric with hyphens and underscores
    # Remove the length restriction as IDs can vary
    if (!grepl("^[a-zA-Z0-9_-]+$", id)) {
      stop("Invalid Google Drive folder ID extracted from URL: ", id)
    }
    id
  } else {
    # Assume it's a direct folder ID
    url
  }
  
  if (verbose) cat("Using folder ID:", folder_id, "\n")

  if (!requireNamespace("googledrive", quietly = TRUE)) {
    stop("Package 'googledrive' is required to download files from Google Drive. ",
         "Please install it with: install.packages('googledrive')")
  }

  # Deauthorize for public access
  googledrive::drive_deauth()
  
  # Get list of files in the folder
  if (verbose) cat("Fetching file list from Google Drive...\n")
  
  tryCatch(
    {
      files <- googledrive::drive_ls(googledrive::as_id(folder_id))
    },
    error = function(e) {
      stop("Failed to list files from Google Drive folder. ",
           "Please ensure the folder is publicly accessible. ",
           "Error: ", e$message)
    }
  )
  
  # Filter files by title patterns if provided
  if (!is.null(title)) {
    pattern <- paste(title, collapse = "|")
    files <- files[grepl(pattern, files$name, ignore.case = TRUE), ]
    if (verbose && nrow(files) > 0) {
      cat("Filtered to", nrow(files), "files matching pattern(s):", paste(title, collapse = ", "), "\n")
    }
  }
  
  if (nrow(files) == 0) {
    if (verbose) cat("No files found matching the criteria.\n")
    return(character(0))
  }
  
  # Set download directory
  download_dir <- if (is.null(path)) getwd() else path
  
  # Create directory if it doesn't exist
  if (!dir.exists(download_dir)) {
    if (verbose) cat("Creating directory:", download_dir, "\n")
    dir.create(download_dir, recursive = TRUE, showWarnings = FALSE)
  }
  
  # Download files with progress feedback
  if (verbose) {
    cat("Found", nrow(files), "file(s) to download\n")
    cat("Download directory:", download_dir, "\n")
    cat("Starting downloads...\n")
  }
  
  downloaded <- 0
  skipped <- 0
  failed <- 0
  
  for (i in seq_len(nrow(files))) {
    file_path <- file.path(download_dir, files$name[i])
    
    if (!overwrite && file.exists(file_path)) {
      if (verbose) cat("  [", i, "/", nrow(files), "] Skipping (exists):", files$name[i], "\n")
      skipped <- skipped + 1
      next
    }
    
    tryCatch(
      {
        if (verbose) cat("  [", i, "/", nrow(files), "] Downloading:", files$name[i], "...")
        googledrive::drive_download(files[i, ], path = file_path, overwrite = overwrite)
        if (verbose) cat(" Done\n")
        downloaded <- downloaded + 1
      },
      error = function(e) {
        if (verbose) cat(" Failed -", e$message, "\n")
        failed <<- failed + 1
      }
    )
  }
  
  if (verbose) {
    cat("\nDownload complete:\n")
    cat("  - Downloaded:", downloaded, "file(s)\n")
    if (skipped > 0) cat("  - Skipped:", skipped, "existing file(s)\n")
    if (failed > 0) cat("  - Failed:", failed, "file(s)\n")
  }
  
  return(download_dir)
}

# This file is part of the standard setup for testthat. It is recommended that
# you do not modify it.
#
# Where should you do additional test configuration? Learn more about the roles
# of various files in:
# * https://r-pkgs.org/tests.html
# * https://testthat.r-lib.org/reference/test_package.html#special-files

# Need to think harder about how we are doing testing. In particular, there are
# a lot of functions here which test all the tutorials in a package. But how
# should we test them here? And how do we use those functions in other tutorial
# packages? This was all easier when everything was bundled together.

library(testthat)
library(tutorial.helpers)

test_check("tutorial.helpers")

#' Finds the number of the next exercise in a tutorial
#'
#' @param file_path Character string of the file path to the tutorial
#'
#' @return The next exercise number based on the file argument or the active document.
#' @export
determine_exercise_number <- function(file_path = NULL){

  # Similar to determine_code_chunk_name(), getActiveDocumentContext() is not currently tested and uses file_path as
  # an alternative. There are issues to this due to cursor placement. Currently, the function returns an integer
  # explore perhaps the typing of the return object, and note that cut_content represents the tutorial's text but reversed.

  if (is.null(file_path))
  {

    ctx <- rstudioapi::getActiveDocumentContext()
    row <- ctx$selection[[1]]$range$end[["row"]]

    cut_content <- rev(ctx$contents[1:row])

  }
  else
  {
    cut_content <- rev(readLines(file_path))

  }

  for (l in cut_content){

    # Find the latest exercise header. The match is anchored so that prose or
    # code merely mentioning "### Exercise" does not alter the numbering.

    if (grepl("^### Exercise\\b", l)){

      # Extract only the number immediately following "Exercise", so headers
      # like "### Exercise 2 (part 3)" yield 2, not 23. A bare "### Exercise"
      # header has no number; treat it as exercise 0 so we return 1.

      num <- sub("^### Exercise\\s*", "", l)
      num <- sub("^(\\d*).*$", "\\1", num)
      latest <- if (nzchar(num)) as.integer(num) else 0L

      return(latest + 1L)
    }

    # Find the latest section. A section header before any exercise means
    # the next exercise is the first one.

    if (grepl("^## ", l)){
      return(1L)
    }
  }

  # No exercise or section header found above this point in the file.

  return(1L)
}

test_file_1 <- test_path("fixtures", "tutorial_examples", "exercise-number-test-tutorial-1.Rmd")

test_file_2 <- test_path("fixtures", "tutorial_examples", "exercise-number-test-tutorial-2.Rmd")

test_that("determine exercise numbers", {
  expect_equal(tutorial.helpers::determine_exercise_number(test_file_1),
               1)
  
  expect_equal(tutorial.helpers::determine_exercise_number(test_file_2),
               4)
})




# Regression tests: the number extraction used to concatenate every digit in
# the heading ("### Exercise 2 (part 3)" gave 24), unanchored matching let
# prose alter the numbering, and a file with no headers returned NULL.

test_that("only the number directly after 'Exercise' is used", {
  f <- tempfile(fileext = ".Rmd")
  on.exit(unlink(f))

  writeLines(c("## Section", "### Exercise 2 (part 3)"), f)
  expect_equal(determine_exercise_number(f), 3L)

  writeLines(c("## Section", "### Exercise"), f)
  expect_equal(determine_exercise_number(f), 1L)

  writeLines(c("## Section", 'str_detect(x, "### Exercise 5")'), f)
  expect_equal(determine_exercise_number(f), 1L)

  writeLines("Just some prose.", f)
  expect_equal(determine_exercise_number(f), 1L)
})

# R/utils.R
#' Package build dependencies
#'
#' This function ensures renv detects quarto as a dependency
#' while keeping it in Suggests rather than Imports. It is deliberately
#' never called: its mere presence (referencing the quarto namespace) is
#' what makes renv record quarto. Do not delete it.
#' @keywords internal
.check_quarto <- function() {
  if (requireNamespace("quarto", quietly = TRUE)) {
    return(TRUE)
  }
  return(FALSE)
}

#' Process Submissions Summary
#'
#' This function processes submissions from a local directory or Google Drive folder containing HTML/XML files.
#' It extracts tables from the files, filters them based on a pattern and key variables,
#' and returns either a summary tibble or a combined tibble with all the data.
#'
#' @param path The path to the local directory containing the HTML/XML files, or a Google Drive folder URL.
#'        If it's a Google Drive URL, the function will download individual files to a temporary directory.
#' @param title A character vector of patterns to match against the file names (default: ".").
#'        Each pattern is processed separately and results are combined.
#' @param emails A character vector of email addresses to filter results by, "*" to include all emails, or NULL to skip email filtering (default: NULL).
#' @param return_value The type of value to return. Allowed values are "Summary" (default) or "All".
#' @param vars A character vector of key variables to extract from the "id" column (default: NULL).
#' @param verbose A logical value (TRUE or FALSE) specifying verbosity level.
#'        If TRUE, reports files that are removed during processing.
#' @param keep_file_name Specifies whether to keep the file name in the summary tibble. Allowed values are NULL (default), "All" (keep entire file name), "Space" (keep up to first space), or "Underscore" (keep up to first underscore). Only used when `return_value` is "Summary".
#'
#' @return If `return_value` is "Summary", returns a tibble with one row for each file, columns corresponding to the `vars`,
#'         and an additional "answers" column indicating the number of rows in each tibble.
#'         If `return_value` is "All", returns a tibble with all the data combined from all the files.
#'
#' @examples
#' # Process submissions from local directory
#' path <- system.file("extdata", "answers_html", package = "tutorial.helpers")
#'
#' result <- submissions_summary(path = path,
#'                              vars = "email",
#'                              title = "stop")
#'
#' \dontrun{
#' drive_url <- "https://drive.google.com/drive/folders/10do12t0fZsfrIrKePxwjpH8IqBNVO86N"
#' x <- submissions_summary(
#'   path = drive_url, 
#'   title = c("positron"),
#'   vars = c("email", "name")
#' )
#' 
#' 
#' }
#' @export
submissions_summary <- function(path, 
                                title = ".", 
                                return_value = "Summary", 
                                vars = NULL, 
                                verbose = FALSE, 
                                keep_file_name = NULL,
                                emails = NULL) {
  
  # Validation: path must be provided
  if (missing(path) || is.null(path)) {
    stop("'path' must be provided.")
  }
  
  # Validate verbose parameter
  if (!is.logical(verbose) || length(verbose) != 1) {
    stop("'verbose' must be a single logical value (TRUE or FALSE).")
  }
  
  # Check if return_value is valid
  if (!(return_value %in% c("Summary", "All"))) {
    stop("Invalid return_value. Allowed values are 'Summary' or 'All'.")
  }
  
  # Check if vars is provided when return_value is "Summary"
  if (return_value == "Summary" && is.null(vars)) {
    stop("vars must be provided when return_value is 'Summary'.")
  }
  
  if (!is.null(keep_file_name) && return_value != "Summary") {
    stop("keep_file_name can only be used when return_value is 'Summary'.")
  }
  
  if (!is.null(keep_file_name) && !(keep_file_name %in% c("All", "Space", "Underscore"))) {
    stop("Invalid keep_file_name. Allowed values are NULL, 'All', 'Space', or 'Underscore'.")
  }
  
  # Call gather_submissions to get the list of tibbles. It has already
  # filtered by every pattern in `title` and deduplicated the files, so we
  # process each file exactly once; re-looping over the patterns here would
  # duplicate files matched by more than one pattern.
  tibble_list <- gather_submissions(path = path, title = title, verbose = verbose)

  filtered_tibble_list <- list()
  removed_files <- character()
  removal_reasons <- character()

  for (file_name in names(tibble_list)) {
    tibble_data <- tibble_list[[file_name]]

    # Answers are stored in an 'answer' column or, in older files, a 'data'
    # column.
    answer_col <- if ("answer" %in% colnames(tibble_data)) {
      "answer"
    } else if ("data" %in% colnames(tibble_data)) {
      "data"
    } else {
      NA_character_
    }

    if (!"id" %in% colnames(tibble_data) || is.na(answer_col)) {
      removed_files <- c(removed_files, file_name)
      removal_reasons <- c(removal_reasons, "no 'id' and 'answer' (or 'data') columns")
      next
    }

    if (!is.null(vars) && !all(vars %in% tibble_data$id)) {
      missing_vars <- setdiff(vars, tibble_data$id)
      removed_files <- c(removed_files, file_name)
      removal_reasons <- c(removal_reasons, paste("missing key variables:", paste(missing_vars, collapse = ", ")))
      next
    }

    if (!is.null(emails) && !identical(emails, "*")) {
      email_value <- tibble_data[[answer_col]][tibble_data$id == "email"][1]
      if (is.na(email_value) || !email_value %in% emails) {
        removed_files <- c(removed_files, file_name)
        removal_reasons <- c(removal_reasons, "email not in 'emails'")
        next
      }
    }

    for (key_var in vars) {
      key_var_value <- tibble_data[[answer_col]][tibble_data$id == key_var][1]
      tibble_data[[key_var]] <- key_var_value
    }
    filtered_tibble_list[[file_name]] <- tibble_data
  }

  # Report removed files if verbose
  if (verbose && length(removed_files) > 0) {
    for (i in seq_along(removed_files)) {
      message("Removed '", removed_files[i], "': ", removal_reasons[i])
    }
  }

  if (length(filtered_tibble_list) == 0) {
    return(tibble::tibble())
  }

  if (return_value == "Summary") {
    summary_rows <- lapply(names(filtered_tibble_list), function(file_name) {
      tibble_data <- filtered_tibble_list[[file_name]]
      answers <- nrow(tibble_data)

      if (!is.null(keep_file_name)) {
        if (keep_file_name == "All") {
          source_name <- file_name
        } else if (keep_file_name == "Space") {
          source_name <- sub("\\s.*", "", file_name)
        } else if (keep_file_name == "Underscore") {
          source_name <- sub("_.*", "", file_name)
        }
      } else {
        source_name <- NULL
      }

      # First row, keeping only the requested vars, plus an answers count.
      summary_row <- tibble_data[1, vars, drop = FALSE]
      summary_row[["answers"]] <- answers

      if (!is.null(source_name)) {
        summary_row <- tibble::add_column(summary_row, source = source_name, .before = 1)
      }

      summary_row
    })
    combined_results <- do.call(rbind, summary_rows)
  } else {
    combined_results <- do.call(rbind, filtered_tibble_list)
  }

  tibble::as_tibble(combined_results)
}

test_dir <- test_path("fixtures", "process_submissions_dir") 
# If test_dir does not work try using this path = test_path("fixtures", "process_submissions_dir") 

# Currently, I do not have tests which explore the verbose argument in detail.

# Tests which confirm that the returned tibble is correct.

test_that("submissions_summary returns the expected summary tibble", {
  expected_output <- tibble::tibble(
    `information-name` = c("Areeb Atif", "Gitanjali Sheth"),
    answers = c(11L, 11L)
  )
  
  actual_output <- submissions_summary(
    path = test_dir,
    title = "getting",
    vars = c("information-name")
  )
  
  expect_equal(actual_output, expected_output)
})

test_that("submissions_summary returns the expected summary tibble with email vars", {
  expected_output <- tibble::tibble(
    `information-email` = c("dave.kane@gmail.com", "areebatif2007@gmail.com", "gbhatia1@yahoo.com"),
    answers = c(6L, 11L, 11L)
  )
  
  actual_output <- submissions_summary(
    path = test_dir,
    title = "get",
    vars = c("information-email")
  )
  
  expect_equal(actual_output, expected_output)
})

test_that("submissions_summary returns the expected summary tibble with name and email vars", {
  expected_output <- tibble::tibble(
    `information-name` = c("David Kane", "Areeb Atif", "Gitanjali Sheth"),
    `information-email` = c("dave.kane@gmail.com", "areebatif2007@gmail.com", "gbhatia1@yahoo.com"),
    answers = c(6L, 11L, 11L)
  )
  
  actual_output <- submissions_summary(
    path = test_dir,
    title = "get",
    vars = c("information-name", "information-email")
  )
  
  expect_equal(actual_output, expected_output)
})

test_that("submissions_summary returns a tibble of the expected size with return_value 'All'", {
  actual_output <- submissions_summary(
    path = test_dir,
    title = "getting",
    return_value = "All"
  )
  
  expect_s3_class(actual_output, "tbl_df")
  expect_equal(dim(actual_output), c(22L, 3L))
})

test_that("submissions_summary returns a tibble of the expected size with return_value 'All' and title 'get'", {
  actual_output <- submissions_summary(
    path = test_dir,
    title = "get",
    return_value = "All"
  )
  
  expect_s3_class(actual_output, "tbl_df")
  expect_equal(dim(actual_output), c(28L, 3L))
})

test_that("submissions_summary with one key_var", {
  actual_output <- submissions_summary(
    path = test_dir,
    title = "get",
    return_value = "All",
    vars = "information-name"
  )
  
  expect_s3_class(actual_output, "tbl_df")
  expect_equal(dim(actual_output), c(28L, 4L))
})

test_that("submissions_summary with two vars", {
  actual_output <- submissions_summary(
    path = test_dir,
    title = "get",
    return_value = "All",
    vars = c("information-name", "information-email")
  )
  
  expect_s3_class(actual_output, "tbl_df")
  expect_equal(dim(actual_output), c(28L, 5L))
})

# Tests which confirm various errors.

test_that("submissions_summary stops with an error when return_value is 'Summary' and no vars are provided", {
  expect_error(
    submissions_summary(
      path = test_dir,
      title = "get",
      return_value = "Summary"
    ),
    "vars must be provided when return_value is 'Summary'."
  )
})

test_that("submissions_summary stops with an error when the specified directory does not exist", {
  expect_error(
    submissions_summary(
      path = "z",
      title = "get",
      return_value = "All"
    ),
    "The specified directory does not exist."
  )
})

test_that("submissions_summary stops with an error when an invalid return_value is provided", {
  expect_error(
    submissions_summary(
      path = test_dir,
      title = "get",
      return_value = "wrong"
    ),
    "Invalid return_value. Allowed values are 'Summary' or 'All'."
  )
})

## keep_file_name tests

test_that("submissions_summary stops with an error when an invalid keep_file_name is provided", {
  expect_error(
    submissions_summary(
      path = test_dir,
      title = "getting",
      vars = c("information-name"),
      keep_file_name = "Hey"
    ),
    "Invalid keep_file_name. Allowed values are NULL, 'All', 'Space', or 'Underscore'."
  )
})

test_that("submissions_summary stops with an error when keep_file_name is used with return_value 'All'", {
  expect_error(
    submissions_summary(
      path = test_dir,
      title = "getting",
      return_value = "All",
      keep_file_name = "All"
    ),
    "keep_file_name can only be used when return_value is 'Summary'."
  )
})

# Testing files with spaces required the use of setup/teardown scaffolding. Note
# the use Sys.getenv("TEST_DIR") as the value of the path argument.

test_that("submissions_summary returns the expected summary tibble with keep_file_name 'All'", {
  expected_output <- tibble::tibble(
    source = c(
      "introduction_answers -- Aadi.html",
      "introduction_answers -- astrxr.html",
      "ivy-introduction -- Ivy S.html"
    ),
    `information-name` = c("Aaditya Gupta", "Mithru Narayan Naidu", "Ivy Spratt"),
    answers = c(29L, 29L, 29L)
  )
  expected_output <- expected_output[order(expected_output$source), ]

  actual_output <- submissions_summary(
    path = Sys.getenv("TEST_DIR"),
    title = "introduction",
    vars = c("information-name"),
    keep_file_name = "All"
  )
  actual_output <- actual_output[order(actual_output$source), ]

  expect_equal(actual_output, expected_output)
})


test_that("submissions_summary returns the expected summary tibble with keep_file_name 'Space' and mixed file names", {
  expected_output <- tibble::tibble(
    source = c(
      "introduction_answers",
      "introduction_answers",
      "ivy-introduction"
    ),
    `information-name` = c("Aaditya Gupta", "Mithru Narayan Naidu", "Ivy Spratt"),
    answers = c(29L, 29L, 29L)
  )
  expected_output <- expected_output[order(expected_output$`information-name`), ]

  actual_output <- submissions_summary(
    path = Sys.getenv("TEST_DIR"),
    title = "introduction",
    vars = c("information-name"),
    keep_file_name = "Space"
  )
  actual_output <- actual_output[order(actual_output$`information-name`), ]

  expect_equal(actual_output, expected_output)
})

test_that("submissions_summary returns the expected summary tibble with keep_file_name 'Underscore' and mixed file names", {
  expected_output <- tibble::tibble(
    source = c(
      "introduction",
      "introduction",
      "ivy-introduction -- Ivy S.html"
    ),
    `information-name` = c("Aaditya Gupta", "Mithru Narayan Naidu", "Ivy Spratt"),
    answers = c(29L, 29L, 29L)
  )
  expected_output <- expected_output[order(expected_output$`information-name`), ]

  actual_output <- submissions_summary(
    path = Sys.getenv("TEST_DIR"),
    title = "introduction",
    vars = c("information-name"),
    keep_file_name = "Underscore"
  )
  actual_output <- actual_output[order(actual_output$`information-name`), ]

  expect_equal(actual_output, expected_output)
})

# Tests using new variable names.

test_that("submissions_summary returns the expected summary tibble", {
  expected_output <- tibble::tibble(
    name = c("David Kane"),
    answers = c(10)
  )
  
  actual_output <- submissions_summary(
    path = "fixtures/process_submissions_dir2/",
    title = "new-labels",
    vars = c("name")
  )
  
  expect_equal(actual_output, expected_output)
})

test_that("submissions_summary returns the expected summary tibble", {
  expected_output <- tibble::tibble(
    name = "David Kane",
    email = "dave.kane@gmail.com",
    minutes = "9",
    answers = c(10)
  )
  
  actual_output <- submissions_summary(
    path = "fixtures/process_submissions_dir2/",
    title = "new-labels",
    vars = c("name", "email", "minutes")
  )
  
  expect_equal(actual_output, expected_output)
})

# Tests for title as vector

test_that("submissions_summary works with multiple titles in vector - All mode", {
  actual_output <- submissions_summary(
    path = test_path("fixtures", "process_submissions_dir"),
    title = c("getting", "astrxr"),
    return_value = "All"
  )
  
  expect_s3_class(actual_output, "tbl_df")
  expect_equal(dim(actual_output), c(51L, 3L))
})

test_that("submissions_summary works with non-overlapping titles", {
  expected_output <- tibble::tibble(
    `information-name` = c("Areeb Atif", "Gitanjali Sheth", "Mithru Narayan Naidu"),
    `information-email` = c("areebatif2007@gmail.com", "gbhatia1@yahoo.com", "conflict454@gmail.com"),
    answers = c(11L, 11L, 29L)
  )
  
  actual_output <- submissions_summary(
    path = test_path("fixtures", "process_submissions_dir"),
    title = c("getting", "astrxr"),  # These should be non-overlapping
    vars = c("information-name", "information-email")
  )
  
  expect_equal(actual_output, expected_output)
})

test_that("submissions_summary works with empty title matches", {
  # Test with titles that don't match any files
  actual_output <- submissions_summary(
    path = test_path("fixtures", "process_submissions_dir"),
    title = c("email", "none"),
    vars = c("information-name", "information-email")
  )
  
  expect_s3_class(actual_output, "tbl_df")
  expect_equal(nrow(actual_output), 0L)
})

test_that("submissions_summary works with mixed existing and non-existing titles", {
  # Test with some titles that match and some that don't
  actual_output <- submissions_summary(
    path = test_path("fixtures", "process_submissions_dir"),
    title = c("none", "none", "getting", "none", "none"),
    vars = c("information-name", "information-email")
  )
  
  expect_s3_class(actual_output, "tbl_df")
  # Should only return results from the matching title
  expect_equal(nrow(actual_output), 2L)  # Only "getting" title matches
})

# Tests for emails parameter

test_that("submissions_summary works with default emails parameter", {
  result_default <- submissions_summary(
    path = test_path("fixtures", "process_submissions_dir"),
    title = "get",
    vars = c("information-email")
  )
  
  result_explicit <- submissions_summary(
    path = test_path("fixtures", "process_submissions_dir"),
    title = "get",
    vars = c("information-email"),
    emails = "*"
  )
  
  expect_equal(result_default, result_explicit)
})

# Regression tests: the `emails` argument used to be a silent no-op,
# overlapping `title` patterns used to duplicate rows, and old-format files
# storing answers in a `data` column used to fail with an opaque error.

extdata_dir <- system.file("extdata", "answers_html", package = "tutorial.helpers")

test_that("emails argument filters the summary", {
  all_rows <- submissions_summary(path = extdata_dir, title = "stop", vars = "email")
  expect_equal(nrow(all_rows), 3)

  one <- submissions_summary(path = extdata_dir, title = "stop", vars = "email",
                             emails = "bluebird.jack.xu@gmail.com")
  expect_equal(one$email, "bluebird.jack.xu@gmail.com")

  star <- submissions_summary(path = extdata_dir, title = "stop", vars = "email",
                              emails = "*")
  expect_equal(nrow(star), 3)
})

test_that("overlapping title patterns do not duplicate files", {
  res <- submissions_summary(path = extdata_dir, title = c("stop", "st"), vars = "email")
  expect_equal(nrow(res), 3)
  expect_equal(anyDuplicated(res$email), 0L)
})

test_that("old-format files with a 'data' column are summarized", {
  old_dir <- file.path(tempdir(), "old-format-summary")
  dir.create(old_dir, showWarnings = FALSE)
  on.exit(unlink(old_dir, recursive = TRUE))
  writeLines(
    c("<html><body><table>",
      "<tr><th>id</th><th>submission_type</th><th>data</th></tr>",
      "<tr><td>email</td><td>question</td><td>old.format@example.com</td></tr>",
      "<tr><td>intro-1</td><td>question</td><td>An answer</td></tr>",
      "</table></body></html>"),
    file.path(old_dir, "sample_answers.html"))

  res <- submissions_summary(path = old_dir, title = "sample", vars = "email")
  expect_equal(res$email, "old.format@example.com")
  expect_equal(res$answers, 2L)
})

#' Match Questions by Pattern
#'
#' This function takes a single HTML file or tibble and finds all questions/answers
#' that contain a specified pattern. It returns the question IDs (from the 'id' column)
#' for rows where the answer contains the pattern.
#'
#' @param x Either a file path to an HTML file or a tibble with 'id' and 'answer'/'data' columns
#' @param pattern A character string to search for in the answers
#' @param ignore.case Logical; should the search be case-insensitive? (default: TRUE)
#'
#' @return A character vector of question IDs where the answer contains the pattern
#'
#' @examples
#' # Search in a tibble
#' path <- system.file("extdata", "answers_html", package = "tutorial.helpers")
#'
#' tibble <- gather_submissions(path, title = "stop")[[1]]
#'
#' result <- match_questions(tibble, "http")
#'
#' \dontrun{
#' # Search in an HTML file
#' question_ids <- match_questions("path/to/submission.html", "temperance")
#' # Returns: c("temperance-16", "temperance-19")
#' }
#' @export
match_questions <- function(x, pattern, ignore.case = TRUE) {
  
  # Input validation
  if (missing(x) || is.null(x)) {
    stop("'x' must be provided (either a file path or tibble).")
  }
  
  if (missing(pattern) || is.null(pattern)) {
    stop("'pattern' must be provided.")
  }
  
  if (!is.logical(ignore.case) || length(ignore.case) != 1) {
    stop("'ignore.case' must be a single logical value (TRUE or FALSE).")
  }
  
  # Determine if x is a file path or tibble
  if (is.character(x) && length(x) == 1) {
    # x is a file path
    tibble_data <- read_html_file(x)
  } else if (is.data.frame(x)) {
    # x is already a tibble/data.frame
    tibble_data <- x
  } else {
    stop("'x' must be either a file path (character string) or a tibble/data.frame.")
  }
  
  # Validate tibble structure
  if (is.null(tibble_data)) {
    return(character(0))
  }
  
  if (!"id" %in% colnames(tibble_data)) {
    stop("The data must have an 'id' column.")
  }
  
  # Find the answer column (either 'answer' or 'data')
  answer_col <- NULL
  if ("answer" %in% colnames(tibble_data)) {
    answer_col <- "answer"
  } else if ("data" %in% colnames(tibble_data)) {
    answer_col <- "data"
  } else {
    stop("The data must have either an 'answer' or 'data' column.")
  }
  
  # Search for pattern in answers
  matching_rows <- grepl(pattern, tibble_data[[answer_col]], ignore.case = ignore.case)
  
  # Return the IDs for matching rows
  matching_ids <- tibble_data$id[matching_rows]
  
  # Convert to character vector and remove any NAs
  result <- as.character(matching_ids[!is.na(matching_ids)])
  
  return(result)
}

# Helper function to read HTML file
read_html_file <- function(file_path) {
  # Check if file exists
  if (!file.exists(file_path)) {
    stop("File does not exist: ", file_path)
  }
  
  # Check if rvest package is available
  if (!requireNamespace("rvest", quietly = TRUE)) {
    stop("Package 'rvest' is required. Please install it with: install.packages('rvest')")
  }
  
  # Check if tibble package is available
  if (!requireNamespace("tibble", quietly = TRUE)) {
    stop("Package 'tibble' is required. Please install it with: install.packages('tibble')")
  }
  
  # Read HTML content
  html_content <- tryCatch({
    rvest::read_html(file_path)
  }, error = function(e) {
    stop("Failed to read HTML file: ", e$message)
  })
  
  # Extract table data
  table_data <- tryCatch({
    tables <- rvest::html_table(html_content)
    if (length(tables) == 0) {
      return(NULL)
    }
    tables[[1]]
  }, error = function(e) {
    stop("Failed to extract table from HTML: ", e$message)
  })
  
  if (is.null(table_data) || nrow(table_data) == 0) {
    return(NULL)
  }
  
  # Convert to tibble
  tibble_data <- tryCatch({
    tibble::as_tibble(table_data)
  }, error = function(e) {
    stop("Failed to convert to tibble: ", e$message)
  })
  
  return(tibble_data)
}

test_dir <- test_path("fixtures", "answers_html")

test_that("Real example - search for http pattern", {
  # Get a tibble from real test data
  tibble_list <- gather_submissions(test_dir, title = "stop")
  tibble_data <- tibble_list[[1]]
  
  result <- match_questions(tibble_data, "http")
  
  # Based on your real result
  expected_ids <- c("justice-12", "temperance-5", "temperance-16", "temperance-17")
  
  expect_equal(result, expected_ids)
  expect_length(result, 4)
  expect_true(all(c("justice-12", "temperance-5", "temperance-16", "temperance-17") %in% result))
})

# Test 2: Search with different pattern
test_that("Search with different pattern", {
  tibble_list <- gather_submissions(test_dir, title = "stop")
  tibble_data <- tibble_list[[1]]

  # Test with a different pattern
  result <- match_questions(tibble_data, "temperance")

  expect_true(is.character(result))
  expect_true(length(result) >= 0)  # Could be 0 or more matches
})

# Test 3: Input validation.
test_that("match_questions validates its arguments", {
  df <- tibble::tibble(id = "q-1", answer = "hello")

  expect_error(match_questions(), "'x' must be provided")
  expect_error(match_questions(NULL), "'x' must be provided")
  expect_error(match_questions(df), "'pattern' must be provided")
  expect_error(match_questions(df, "hello", ignore.case = "yes"),
               "'ignore.case' must be a single logical value")
  expect_error(match_questions(42, "hello"),
               "'x' must be either a file path")
})

# Test 4: A tibble whose answers live in a 'data' column (not 'answer').
test_that("match_questions handles a 'data' answer column", {
  df <- tibble::tibble(
    id = c("a-1", "a-2", "a-3"),
    data = c("see https://x.com", "no link here", "also https://y.org"))

  expect_equal(match_questions(df, "https"), c("a-1", "a-3"))
})

# Test 5: Missing required columns error.
test_that("match_questions errors on missing required columns", {
  no_id <- tibble::tibble(answer = "hello")
  expect_error(match_questions(no_id, "hello"), "must have an 'id' column")

  no_answer <- tibble::tibble(id = "q-1", other = "hello")
  expect_error(match_questions(no_answer, "hello"),
               "must have either an 'answer' or 'data' column")
})

# Test 6: ignore.case toggles case sensitivity; non-matches return empty.
test_that("match_questions respects ignore.case", {
  df <- tibble::tibble(id = c("q-1", "q-2"),
                       answer = c("Hello World", "goodbye"))

  expect_equal(match_questions(df, "hello", ignore.case = TRUE), "q-1")
  expect_equal(match_questions(df, "hello", ignore.case = FALSE), character(0))
  expect_equal(match_questions(df, "zzz"), character(0))
})

#' Knit a set of tutorials
#'
#' @description We define "testing" a tutorial as (successfully) running
#'   `render()` on it. This function renders all the tutorials provided in
#'   `tutorial_paths`. There is no check to see if the rendered file looks OK.
#'   If a tutorial fails to render, then an error will be generated which will
#'   propagate to the caller.
#'
#' @param tutorial_paths Character vector of the paths to the tutorials to be
#'   knitted.
#'
#' @returns No return value, called for side effects.
#'
#' @examples
#' \dontrun{
#'   knit_tutorials(tutorial_paths = return_tutorial_paths("tutorial.helpers"))
#' }
#'
#' @export

knit_tutorials <- function(tutorial_paths){
  
  stopifnot(all(file.exists(tutorial_paths)))

  # Might we do more here? For example, what we really want to confirm is that,
  # when a student presses the "Start Tutorial" button, things will work. I am not
  # sure if render() is the same thing. But, the good news is that this test seems
  # much more robust than that. In other words, it catches things that do not
  # cause (immediate) failures with Start Tutorial.
  
  # Note that the Debian setup on CRAN does not allow for writing files to any
  # location other than the temporary directory, which is why we must specify
  # tempdir() in the two dir arguments.
  
  # Would be nice to have more flexibility with knit_tutorials(). The problem
  # arises when knitting a collection of tutorial paths takes too long,
  # especially on CRAN, where all tests should only take 10 minutes total. Might
  # be nice if tutorial paths:
    
  # 1) Had an option to report the time (or start/end time) which each knitting
  # used. We want to identify which tutorials take too long to knit. Right now,
  # there is no easy way to do so.
  
  # 2) Provide a `skip` argument to knit_tutorials which allows it to skip any
  # tutorial path which includes a specific string. This will generally be used
  # like skip = c("06-data-tidying", "08-data-import"). Note that these are not
  # the names of tutorial files (which are mostly "tutorial.Rmd") nor the full
  # path to those tutorials (which we don't know until we run
  # `return_tutorial_paths()`). Instead, they are strings from within the path.
  # (Maybe require that they be full parts of the path? Or maybe any match is
  # fine?)

  for(i in tutorial_paths){
    message("Rendering: ", i)
    tryCatch({
      rmarkdown::render(input = i, 
                        output_dir = tempdir(),
                        intermediates_dir = tempdir())
      message("Successfully rendered: ", i)
    }, error = function(e) {
      stop("Failed to render ", i, ": ", e$message, call. = FALSE)
    })
  }
  
  message("Successfully rendered ", length(tutorial_paths), " tutorial(s)")
  invisible(NULL)
}

# Test suite for knit_tutorials function

test_that("knit_tutorials exists and has correct signature", {
  expect_true(exists("knit_tutorials"))
  expect_true(is.function(knit_tutorials))
  
  # Check function arguments
  args <- formals(knit_tutorials)
  expect_true("tutorial_paths" %in% names(args))
  expect_equal(length(args), 1)
})

test_that("knit_tutorials validates file existence", {
  # Test with non-existent file
  expect_error(
    knit_tutorials("definitely_does_not_exist.Rmd"),
    "all\\(file\\.exists\\(tutorial_paths\\)\\) is not TRUE"
  )
  
  # Test with mix of existing and non-existing files
  temp_file <- tempfile(fileext = ".Rmd")
  writeLines("# Test", temp_file)
  on.exit(unlink(temp_file))
  
  expect_error(
    knit_tutorials(c(temp_file, "nonexistent.Rmd")),
    "all\\(file\\.exists\\(tutorial_paths\\)\\) is not TRUE"
  )
})

test_that("knit_tutorials handles empty input correctly", {
  # Empty character vector should work
  expect_message(
    result <- knit_tutorials(character(0)),
    "Successfully rendered 0 tutorial\\(s\\)"
  )
  expect_null(result)
})

test_that("knit_tutorials renders simple R Markdown file", {
  # Create a minimal but valid R Markdown file
  temp_tutorial <- tempfile(fileext = ".Rmd")
  on.exit(unlink(temp_tutorial))
  
  simple_content <- c(
    "---",
    "title: 'Test Tutorial'",
    "output: html_document",
    "---",
    "",
    "# Simple Test",
    "",
    "This is a basic test document.",
    "",
    "```{r}",
    "2 + 2",
    "```"
  )
  
  writeLines(simple_content, temp_tutorial)
  
  # Should render successfully
  expect_message(
    expect_message(
      result <- knit_tutorials(temp_tutorial),
      "Rendering:"
    ),
    "Successfully rendered:"
  )
  
  expect_null(result)
})

test_that("knit_tutorials renders multiple files", {
  # Create two simple R Markdown files
  temp_tutorial1 <- tempfile(fileext = ".Rmd")
  temp_tutorial2 <- tempfile(fileext = ".Rmd")
  on.exit(unlink(c(temp_tutorial1, temp_tutorial2)))
  
  content1 <- c(
    "---",
    "title: 'Tutorial 1'",
    "output: html_document",
    "---",
    "",
    "# Tutorial One",
    "First tutorial content."
  )
  
  content2 <- c(
    "---",
    "title: 'Tutorial 2'",
    "output: html_document",
    "---",
    "",
    "# Tutorial Two",
    "Second tutorial content."
  )
  
  writeLines(content1, temp_tutorial1)
  writeLines(content2, temp_tutorial2)
  
  # Should render both successfully
  expect_message(
    result <- knit_tutorials(c(temp_tutorial1, temp_tutorial2)),
    "Successfully rendered 2 tutorial\\(s\\)"
  )
  
  expect_null(result)
})

test_that("knit_tutorials fails gracefully on invalid R Markdown", {
  # Create an R Markdown file with invalid R code
  temp_tutorial <- tempfile(fileext = ".Rmd")
  on.exit(unlink(temp_tutorial))
  
  invalid_content <- c(
    "---",
    "title: 'Invalid Tutorial'",
    "output: html_document",
    "---",
    "",
    "# Invalid Code Test",
    "",
    "```{r}",
    "this_function_does_not_exist()",
    "```"
  )
  
  writeLines(invalid_content, temp_tutorial)
  
  # Should fail with clear error message
  expect_error(
    knit_tutorials(temp_tutorial),
    "Failed to render.*could not find function"
  )
})

test_that("knit_tutorials works with learnr tutorial format", {
  skip_if_not_installed("learnr")
  
  # Create a basic learnr tutorial
  temp_tutorial <- tempfile(fileext = ".Rmd")
  on.exit(unlink(temp_tutorial))
  
  learnr_content <- c(
    "---",
    "title: 'Test Tutorial'",
    "output: learnr::tutorial",
    "runtime: shiny_prerendered",
    "---",
    "",
    "```{r setup, include=FALSE}",
    "library(learnr)",
    "```",
    "",
    "## Introduction",
    "",
    "This is a test tutorial.",
    "",
    "```{r exercise-1, exercise=TRUE}",
    "1 + 1",
    "```"
  )
  
  writeLines(learnr_content, temp_tutorial)
  
  # Should render successfully
  expect_message(
    result <- knit_tutorials(temp_tutorial),
    "Successfully rendered"
  )
  
  expect_null(result)
})

test_that("knit_tutorials respects output directory", {
  # Create a simple tutorial
  temp_tutorial <- tempfile(fileext = ".Rmd")
  on.exit(unlink(temp_tutorial))
  
  simple_content <- c(
    "---",
    "title: 'Directory Test'",
    "output: html_document",
    "---",
    "",
    "# Test",
    "Testing output directory."
  )
  
  writeLines(simple_content, temp_tutorial)
  
  # Render and check that output goes to tempdir
  initial_temp_files <- list.files(tempdir(), pattern = "\\.html$")
  
  knit_tutorials(temp_tutorial)
  
  final_temp_files <- list.files(tempdir(), pattern = "\\.html$")
  
  # Should have created a new HTML file in tempdir
  expect_true(length(final_temp_files) > length(initial_temp_files))
})

test_that("knit_tutorials input validation", {
  # Test with NULL input
  expect_error(knit_tutorials(NULL))
  
  # Test with non-character input
  expect_error(knit_tutorials(123))
  expect_error(knit_tutorials(list("file.Rmd")))
  
  # Test with NA values
  expect_error(knit_tutorials(c("valid.Rmd", NA)))
})

# For now, we will do all our tutorial testing in this one script. Skipped on
# CRAN because rendering the full learnr tutorial is slow; test-knit_tutorials.R
# covers the same code path.

test_that("all installed tutorials knit and have the default components", {
  skip_on_cran()

  tutorial_paths <- tutorial.helpers::return_tutorial_paths(package = "tutorial.helpers")

  # First, we make sure that all the tutorials can be knitted.

  expect_no_error(tutorial.helpers::knit_tutorials(tutorial_paths))

  # Second, ensure that all the tutorials have the default components.

  expect_no_error(tutorial.helpers::check_tutorial_defaults(tutorial_paths))
})

#' Return all the paths to the tutorials in a package
#'
#' @param package Character string of the package name to be tested.
#' 
#' @description Takes a package name and returns a character vector of all the
#'   paths to tutorials in the installed package. This function looks for all
#'   R Markdown files (.Rmd) in the inst/tutorials/ subdirectories of the
#'   specified package. It uses learnr::available_tutorials() to identify
#'   tutorial directories, with a fallback to directory scanning if that fails.
#'   
#' @details The function first checks if the package is installed and has a
#'   tutorials directory. It then attempts to use learnr::available_tutorials()
#'   to get the official list of tutorial directories. If that fails (e.g., if
#'   the package doesn't properly register its tutorials with learnr), it falls
#'   back to scanning all subdirectories under inst/tutorials/. Finally, it
#'   collects all .Rmd files from these directories.
#'   
#'   Returns an empty character vector if the package has no tutorials or
#'   doesn't exist, rather than throwing an error.
#'   
#' @returns Character vector of the full paths to all installed tutorials in
#'   `package`. Returns character(0) if no tutorials are found or if the
#'   package doesn't exist.
#'   
#' @examples
#' \dontrun{
#'   # Get all learnr tutorial paths
#'   return_tutorial_paths('learnr')
#'   
#'   # Get tutorial paths from your own package
#'   return_tutorial_paths('tutorial.helpers')
#'   
#'   # Returns empty vector for packages without tutorials
#'   return_tutorial_paths('base')
#' }
#'   
#' @export

return_tutorial_paths <- function(package) {
  
  # Input validation
  if (!is.character(package) || length(package) != 1) {
    stop("'package' must be a single character string", call. = FALSE)
  }
  
  # Check if package exists
  if (!requireNamespace(package, quietly = TRUE)) {
    warning("Package '", package, "' is not installed", call. = FALSE)
    return(character(0))
  }
  
  # Get base tutorials directory
  tutorials_base <- system.file("tutorials", package = package)
  if (tutorials_base == "") {
    # Package exists but has no tutorials directory
    return(character(0))
  }
  
  # Get tutorial subdirectories - try learnr first, fallback to manual scan
  tutorial_dirs <- tryCatch({
    # Attempt to use learnr's official tutorial registry
    available <- learnr::available_tutorials(package)
    if (is.data.frame(available) && "name" %in% names(available)) {
      available$name
    } else {
      character(0)
    }
  }, error = function(e) {
    # learnr method failed, fall back to directory scanning
    character(0)
  })
  
  # If learnr method failed or returned nothing, scan directories manually
  if (length(tutorial_dirs) == 0) {
    all_dirs <- list.dirs(tutorials_base, recursive = FALSE, full.names = FALSE)
    tutorial_dirs <- all_dirs[all_dirs != ""]
  }
  
  if (length(tutorial_dirs) == 0) {
    return(character(0))
  }
  
  # Build full paths to tutorial directories
  full_dirs <- file.path(tutorials_base, tutorial_dirs)
  
  # Find all .Rmd files in these directories
  tutorial_files <- character(0)
  
  for (dir in full_dirs) {
    if (dir.exists(dir)) {
      rmd_files <- list.files(dir, pattern = "\\.Rmd$", full.names = TRUE, 
                             recursive = FALSE, ignore.case = TRUE)
      tutorial_files <- c(tutorial_files, rmd_files)
    }
  }
  
  # Return sorted paths for consistent output
  sort(tutorial_files)
}

# Test suite for return_tutorial_paths function

test_that("return_tutorial_paths function exists and has correct signature", {
  expect_true(exists("return_tutorial_paths"))
  expect_true(is.function(return_tutorial_paths))
  
  # Check function arguments
  args <- formals(return_tutorial_paths)
  expect_true("package" %in% names(args))
  expect_equal(length(args), 1)
})

test_that("return_tutorial_paths validates input correctly", {
  # Test with invalid input types
  expect_error(return_tutorial_paths(123), "'package' must be a single character string")
  expect_error(return_tutorial_paths(NULL), "'package' must be a single character string")
  expect_error(return_tutorial_paths(c("pkg1", "pkg2")), "'package' must be a single character string")
  expect_error(return_tutorial_paths(character(0)), "'package' must be a single character string")
})

test_that("return_tutorial_paths handles non-existent packages gracefully", {
  # Should return empty vector with warning for non-existent package
  expect_warning(
    result <- return_tutorial_paths("this_package_definitely_does_not_exist_12345"),
    "Package .* is not installed"
  )
  expect_equal(result, character(0))
  expect_type(result, "character")
})

test_that("return_tutorial_paths handles packages without tutorials", {
  # Base R packages like 'stats' don't have tutorials
  # Should return empty vector, not error
  result <- return_tutorial_paths("stats")
  expect_equal(result, character(0))
  expect_type(result, "character")
})

test_that("return_tutorial_paths works with learnr package", {
  skip_if_not_installed("learnr")
  
  result <- return_tutorial_paths("learnr")
  
  # Should return character vector (may be empty if no tutorials)
  expect_type(result, "character")
  
  # If tutorials are found, they should be valid file paths
  if (length(result) > 0) {
    expect_true(all(file.exists(result)))
    expect_true(all(grepl("\\.Rmd$", result, ignore.case = TRUE)))
    # Should be sorted
    expect_equal(result, sort(result))
  }
})

test_that("return_tutorial_paths works with tutorial.helpers", {
  skip_if_not_installed("tutorial.helpers")
  
  # Check if tutorials directory exists (only available after installation)
  tutorials_dir <- system.file("tutorials", package = "tutorial.helpers")
  skip_if(tutorials_dir == "", "tutorial.helpers not fully installed")
  
  result <- return_tutorial_paths("tutorial.helpers")
  
  expect_type(result, "character")
  
  # If tutorials are found, validate them
  if (length(result) > 0) {
    expect_true(all(file.exists(result)))
    expect_true(all(grepl("\\.Rmd$", result, ignore.case = TRUE)))
    # Should be sorted
    expect_equal(result, sort(result))
    
    # Each path should contain "tutorial.helpers" and "tutorials"
    expect_true(all(grepl("tutorial\\.helpers", result)))
    expect_true(all(grepl("tutorials", result)))
  }
})

test_that("return_tutorial_paths returns consistent output format", {
  skip_if_not_installed("learnr")
  
  result <- return_tutorial_paths("learnr")
  
  # Should always return character vector
  expect_type(result, "character")
  
  # Should be sorted
  expect_equal(result, sort(result))
  
  # No duplicates
  expect_equal(length(result), length(unique(result)))
  
  # If any results, they should be valid paths
  if (length(result) > 0) {
    expect_true(all(file.exists(result)))
    expect_true(all(nzchar(result)))
  }
})

test_that("return_tutorial_paths handles case sensitivity", {
  skip_if_not_installed("learnr")
  
  result <- return_tutorial_paths("learnr")
  
  # Should find .Rmd files regardless of case (.Rmd, .rmd, .RMD)
  if (length(result) > 0) {
    expect_true(all(grepl("\\.[Rr][Mm][Dd]$", result)))
  }
})

test_that("return_tutorial_paths fallback mechanism", {
  # Test that function can handle cases where learnr::available_tutorials fails
  # This is hard to test directly, but we can test with a package that might
  # not register properly with learnr but still has .Rmd files
  
  # Test with a base package - should return empty gracefully
  result <- return_tutorial_paths("utils")
  expect_type(result, "character")
  expect_equal(length(result), 0)
})

test_that("return_tutorial_paths file path construction", {
  skip_if_not_installed("learnr")
  
  result <- return_tutorial_paths("learnr")
  
  if (length(result) > 0) {
    # All paths should be non-empty strings
    expect_true(all(nzchar(result)))
    
    # All paths should end with .Rmd (case insensitive)
    expect_true(all(grepl("\\.Rmd$", result, ignore.case = TRUE)))
    
    # All paths should contain "tutorials" directory
    expect_true(all(grepl("tutorials", result)))
    
    # All files should actually exist
    expect_true(all(file.exists(result)))
  }
})

#' Extract Answers from Submissions with Filtering
#'
#' This function gathers submissions matching a title pattern, filters them by membership,
#' and extracts specified variables, returning a tibble with one row per valid submission
#' and one column per variable.
#'
#' @param path The path to the local directory or Google Drive folder URL containing submissions
#' @param title A character vector of patterns to match against file names (passed to gather_submissions)
#' @param key_var A character string specifying the key variable to check for membership (e.g., "email").
#'        If NULL (default), no membership filtering is applied.
#' @param membership A character vector of allowed values for the key variable, or "*" to include all submissions.
#'        If NULL (default), no membership filtering is applied. Ignored if key_var is NULL.
#' @param vars A character vector of variables/questions to extract, or "*" to extract all available variables
#' @param keep_file_name How to handle file names: NULL (don't include), "All" (full name),
#'        "Space" (up to first space), "Underscore" (up to first underscore)
#' @param verbose A logical value (TRUE or FALSE) specifying verbosity level.
#'        If TRUE, reports files that are removed during processing.
#'
#' @return A tibble with one row per valid submission, columns for each variable,
#'         and optionally a "source" column
#'
#' @examples
#' # Extract specific variables from submissions matching title pattern
#' path <- system.file("extdata", "answers_html", package = "tutorial.helpers")
#'
#' result <- submissions_answers(
#'   path = path,
#'   title = c("stop"), 
#'   key_var = "email",
#'   membership = c("bluebird.jack.xu@gmail.com", "abdul.hannan20008@gmail.com"),
#'   vars = c("name", "email", "introduction-1"),
#'   verbose = TRUE
#' )
#' 
#' # Extract all variables from submissions
#' result_all <- submissions_answers(
#'   path = path,
#'   title = c("stop"), 
#'   key_var = "email",
#'   membership = c("bluebird.jack.xu@gmail.com", "abdul.hannan20008@gmail.com"),
#'   vars = "*",
#'   verbose = TRUE
#' )
#'
#' \dontrun{
#' drive_url <- "https://drive.google.com/drive/folders/10do12t0fZsfrIrKePxwjpH8IqBNVO86N"
#' x <- submissions_answers(
#'   path = drive_url, 
#'   title = c("introduction"),
#'   key_var = "email",
#'   membership = c("fmehmud325@gmail.com"),
#'   vars = c("email", "name", "what-you-will-learn-15")
#' )
#' }
#' @export

submissions_answers <- function(path, title, key_var = NULL, membership = NULL, vars, 
                               keep_file_name = NULL, verbose = FALSE) {
  
  # Input validation
  if (missing(path) || is.null(path)) {
    stop("'path' must be provided.")
  }
  
  if (missing(title) || is.null(title)) {
    stop("'title' must be provided.")
  }
  
  if (missing(vars) || is.null(vars)) {
    stop("'vars' must be provided.")
  }
  
  # Validate key_var and membership consistency
  if (!is.null(key_var) && is.null(membership)) {
    stop("'membership' must be provided when 'key_var' is specified.")
  }
  
  if (is.null(key_var) && !is.null(membership)) {
    stop("'key_var' must be provided when 'membership' is specified.")
  }
  
  # Validate verbose parameter
  if (!is.logical(verbose) || length(verbose) != 1) {
    stop("'verbose' must be a single logical value (TRUE or FALSE).")
  }
  
  # Validate keep_file_name parameter
  if (!is.null(keep_file_name) && !(keep_file_name %in% c("All", "Space", "Underscore"))) {
    stop("Invalid keep_file_name. Allowed values are NULL, 'All', 'Space', or 'Underscore'.")
  }
  
  # Step 1: Gather submissions matching the title pattern
  tibble_list <- gather_submissions(path = path, title = title, verbose = FALSE)
  
  if (verbose && length(tibble_list) > 0) {
    message("Found ", length(tibble_list), " submission(s) matching title pattern '", paste(title, collapse = "|"), "'")
  }
  
  if (length(tibble_list) == 0) {
    if (verbose) {
      message("No submissions found matching the title pattern.")
    }
    return(tibble::tibble())
  }
  
  # Step 2: Filter by membership using check_membership, or skip if no filtering requested
  if (is.null(key_var) || is.null(membership)) {
    # No membership filtering - include all submissions
    valid_tibbles <- tibble_list
    if (verbose) {
      message("No membership filtering applied - including all ", length(tibble_list), " submission(s)")
    }
  } else if (length(membership) == 1 && membership == "*") {
    # Include all submissions - no membership filtering
    valid_tibbles <- tibble_list
    if (verbose) {
      message("Membership set to '*' - including all ", length(tibble_list), " submission(s)")
    }
  } else {
    # Apply membership filtering
    valid_tibbles <- check_membership(tibble_list, key_var, membership, verbose = FALSE)
    
    if (verbose) {
      message("After membership filtering: ", length(valid_tibbles), " submission(s) retained")
    }
  }
  
  if (length(valid_tibbles) == 0) {
    if (verbose) {
      message("No submissions passed the membership check.")
    }
    return(tibble::tibble())
  }
  
  # Step 2.5: Determine which variables to extract
  if (length(vars) == 1 && vars == "*") {
    # Extract all available variables from all submissions
    all_vars <- character(0)
    for (tibble_data in valid_tibbles) {
      if ("id" %in% colnames(tibble_data)) {
        all_vars <- unique(c(all_vars, tibble_data$id))
      }
    }
    vars_to_extract <- all_vars
    
    if (verbose) {
      message("Extracting all available variables: ", paste(vars_to_extract, collapse = ", "))
    }
  } else {
    vars_to_extract <- vars
  }
  
  # Step 3: Extract answers and create final tibble
  result_rows <- lapply(names(valid_tibbles), function(file_name) {
    tibble_data <- valid_tibbles[[file_name]]
    
    # Create base row with source information if requested
    if (!is.null(keep_file_name)) {
      if (keep_file_name == "All") {
        source_name <- file_name
      } else if (keep_file_name == "Space") {
        source_name <- sub("\\s.*", "", file_name)
      } else if (keep_file_name == "Underscore") {
        source_name <- sub("_.*", "", file_name)
      }
    }
    
    # Initialize result row
    row_data <- list()
    
    # Add source column if requested
    if (!is.null(keep_file_name)) {
      row_data[["source"]] <- source_name
    }
    
    # Extract each variable
    for (var in vars_to_extract) {
      if ("id" %in% colnames(tibble_data) && var %in% tibble_data$id) {
        # Get the answer for this variable - check for both 'answer' and 'data' columns
        if ("answer" %in% colnames(tibble_data)) {
          answer_value <- tibble_data$answer[tibble_data$id == var]
        } else if ("data" %in% colnames(tibble_data)) {
          answer_value <- tibble_data$data[tibble_data$id == var]
        } else {
          answer_value <- character(0)
        }
        # Take first answer if multiple exist
        row_data[[var]] <- if (length(answer_value) > 0) answer_value[1] else NA
      } else {
        # Variable not found, set to NA
        row_data[[var]] <- NA
      }
    }
    
    # Convert to tibble
    tibble::as_tibble(row_data)
  })

  result <- do.call(rbind, result_rows)

  return(tibble::as_tibble(result))
}

test_dir <- test_path("fixtures", "answers_html")

# Test 1: Real data - existing variables work correctly
test_that("Existing variables return correct values", {
  result <- submissions_answers(
    test_dir, 
    title = "stop", 
    key_var = "email", 
    membership = c("bluebird.jack.xu@gmail.com", "abdul.hannan20008@gmail.com"), 
    vars = c("name", "email")
  )
  
  expect_equal(nrow(result), 2)
  expect_equal(ncol(result), 2)
  expect_false(any(is.na(result)))
  expect_true("name" %in% colnames(result))
  expect_true("email" %in% colnames(result))
  expect_false("answers" %in% colnames(result))
  
  # Test specific values from your real data
  expect_true("Abdul Hannan" %in% result$name)
  expect_true("Jack Xu" %in% result$name)
  expect_true("abdul.hannan20008@gmail.com" %in% result$email)
  expect_true("bluebird.jack.xu@gmail.com" %in% result$email)
})

# Test 2: Missing variables return NA
test_that("Missing variables return NA", {
  result <- submissions_answers(
    test_dir, 
    title = "stop", 
    key_var = "email", 
    membership = c("bluebird.jack.xu@gmail.com", "abdul.hannan20008@gmail.com"), 
    vars = c("name", "email2")  # email2 doesn't exist
  )
  
  expect_equal(nrow(result), 2)
  expect_equal(ncol(result), 2)
  expect_false(any(is.na(result$name)))
  expect_true(all(is.na(result$email2)))
  expect_false("answers" %in% colnames(result))
})

# Test 3: Mixed existing and missing variables
test_that("Mixed variables work correctly", {
  result <- submissions_answers(
    test_dir, 
    title = "stop", 
    key_var = "email", 
    membership = c("bluebird.jack.xu@gmail.com", "abdul.hannan20008@gmail.com"), 
    vars = c("name", "email2", "email")  # email2 missing, others exist
  )
  
  expect_equal(nrow(result), 2)
  expect_equal(ncol(result), 3)
  expect_false(any(is.na(result$name)))
  expect_true(all(is.na(result$email2)))
  expect_false(any(is.na(result$email)))
  expect_false("answers" %in% colnames(result))
  
  # Test specific values are preserved
  expect_true("Abdul Hannan" %in% result$name)
  expect_true("Jack Xu" %in% result$name)
  expect_true("abdul.hannan20008@gmail.com" %in% result$email)
  expect_true("bluebird.jack.xu@gmail.com" %in% result$email)
})

#' Determine the code chunk name of a new exercise in a tutorial.
#'
#' @param file_path Character string of the file path to the tutorial
#'
#' @return The section id of the exercise based on its section
#' @export

determine_code_chunk_name <- function(file_path = NULL){

  # Unable to test function using the getActiveDocumentContext() function, use
  # the file_path as an alternative. Issue with using file_path is the cursor placement. It will
  # always run the function as if the next exercise to create is at the bottom of the page.
  # Also, note to explore options of including special characters in code chunk names.
  # Currently, most special characters like "{#(.*)}" are excluded and check for the neccesity of some of the
  # trimws functions. Currently, cut_content represents a all the lines in a file in a list which is then reversed.
  # Also note to perhaps explore expanding the 30 character limit for exercises headers. Explore splitting this function up
  # by creating a function that takes a section header and returns a code chunk name (both being character strings).
  # Ultimately this would greatly improve testing but may not be completely necessary.

  if (is.null(file_path))
  {

    ctx <- rstudioapi::getActiveDocumentContext()
    row <- ctx$selection[[1]]$range$end[["row"]]

    cut_content <- rev(ctx$contents[1:row])

  }
  else
  {
    cut_content <- rev(readLines(file_path))

  }

  # Get everything until the current row as a list of lines
  # and reverse that list

  section_id <- "NONE_SET"

  for (l in cut_content){
    # Find the latest section

    if (grepl("^## ", l)){

      # Remove the pattern "{#...}" and non-alphanumeric characters except
      # spaces and slashes
      
      cleaned_l <- gsub("\\{#(.*)\\}", "", l)
      cleaned_l <- gsub("[^a-zA-Z0-9 /]", "", cleaned_l)
      
      # Convert to lowercase, replace spaces and slashes with hyphens, trim to
      # 30 characters, and trim whitespace
      
      section_id <- trimws(cleaned_l)
      section_id <- substr(gsub("[ /]", "-", tolower(section_id)), 1, 30)
      section_id <- gsub("-+$", "", section_id)
      section_id <- gsub("^-+", "", section_id)

      # After finding a section, stop looping immediately

      return(section_id)
    }
  }

  # No section header found above this point in the file.

  return(section_id)
}

test_that("Reading/Writing Excel section", {
  rmd_path <- testthat::test_path(
    "fixtures", "tutorial_examples", "code-chunk-name-test-tutorial-1.Rmd"
  )
  expect_equal(
    tutorial.helpers::determine_code_chunk_name(rmd_path),
    "reading-writing-excel"
  )
})

test_that("Testing name cut off", {
  rmd_path <- testthat::test_path(
    "fixtures", "tutorial_examples", "code-chunk-name-test-tutorial-2.Rmd"
  )
  expect_equal(
    tutorial.helpers::determine_code_chunk_name(rmd_path),
    "really-long-name-to-cut-off-an"
  )
})

# Regression test: a file with no section headers used to return NULL.

test_that("a file with no sections returns the default", {
  f <- tempfile(fileext = ".Rmd")
  on.exit(unlink(f))
  writeLines("No sections here.", f)
  expect_equal(determine_code_chunk_name(f), "NONE_SET")
})

tutorial_paths <- tutorial.helpers::return_tutorial_paths("learnr")

# One problem is that the tutorials in the learnr package have various problems.
# So, this "test" would produce warnings/errors which are not our fault! We
# could restrict our tests to just tutorials which we know are well-formatted.
# Or we could try to clean up learnr itself. I have submitted one PR but, for
# now, we just comment out the last set of checks which rely on parse_rmd().

for(i in tutorial_paths){

  # With regard to this first comment, maybe we should just fail from the start
  # if parse_rmd fail.

  # Note that parsermd, despite being wonderful, can not check for certain
  # errors because it can't parse a file with those errors in the first place.
  # So, we need to check, by hand, for code chunks which do not begin with an
  # "r" followed by a space. Gets the first line (the chunk label) of each code
  # chunk in the file.

  # The ex-data-mutate.Rmd has a missing last EOL. I submitted a PR to learnr.
  # Who knows how long that will take to fix. In the meantime, I set warn =
  # FALSE.

  lines <- readLines(i, warn = FALSE)
  labels <- lines[grepl("^```\\{.*\\}", lines)]

  # Gets the labels that don't have r (or other allowed languages) at the
  # beginning. By the way, the =html tag is used to render a gist in some
  # tutorials.

  no_r_labels <- labels[!grepl("```\\{(r|=html|bash|python|sql)[ ,\\}].*", labels)]

  if(length(no_r_labels) > 0){
    stop("From test-code-chunks.R. Missing `r` at beginning of code chunk labels: ",
         toString(no_r_labels), " Found in file ", i, "\n")
  }

  # Gets the labels that don't have a space, a comma, or a }. This is because
  # labels like {r chunk-name}, {r}, and {r, include = FALSE} are all valid and
  # used throughout the tutorial. parsermd also doesn't detect chunks if there's
  # no space so this is easier.

  no_space_labels <- labels[grepl("```\\{r[^\ },]", labels)]
  if(length(no_space_labels) > 0){
    stop("From test-code-chunks.R. Missing space or comma at beginning of code chunk labels: ",
         toString(no_space_labels), " Found in file ", i, "\n")
  }

  # Gets the labels that don't have a } at the end. This is so that everything
  # parses correctly. Problem is that some tutorials still make use of the `r
  # ''`, so we need to check for both } and that at the end.

  no_end_labels <- labels[!grepl("}|}`r ''`", labels)]
  if(length(no_end_labels) > 0){
    stop("From test-code-chunks.R. Missing `}` at end of code chunk labels: ",
         toString(no_end_labels), " Found in file ", i, "\n")
  }

}

#' Gather Submissions
#'
#' @description This function finds and reads HTML/XML files from a local directory 
#' or Google Drive folder that match specified patterns. It extracts tables from the 
#' files and returns a list of tibbles containing the submission data.
#'
#' @param path The path to the local directory containing the HTML/XML files, or a Google Drive folder URL.
#'        If it's a Google Drive URL, the function will download the entire folder to a temporary directory.
#' @param title A character vector of patterns to match against the file names.
#'        Each pattern is processed separately and results are combined.
#' @param keep_loc A character string specifying where to save downloaded files (only for Google Drive URLs). 
#'        If NULL (default), files are downloaded to a temporary directory and deleted after processing.
#'        If specified, files are downloaded to this location and kept.
#' @param verbose A logical value (TRUE or FALSE) specifying verbosity level. 
#'        If TRUE, reports files that are removed during processing.
#'
#' @details Google Drive allows for more than one file with the exact same name. 
#' If you download files manually ("by hand"), you will get both files but with 
#' one of them automatically renamed by your browser. However, if you use the 
#' Google Drive functionality in this function, the second file will overwrite 
#' the first, potentially resulting in data loss.
#'
#' @return A named list of tibbles, where each tibble contains the data from one HTML/XML file
#'         that matches any of the specified patterns and has valid table structure.
#'
#' @examples
#' # Find submissions from local directory
#' path <- system.file("extdata", "answers_html", package = "tutorial.helpers")
#'
#' tibble_list <- gather_submissions(path = path, title = "stop", verbose = TRUE)
#'
#' \dontrun{
#' # Find submissions from Google Drive folder (temporary download)
#' drive_url <- "https://drive.google.com/drive/folders/10do12t0fZsfrIrKePxwjpH8IqBNVO86N"
#' tibble_list <- gather_submissions(
#'   path = drive_url, 
#'   title = c("positron")
#' )
#'
#' # Find submissions from Google Drive folder (keep files)
#' tibble_list <- gather_submissions(
#'   path = drive_url, 
#'   title = c("introduction"),
#'   keep_loc = "temp_file"
#' )
#' }
#' @export

gather_submissions <- function(path, title, keep_loc = NULL, verbose = FALSE) {
  
  # Input validation
  if (missing(path) || is.null(path)) {
    stop("'path' must be provided.")
  }
  
  if (missing(title) || is.null(title)) {
    stop("'title' must be provided.")
  }
  
  # Validate verbose parameter
  if (!is.logical(verbose) || length(verbose) != 1) {
    stop("'verbose' must be a single logical value (TRUE or FALSE).")
  }
  
  # Determine if path is a Google Drive URL or local directory
  is_drive_url <- grepl("^https?://", path)
  
  # Validate keep_loc usage
  if (!is.null(keep_loc) && !is_drive_url) {
    stop("'keep_loc' can only be used with Google Drive URLs.")
  }
  
  # Track temporary directory for cleanup
  temp_dir_to_cleanup <- NULL
  
  if (is_drive_url) {
    result <- handle_google_drive_url(path, title, keep_loc, verbose)
    real_path <- result$path
    temp_dir_to_cleanup <- result$temp_dir
  } else {
    real_path <- handle_local_directory(path, verbose)
  }
  
  # Get all files in directory
  all_files <- list.files(real_path, full.names = FALSE, recursive = TRUE)
  
  # Filter HTML/XML files
  html_xml_files <- filter_html_xml_files(all_files, real_path, verbose)
  
  # Filter by title patterns
  matching_files <- filter_by_title_patterns(html_xml_files, title, verbose)
  
  # Process matching files
  tibble_list <- process_matching_files(matching_files, real_path, verbose)
  
  # Cleanup temporary directory if needed
  if (!is.null(temp_dir_to_cleanup) && dir.exists(temp_dir_to_cleanup)) {
    unlink(temp_dir_to_cleanup, recursive = TRUE)
    if (verbose) {
      message("Cleaned up temporary directory: ", temp_dir_to_cleanup)
    }
  }
  
  return(tibble_list)
}

# Handle Google Drive URL
handle_google_drive_url <- function(path, title, keep_loc, verbose) {
  if (verbose) {
    message("Detected Google Drive URL. Downloading entire folder...")
  }
  
  # Check if required packages are available
  if (!requireNamespace("googledrive", quietly = TRUE)) {
    stop("Package 'googledrive' is required for Google Drive functionality. ",
         "Please install it with: install.packages('googledrive')")
  }
  
  # Validate Google Drive link format
  if (!grepl("drive\\.google\\.com", path)) {
    stop("Invalid Google Drive link format.")
  }
  
  if (!grepl("/folders/", path)) {
    stop("Only Google Drive folder URLs are supported. ",
         "Please provide a folder URL in the format: ",
         "https://drive.google.com/drive/folders/FOLDER_ID")
  }
  
  # Determine download location
  if (!is.null(keep_loc)) {
    # Use specified location - download_google_drive will create a subdirectory there
    download_base_path <- keep_loc
    temp_dir_to_cleanup <- NULL
  } else {
    # Use temporary directory - download_google_drive will create a subdirectory there
    download_base_path <- tempdir()
    # We'll need to clean up the subdirectory that download_google_drive creates
    temp_dir_to_cleanup <- NULL  # Will be set after download
  }
  
  # Capture output to suppress the cat() messages if not verbose
  if (!verbose) {
    invisible(utils::capture.output({
      downloaded_path <- download_google_drive(url = path, path = download_base_path, title = title)
    }))
  } else {
    downloaded_path <- download_google_drive(url = path, path = download_base_path, title = title)
  }
  
  # Validate that the download was successful
  if (is.null(downloaded_path) || !dir.exists(downloaded_path)) {
    stop("Google Drive download failed - no valid directory returned.")
  }
  
  if (verbose) {
    message("Google Drive folder successfully downloaded to: ", downloaded_path)
  }
  
  # Set temp directory for cleanup if using temporary location
  if (is.null(keep_loc)) {
    # The downloaded_path is the full path to the created subdirectory (e.g., /tmp/gdrive_download_2025-08-31)
    # This is what we want to clean up, not its parent
    temp_dir_to_cleanup <- downloaded_path
  }
  
  return(list(path = downloaded_path, temp_dir = temp_dir_to_cleanup))
}

# Handle Local Directory
handle_local_directory <- function(path, verbose) {
  if (!dir.exists(path)) {
    stop("The specified directory does not exist: ", path)
  }
  
  return(normalizePath(path))
}

# Filter HTML/XML Files
filter_html_xml_files <- function(all_files, real_path, verbose) {
  if (length(all_files) == 0) {
    return(character(0))
  }
  
  full_file_paths <- file.path(real_path, all_files)

  html_xml_files <- sapply(full_file_paths, function(file) {
    if (!file.exists(file)) return(FALSE)

    # Identify HTML/XML files by their extension.
    grepl("\\.(html?|xml)$", basename(file), ignore.case = TRUE)
  })
  
  html_xml_file_names <- all_files[html_xml_files]
  non_html_xml_file_names <- all_files[!html_xml_files]
  
  # Only report removed files when verbose = TRUE
  if (verbose && length(non_html_xml_file_names) > 0) {
    if (!requireNamespace("utils", quietly = TRUE)) {
      stop("Package 'utils' is required. Please install it with: install.packages('utils')")
    }
    files_to_show <- utils::head(non_html_xml_file_names, 3)
    message("Removed ", length(non_html_xml_file_names), " non-HTML/XML file(s): '", 
           paste(files_to_show, collapse = "', '"), 
           if(length(non_html_xml_file_names) > 3) "', ..." else "'")
  }
  
  return(html_xml_file_names)
}

# Filter Files by Title Patterns
filter_by_title_patterns <- function(html_xml_files, title, verbose) {
  if (length(html_xml_files) == 0) {
    return(character(0))
  }
  
  all_matching_files <- character()
  
  for (current_title in title) {
    pattern_matches <- grep(current_title, html_xml_files, value = TRUE)
    all_matching_files <- c(all_matching_files, pattern_matches)
  }
  
  all_matching_files <- unique(all_matching_files)
  
  # Report files that matched patterns when verbose = TRUE
  if (verbose && length(all_matching_files) > 0) {
    if (!requireNamespace("utils", quietly = TRUE)) {
      stop("Package 'utils' is required. Please install it with: install.packages('utils')")
    }
    files_to_show <- utils::head(all_matching_files, 3)
    message("Found ", length(all_matching_files), " file(s) matching pattern '", 
           paste(title, collapse = "|"), "': '", 
           paste(files_to_show, collapse = "', '"), 
           if(length(all_matching_files) > 3) "', ..." else "'")
  }
  
  return(all_matching_files)
}

# Process Matching Files
process_matching_files <- function(matching_files, real_path, verbose) {
  if (length(matching_files) == 0) {
    return(list())
  }
  
  tibble_list <- list()
  failed_files <- character()
  
  for (file_name in matching_files) {
    result <- process_single_file(file_name, real_path, verbose)
    
    if (!is.null(result)) {
      tibble_list[[file_name]] <- result
    } else {
      failed_files <- c(failed_files, file_name)
    }
  }
  
  # Report files that failed processing when verbose = TRUE
  if (verbose && length(failed_files) > 0) {
    if (!requireNamespace("utils", quietly = TRUE)) {
      stop("Package 'utils' is required. Please install it with: install.packages('utils')")
    }
    files_to_show <- utils::head(failed_files, 3)
    message("Removed ", length(failed_files), " file(s) due to processing failures: '", 
           paste(files_to_show, collapse = "', '"), 
           if(length(failed_files) > 3) "', ..." else "'")
  }
  
  return(tibble_list)
}

# Process Single File
# @param file_name Name of the file to process
# @param real_path Base directory path
# @param verbose Verbosity level
# @return Tibble or NULL if processing failed
process_single_file <- function(file_name, real_path, verbose) {
  file_path <- file.path(real_path, file_name)
  
  # Check if rvest package is available
  if (!requireNamespace("rvest", quietly = TRUE)) {
    stop("Package 'rvest' is required. Please install it with: install.packages('rvest')")
  }
  
  # Read HTML content
  html_content <- tryCatch({
    rvest::read_html(file_path)
  }, error = function(e) {
    return(NULL)
  })
  
  if (is.null(html_content)) {
    return(NULL)
  }
  
  # Extract table data
  table_data <- tryCatch({
    tables <- rvest::html_table(html_content)
    if (length(tables) == 0) {
      return(NULL)
    }
    tables[[1]]
  }, error = function(e) {
    return(NULL)
  })
  
  if (is.null(table_data) || nrow(table_data) == 0) {
    return(NULL)
  }
  
  # Check if tibble package is available
  if (!requireNamespace("tibble", quietly = TRUE)) {
    stop("Package 'tibble' is required. Please install it with: install.packages('tibble')")
  }
  
  # Convert to tibble
  tibble_data <- tryCatch({
    tibble::as_tibble(table_data)
  }, error = function(e) {
    return(NULL)
  })
  
  return(tibble_data)
}

test_that("gather_submissions returns a list", {
  result <- gather_submissions(
    path = test_path("fixtures", "process_submissions_dir"),
    title = "."
  )
  expect_type(result, "list")
})

test_that("gather_submissions returns tibbles in the list", {
  result <- gather_submissions(
    path = test_path("fixtures", "process_submissions_dir"),
    title = "."
  )
  expect_true(all(sapply(result, function(x) inherits(x, "tbl_df"))))
})

test_that("gather_submissions works with single pattern", {
  result <- gather_submissions(
    path = test_path("fixtures", "process_submissions_dir"),
    title = "getting"
  )
  expect_true(all(grepl("getting", names(result))))
})

# Better tests using real answer files
library(tibble)

test_that("gather_submissions works with single pattern on real answers", {
  result <- gather_submissions(
    path = test_path("fixtures", "answers_html"),
    title = "stops"
  )

  expect_true(is.list(result))
  expect_true(length(result) == 25)
  expect_true(all(sapply(result, tibble::is_tibble)))
  expect_true(min(sapply(result, nrow)) == 28)
  expect_true(max(sapply(result, nrow)) == 80)
})

# tibble_list_test <- gather_submissions("~/Downloads/responses", title = "probability", email = c("ryanbansal04@gmail.com", "surya.fraser@gmail.com", "darakhshan.fatima110@gmail.com"))
# saveRDS(tibble_list_test, test_path("fixtures", "check_test_dir", "tibble_list_test.rds"))

tibble_list_test <- readRDS(test_path("fixtures", "check_test_dir", "tibble_list_test.rds"))

# Test 1: Basic functionality - single key variable that exists
test_that("check_key_vars returns tibbles with required single key variable", {
  result <- check_key_vars(tibble_list_test, key_vars = "email")
  
  # Should return all 3 tibbles since they all have "email"
  expect_length(result, 3)
  expect_setequal(names(result), names(tibble_list_test))
  
  # Check that all returned tibbles have the email variable
  for (tibble_data in result) {
    expect_true("email" %in% tibble_data$id)
  }
})

# Test 2: Multiple key variables that all exist
test_that("check_key_vars returns tibbles with all required multiple key variables", {
  result <- check_key_vars(tibble_list_test, key_vars = c("email", "name"))
  
  # Should return all 3 tibbles since they all have both "email" and "name"
  expect_length(result, 3)
  expect_setequal(names(result), names(tibble_list_test))
  
  # Check that all returned tibbles have both variables
  for (tibble_data in result) {
    expect_true(all(c("email", "name") %in% tibble_data$id))
  }
})

# Test 3: Key variable that doesn't exist in any tibble
test_that("check_key_vars returns empty list when key variable doesn't exist", {
  result <- check_key_vars(tibble_list_test, key_vars = "phone")
  
  expect_length(result, 0)
  expect_equal(result, list())
})

# tibble_list_test <- gather_submissions("~/Downloads/responses", title = "probability", email = c("ryanbansal04@gmail.com", "surya.fraser@gmail.com", "darakhshan.fatima110@gmail.com"))
# saveRDS(tibble_list_test, test_path("fixtures", "check_test_dir", "tibble_list_test.rds"))

tibble_list_test <- readRDS(test_path("fixtures", "check_test_dir", "tibble_list_test.rds"))

# Test 1: Basic functionality - should return Ryan's tibble
test_that("check_membership returns correct tibble for single email", {
  result <- check_membership(tibble_list_test, 
                           key_var = "email", 
                           membership = "ryanbansal04@gmail.com", 
                           verbose = FALSE)
  
  # Should return exactly 1 tibble (Ryan's)
  expect_length(result, 1)
  expect_true("probability_answers - Ryan.html" %in% names(result))
  
  # Check that the returned tibble has the correct email
  email_row <- which(result[[1]]$id == "email")
  expect_equal(result[[1]]$answer[email_row], "ryanbansal04@gmail.com")
})

# Test 2: Multiple emails - should return all 3 tibbles
test_that("check_membership returns all tibbles when all emails are included", {
  result <- check_membership(tibble_list_test, 
                           key_var = "email", 
                           membership = c("ryanbansal04@gmail.com", "surya.fraser@gmail.com", "darakhshan.fatima110@gmail.com"), 
                           verbose = FALSE)
  
  expect_length(result, 3)
  expect_setequal(names(result), names(tibble_list_test))
})

# Test 3: Non-existent key variable - should return empty list
test_that("check_membership returns empty list for non-existent key variable", {
  result <- check_membership(tibble_list_test,
                           key_var = "nonexistent_var",
                           membership = "umaira.nazar09@gmail.com",
                           verbose = FALSE)

  expect_length(result, 0)
})

# Test 4: verbose = TRUE reports the email summary while returning the same
# result as verbose = FALSE.
test_that("check_membership verbose output reports the email summary", {
  membership <- c("ryanbansal04@gmail.com", "not.a.real.student@gmail.com")

  expect_message(
    result <- check_membership(tibble_list_test, key_var = "email",
                               membership = membership, verbose = TRUE),
    "Email summary:")

  # The kept set still resolves to exactly Ryan's submission.
  expect_length(result, 1)
  expect_true("probability_answers - Ryan.html" %in% names(result))

  # Specific summary lines are emitted.
  expect_message(
    check_membership(tibble_list_test, key_var = "email",
                     membership = membership, verbose = TRUE),
    "Membership emails not found")
  expect_message(
    check_membership(tibble_list_test, key_var = "email",
                     membership = membership, verbose = TRUE),
    "Final result: 1 tibble")
})

# Test 5: verbose = TRUE on a key variable that no tibble has reports the
# early-exit message and returns an empty list.
test_that("check_membership verbose reports when no tibble has the key var", {
  expect_message(
    result <- check_membership(tibble_list_test, key_var = "nonexistent_var",
                               membership = "ryanbansal04@gmail.com",
                               verbose = TRUE),
    "No tibbles contain the required key variable")
  expect_length(result, 0)
})
# Regression tests: tibbles storing answers in an old-format `data` column
# used to be silently dropped, and a duplicated key variable in one file used
# to abort the whole batch.

test_that("tibbles with a 'data' column pass membership checks", {
  tibble_list <- list(
    old_format = tibble::tibble(
      id = c("name", "email"),
      submission_type = c("question", "question"),
      data = c("Old Format", "old.format@example.com")))

  result <- check_membership(tibble_list,
                             key_var = "email",
                             membership = "old.format@example.com")
  expect_equal(names(result), "old_format")
})

test_that("a duplicated key variable does not abort the batch", {
  tibble_list <- list(
    dup = tibble::tibble(
      id = c("email", "email"),
      submission_type = "question",
      answer = c("first@example.com", "second@example.com")),
    ok = tibble::tibble(
      id = "email",
      submission_type = "question",
      answer = "kept@example.com"))

  result <- check_membership(tibble_list,
                             key_var = "email",
                             membership = c("first@example.com", "kept@example.com"))
  expect_setequal(names(result), c("dup", "ok"))
})

#' Write tutorial answers to file
#'
#' @description
#'   Take a tutorial session (or a submission list), extract all submitted answers,
#'   and write out an HTML file with those answers.
#'
#' @param file Output file path (should end in .html).
#' @param obj Either a Shiny session object (from learnr) or a list of submissions
#'   (as returned by get_submissions_from_learnr_session()).
#' @returns No return value, called for the side effect of writing an HTML
#'   file of submitted answers to `file`.
#' @export

write_answers <- function(file, obj) {
  if (is.environment(obj)) {
    objs <- get_submissions_from_learnr_session(obj)
    tutorial_id <- learnr::get_tutorial_info()$tutorial_id
  } else if (is.list(obj) && !is.data.frame(obj) && length(obj) > 0 &&
             is.list(obj[[1]]) && !is.null(obj[[1]]$id)) {
    objs <- obj
    tutorial_id <- "test-tutorial"
  } else {
    stop("obj must be a Shiny session or a submissions list.")
  }

  out <- tibble::tibble(
    id = vapply(objs, function(.x) if (is.null(.x$id) || length(.x$id) == 0) NA_character_ else as.character(.x$id), character(1)),
    submission_type = vapply(objs, function(.x) if (is.null(.x$type) || length(.x$type) == 0) NA_character_ else as.character(.x$type), character(1)),
    answer = vapply(objs, function(.x) {
      ans <- .x$answer
      if (is.null(ans) || length(ans) == 0) {
        NA_character_
      } else if (length(ans) > 1) {
        paste(as.character(ans), collapse = ", ")
      } else {
        as.character(ans)
      }
    }, character(1))
  )

  out <- rbind(c(id = "tutorial-id", submission_type = "none", answer = tutorial_id), out)

  z <- knitr::kable(out, format = "html")
  write(as.character(z), file = file)
  invisible(NULL)
}

#' Check Key Variables in List of Tibbles
#'
#' This function checks if specified key variables are present in each tibble's "id" column
#' and returns only tibbles that contain all required key variables.
#'
#' @param tibble_list A named list of tibbles, each containing an "id" column with question identifiers
#' @param key_vars A character vector of key variables to check for
#' @param verbose A logical value (TRUE or FALSE) specifying verbosity level. 
#'        If TRUE, reports tibbles that are removed and why.
#'
#' @return A list of tibbles that contain all required key variables
#'
#' @examples
#' # Create sample data
#' path <- system.file("extdata", "answers_html", package = "tutorial.helpers")
#'
#' tibble_list <- gather_submissions(path, "stop")
#'
#' result <- check_key_vars(tibble_list,
#'                          key_vars = c("name", "email"))
#' @export
check_key_vars <- function(tibble_list, key_vars, verbose = FALSE) {
  
  # Validate verbose parameter
  if (!is.logical(verbose) || length(verbose) != 1) {
    stop("'verbose' must be a single logical value (TRUE or FALSE).")
  }
  
  # Initialize list to store valid tibbles
  valid_tibbles <- list()
  removed_files <- character()
  removal_reasons <- character()
  
  # Get names of tibbles (use indices if no names provided)
  if (is.null(names(tibble_list))) {
    tibble_names <- paste0("tibble_", seq_along(tibble_list))
  } else {
    tibble_names <- names(tibble_list)
  }
  
  # Check each tibble
  for (i in seq_along(tibble_list)) {
    tibble_data <- tibble_list[[i]]
    file_name <- tibble_names[i]
    
    # Check if tibble has an 'id' column
    if (!"id" %in% colnames(tibble_data)) {
      removed_files <- c(removed_files, file_name)
      removal_reasons <- c(removal_reasons, "no 'id' column")
      next
    }
    
    # Check which key variables are present
    missing_vars <- setdiff(key_vars, tibble_data$id)
    
    # If all key variables are present, keep the tibble
    if (length(missing_vars) == 0) {
      valid_tibbles[[file_name]] <- tibble_data
    } else {
      # Track removal
      removed_files <- c(removed_files, file_name)
      removal_reasons <- c(removal_reasons, paste("missing key variables:", paste(missing_vars, collapse = ", ")))
    }
  }
  
  # Report removed files if verbose
  if (verbose && length(removed_files) > 0) {
    for (i in seq_along(removed_files)) {
      message("Removed '", removed_files[i], "': ", removal_reasons[i])
    }
    message("Summary: ", length(removed_files), " tibble(s) removed, ", length(valid_tibbles), " tibble(s) retained")
  }
  
  return(valid_tibbles)
}

#' Return a list of tutorial answers
#'
#' @description Grabs information from the `learnr` session environment, not
#'   directly from the session object itself. Since we are using the session
#'   environment, we currently don't (?) have a way to save the environment and
#'   hence can't test this function.
#'
#' @param sess session object from shiny with learnr
#'
#' @returns a list which includes the exercise submissions of tutorial
#' 
#' @export
get_submissions_from_learnr_session <- function(sess){

  # Why not just make this a tibble, rather than a nested list? Certainly  would
  # make later code easier.
  
  # Note that get_tutorial_state(), with its `label` argument, allows you to
  # grab the answer to a single question. Also, the returned object has a
  # `timestamp` entry. Perhaps these would be of use somehow?
  
  # As best I can tell, get_tutorial_state() can only return questions which
  # students have answered. You need to do something else, get_tutorial_info(),
  # to get a list of all the questions.
  
  # get_tutorial_info() returns a list, one element of which is `items`

  curr_state <- learnr::get_tutorial_state(session = sess)

  label_names <- names(curr_state)

  obj_list <- list()

  for (n in label_names){
    obj_list[[n]] <- list(
      id = n,
      type = curr_state[[n]]$type,
      answer = curr_state[[n]]$answer
    )
  }

  obj_list
}

## ```{r info-section, child = system.file("child_documents/info_section.Rmd", package = "tutorial.helpers")}
## ```

## ```{r download-answers, child = system.file("child_documents/download_answers.Rmd", package = "tutorial.helpers")}
## ```

## ### Exercise 6
## 
## Explain potential outcomes in about two sentences.
## 
## ```{r definitions-6}
## question_text(NULL,
##     message = "This is where we place the correct answer. It will appear only after
##     students have submitted their own answers. Note that we do not need to wrap the
##     answer text by hand.",
##     answer(NULL,
##            correct = TRUE),
##     allow_retry = FALSE,
##     incorrect = NULL,
##     rows = 6)
## ```

## ### Exercise 7
## 
## 
## ```{r exercise-7}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 5)
## ```
## 
## ###
## 
## ```{r exercise-7-test}
## 
## ```
## 
## ###
## 

## median_age <- get_acs(geography = "county",
##                       variables = "B01002_001",
##                       year = 2020)
## write_rds(median_age, "data/median_age.rds")
## 
## median_age <- read_rds("data/median_age.rds")

## fit_gauss <- brm(formula = biden ~ 1,
##                 data = poll_data,
##                 refresh = 0,
##                 silent = 2,
##                 seed = 9)
## write_rds(fit_gauss, "data/fit_gauss.rds")
## 
## fit_gauss <- read_rds("data/fit_gauss.rds")

## ```{r}
## include_graphics("images/example.png")
## ```

## <pre><code>```{r}
## 1 + 1
## ```</code></pre>

## <pre><code>
## > tutorial.helpers::show_file("quarto-1.qmd")
## ---
## title: "Quarto 1"
## author: David Kane
## format: html
## execute:
##   echo: false
## ---
## <pre><code>```{r}
## suppressPackageStartupMessages(library(tidyverse))
## library(palmerpenguins)
## ```</code></pre>
## >
## </code></pre>

## ### Exercise 1
## 
## You should be connected to a repo named `r4ds-1`. If you are not, create one and connect to it.
## 
## You need two R Terminals: one for running this tutorial and one for your exercises.
## 
## Create a new Quarto Document and save it as `analysis.qmd`. Add a YAML header at the top with a title (`"Analyzing the Billboard 100"`) and your name as author.
## 
## Render the document by running `quarto render` in the bash Terminal. Open the rendered `analysis.html` file in a new tab by right clicking it in the Explorer and selecting "Open with Live Server". When you run `quarto render` from now on that tab will be updated with the newly rendered file.
## 
## Create a `.gitignore` file with `analysis_files` on the first line and then a blank line. Commit and push.
## 
## In the R Terminal, run:
## 
## ```
## show_file(".gitignore")
## ```
## 
## If that fails, it is probably because you have not yet loaded `library(tutorial.helpers)` in the R Terminal.
## 
## CP/CR.
## 
## ```{r introduction-1}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 3)
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->
## 
## <!-- XX: By default, the next two questions just mention tidyverse. If there is more than one required library, you can just add those libraries to these steps, adjusting the grammar as needed. Or, you might have separate questions for each library, thereby providing more room for knowledge drops. -->

## ### Exercise 2
## 
## In your QMD, put `library(tidyverse)` in a new code chunk. Render the file by running `quarto render` in the bash Terminal.
## 
## Notice that the file does not look good because the code is visible and there are annoying messages. To take care of this, add `#| message: false` to remove all the messages in this setup chunk. Also, add the following to the YAML header to remove all code echoes from the HTML:
## 
## ```
## execute:
##   echo: false
## ```
## 
## Render the file again, using `quarto render`. Only the title and author should appear in the HTML.
## 
## In the R Terminal, run:
## 
## ```
## show_file("analysis.qmd", chunk = "Last")
## ```
## 
## CP/CR.
## 
## ```{r introduction-2}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 6)
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 3
## 
## Place your cursor in the QMD file on the `library(tidyverse)` line. Use `Cmd/Ctrl + Enter` to execute that line.
## 
## Note how this command causes `library(tidyverse)` to be copied down to the R Terminal and then executed.
## 
## CP/CR.
## 
## ```{r introduction-3}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 8)
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 4
## 
## <!-- XX: Delete this question if you do not make use of a `data` directory in this tutorial. -->
## 
## Create a `data` directory at the top level of the `XX` repo. This is a good place to store any data that you are working with.
## 
## In the bash Terminal, run `ls`.
## 
## CP/CR.
## 
## ```{r introduction-4}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 5)
## ```
## 
## ###
## 
## You answer should look something like this:
## 
## ```
## $ ls
## README.md  analysis.html  analysis.qmd  analysis_files  data
## ```
## 
## ###
## 
## <!-- XX: If you have downloaded some data, then you might want to create the tibbles that you will use here. (Of course, you also need to create those tibbles in the setup chunk so that any test code will run.) -->

## ## XX: First topic (use sentence case)
## ###
## 
## <!-- XX: Mention the packages/functions which you plan on covering in this Topic. Not everything mentioned here is specified in the Introduction/Summary, but everything in Introduction/Summary is referenced in one of these topic introductions, the space before Exercise 1 in each topic. -->

## ### Exercise 1
## 
## <!-- XX: In this question, "XX" is the full name of the file, like "cheeses.xlsx". -->
## 
## We begin by downloading XX directly from GitHub to the `data` directory using `download.file()`.
## 
## In the R Terminal, run:
## 
## ```
## download.file(
##   "https://github.com/PPBDS/ai.tutorials/raw/refs/heads/main/inst/tutorials/r4ds-2/data/cheeses.xlsx",
##   destfile = "data/XX"
## )
## ```
## 
## CP/CR.
## 
## ```{r cheese-1}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 8)
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 6
## 
## Prompt AI to generate R code that ... Add the code to your QMD in a new chunk. Place your cursor on the first line of the code and run `Cmd/Ctrl + Enter`.
## 
## In the R Terminal, run:
## 
## ```
## show_file("analysis.qmd", chunk = "Last")
## ```
## 
## CP/CR.
## 
## ```{r something-1}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 4)
## ```
## 
## ###
## 
## ```{r}
## # XX: Insert code that produces the correct result — a plot, summary statistics, or printed tibble. With echo = FALSE (the default), students see only the output; add #| echo: true when you want them to study the code too.
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 3
## 
## Before creating a plot, we need to ensure that your data matches ours. Run your pipe in the R Terminal and compare your output to the result shown below.
## 
## In the R Terminal, run:
## 
## ```
## show_file("analysis.qmd", chunk = "Last")
## ```
## 
## CP/CR.
## 
## ```{r xx-first-section-use-sentence-case-3}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 6)
## ```
## 
## ###
## 
## ```{r xx-first-section-use-sentence-case-3-test}
## # XX: Insert code that prints the tibble students should have at this stage.
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 4
## 
## Within the latest code chunk, assign the result of the pipe to `x`.
## 
## Place your cursor on the line where the pipe is assigned to `x`, run `Cmd/Ctrl + Enter`, thus ensuring that the workspace also includes a copy of `x`.
## 
## CP/CR.
## 
## ```{r xx-first-section-use-sentence-case-4}
## question_text(NULL,
##     answer(NULL, correct = TRUE),
##     allow_retry = TRUE,
##     try_again_button = "Edit Answer",
##     incorrect = NULL,
##     rows = 8)
## ```
## 
## <!-- XX: Show x here. -->
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 5
## 
## Add the option `#| cache: true` to the chunk which creates `x`. This causes Quarto to cache the results of the chunk. The next time you render your QMD, as long as you have not changed the code, Quarto will just load up the saved object.
## 
## Run `quarto render` in the bash Terminal. Rendering creates an `analysis_cache` directory next to the QMD.
## 
## In the bash Terminal, run `ls`.
## 
## CP/CR.
## 
## ```{r xx-first-section-use-sentence-case-5}
## question_text(NULL,
##     answer(NULL, correct = TRUE),
##     allow_retry = TRUE,
##     try_again_button = "Edit Answer",
##     incorrect = NULL,
##     rows = 5)
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 6
## 
## Add `analysis_cache` to the `.gitignore`. The contents of the cache directory do not belong on GitHub.
## 
## In the R Terminal, run:
## 
## ```
## show_file(".gitignore")
## ```
## 
## CP/CR.
## 
## ```{r xx-first-section-use-sentence-case-6}
## question_text(NULL,
##     answer(NULL, correct = TRUE),
##     allow_retry = TRUE,
##     try_again_button = "Edit Answer",
##     incorrect = NULL,
##     rows = 3)
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 7
## 
## Within the R Terminal, type `x`, which we previously assigned to a pipe and ran in the R Terminal. Hit `Enter`.
## 
## CP/CR.
## 
## ```{r xx-first-section-use-sentence-case-7}
## question_text(NULL,
##     answer(NULL, correct = TRUE),
##     allow_retry = TRUE,
##     try_again_button = "Edit Answer",
##     incorrect = NULL,
##     rows = 8)
## ```
## 
## ###
## 
## Our result:
## 
## ```{r}
## # x
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 8
## 
## Ask AI to generate R code that uses `x` to plot a basic graph or calculate and present summary statistics showing XX ... Mention you want to use the data from `x`. If using a chat interface copy/paste the `x` you ran in the R Terminal with the resulting tibble. You only need the top 3 lines, mainly to include column names.
## 
## Consider adding a title, subtitle, and caption. If axis labels would be useful, add them, but if unnecessary, don't bother. Don't assign the code for the plot to any variable. Put the plot code in a new code chunk. Run `quarto render` to ensure that everything works. Make your plot look nice.
## 
## In the R Terminal, run:
## 
## ```
## show_file("analysis.qmd", chunk = "Last")
## ```
## 
## CP/CR.
## 
## ```{r xx-first-section-use-sentence-case-8}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 12)
## ```
## 
## ###
## 
## Our result:
## 
## ```{r}
## # XX: Insert plotting code. With echo = FALSE (the default), students see only the plot, not this code. Make sure the plot has a subtitle that states the take-away message and uses clean, concise code.
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ## Summary
## ###
## 
## <!-- XX: The exact same two to four sentences about the main packages/functions used in the Introduction, but written here in the past tense. You made a promise and you kept it.  -->

## ### Exercise 1
## 
## Run `quarto render` to ensure that everything works.  The resulting HTML page should be attractive, showing clean versions of your plot(s).
## 
## At the R Terminal, run:
## 
## ```
## show_file("analysis.qmd")
## ```
## 
## CP/CR.
## 
## ```{r summary-1}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 30)
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->

## ### Exercise 2
## 
## Publish your rendered QMD to GitHub Pages. In the bash Terminal --- not the R Terminal! --- run:
## 
## ```
## quarto publish gh-pages analysis.qmd
## ```
## 
## Copy/paste the resulting URL below.
## 
## ```{r summary-2}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 1)
## ```
## 
## ###
## 
## <!-- XX: Insert a knowledge drop related to this project. -->
## 

## ### Exercise 3
## 
## Commit and push all your files. Copy/paste the URL to your Github repo.
## 
## ```{r summary-3}
## question_text(NULL,
## 	answer(NULL, correct = TRUE),
## 	allow_retry = TRUE,
## 	try_again_button = "Edit Answer",
## 	incorrect = NULL,
## 	rows = 3)
## ```
## 
## ###
## 
## <!-- XX: The tutorial is now over. Add any necessary acknowledgements and/or provide a link to further high quality readings, ideally readings which you mentioned in at least one knowledge drop above. -->

#' Format RMarkdown tutorial code chunks
#'
#' This function processes an R Markdown tutorial file to standardize code chunk labels
#' based on section names and exercise numbers. It also renumbers exercises sequentially
#' within each section and fixes spacing in topic headers.
#'
#' @param file_path Character string. Path to the R Markdown file to process.
#'
#' @details
#' The function applies the following formatting rules:
#' \itemize{
#'   \item Topic headers (# headers) have their spacing standardized
#'   \item Exercises are renumbered sequentially within each section
#'   \item Code chunks are relabeled according to the pattern: section-name-exercise-number
#'   \item Chunks with `eval = FALSE` receive a `-hint-N` suffix
#'   \item Chunks with `include = FALSE` receive a `-test` suffix
#'   \item Chunks with label "setup" are not modified
#'   \item Chunks with the "file" option are not modified
#'   \item Unlabeled chunks without key options are not modified
#'   \item All formatted chunks preserve their original options
#'   \item Content between quadruple backticks (```` ````) is preserved untouched
#' }
#'
#' @return Character string containing the formatted R Markdown content.
#' @export
format_tutorial <- function(file_path) {
  # Helper to always format chunk header with clean spaces and exactly one closing }
  format_chunk_header <- function(label, opts) {
    opts <- trimws(opts)
    opts <- sub("\\}+$", "", opts)               # Remove all closing }
    opts <- gsub("^\\s*,\\s*", "", opts)
    opts <- gsub("\\s+", " ", opts)
    opts <- gsub(",\\s+", ", ", opts)
    opts <- trimws(opts)
    if (nchar(opts) > 0) {
      line <- paste0("```{r ", label, ", ", opts, "}")
    } else {
      line <- paste0("```{r ", label, "}")
    }
    # Guarantee one closing }
    line <- sub("\\}+$", "}", line)
    return(line)
  }
  
  lines <- readLines(file_path)
  section_name <- ""
  exercise_counter <- 0
  in_yaml <- FALSE
  in_verbatim <- FALSE
  in_section <- FALSE
  verbatim_count <- 0
  hint_counters <- list()
  force_exercise_chunk <- FALSE

  # First pass: Fix topic header spacing and identify/renumber exercises
  i <- 1
  while (i <= length(lines)) {
    current_line <- lines[i]
    if (grepl("^````\\s*$", trimws(current_line))) {
      verbatim_count <- verbatim_count + 1
      in_verbatim <- verbatim_count %% 2 != 0
      i <- i + 1; next
    }
    if (in_verbatim) { i <- i + 1; next }
    if (grepl("^---$", current_line)) {
      in_yaml <- !in_yaml; i <- i + 1; next
    }
    if (in_yaml) { i <- i + 1; next }
    if (grepl("^#\\s+", current_line)) {
      topic_title <- sub("^#\\s+", "", current_line)
      topic_title <- gsub("\\s+", " ", trimws(topic_title))
      lines[i] <- paste0("# ", topic_title)
      i <- i + 1; next
    }
    if (grepl("^##\\s+", current_line)) {
      section_title <- sub("^##\\s+", "", current_line)
      section_title <- gsub("\\s+", " ", trimws(section_title))
      lines[i] <- paste0("## ", section_title)
      section_name <- tolower(section_title)
      section_name <- gsub("\\s+", "-", section_name)
      section_name <- gsub("[^a-z0-9-]", "", section_name)
      section_name <- gsub("-+", "-", section_name)
      if (grepl("-$", section_name)) section_name <- sub("-$", "", section_name)
      exercise_counter <- 0
      in_section <- TRUE
      i <- i + 1; next
    }
    if (in_section && grepl("^### Exercise", current_line)) {
      exercise_counter <- exercise_counter + 1
      lines[i] <- paste0("### Exercise ", exercise_counter)
      hint_key <- paste0(section_name, "-", exercise_counter)
      hint_counters[[hint_key]] <- 0
      i <- i + 1; next
    }
    i <- i + 1
  }

  # Second pass: Update code chunk labels
  section_name <- ""
  exercise_counter <- 0
  in_yaml <- FALSE
  in_verbatim <- FALSE
  verbatim_count <- 0
  hint_counters <- list()
  force_exercise_chunk <- FALSE
  i <- 1
  while (i <= length(lines)) {
    current_line <- lines[i]
    if (grepl("^````\\s*$", trimws(current_line))) {
      verbatim_count <- verbatim_count + 1
      in_verbatim <- verbatim_count %% 2 != 0
      i <- i + 1; next
    }
    if (in_verbatim) { i <- i + 1; next }
    if (grepl("^---$", current_line)) {
      in_yaml <- !in_yaml; i <- i + 1; next
    }
    if (in_yaml) { i <- i + 1; next }
    if (grepl("^#\\s+", current_line)) { i <- i + 1; next }
    if (grepl("^##\\s+", current_line)) {
      section_title <- sub("^##\\s+", "", current_line)
      section_title <- gsub("\\s+", " ", trimws(section_title))
      section_name <- tolower(section_title)
      section_name <- gsub("\\s+", "-", section_name)
      section_name <- gsub("[^a-z0-9-]", "", section_name)
      section_name <- gsub("-+", "-", section_name)
      if (grepl("-$", section_name)) section_name <- sub("-$", "", section_name)
      exercise_counter <- 0
      i <- i + 1; next
    }
    if (grepl("^### Exercise", current_line)) {
      exercise_counter <- exercise_counter + 1
      hint_key <- paste0(section_name, "-", exercise_counter)
      hint_counters[[hint_key]] <- 0
      force_exercise_chunk <- TRUE
      i <- i + 1; next
    }

    # ---- Code chunk logic starts here ----
    if (grepl("^```\\{r", current_line)) {
      # (1) SKIP completely unlabeled chunks {r}
      if (grepl("^```\\{r\\s*\\}$", current_line) || grepl("^```\\{r\\s*$", current_line)) {
        i <- i + 1; next
      }
      # (2) SKIP any chunk with file = ... (anywhere in the header)
      header_only <- gsub("\\}.*$", "", current_line)
      if (grepl("file\\s*=", header_only)) {
        i <- i + 1; next
      }
      # (3) SKIP "setup" chunks
      if (grepl("^```\\{r\\s+setup", current_line)) {
        i <- i + 1; next
      }

      # Forced relabel after Exercise header
      if (force_exercise_chunk) {
        new_base_label <- paste0(section_name, "-", exercise_counter)
        chunk_options <- sub("^```\\{r\\s*[^,}]*\\s*(,.*)?\\s*\\}?\\s*$", "\\1", current_line)
        lines[i] <- format_chunk_header(new_base_label, chunk_options)
        force_exercise_chunk <- FALSE
        i <- i + 1; next
      }

      # (4) Skip if not in a section or exercise. This must come before the
      # -test/-hint relabeling so that pre-labeled chunks outside any
      # section/exercise are left untouched rather than corrupted.
      if (section_name == "" || exercise_counter == 0) {
        i <- i + 1; next
      }

      # -test and -hint labels (fix only label, preserve all options)
      if (grepl("-test(\\}|,|$)", current_line)) {
        new_base_label <- paste0(section_name, "-", exercise_counter, "-test")
        chunk_options <- sub("^```\\{r\\s*[^,}]*\\s*(,.*)?\\s*\\}?\\s*$", "\\1", current_line)
        lines[i] <- format_chunk_header(new_base_label, chunk_options)
        i <- i + 1; next
      }
      if (grepl("-hint-\\d+(\\}|,|$)", current_line)) {
        hint_key <- paste0(section_name, "-", exercise_counter)
        if (is.null(hint_counters[[hint_key]])) hint_counters[[hint_key]] <- 0
        hint_counters[[hint_key]] <- hint_counters[[hint_key]] + 1
        hint_number <- hint_counters[[hint_key]]
        new_base_label <- paste0(section_name, "-", exercise_counter, "-hint-", hint_number)
        chunk_options <- sub("^```\\{r\\s*[^,}]*\\s*(,.*)?\\s*\\}?\\s*$", "\\1", current_line)
        lines[i] <- format_chunk_header(new_base_label, chunk_options)
        i <- i + 1; next
      }

      # (5) If already has the correct label, skip
      current_label_pattern <- paste0("^```\\{r\\s+", section_name, "-", exercise_counter, "(,|\\})")
      if (grepl(current_label_pattern, current_line)) {
        i <- i + 1; next
      }

      # --- Decide label type ---
      new_base_label <- paste0(section_name, "-", exercise_counter)
      # Special suffixes
      if (grepl("include\\s*=\\s*FALSE", current_line)) {
        new_base_label <- paste0(new_base_label, "-test")
      } else if (grepl("eval\\s*=\\s*FALSE", current_line)) {
        hint_key <- paste0(section_name, "-", exercise_counter)
        hint_counters[[hint_key]] <- hint_counters[[hint_key]] + 1
        hint_number <- hint_counters[[hint_key]]
        new_base_label <- paste0(new_base_label, "-hint-", hint_number)
      } else if (grepl("exercise\\s*=\\s*TRUE", current_line)) {
        # keep base label
      } else {
        # skip chunks without key options
        i <- i + 1; next
      }
      # Actually rebuild chunk header
      chunk_options <- sub("^```\\{r\\s*[^,}]*\\s*(,.*)?\\s*\\}?\\s*$", "\\1", current_line)
      lines[i] <- format_chunk_header(new_base_label, chunk_options)
      i <- i + 1; next
    }
    i <- i + 1
  }
  return(paste(lines, collapse = "\n"))
}

library(testthat)

f_1_test <- tutorial.helpers::format_tutorial(
  test_path("fixtures", "addin_test_inputs", "format_input_1.Rmd")
)

# writeLines(
#   f_1_test,
#   test_path("fixtures", "addin_test_outputs", "format_output_1.Rmd")
# )

f_1_truth <- paste(
  readLines(test_path("fixtures", "addin_test_outputs", "format_output_1.Rmd")),
  collapse = "\n"
)

test_that("Format 1 works", {
  expect_equal(f_1_test, f_1_truth)
})


# Regression tests: section names with special characters used to produce
# labels with doubled hyphens in the rewritten chunk headers, and pre-labeled
# -test/-hint chunks before the first exercise used to be corrupted.

test_that("special characters in section names collapse to single hyphens", {
  lines <- c("## Data & plots",
             "### Exercise 1",
             "```{r, exercise = TRUE}",
             "```")
  f <- tempfile(fileext = ".Rmd")
  on.exit(unlink(f))
  writeLines(lines, f)

  out <- strsplit(format_tutorial(f), "\n")[[1]]
  expect_true(any(grepl("^```\\{r data-plots-1", out)))
  expect_false(any(grepl("--", out[grepl("^```\\{r", out)])))
})

test_that("pre-labeled -test/-hint chunks before the first exercise are untouched", {
  lines <- c("```{r my-test, include = FALSE}",
             "```",
             "```{r old-hint-1, eval = FALSE}",
             "```",
             "## Section",
             "### Exercise 1")
  f <- tempfile(fileext = ".Rmd")
  on.exit(unlink(f))
  writeLines(lines, f)

  out <- strsplit(format_tutorial(f), "\n")[[1]]
  expect_true("```{r my-test, include = FALSE}" %in% out)
  expect_true("```{r old-hint-1, eval = FALSE}" %in% out)
})

#' @title Add a tutorial code exercise or question to the active document
#'
#' @rdname exercise_creation
#'
#' @description When writing tutorials, it is handy to be able to insert the
#'   skeleton for a new code exercise or question. Note that the
#'   function determines the correct exercise number to use and also adds
#'   appropriate code chunk labels, based on the exercise number and section
#'   title.
#'
#' @param type Character of question type. Must be one of "code", "no-answer",
#'   or "yes-answer". Abbreviations such as "no", "yes", and "co" are allowed.
#'
#' @param file_path Character path to a file. If NULL, the RStudio active
#'   document is used, which is the default behavior. An actual file path is
#'   used for testing.
#'
#' @returns Exercise skeleton corresponding to the `type` argument.
#'
#' @export

make_exercise <- function(type = "no-answer", file_path = NULL) {
  
  # --- Step 1: Determine the Content Context ---
  
  # We need to determine what "text" we are analyzing to generate the numbers.
  # If file_path is provided, we use that file (Testing mode).
  # If file_path is NULL, we use the Active Document (Interactive mode).
  
  # Variable to track if we created a temp file that needs deletion
  temp_file_created <- NULL
  
  if (is.null(file_path)) {
    
    # Get the active document context
    ctx <- rstudioapi::getActiveDocumentContext()
    
    if (is.null(ctx)) {
      stop("No active document. Please open an Rmd file.")
    }
    
    # --- THE TRICK ---
    # Instead of using the file path on disk (which includes the WHOLE file),
    # we grab the content from the editor buffer and slice it at the cursor.
    # This ensures we only count exercises *before* the insertion point.
    
    # 1. Get all lines from the buffer
    all_contents <- ctx$contents
    
    # 2. Identify the cursor row (start of the selection range)
    if (length(ctx$selection) > 0 && !is.null(ctx$selection[[1]])) {
      cursor_row <- ctx$selection[[1]]$range$start["row"]
    } else {
      cursor_row <- length(all_contents) # Fallback to end if no cursor found
    }
    
    # 3. Slice content: Keep only lines 1 up to the cursor
    # We ensure we take at least one line to avoid errors
    if (cursor_row < 1) cursor_row <- 1
    preceding_content <- all_contents[1:cursor_row]
    
    # 4. Write this partial content to a temporary file.
    # We do this because determine_exercise_number() expects a file path argument.
    temp_file <- tempfile(fileext = ".Rmd")
    writeLines(preceding_content, temp_file)
    
    # Set the working path to this temp file for the calculation steps
    working_path <- temp_file
    temp_file_created <- temp_file
    
  } else {
    # Testing mode: Use the provided file path exactly as is
    working_path <- file_path
  }
  
  
  # --- Step 2: Calculate Numbers and IDs ---
  
  # Now we determine exercise number and section ID using the working path
  # (which is either the full test file or the partial temp file)
  
  exercise_number <- determine_exercise_number(working_path)
  section_id <- determine_code_chunk_name(working_path)
  
  # Clean up the temp file if we made one
  if (!is.null(temp_file_created) && file.exists(temp_file_created)) {
    unlink(temp_file_created)
  }
  
  
  # --- Step 3: Handle Defaults and Validation ---
  
  # Handle empty section_id - use a default if it's empty
  if (is.null(section_id) || section_id == "" || is.na(section_id)) {
    section_id <- "exercise"
  }
  
  # Ensure exercise_number is valid
  if (is.null(exercise_number) || is.na(exercise_number)) {
    exercise_number <- 1
  }
  
  # Allow abbreviated type inputs (e.g., "no" for "no-answer", "ye" for "yes-answer")
  allowed_types <- c("no-answer", "yes-answer", "code")
  
  # Expand abbreviations
  if (!type %in% allowed_types) {
    match_idx <- pmatch(type, allowed_types)
    if (is.na(match_idx)) stop("Invalid type argument.")
    type <- allowed_types[match_idx]
  }
  
  
  # --- Step 4: Compose the Exercise Skeleton ---
  
  new_exercise <- switch(type,
    "no-answer" = sprintf(
      "### Exercise %s\n\n\n```{r %s-%s}\nquestion_text(NULL,\n\tanswer(NULL, correct = TRUE),\n\tallow_retry = TRUE,\n\ttry_again_button = \"Edit Answer\",\n\tincorrect = NULL,\n\trows = 5)\n```\n\n###\n\n```{r %s-%s-test, echo = TRUE}\n\n```\n\n###\n\n",
      exercise_number,
      section_id,
      exercise_number,
      section_id,
      exercise_number
    ),
    "yes-answer" = sprintf(
      "### Exercise %s\n\n\n```{r %s-%s}\nquestion_text(NULL,\n\tmessage = \"Place correct answer here.\",\n\tanswer(NULL, correct = TRUE),\n\tallow_retry = FALSE,\n\tincorrect = NULL,\n\trows = 6)\n```\n\n###\n\n",
      exercise_number,
      section_id,
      exercise_number
    ),
    "code" = sprintf(
      "### Exercise %s\n\n\n```{r %s-%s, exercise = TRUE}\n\n```\n\n<button onclick = \"transfer_code(this)\">Copy previous code</button>\n\n```{r %s-%s-hint, eval = FALSE}\n\n```\n\n```{r %s-%s-test, include = FALSE}\n\n```\n\n###\n\n",
      exercise_number,
      section_id,
      exercise_number,
      section_id,
      exercise_number,
      section_id,
      exercise_number
    ),
    stop("Unknown type argument to make_exercise()")
  )
  
  
  # --- Step 5: Insertion Logic ---
  
  if (is.null(file_path)) {
    
    # Interactive Mode: Insert into the active document
    
    # Recalculate context to be safe (though usually unchanged)
    ctx <- rstudioapi::getActiveDocumentContext()
    
    if (length(ctx$selection) > 0 && !is.null(ctx$selection[[1]])) {
      
      # Extract the range from the current selection
      cursor_range <- ctx$selection[[1]]$range
      
      # Insert the text at the cursor position
      rstudioapi::insertText(location = cursor_range, text = new_exercise)
      
    } else {
      stop("Could not determine cursor position in the active document.")
    }
    
  } else {
    
    # Test Mode: Write or append to the provided file path
    cat(new_exercise, file = file_path, append = TRUE)
    invisible(new_exercise)
  }
}

# make_exercise.R is still rough because it behaves poorly if you execute it in
# a document which is not set up like a tutorial, especially one which lacks a
# Section header. 

library(testthat)

create_test_file <- function(path, add_headers = TRUE) {
  if (add_headers) {
    writeLines(
      c("## My section", "###", "", "### Exercise 1", "", "### Exercise 2", ""),
      con = path
    )
  } else {
    file.create(path)
  }
}

# ---- MAIN TEST (APPEND MODE) ----

test_that("make_exercise() generates correct output for all types", {
  # Create temp file and add section headers for proper numbering
  tmp <- tempfile(fileext = ".Rmd")
  create_test_file(tmp, add_headers = TRUE)
  
  tutorial.helpers::make_exercise(type = "no", file_path = tmp)
  tutorial.helpers::make_exercise(type = "yes", file_path = tmp)
  tutorial.helpers::make_exercise(type = "co", file_path = tmp)
  
  output <- paste(readLines(tmp), collapse = "\n")
  
  truth <- paste(
    readLines(testthat::test_path("fixtures", "tutorial_examples", "make_exercise_expected.Rmd")),
    collapse = "\n"
  )
  
  # Test: do they match?
  expect_equal(output, truth)
})

# ---- ERROR HANDLING TEST ----

test_that("make_exercise() errors with bad input", {
  tmp <- tempfile(fileext = ".Rmd")
  create_test_file(tmp)
  expect_error(tutorial.helpers::make_exercise(type = "bad-input", file_path = tmp))
})

# ---- INTERACTIVE CURSOR TEST (THE TRICK) ----

test_that("make_exercise() ignores exercises after the cursor", {
  
  # This test simulates the "Trick": We are in the middle of a document.
  # Ex 1 is above us. Ex 10 is below us.
  # The function should suggest "Exercise 2", NOT "Exercise 11".
  
  fake_contents <- c(
    "## Data Wrangling",
    "",
    "### Exercise 1",
    "",
    "", # <--- Cursor will be simulated here (Row 5)
    "### Exercise 10", 
    ""
  )
  
  # Mock the Context Object that rstudioapi normally returns
  fake_ctx <- list(
    contents = fake_contents,
    path = "dummy_path.Rmd",
    selection = list(list(range = list(start = c(row = 5))))
  )
  
  # Variable to capture the text that make_exercise tries to insert
  captured_insertion <- NULL
  
  # Mock rstudioapi functions
  local_mocked_bindings(
    getActiveDocumentContext = function() fake_ctx,
    insertText = function(location, text) { captured_insertion <<- text },
    .package = "rstudioapi"
  )
  
  # Run function in interactive mode (file_path = NULL)
  # We specify type = "code" so we can verify the copy button generation
  tutorial.helpers::make_exercise(type = "code", file_path = NULL)
  
  # 1. Check Numbering Logic (The "Trick")
  # It correctly identified that we are after Ex 1, ignoring Ex 10.
  expect_match(captured_insertion, "### Exercise 2")
  
  # 2. Check Section Logic
  # The pattern is {slug}-{number}, so "data-wrangling-2"
  expect_match(captured_insertion, "data-wrangling-2")
  
  # 3. Check Structure
  # Since we requested type="code", the transfer button should exist
  expect_match(captured_insertion, "transfer_code")
})

# Create a temporary directory for test files

temp_dir <- tempdir()

# Create test files with non-portable file names

file.copy(from = "fixtures/process_submissions_dir/Aadi.html",
          to = file.path(temp_dir, "introduction_answers -- Aadi.html"),
          overwrite = TRUE)
file.copy(from = "fixtures/process_submissions_dir/astrxr.html",
          to = file.path(temp_dir, "introduction_answers -- astrxr.html"),
          overwrite = TRUE)
file.copy(from = "fixtures/process_submissions_dir/Ivy.html",
          to = file.path(temp_dir, "ivy-introduction -- Ivy S.html"),
          overwrite = TRUE)

# Set the test directory path as an environment variable
Sys.setenv("TEST_DIR" = temp_dir)

#' Check Membership in List of Tibbles
#'
#' This function filters a list of tibbles based on whether a key variable's value
#' is among the membership values. It first uses check_key_vars() to ensure the 
#' key variable exists, then checks membership. Useful for keeping only specific 
#' students or participants.
#'
#' @param tibble_list A named list of tibbles, each containing an "id" column and an "answer" column
#' @param key_var A character string specifying the key variable to check
#' @param membership A character vector of allowed values for the key variable
#' @param verbose Logical indicating whether to report removed items (default: FALSE)
#'
#' @return A list of tibbles where the key variable exists and its value is in the membership list
#'
#' @examples
#' # Create sample data with student emails
#' path <- system.file("extdata", "answers_html", package = "tutorial.helpers")
#'
#' tibble_list <- gather_submissions(path, "stop")
#'
#' result <- check_membership(tibble_list,
#'                            key_var = "email",
#'                            membership = c("bluebird.jack.xu@gmail.com",
#'                                           "hassan.alisoni007@gmail.com"))
#' @export
check_membership <- function(tibble_list, key_var, membership, verbose = FALSE) {
  
  # First, use check_key_vars to filter tibbles that have the required key variable
  if (verbose) {
    message("Step 1: Checking for required key variable '", key_var, "'...")
  }
  
  # Use check_key_vars without verbose output to avoid duplicate messages
  tibbles_with_key <- check_key_vars(tibble_list, key_var, verbose = FALSE)
  
  if (length(tibbles_with_key) == 0) {
    if (verbose) {
      message("No tibbles contain the required key variable '", key_var, "'")
    }
    return(list())
  }
  
  if (verbose) {
    message("Step 2: Checking membership for key variable '", key_var, "'...")
  }
  
  # Initialize list to store valid tibbles
  valid_tibbles <- list()
  
  # Track emails found for verbose reporting
  emails_found <- character()
  emails_kept <- character()
  emails_removed <- character()
  
  # Check membership for each tibble that passed the key variable check
  for (file_name in names(tibbles_with_key)) {
    tibble_data <- tibbles_with_key[[file_name]]
    
    # Get the value of the key variable
    key_var_row <- which(tibble_data$id == key_var)
    
    # Check if we found the key variable
    if (length(key_var_row) == 0) {
      next
    }
    
    # Values are stored in an 'answer' column or, in older files, a 'data'
    # column.
    if ("answer" %in% colnames(tibble_data)) {
      key_var_value <- tibble_data$answer[key_var_row]
    } else if ("data" %in% colnames(tibble_data)) {
      key_var_value <- tibble_data$data[key_var_row]
    } else {
      next
    }

    # A malformed file may repeat the key variable; use the first value.
    key_var_value <- key_var_value[1]
    emails_found <- c(emails_found, key_var_value)
    
    # Check if the value is in the membership list
    if (key_var_value %in% membership) {
      valid_tibbles[[file_name]] <- tibble_data
      emails_kept <- c(emails_kept, key_var_value)
    } else {
      emails_removed <- c(emails_removed, key_var_value)
    }
  }
  
  # Report summary if verbose
  if (verbose) {
    n_original <- length(tibble_list)
    n_with_key <- length(tibbles_with_key)
    n_final <- length(valid_tibbles)
    n_removed_key <- n_original - n_with_key
    n_removed_membership <- n_with_key - n_final
    
    message("Email summary:")
    
    # Show emails that were kept (present in membership)
    if (length(emails_kept) > 0) {
      message("- Emails found and kept (", length(emails_kept), "): ", paste(emails_kept, collapse = ", "))
    } else {
      message("- Emails found and kept: none")
    }
    
    # Show emails that were removed (not in membership)
    if (length(emails_removed) > 0) {
      message("- Emails found but not in membership (", length(emails_removed), "): ", paste(emails_removed, collapse = ", "))
    } else {
      message("- Emails found but not in membership: none")
    }
    
    # Show membership emails that were not found
    emails_not_found <- setdiff(membership, emails_found)
    if (length(emails_not_found) > 0) {
      message("- Membership emails not found (", length(emails_not_found), "): ", paste(emails_not_found, collapse = ", "))
    } else {
      message("- All membership emails were found")
    }
    
    message("Final summary:")
    message("- Started with: ", n_original, " tibble(s)")
    message("- Removed ", n_removed_key, " tibble(s) for missing key variable '", key_var, "'")
    message("- Removed ", n_removed_membership, " tibble(s) for membership check")
    message("- Final result: ", n_final, " tibble(s) retained")
  }
  
  return(valid_tibbles)
}

#' Check current tutorial
#'
#' @description
#'
#' A function which formats the exercise numbers and code chunk
#' labels correctly. `check_current_tutorial()` reads in the tutorial
#' which is in the active editor window in Positron. It determines what 
#' the number of each exercise should be and fixes any mistakes. It
#' ensures that all code chunk labels are the correct function of the 
#' section title and exercise number.
#'
#' @returns No return value, called for the side effect of rewriting the
#'   tutorial file in the active editor window.
#'
#' @export

check_current_tutorial <- function(){

  file_path <- rstudioapi::getSourceEditorContext()$path

  new_doc <- format_tutorial(file_path)

  # This part just writes that reformatted document
  # to the file.

  out <- file(file_path, "w")
  write(trimws(new_doc), file = out)
  close(out)
}

#' Confirm that a tutorial has the recommended components
#'
#' @description Checks that tutorials contain required libraries and child 
#'   documents. The function looks for library() calls and child document
#'   inclusions in the tutorial files.
#'
#' @param tutorial_paths Character vector of the paths to the tutorials to be
#'   examined.
#'   
#' @param libraries Character vector of library names that should be loaded
#'   in the tutorial. The function looks for `library(name)` calls. Default
#'   is `c("learnr", "tutorial.helpers")`.
#'   
#' @param children Character vector of child document names (without the .Rmd
#'   extension) that should be included in the tutorial. The function looks 
#'   for these in child document inclusion chunks. Default is 
#'   `c("info_section", "download_answers")`.
#'
#' @examples
#'   # Check with default requirements
#'   check_tutorial_defaults(
#'     tutorial_paths = return_tutorial_paths("tutorial.helpers")
#'   )
#'   
#'   # Check for specific libraries and child documents
#'   check_tutorial_defaults(
#'     tutorial_paths = return_tutorial_paths("tutorial.helpers"),
#'     libraries = c("learnr", "knitr"),
#'     children = c("info_section")
#'   )
#'
#' @returns No return value, called for side effects. 
#'
#' @export

check_tutorial_defaults <- function(tutorial_paths,
                                    libraries = c("learnr", "tutorial.helpers"),
                                    children = c("info_section", "download_answers")) {
  
  stopifnot(all(file.exists(tutorial_paths)))
  
  for(tutorial_path in tutorial_paths) {
    
    # Read the tutorial content
    
    tutorial_lines <- readLines(tutorial_path)
    
    
    # Check for required libraries
    # Look for lines like: library(learnr) or library("learnr") or library('learnr')
    
    if(length(libraries) > 0) {
      for(lib in libraries) {
        
        # Create regex pattern to match library calls with or without quotes
        
        lib_pattern <- sprintf("library\\(['\"]?%s['\"]?\\)", lib)
        
        if(!any(grepl(lib_pattern, tutorial_lines))) {
          stop("Missing library(", lib, ") call in file ", tutorial_path, "\n")
        }
      }
    }
    
    
    # Check for required child documents
    # Look for lines containing child document inclusions
    # Pattern matches: child = system.file("child_documents/[name].Rmd", package = "tutorial.helpers")
    # Also handles variations in spacing and quotes
    
    if(length(children) > 0) {
      for(child in children) {
        
        # Create regex pattern to match child document inclusion
        # This pattern is flexible about quotes and spacing
        
        child_pattern <- sprintf(
          "child\\s*=\\s*system\\.file\\(['\"].*/%s\\.Rmd['\"]",
          child
        )
        
        if(!any(grepl(child_pattern, tutorial_lines))) {
          stop("Missing child document '", child, "' in file ", tutorial_path, "\n")
        }
      }
    }
  }
  
  invisible(NULL)
}

# Key insight is that we can test tutorial files without installing them as
# tutorials. But maybe we should also have an installed tutorial . . .

test_that("check_tutorial() works on legal tutorials", {
  expect_null(
    check_tutorial_defaults(test_path("fixtures", "tutorial_examples", "good-tutorial.Rmd"))
  )
  expect_null(
    check_tutorial_defaults(
      c(test_path("fixtures", "tutorial_examples", "good-tutorial.Rmd"),
        test_path("fixtures", "tutorial_examples", "good-tutorial-2.Rmd"))
    )
  )
})


test_that("check_tutorial() fails on illegal tutorials", {
  expect_error(
    check_tutorial_defaults(test_path("fixtures", "tutorial_examples", "no-exist-tutorial.Rmd"))
  )
  expect_error(
    check_tutorial_defaults(test_path("fixtures", "tutorial_examples", "no-info-tutorial.Rmd"))
    )
  expect_error(
    check_tutorial_defaults(test_path("fixtures", "tutorial_examples", "no-tutorialhelpers-tutorial.Rmd"))
    )
})

## Download answers

```{r minutes}
question_numeric(
	"How many minutes, approximately, did it take you to complete this tutorial? For example, an hour and a half would be 90 minutes.",
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again = "Edit Answer",
	incorrect = NULL,
  min = 0,
	max = 360,
	step = 1)
```

###

```{r context = "setup"}
tutorial.helpers::submission_ui
```

```{r context = "server"}
tutorial.helpers::submission_server()
```

---
params:
  title: "Week 0 Tutorial Report"
  output: !r tibble::tibble(id = "test", checked = FALSE, correct = NA)
title: "`r params$title`"
---

```{r setup, include = FALSE}
library(gt)
```

```{r, echo = FALSE, center = TRUE}
gt(params$output)
```

<!-- This file is used in some testing. -->

---
title: Introduction to Python
author: Satvika Upperla and David Kane
tutorial:
  id: introduction-to-python
output:
  learnr::tutorial:
    df_print: default
    progressive: yes
    allow_skip: yes
runtime: shiny_prerendered
description: Tutorial for introducing students to the Python language.
---

```{r setup, include=FALSE}
library(learnr)
library(tutorial.helpers)
library(knitr)

knitr::opts_chunk$set(echo = FALSE)
knitr::opts_chunk$set(out.width = '90%')
options(tutorial.exercise.timelimit = 60, 
        tutorial.storage = "local") 
```

```{r copy-code-chunk, child = system.file("child_documents/copy_button.Rmd", package = "tutorial.helpers")}
```

```{r info-section, child = system.file("child_documents/info_section.Rmd", package = "tutorial.helpers")}
```

### Introduction
##

This is tutorial introduces you to using [Python](https://www.python.org/) for data science. You will learn how to work with data sets using **polars**, a DataFrame library that uses a syntax similar to R's **tidyverse**. You will learn how to chain operations using method chaining with pipes, and how to make plots using **plotnine**, which implements the grammar of graphics just like **ggplot2** in R.

###

This tutorial assumes that you have already completed the "Getting Started" tutorial in the [**tutorial.helpers**](https://ppbds.github.io/tutorial.helpers/) package. If you haven't, do so now. It is quick!

###

To complete this tutorial, you must either install Python on your local machine or work with [GitHub Codespaces](https://github.com/codespaces) in the cloud. Because Python installation is a tricky, finicky process, we do not provide any guidance. We recommend you use GitHub Codespaces instead. But all our instructions should work just as well on your machine.


## GitHub Codespaces
###

Professionals store their work on GitHub, or a similar *source control* tool. If your computer blows up, you don't want to lose your work. GitHub is like Google Drive --- both live in the cloud --- but for your computational work rather than your documents.

### Exercise 1

Create a GitHub account by following the instructions at the [GitHub homepage](https://github.com/). 

**Follow [this advice](https://happygitwithr.com/github-acct.html#username-advice) when choosing your username.** 

We recommend using a permanent email address for this account, not one which you lose access to when, for example, you change schools or jobs.

Copy your GitHub account URL in the field below.

```{r github-codespaces-1}
question_text(NULL,
    answer(NULL, correct = TRUE),
    allow_retry = TRUE,
    try_again_button = "Edit Answer",
    incorrect = NULL,
    rows = 2)
```

### 

Your answer should look like this:

```         
https://github.com/your-username
```

[Git](https://en.wikipedia.org/wiki/Git) is "software for tracking changes in any set of files, usually used for coordinating work among programmers collaboratively developing source code during software development."

### Exercise 2

Once you are logged in to GitHub, go to https://github.com/codespaces. (This is also accessible via the lefthand-side pull-down menu.)

```{r}
include_graphics("images/codespaces-1.png")
```

Press the "Use this template" button for the "Blank" template. This will create a new Codespace which looks like this:

```{r}
include_graphics("images/codespaces-2.png")
```

Copy/paste the URL for your Codespace

```{r github-codespaces-2}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

A new URL is created for each Codespace. Ours was:

```
https://congenial-computing-machine-65ppr9j7wrhqq9.github.dev/
```

Yours will look similar.

### Exercise 3

Place your cursor in the Terminal, type `pwd` and hit `return/enter`. Copy/paste the command and return, an instruction which we will henceforth abbreviate as CP/CR.

```{r github-codespaces-3}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```
@davidkane9 ➜ /workspaces/codespaces-blank $ pwd
/workspaces/codespaces-blank
@davidkane9 ➜ /workspaces/codespaces-blank $ 
```

Your answer should look similar, except with your GitHub ID in place of mine, which is `davidkane9`. In this tutorial, we will just work at the Terminal because our goal is to learn from Python, not learn about Codespaces. We just use Codespaces as a handy, free location to use Python.


### Exercise 4

Type `python` at the Terminal and hit `return/enter`.

CP/CR.

```{r github-codespaces-4}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```
@davidkane9 ➜ /workspaces/codespaces-blank $ python
Python 3.12.1 (main, Nov 27 2025, 10:47:52) [GCC 13.3.0] on linux
Type "help", "copyright", "credits" or "license" for more information.
>>> 
```

This is the Python "shell" or "interpreter" or "console." It allows us to work with Python interactively. We will use the term "Console" and capitalize it for clarity.

### Exercise 5

At the Console,  type `2 + 2` and hit `enter/return`. Remember, the Console is just the Python shell opened up in the terminal after typing "Python" in.

CP/CR.

```{r github-codespaces-5}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```
>>> 2 + 2
4
>>> 
```

You can exit the Console by typing `exit()`. Make sure to quit a codespace when you stop using it so that you don't use up your free credits.


## Working with data
###

Learn how to explore a data set using functions like `describe()`, `info()`, and `sample()`.

### Exercise 1

Before you start doing data science, you must import the libraries you are going to use. Let's start with the **polars** library. At the Console, type:

`import polars as pl`

Going forward, we won't remind you to hit `return/enter` after every command.

CP/CR.

```{r working-with-data-1}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

You will probably get an error which looks like this:

```
>>> import polars as pl
Traceback (most recent call last):
  File "<stdin>", line 1, in <module>
ModuleNotFoundError: No module named 'polars'
>>> 
```

###

The **polars** library is not installed by default in Codespaces. So we need to install it, along with the **plotnines** and **pyarrow** libraries ourselves. Note that, in Python, the terms "library," "package," and "module" are used (mostly) interchangeably. 


### Exercise 2

Run `exit()` at the Console. This ends your Python session and dumps you back into the shell. Run:

`pip install polars plotnine pyarrow`

Note that the command "run" means to type the provided command and then hit the `return/enter` key.

CP/CR. (But just the first few lines.)


```{r working-with-data-2}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```
>>> exit()
@davidkane9 ➜ /workspaces/codespaces-blank $ pip install polars plotnine
Collecting polars
  Downloading polars-1.36.1-py3-none-any.whl.metadata (10 kB)
Collecting plotnine
  Downloading plotnine-0.15.2-py3-none-any.whl.metadata (9.5 kB)
Collecting polars-runtime-32==1.36.1 (from polars)
  Downloading polars_runtime_32-1.36.1-cp39-abi3-manylinux_2_17_x86_64.manylinux2014_x86_64.whl.metadata (1.5 kB)
...
```

One the of the most important parts of data science is keeping track of all the packages which you need and ensuring that they are installed when you need them.

### Exercise 3

Run `python` at the Terminal. This starts the Python Console again. At the Console, type:

`import polars as pl`

Going forward, we won't remind you to hit `return/enter` after every command.

CP/CR.

```{r working-with-data-3}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```
@davidkane9 ➜ /workspaces/codespaces-blank $ python
Python 3.12.1 (main, Nov 27 2025, 10:47:52) [GCC 13.3.0] on linux
Type "help", "copyright", "credits" or "license" for more information.
>>> import polars as pl
>>> 
```

In Python, we import libraries to access their functions and data. The `import` statement makes the library available, and using `as pl` creates a short alias so we can type `pl` instead of `polars` every time we use a function from the **polars** library.

Your answer never needs to match ours perfectly. Our goal is just to ensure that you are actually following the instructions.

### Exercise 4

DataFrames are spreadsheet-like data structures in **polars**. Let's load the famous iris dataset. We'll read it from a URL.

In the Console, run:

```
iris = pl.read_csv("https://raw.githubusercontent.com/mwaskom/seaborn-data/master/iris.csv")
iris
```

CP/CR.

```{r working-with-data-4}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 15)
```

###

```
>>> iris = pl.read_csv("https://raw.githubusercontent.com/mwaskom/seaborn-data/master/iris.csv")
>>> iris
shape: (150, 5)
┌──────────────┬─────────────┬──────────────┬─────────────┬───────────┐
│ sepal_length ┆ sepal_width ┆ petal_length ┆ petal_width ┆ species   │
│ ---          ┆ ---         ┆ ---          ┆ ---         ┆ ---       │
│ f64          ┆ f64         ┆ f64          ┆ f64         ┆ str       │
╞══════════════╪═════════════╪══════════════╪═════════════╪═══════════╡
│ 5.1          ┆ 3.5         ┆ 1.4          ┆ 0.2         ┆ setosa    │
│ 4.9          ┆ 3.0         ┆ 1.4          ┆ 0.2         ┆ setosa    │
│ 4.7          ┆ 3.2         ┆ 1.3          ┆ 0.2         ┆ setosa    │
│ 4.6          ┆ 3.1         ┆ 1.5          ┆ 0.2         ┆ setosa    │
│ 5.0          ┆ 3.6         ┆ 1.4          ┆ 0.2         ┆ setosa    │
│ …            ┆ …           ┆ …            ┆ …           ┆ …         │
│ 6.7          ┆ 3.0         ┆ 5.2          ┆ 2.3         ┆ virginica │
│ 6.3          ┆ 2.5         ┆ 5.0          ┆ 1.9         ┆ virginica │
│ 6.5          ┆ 3.0         ┆ 5.2          ┆ 2.0         ┆ virginica │
│ 6.2          ┆ 3.4         ┆ 5.4          ┆ 2.3         ┆ virginica │
│ 5.9          ┆ 3.0         ┆ 5.1          ┆ 1.8         ┆ virginica │
└──────────────┴─────────────┴──────────────┴─────────────┴───────────┘
>>> 
```

Whenever we show outputs like this after a question, then we are showing our answer to the previous question, even if we do not label it as such.

### Exercise 5

In the Console, run `iris.describe()`. This provides summary statistics for numerical columns.

CP/CR.

```{r working-with-data-5}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 10)
```

###

```
>>> iris.describe()
shape: (9, 6)
┌────────────┬──────────────┬─────────────┬──────────────┬─────────────┬───────────┐
│ statistic  ┆ sepal_length ┆ sepal_width ┆ petal_length ┆ petal_width ┆ species   │
│ ---        ┆ ---          ┆ ---         ┆ ---          ┆ ---         ┆ ---       │
│ str        ┆ f64          ┆ f64         ┆ f64          ┆ f64         ┆ str       │
╞════════════╪══════════════╪═════════════╪══════════════╪═════════════╪═══════════╡
│ count      ┆ 150.0        ┆ 150.0       ┆ 150.0        ┆ 150.0       ┆ 150       │
│ null_count ┆ 0.0          ┆ 0.0         ┆ 0.0          ┆ 0.0         ┆ 0         │
│ mean       ┆ 5.843333     ┆ 3.057333    ┆ 3.758        ┆ 1.199333    ┆ null      │
│ std        ┆ 0.828066     ┆ 0.435866    ┆ 1.765298     ┆ 0.762238    ┆ null      │
│ min        ┆ 4.3          ┆ 2.0         ┆ 1.0          ┆ 0.1         ┆ setosa    │
│ 25%        ┆ 5.1          ┆ 2.8         ┆ 1.6          ┆ 0.3         ┆ null      │
│ 50%        ┆ 5.8          ┆ 3.0         ┆ 4.4          ┆ 1.3         ┆ null      │
│ 75%        ┆ 6.4          ┆ 3.3         ┆ 5.1          ┆ 1.8         ┆ null      │
│ max        ┆ 7.9          ┆ 4.4         ┆ 6.9          ┆ 2.5         ┆ virginica │
└────────────┴──────────────┴─────────────┴──────────────┴─────────────┴───────────┘
>>> 
```

###

This method provides a quick statistical overview of each numerical variable in the dataset. In some cases, the tutorial displays the same object differently from what you were able to copy/paste. And that is OK! Your answer does not need to match our answer.

### Exercise 6

In the Console, run `iris.sample()`. This selects a random row from the dataset.

CP/CR.

```{r working-with-data-6}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 7)
```

###

```
>>> iris.sample()
shape: (1, 5)
┌──────────────┬─────────────┬──────────────┬─────────────┬───────────┐
│ sepal_length ┆ sepal_width ┆ petal_length ┆ petal_width ┆ species   │
│ ---          ┆ ---         ┆ ---          ┆ ---         ┆ ---       │
│ f64          ┆ f64         ┆ f64          ┆ f64         ┆ str       │
╞══════════════╪═════════════╪══════════════╪═════════════╪═══════════╡
│ 6.3          ┆ 2.8         ┆ 5.1          ┆ 1.5         ┆ virginica │
└──────────────┴─────────────┴──────────────┴─────────────┴───────────┘
>>> 
```

Your answer will differ from this answer because of the inherent randomness in methods like `sample()`.

### Exercise 7

In the Console, hit the Up Arrow to retrieve the previous command. Edit it to add the argument `n = 4` to `iris.sample()`. This will return 4 random rows from the `iris` dataset.

CP/CR.

```{r working-with-data-7}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 10)
```

###

```
>>> iris.sample(n = 4)
shape: (4, 5)
┌──────────────┬─────────────┬──────────────┬─────────────┬───────────┐
│ sepal_length ┆ sepal_width ┆ petal_length ┆ petal_width ┆ species   │
│ ---          ┆ ---         ┆ ---          ┆ ---         ┆ ---       │
│ f64          ┆ f64         ┆ f64          ┆ f64         ┆ str       │
╞══════════════╪═════════════╪══════════════╪═════════════╪═══════════╡
│ 5.0          ┆ 3.3         ┆ 1.4          ┆ 0.2         ┆ setosa    │
│ 6.3          ┆ 2.9         ┆ 5.6          ┆ 1.8         ┆ virginica │
│ 5.6          ┆ 2.8         ┆ 4.9          ┆ 2.0         ┆ virginica │
│ 5.2          ┆ 3.4         ┆ 1.4          ┆ 0.2         ┆ setosa    │
└──────────────┴─────────────┴──────────────┴─────────────┴───────────┘
>>> 
```

###

Editing code directly in the Console quickly becomes annoying. See the **[positron.tutorials](https://ppbds.github.io/positron.tutorials/)** package for tutorials about using the Positron IDE to write and organize your code.

### Exercise 8

In the Console, run `print(iris)`. This returns the same result as typing `iris`.

CP/CR.

```{r working-with-data-8}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 10)
```

###

```
>>> print(iris)
shape: (150, 5)
┌──────────────┬─────────────┬──────────────┬─────────────┬───────────┐
│ sepal_length ┆ sepal_width ┆ petal_length ┆ petal_width ┆ species   │
│ ---          ┆ ---         ┆ ---          ┆ ---         ┆ ---       │
│ f64          ┆ f64         ┆ f64          ┆ f64         ┆ str       │
╞══════════════╪═════════════╪══════════════╪═════════════╪═══════════╡
│ 5.1          ┆ 3.5         ┆ 1.4          ┆ 0.2         ┆ setosa    │
│ 4.9          ┆ 3.0         ┆ 1.4          ┆ 0.2         ┆ setosa    │
│ 4.7          ┆ 3.2         ┆ 1.3          ┆ 0.2         ┆ setosa    │
│ 4.6          ┆ 3.1         ┆ 1.5          ┆ 0.2         ┆ setosa    │
│ 5.0          ┆ 3.6         ┆ 1.4          ┆ 0.2         ┆ setosa    │
│ …            ┆ …           ┆ …            ┆ …           ┆ …         │
│ 6.7          ┆ 3.0         ┆ 5.2          ┆ 2.3         ┆ virginica │
│ 6.3          ┆ 2.5         ┆ 5.0          ┆ 1.9         ┆ virginica │
│ 6.5          ┆ 3.0         ┆ 5.2          ┆ 2.0         ┆ virginica │
│ 6.2          ┆ 3.4         ┆ 5.4          ┆ 2.3         ┆ virginica │
│ 5.9          ┆ 3.0         ┆ 5.1          ┆ 1.8         ┆ virginica │
└──────────────┴─────────────┴──────────────┴─────────────┴───────────┘
>>> 
```

You can control how many rows to display using `iris.head(n)` for the first n rows or `iris.tail(n)` for the last n rows.

### Exercise 9

In the Console, run `iris.head(3)`. This returns the first 3 rows of the `iris` dataset.

CP/CR.

```{r working-with-data-9}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 8)
```

###

```
>>> iris.head(3)
shape: (3, 5)
┌──────────────┬─────────────┬──────────────┬─────────────┬─────────┐
│ sepal_length ┆ sepal_width ┆ petal_length ┆ petal_width ┆ species │
│ ---          ┆ ---         ┆ ---          ┆ ---         ┆ ---     │
│ f64          ┆ f64         ┆ f64          ┆ f64         ┆ str     │
╞══════════════╪═════════════╪══════════════╪═════════════╪═════════╡
│ 5.1          ┆ 3.5         ┆ 1.4          ┆ 0.2         ┆ setosa  │
│ 4.9          ┆ 3.0         ┆ 1.4          ┆ 0.2         ┆ setosa  │
│ 4.7          ┆ 3.2         ┆ 1.3          ┆ 0.2         ┆ setosa  │
└──────────────┴─────────────┴──────────────┴─────────────┴─────────┘
>>> 
```

###

`head()` by default gives the top of the DataFrame, so your answer should match our answer. `sample()`, on the other hand, picks random rows to return. But, in both cases, the result is a DataFrame.

A central organizing principle of **polars** is that most methods take a DataFrame and return a DataFrame. This allows us to "chain" commands together, one after the other, creating a pipeline very similar to R's pipe operator `|>`.

### Exercise 10

In the Console, run `help(iris)`. CP/CR (but just the first few rows).


```{r working-with-data-10}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 10)
```

###

```
Help on DataFrame in module polars.dataframe.frame object:

class DataFrame(builtins.object)
 |  DataFrame(data: 'FrameInitTypes | None' = None, schema: 'SchemaDefinition | None' = None, *, schema_overrides: 'SchemaDict | N
one' = None, strict: 'bool' = True, orient: 'Orientation | None' = None, infer_schema_length: 'int | None' = 100, nan_to_null: 'bo
ol' = False) -> 'None'
...
```

To exit from the help interface, type `q` and hit `return/enter`.

### Exercise 11

In the Console, run `iris.schema`. This shows the column names and data types. CP/CR.

```{r working-with-data-11}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 10)
```

###

```
>>> iris.schema
Schema({'sepal_length': Float64, 'sepal_width': Float64, 'petal_length': Float64, 'petal_width': Float64, 'species': String})
>>> 
```

The `schema` attribute displays information about the DataFrame's structure including the data types of each column. For example, `sepal_length` is listed as `Float64`, meaning it's a 64-bit floating-point number. You can also use `iris.dtypes` to see just the data types, or `iris.columns` to see just the column names.

### Exercise 12

In the Console, run `import math` then `math.sqrt(144)`.

CP/CR.

```{r working-with-data-12}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 5)
```

###

```
>>> import math
>>> math.sqrt(144)
12.0
>>> 
```

The square root function is one of many built-in functions in Python's **math** package. Most return their result, which Python then, by default, prints out. We did not need to use `pip` to install the math module because it is part of the base Python installation.

### Exercise 13

In the Console, run `x = math.sqrt(144)`.

CP/CR.

```{r working-with-data-13}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 3)
```

###

```
x = math.sqrt(144)
```

###

The `=` symbol is the assignment operator in Python. In this case, we are *assigning* the value of `math.sqrt(144)` to the variable `x`. Nothing is printed out because of that assignment.

### Exercise 14

In the Console, run `x` or `print(x)`.

CP/CR.

```{r working-with-data-14}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 3)
```

###

```
>>> x
12.0
>>> 
```

Now that `x` has been defined in the Console, it is available for your use. Above, we just print it out. But we could also use it in other calculations, e.g., `x + 5`.


## Method chaining and plots
###

Although **polars** includes hundreds of methods for data manipulation, the most important are `filter()`, `select()`, `sort()`, `with_columns()`, and `group_by()` with `agg()`. These work very similarly to their R **tidyverse** equivalents.

### Exercise 1

Let's warm up by examining the `tips` dataset. Run:

```
tips = pl.read_csv("https://raw.githubusercontent.com/mwaskom/seaborn-data/master/tips.csv")
tips
```

CP/CR.

```{r method-chaining-and-plots-1}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 20)
```

###

```
>>> tips = pl.read_csv("https://raw.githubusercontent.com/mwaskom/seaborn-data/master/tips.csv")
>>> tips
shape: (244, 7)
┌────────────┬──────┬────────┬────────┬──────┬────────┬──────┐
│ total_bill ┆ tip  ┆ sex    ┆ smoker ┆ day  ┆ time   ┆ size │
│ ---        ┆ ---  ┆ ---    ┆ ---    ┆ ---  ┆ ---    ┆ ---  │
│ f64        ┆ f64  ┆ str    ┆ str    ┆ str  ┆ str    ┆ i64  │
╞════════════╪══════╪════════╪════════╪══════╪════════╪══════╡
│ 16.99      ┆ 1.01 ┆ Female ┆ No     ┆ Sun  ┆ Dinner ┆ 2    │
│ 10.34      ┆ 1.66 ┆ Male   ┆ No     ┆ Sun  ┆ Dinner ┆ 3    │
│ 21.01      ┆ 3.5  ┆ Male   ┆ No     ┆ Sun  ┆ Dinner ┆ 3    │
│ 23.68      ┆ 3.31 ┆ Male   ┆ No     ┆ Sun  ┆ Dinner ┆ 2    │
│ 24.59      ┆ 3.61 ┆ Female ┆ No     ┆ Sun  ┆ Dinner ┆ 4    │
│ …          ┆ …    ┆ …      ┆ …      ┆ …    ┆ …      ┆ …    │
│ 29.03      ┆ 5.92 ┆ Male   ┆ No     ┆ Sat  ┆ Dinner ┆ 3    │
│ 27.18      ┆ 2.0  ┆ Female ┆ Yes    ┆ Sat  ┆ Dinner ┆ 2    │
│ 22.67      ┆ 2.0  ┆ Male   ┆ Yes    ┆ Sat  ┆ Dinner ┆ 2    │
│ 17.82      ┆ 1.75 ┆ Male   ┆ No     ┆ Sat  ┆ Dinner ┆ 2    │
│ 18.78      ┆ 3.0  ┆ Female ┆ No     ┆ Thur ┆ Dinner ┆ 2    │
└────────────┴──────┴────────┴────────┴──────┴────────┴──────┘
>>> 
```

The `tips` dataset contains information about restaurant tips, including total bill, tip amount, and other variables.

### Exercise 2

Run `tips.describe()` to see summary statistics.


```{r method-chaining-and-plots-2}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```
>>> tips.describe()
shape: (9, 8)
┌────────────┬────────────┬──────────┬────────┬────────┬──────┬────────┬──────────┐
│ statistic  ┆ total_bill ┆ tip      ┆ sex    ┆ smoker ┆ day  ┆ time   ┆ size     │
│ ---        ┆ ---        ┆ ---      ┆ ---    ┆ ---    ┆ ---  ┆ ---    ┆ ---      │
│ str        ┆ f64        ┆ f64      ┆ str    ┆ str    ┆ str  ┆ str    ┆ f64      │
╞════════════╪════════════╪══════════╪════════╪════════╪══════╪════════╪══════════╡
│ count      ┆ 244.0      ┆ 244.0    ┆ 244    ┆ 244    ┆ 244  ┆ 244    ┆ 244.0    │
│ null_count ┆ 0.0        ┆ 0.0      ┆ 0      ┆ 0      ┆ 0    ┆ 0      ┆ 0.0      │
│ mean       ┆ 19.785943  ┆ 2.998279 ┆ null   ┆ null   ┆ null ┆ null   ┆ 2.569672 │
│ std        ┆ 8.902412   ┆ 1.383638 ┆ null   ┆ null   ┆ null ┆ null   ┆ 0.9511   │
│ min        ┆ 3.07       ┆ 1.0      ┆ Female ┆ No     ┆ Fri  ┆ Dinner ┆ 1.0      │
│ 25%        ┆ 13.37      ┆ 2.0      ┆ null   ┆ null   ┆ null ┆ null   ┆ 2.0      │
│ 50%        ┆ 17.81      ┆ 2.92     ┆ null   ┆ null   ┆ null ┆ null   ┆ 2.0      │
│ 75%        ┆ 24.08      ┆ 3.55     ┆ null   ┆ null   ┆ null ┆ null   ┆ 3.0      │
│ max        ┆ 50.81      ┆ 10.0     ┆ Male   ┆ Yes    ┆ Thur ┆ Lunch  ┆ 6.0      │
└────────────┴────────────┴──────────┴────────┴────────┴──────┴────────┴──────────┘
>>> 
```

Note that this gives us statistics for the numerical columns in the dataset.

### Exercise 3

Use `.drop_nulls()` to remove rows with missing values. In **polars**, we pipe operations by calling methods one after another, just like in R. Try:

```
tips.drop_nulls()
```

CP/CR.

```{r method-chaining-and-plots-3}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 10)
```

###

```
>>> tips.drop_nulls()
shape: (244, 7)
┌────────────┬──────┬────────┬────────┬──────┬────────┬──────┐
│ total_bill ┆ tip  ┆ sex    ┆ smoker ┆ day  ┆ time   ┆ size │
│ ---        ┆ ---  ┆ ---    ┆ ---    ┆ ---  ┆ ---    ┆ ---  │
│ f64        ┆ f64  ┆ str    ┆ str    ┆ str  ┆ str    ┆ i64  │
╞════════════╪══════╪════════╪════════╪══════╪════════╪══════╡
│ 16.99      ┆ 1.01 ┆ Female ┆ No     ┆ Sun  ┆ Dinner ┆ 2    │
│ 10.34      ┆ 1.66 ┆ Male   ┆ No     ┆ Sun  ┆ Dinner ┆ 3    │
│ 21.01      ┆ 3.5  ┆ Male   ┆ No     ┆ Sun  ┆ Dinner ┆ 3    │
│ 23.68      ┆ 3.31 ┆ Male   ┆ No     ┆ Sun  ┆ Dinner ┆ 2    │
│ 24.59      ┆ 3.61 ┆ Female ┆ No     ┆ Sun  ┆ Dinner ┆ 4    │
│ …          ┆ …    ┆ …      ┆ …      ┆ …    ┆ …      ┆ …    │
│ 29.03      ┆ 5.92 ┆ Male   ┆ No     ┆ Sat  ┆ Dinner ┆ 3    │
│ 27.18      ┆ 2.0  ┆ Female ┆ Yes    ┆ Sat  ┆ Dinner ┆ 2    │
│ 22.67      ┆ 2.0  ┆ Male   ┆ Yes    ┆ Sat  ┆ Dinner ┆ 2    │
│ 17.82      ┆ 1.75 ┆ Male   ┆ No     ┆ Sat  ┆ Dinner ┆ 2    │
│ 18.78      ┆ 3.0  ┆ Female ┆ No     ┆ Thur ┆ Dinner ┆ 2    │
└────────────┴──────┴────────┴────────┴──────┴────────┴──────┘
>>> 
```

Note the number of rows in the DataFrame after `drop_nulls()`. This dataset actually has no missing values, so all rows remain.

We can chain methods by writing `tips.drop_nulls().head()` to first drop NA values and then show the first few rows.

### Exercise 4

Chain `.filter()` to filter rows. Use `pl.col("time") == "Dinner"` as the argument. This is very similar to R's `filter()`. In other words, run:

```
tips.filter(pl.col("time") == "Dinner")
```

```{r method-chaining-and-plots-4}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 10)
```

###

```
>>> tips.filter(pl.col("time") == "Dinner")
shape: (176, 7)
┌────────────┬──────┬────────┬────────┬──────┬────────┬──────┐
│ total_bill ┆ tip  ┆ sex    ┆ smoker ┆ day  ┆ time   ┆ size │
│ ---        ┆ ---  ┆ ---    ┆ ---    ┆ ---  ┆ ---    ┆ ---  │
│ f64        ┆ f64  ┆ str    ┆ str    ┆ str  ┆ str    ┆ i64  │
╞════════════╪══════╪════════╪════════╪══════╪════════╪══════╡
│ 16.99      ┆ 1.01 ┆ Female ┆ No     ┆ Sun  ┆ Dinner ┆ 2    │
│ 10.34      ┆ 1.66 ┆ Male   ┆ No     ┆ Sun  ┆ Dinner ┆ 3    │
│ 21.01      ┆ 3.5  ┆ Male   ┆ No     ┆ Sun  ┆ Dinner ┆ 3    │
│ 23.68      ┆ 3.31 ┆ Male   ┆ No     ┆ Sun  ┆ Dinner ┆ 2    │
│ 24.59      ┆ 3.61 ┆ Female ┆ No     ┆ Sun  ┆ Dinner ┆ 4    │
│ …          ┆ …    ┆ …      ┆ …      ┆ …    ┆ …      ┆ …    │
│ 29.03      ┆ 5.92 ┆ Male   ┆ No     ┆ Sat  ┆ Dinner ┆ 3    │
│ 27.18      ┆ 2.0  ┆ Female ┆ Yes    ┆ Sat  ┆ Dinner ┆ 2    │
│ 22.67      ┆ 2.0  ┆ Male   ┆ Yes    ┆ Sat  ┆ Dinner ┆ 2    │
│ 17.82      ┆ 1.75 ┆ Male   ┆ No     ┆ Sat  ┆ Dinner ┆ 2    │
│ 18.78      ┆ 3.0  ┆ Female ┆ No     ┆ Thur ┆ Dinner ┆ 2    │
└────────────┴──────┴────────┴────────┴──────┴────────┴──────┘
>>> 
```

This workflow --- in which we chain DataFrame methods together --- is very common in **polars** and very similar to R's pipe workflow.

The resulting DataFrame has the same number of columns as `tips` because `filter()` only affects the rows. But there are fewer rows now.

### Exercise 5

Use the "Up" arrow to retrieve the command from the previous question in the Console. (We will do this for most exercises in this section.) Continue the chain with `.select()` to choose specific columns. Use `["total_bill", "tip", "sex", "day"]` as the argument. That is, run:

```
tips.filter(pl.col("time") == "Dinner").select(["total_bill", "tip", "sex", "day"])
```

```{r method-chaining-and-plots-5}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 10)
```

###

```
>>> tips.filter(pl.col("time") == "Dinner").select(["total_bill", "tip", "sex", "day"])
shape: (176, 4)
┌────────────┬──────┬────────┬──────┐
│ total_bill ┆ tip  ┆ sex    ┆ day  │
│ ---        ┆ ---  ┆ ---    ┆ ---  │
│ f64        ┆ f64  ┆ str    ┆ str  │
╞════════════╪══════╪════════╪══════╡
│ 16.99      ┆ 1.01 ┆ Female ┆ Sun  │
│ 10.34      ┆ 1.66 ┆ Male   ┆ Sun  │
│ 21.01      ┆ 3.5  ┆ Male   ┆ Sun  │
│ 23.68      ┆ 3.31 ┆ Male   ┆ Sun  │
│ 24.59      ┆ 3.61 ┆ Female ┆ Sun  │
│ …          ┆ …    ┆ …      ┆ …    │
│ 29.03      ┆ 5.92 ┆ Male   ┆ Sat  │
│ 27.18      ┆ 2.0  ┆ Female ┆ Sat  │
│ 22.67      ┆ 2.0  ┆ Male   ┆ Sat  │
│ 17.82      ┆ 1.75 ┆ Male   ┆ Sat  │
│ 18.78      ┆ 3.0  ┆ Female ┆ Thur │
└────────────┴──────┴────────┴──────┘
>>> 
```

Because `select()` doesn't affect rows, we have the same number as after `filter()`. But we only have 4 columns now. 

### Exercise 6

Use the "Up" arrow to get the previous code. Continue the chain with `.describe()`.

```{r method-chaining-and-plots-6}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 10)
```

###

```
>>> tips.filter(pl.col("time") == "Dinner").select(["total_bill", "tip", "sex", "day"]).describe()
shape: (9, 5)
┌────────────┬────────────┬──────────┬────────┬──────┐
│ statistic  ┆ total_bill ┆ tip      ┆ sex    ┆ day  │
│ ---        ┆ ---        ┆ ---      ┆ ---    ┆ ---  │
│ str        ┆ f64        ┆ f64      ┆ str    ┆ str  │
╞════════════╪════════════╪══════════╪════════╪══════╡
│ count      ┆ 176.0      ┆ 176.0    ┆ 176    ┆ 176  │
│ null_count ┆ 0.0        ┆ 0.0      ┆ 0      ┆ 0    │
│ mean       ┆ 20.797159  ┆ 3.10267  ┆ null   ┆ null │
│ std        ┆ 9.142029   ┆ 1.436243 ┆ null   ┆ null │
│ min        ┆ 3.07       ┆ 1.0      ┆ Female ┆ Fri  │
│ 25%        ┆ 14.48      ┆ 2.0      ┆ null   ┆ null │
│ 50%        ┆ 18.43      ┆ 3.0      ┆ null   ┆ null │
│ 75%        ┆ 25.28      ┆ 3.68     ┆ null   ┆ null │
│ max        ┆ 50.81      ┆ 10.0     ┆ Male   ┆ Thur │
└────────────┴────────────┴──────────┴────────┴──────┘
>>> 
```

This gives us summary statistics for our filtered and selected data.



### Exercise 7

Continue the chain with `.sort("tip")` to sort by the tip column.

```{r method-chaining-and-plots-7}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 10)
```

###

```
>>> tips.filter(pl.col("time") == "Dinner").select(["total_bill", "tip", "sex", "day"]).sort("tip")
shape: (176, 4)
┌────────────┬──────┬────────┬─────┐
│ total_bill ┆ tip  ┆ sex    ┆ day │
│ ---        ┆ ---  ┆ ---    ┆ --- │
│ f64        ┆ f64  ┆ str    ┆ str │
╞════════════╪══════╪════════╪═════╡
│ 3.07       ┆ 1.0  ┆ Female ┆ Sat │
│ 5.75       ┆ 1.0  ┆ Female ┆ Fri │
│ 7.25       ┆ 1.0  ┆ Female ┆ Sat │
│ 12.6       ┆ 1.0  ┆ Male   ┆ Sat │
│ 16.99      ┆ 1.01 ┆ Female ┆ Sun │
│ …          ┆ …    ┆ …      ┆ …   │
│ 28.17      ┆ 6.5  ┆ Female ┆ Sat │
│ 48.27      ┆ 6.73 ┆ Male   ┆ Sat │
│ 39.42      ┆ 7.58 ┆ Male   ┆ Sat │
│ 48.33      ┆ 9.0  ┆ Male   ┆ Sat │
│ 50.81      ┆ 10.0 ┆ Male   ┆ Sat │
└────────────┴──────┴────────┴─────┘
>>> 
```

The `sort()` method sorts the rows of a DataFrame. By default, it sorts in ascending order. 

### Exercise 8

Before we can create plots, we need to import the **plotnine** library. Run: `from plotnine import *`

CP/CR.

```{r method-chaining-and-plots-8}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 10)
```

###

```
>>> from plotnine import *
>>> 
```

**plotnine** is Python's implementation of the Grammar of Graphics, similar to R's **ggplot2**. 

Note that this import invocation is different from the one we used with **polars**: `import polars as pl`. That invocation requires us to use `pl` as a suffix whenever we call a function from **polars**. Using `import *` allows as to just call the functions directly. The danger comes if we have two or more packages with functions which have the same name. The `as` invocation is very common because it forces us to specify the function we want. However, doing so can get annoying! So, we will be "lazy" for this plot.

### Exercise 9

Let's create our filtered dataset and assign it to a variable called `dinner_data`. Use the chain from Exercise 7. Run:`dinner_data = tips.filter(pl.col("time") == "Dinner").select(["total_bill", "tip", "sex", "day"]).sort("tip")`. After doing so, just print out `data_dinner` to confirm that it looks correct.

```{r method-chaining-and-plots-9}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 10)
```

###

```
>>> dinner_data = tips.filter(pl.col("time") == "Dinner").select(["total_bill", "tip", "sex", "day"]).sort("tip")
>>> dinner_data
shape: (176, 4)
┌────────────┬──────┬────────┬─────┐
│ total_bill ┆ tip  ┆ sex    ┆ day │
│ ---        ┆ ---  ┆ ---    ┆ --- │
│ f64        ┆ f64  ┆ str    ┆ str │
╞════════════╪══════╪════════╪═════╡
│ 3.07       ┆ 1.0  ┆ Female ┆ Sat │
│ 5.75       ┆ 1.0  ┆ Female ┆ Fri │
│ 7.25       ┆ 1.0  ┆ Female ┆ Sat │
│ 12.6       ┆ 1.0  ┆ Male   ┆ Sat │
│ 16.99      ┆ 1.01 ┆ Female ┆ Sun │
│ …          ┆ …    ┆ …      ┆ …   │
│ 28.17      ┆ 6.5  ┆ Female ┆ Sat │
│ 48.27      ┆ 6.73 ┆ Male   ┆ Sat │
│ 39.42      ┆ 7.58 ┆ Male   ┆ Sat │
│ 48.33      ┆ 9.0  ┆ Male   ┆ Sat │
│ 50.81      ┆ 10.0 ┆ Male   ┆ Sat │
└────────────┴──────┴────────┴─────┘
>>> 
```

Assigning our filtered data to a variable makes it easier to reuse and keeps our plotting code cleaner.

### Exercise 10

Show AI the `dinner_data` DataFrame and ask it to use the **plotnine** package to create a simple but nice looking plot and to save that plot as `myplot.png`. Run that code at the Console using copy/paste. If it fails, show AI the error and try again until it works. Double-click on `myplot.png`.

```{r method-chaining-and-plots-10}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

Your screen should look something like this:

###

```{r}
include_graphics("images/plot-1.png")
```

###

Our code:

```
p = (
    ggplot(dinner_data, aes(x = 'total_bill', y = 'tip', color = 'sex')) +
      geom_point() +
      theme_minimal() +
      labs(x= 'Total Bill', y= 'Tip', color = 'Sex')
)

p.save('myplot.png')
```

You code is probably different. And that is OK! The important skill is how to use AI to create beautiful graphics.


## Generative AI
###

Generative AI --- tools like [ChatGPT](https://chat.openai.com/), [Grok](https://x.ai/), [Claude](https://claude.ai/), [DeepSeek](https://www.deepseek.com/en) and so on --- are the future, of data science and everything else. The more you use these tools, the better off you will be. Unfortunately, the tools are changing so much that it is hard for a tutorial like this to stay up-to-date. This section provides some general advice and practice exercises.

### Exercise 1

Using any AI you like, ask it to write a one-sentence summary about the Python programming language. Copy the answer below.

```{r generative-ai-1}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 5)
```

###

Example answer:

```
Python is a high-level, interpreted programming language known for its simplicity, readability, and versatility, widely used for web development, data analysis, artificial intelligence, and scientific computing.
```

If you do not want to pay for an AI service, then you will probably need to have free accounts with several different services. That way, if one service cuts you off for the day, you can switch to another.

### Exercise 2

Run this in the Python Console:

```
tips.head()
```

CP/CR.

```{r generative-ai-2}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 10)
```

###

```
>>> tips.head()
shape: (5, 7)
┌────────────┬──────┬────────┬────────┬─────┬────────┬──────┐
│ total_bill ┆ tip  ┆ sex    ┆ smoker ┆ day ┆ time   ┆ size │
│ ---        ┆ ---  ┆ ---    ┆ ---    ┆ --- ┆ ---    ┆ ---  │
│ f64        ┆ f64  ┆ str    ┆ str    ┆ str ┆ str    ┆ i64  │
╞════════════╪══════╪════════╪════════╪═════╪════════╪══════╡
│ 16.99      ┆ 1.01 ┆ Female ┆ No     ┆ Sun ┆ Dinner ┆ 2    │
│ 10.34      ┆ 1.66 ┆ Male   ┆ No     ┆ Sun ┆ Dinner ┆ 3    │
│ 21.01      ┆ 3.5  ┆ Male   ┆ No     ┆ Sun ┆ Dinner ┆ 3    │
│ 23.68      ┆ 3.31 ┆ Male   ┆ No     ┆ Sun ┆ Dinner ┆ 2    │
│ 24.59      ┆ 3.61 ┆ Female ┆ No     ┆ Sun ┆ Dinner ┆ 4    │
└────────────┴──────┴────────┴────────┴─────┴────────┴──────┘
>>> 
```

When working with AI, you often need to tell it about the dataset. The easiest way to do that is often to just copy/paste the first few rows. That shows the AI what the column names and types are, which is key information for creating plots and data pipelines.

### Exercise 3

Copy/paste the top of the `tips` DataFrame into your AI interface and ask it to create a chain of methods using **polars** that calculates the average `tip` for each `sex`. Run the provided code in the Console. If it fails, show the AI the error and ask for better code.

CP/CR.

```{r generative-ai-3}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 10)
```

###

Claude gave us this answer:

```
>>> (
    tips
    .group_by("sex")
    .agg(pl.col("tip").mean())
)
shape: (2, 2)
┌────────┬──────────┐
│ sex    ┆ tip      │
│ ---    ┆ ---      │
│ str    ┆ f64      │
╞════════╪══════════╡
│ Male   ┆ 3.089618 │
│ Female ┆ 2.833448 │
└────────┴──────────┘
>>> 
```

There are two differences between this code and the code we have been writing. First, the entire statement is within parantheses. Second, the methods are placed on separate lines, which makes readability easier. Using parantheses allows us to place methods on separate lines.

###

This is a great answer! It uses `group_by()` just like R's tidyverse, and then `agg()` (short for aggregate) with `pl.col()` to specify which column to calculate the mean for. 

Using AI is good. But intelligent use --- use in which you understand what the AI has done and try to improve/clarify its answer --- is even better.

### Exercise 4

Ask AI to create a beautiful plot using the `tips` dataset and the **plotnine** library and to save the resulting plot as `myplot.png`. Run the provided code in the Console. If it fails, show the AI the error and ask for better code.

CP/CR.

```{r generative-ai-4}
question_text(NULL,
answer(NULL, correct = TRUE),
allow_retry = TRUE,
try_again_button = "Edit Answer",
incorrect = NULL,
rows = 20)
```

###

Example code from Claude:

```
p = (
    ggplot(tips, aes(x="total_bill", y="tip", color="sex")) +
    geom_point(alpha=0.7, size=3) +
    geom_smooth(method="lm", se=False) +
    facet_wrap("~time") +
    labs(
        title="Tipping Patterns by Total Bill",
        subtitle="Comparing lunch and dinner service",
        x="Total Bill ($)",
        y="Tip ($)",
        color="Sex"
    ) +
    theme_minimal() +
    theme(
        figure_size=(10, 5),
        plot_title=element_text(size=14, weight="bold"),
        plot_subtitle=element_text(size=10, color="gray")
    )
)
p.save("output_plot.png")
```

It is convenient to have the new plot saved as the same name as the plot which is already open. In this case, the new plot will just replace the old plot without you having to double-click anything. Of course, you can't use that trick if you want to save both plots.


## Summary
###

This tutorial introduced you to the Python language for data science. You learned how to work with datasets using **polars**, a fast DataFrame library with syntax very similar to R's tidyverse. You learned how to chain operations using method chaining (just like R's pipe operator `|>`), and how to make plots using **plotnine**.

The key advantage of **polars** is that it feels familiar to R users while providing the speed and ecosystem of Python. Functions like `filter()`, `select()`, `sort()`, and `group_by()` work almost identically to their R counterparts!

```{r download-answers, child = system.file("child_documents/download_answers.Rmd", package = "tutorial.helpers")}
```

```{js change-code, echo = FALSE}

function transfer_code(elem){
  Shiny.setInputValue("js_to_server", elem.previousElementSibling.dataset.label);
  
}

Shiny.addCustomMessageHandler('set-exercise-code', function(x) {
  var el = $(`.tutorial-exercise[data-label="${x.label}"] .tutorial-exercise-code-editor`)
  var editor = ace.edit($(el).attr('id'));
  editor.getSession().setValue(x.code);
  Shiny.setInputValue("js_to_server", null);
})
```

```{r observe-transfers, context = "server"}

observeEvent(input$js_to_server, {
  ex_next = input$js_to_server
  ex_next_label_data = strsplit(ex_next, "-")[[1]]
  ex_next_num = tail(ex_next_label_data, n = 1)
  str_head = ex_next_label_data |>
    head(n = length(ex_next_label_data) - 1) |>
    paste0(collapse = "-")
  str_num = as.integer(gsub("[^0-9]", "", ex_next_num)) - 1
  ex_prev = paste0(str_head, "-", str_num)
  code <- learnr::get_tutorial_state(ex_prev, session)$answer |> 
    trimws()
  session$sendCustomMessage("set-exercise-code", list(label = ex_next, code = code))
})
```

---
title: "Functions for Working with Tutorials"
author: David Kane
format: html
vignette: >
  %\VignetteEngine{quarto::html}
  %\VignetteIndexEntry{Functions for Working with Tutorials}
  %\VignetteEncoding{UTF-8}
---



## Overview

We have provided two functions which help in the writing of tutorials: `make_exercise()` and `check_current_tutorial()`. The first creates a new exercise in the open RMD at the current location of your cursor.  The key argument of `make_exercise()` is `type`, which is `"code"` by default and produces a coding exercise. The other allowed values of `type` are `"yes-answer"` and `"no-answer"`. `check_current_tutorial()` reformats the entire open tutorial, mainly ordering exercises correctly and ensuring that chunk labels are correct.

When using RStudio, we have mapped variations of these functions to four RStudio addins: "Tutorial Code Exercise", "Tutorial Written Exercise (no Answer)", "Tutorial Written Exercise (with Answer)" and "Format Tutorial Chunk Labels." The first three each insert the skeleton for the next exercise in a tutorial, featuring all the key component parts. We even take a guess at the correct exercise number.  The "Format Tutorial Chunk Labels" addin is the most useful. Always run it before testing your tutorial. It ensures that all the exercises are sequentially numbered and that all the code chunk names are correct and unique.

You can find the addins in the "Addins" tab on the top Rstudio toolbar. Make sure that your cursor is located at the point in your Rmd at which you want to insert the new exercise. 

Make sure that your cursor is located in the correct location.

### Tutorial Code Exercise

`make_exercise()`, by default, it produces a new code exercise skeleton with an exercise title and with auto-generated code chunk labels. The Topic, of which this Exercise is a part, is titled "Plotting".


````{verbatim}
### Exercise 7


```{r plotting-7, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r plotting-7-hint, eval = FALSE}

```

```{r plotting-7-test, include = FALSE}

```

###
````

There are three code chunks included by default: exercise, hint and test. You do not need to include all three, although in most cases you should. The exercise code chunk is the location in which students will type their answers. The hint code chunk will be displayed to students when they click on the Hint button. The test code chunk should include the correct answer, so that you can be sure it will work for students when they enter it in the exercise code chunk.

See [our advice](https://ppbds.github.io/tutorial.helpers/articles/instructions.html#exercises) about writing good tutorial code exercises. First, start with the correct answer, the code you want students to submit. Place that code in the test chunk at the end. Our test process will ensure that this code will work for the students. Second, copy/paste the correct answer into the hint chunk in the middle. Then replace one or more of the key words or function arguments or argument values with `...`. Third, add your actual question to the top of the skeleton. What instructions will cause students to enter the correct answer in the exercise chunk? Fourth, drop some knowledge after the `###` at the end. Each exercise is an opportunity to teach. Make use of it!


### Tutorial Written Exercise (without and with answers)

Both create similar exercise structures with an auto-generated code chunk id and an exercise number. The difference is that the `question_text()` options are filled in differently.

#### Written Exercise (no Answer):

`make_exercise(type = 'no-answer')` is the R function to run, if you are using Positron.

````{verbatim}
### Exercise 9

```{r plotting-9}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###
````

The "no Answer" Exercises are usually used for confirmation that a student has completed a specified task. In that case, there is no need for us to supply a correct answer. And we can allow students to edit their submissions.

#### Written Exercise (with Answer):

`make_exercise(type = 'yes-answer')` is the R function to run, if you are using Positron.

````{verbatim}
### Exercise 8

```{r plotting-8}
question_text(NULL,
	message = "Place correct answer here.",
	answer(NULL, correct = TRUE),
	allow_retry = FALSE,
	incorrect = NULL,
	rows = 6)
```

###
````

The "with Answer" Exercises require the tutorial author to provide an (excellent!) answer to the question. This is harder than it looks, especially for questions without a single "right" answer. But it is also a rare opportunity since students will usually study the supplied answer quite closely. They want to check that their answer matches. We can't allow students to edit their answers to these questions since they might (misuse) that option to just copy/paste/modify our supplied answer.

The **learnr** package does not allow for hints to written exercises. We could add a test chunk, but that rarely makes sense for a written exercise. The main counter-example is a question which asks students to run an R command from the Console and then copy the command and the result of the command as their answer. Even though they are not submitting simple code as their answer, it can make sense to have a test chunk which executes the same R code which you are asking them to run, just to ensure that it will work for them.

### Format Tutorial Chunk Labels

`check_current_tutorial()` is the R function to run, if you are using Positron.

We often need to add a new exercise in the middle of a collection of other exercises. Or, we want to delete one exercise from the middle of the collection. In either scenario, our exercises are now mis-numbered. We either have two Exercise 5's or we go straight from Exercise 4 to Exercise 6. We want to renumber all the remaining exercises so that there are no duplicates or missing numbers.

The "Format Tutorial Chunk Labels" addin accomplishes this renumbering. But it also does more, changing all the code chunk names to be consistent with the new exercise numbers. Finally, it ensures that all code chunk labels follow our standard: begin with (up to 30 characters from) the topic title, remove special characters, replace spaces with dashes, and make all letters lowercase.

Since the code chunk labels (derived from the title of the topic in which the exercise resides) have a hard cutoff at 30 characters, try to make sure that your topic titles are different somewhere in the first 30 characters (including spaces) within a given tutorial. If not, the tutorial will not run since unique code chunk labels are required.

---
title: "Testing Your Package of Tutorials"
author: David Kane
format: html
vignette: >
  %\VignetteEngine{quarto::html}
  %\VignetteIndexEntry{Testing Your Package of Tutorials}
  %\VignetteEncoding{UTF-8}
---

This vignette demonstrates how to use the **tutorial.helpers** package to "test" the tutorials in your package. I place "test" in quotes because the testing is not very extensive. We are merely ensuring that your tutorials can be knitted without error. I *think* that this guarantees that the tutorials will run when students run `learnr::run_tutorial()`, but I am not certain. This testing certainly does nothing to ensure that the substance of your tutorial is correct.

If you are using the [**testthat**](https://testthat.r-lib.org/) framework for testing, the `tests` directory would have a file called `testthat.R` which just contains:

````
library(testthat)
library(your.package)

test_check("your.package")
````

Note how the name of **your.package** is not quoted with `library()` but is quoted with `test_check()`.

Within the `tests/testthat` directory there will be a variety of testing scripts. Create a file called `test-tutorials.R`. (The file can be named whatever you want, consistent with **testthat** requirements.) It might contain:

````
tut_paths <- tutorial.helpers::return_tutorial_paths("your.package")

test_that("All tutorials can be knit without error", {
  expect_null(
    tutorial.helpers::knit_tutorials(tut_paths)
  )
})


test_that("All tutorials have the expected components", {
  expect_null(
    tutorial.helpers::check_tutorial_defaults(tut_paths)
  )
})
````

The first step in testing the tutorials in your package is to determine the full paths to all those tutorials. The `return_tutorial_paths()` returns a vector of those paths.

The second step is to confirm that all your tutorials knit without error. `knit_tutorials()`, perhaps the most useful function in the entire package, accomplishes this. If a tutorial does not knit, an error is generated and the test fails.

The third step is only relevant for tutorial creators who follow [our advice](ai.html) concerning tutorial construction. In particular, `check_tutorial_defaults()` ensures that, somewhere in each tutorial, you have included the required child documents: the `info_section` and `download_answers` chunks.

`check_tutorial_defaults()` also fails if you do not have `library(learnr)` and `library(tutorial.helpers)` in your tutorial. 

Both `knit_tutorials()` and `check_tutorial_defaults()` return `NULL`, which is why we use `expect_null()`.

---
title: "Tutorials in the Age of AI"
author: David Kane
format: html
vignette: >
  %\VignetteEngine{quarto::html}
  %\VignetteIndexEntry{Tutorials in the Age of AI}
  %\VignetteEncoding{UTF-8}
---

### Introduction

AI changes everything.

This document describes the best way to write R tutorials which embrace the student use of AI. Prior to the rise of AI, [this](https://web.archive.org/web/20251008195459/https://ppbds.github.io/tutorial.helpers/articles/instructions.html) [was](https://web.archive.org/web/20251008195807/https://ppbds.github.io/tutorial.helpers/articles/books.html) the best way to write tutorials. Our purpose is no longer to teach students how to code. 

> *Our purpose is to teach students how to use AI to create.* 

The [**tutorial.helpers**}(https://ppbds.github.io/tutorial.helpers/) package will help you to create good tutorials. Make sure that you are using the latest development version. Run `library(tutorial.helpers)`

### Overview

To create a new tutorial, you first need a new directory, located in the `inst/tutorials` directory of your package. Once you create

```
> rmarkdown::draft("tutorial.Rmd", 
                  template = "tutorial_template", 
                  package = "tutorial.helpers",
                  edit = TRUE)
```


Students always need more practice working in a Quarto document (the QMD) and the Console at the same time. Good data scientists go back and forth between these two modes, writing something in the QMD, executing it in the Console, editing the QMD, executing again, and so on. We need to force students to do that more often.

AI tutorials begin with an *Introduction* which provides a summary of the key packages/functions which the tutorial will cover. The Introduction continues with a series of exercises which set up the repo/project/QMD in which most of the tutorial will be completed.

After the Introduction, there are 1 to 3 sections --- the official **learnr** nomenclature is "Topics" --- which are the meat of the tutorial. 

The last section is the *Summary*. It starts with the same overview with which the Introduction began, but in the past tense. It then has a couple exercises which finish up the tutorial by using `quarto publish gh-pages analysis.qmd` to create a webpage featuring the cool plots which the student has created. The URL for this new webpage is usually the answer to the last exercise in Summary, thereby completing the tutorial.

Anytime you ask a student to execute something in the Console, you confirm that they have done so with CP/CR, the abbreviation to **C**opy/**P**aste the **C**ommand/**R**esponse.

The [**tutorial.helpers**}() package will help you to create good tutorials. Make sure that you are using the latest development version. The package includes two important functions: `make_exercise()` and `check_current_tutorial()`. Use `make_exercise()` to add a new exercise to the current tutorial. It will number the exercise, and the code chunk labels, automatically. Use `check_current_tutorial()` to renumber all the exercises and relabel the code chunks. This is especially useful if you add or delete an exercise in the middle of a section.

### Tutorial Introduction

Students need some background in order to complete these sorts of tutorials, some familiarity with R, GitHub and so on. Students should have completed the "Getting Started" and "Introduction to R" tutorials in the [**tutorial.helpers**](https://CRAN.R-project.org/package=tutorial.helpers) package. They should also complete the first four tutorials (i.e., through "Positron and GitHub Introduction") from [**positron.tutorials**](https://ppbds.github.io/positron.tutorials/).

Always begin by having students set up a repo and Quarto document to work in. Again, you must replace `XX` with something sensible and usually different from the other `XX`'s. So, in the below example, the name of the repo and the title of the QMD will be different even though we hold their places with `XX` in both cases.

````{verbatim}
### Exercise 1

Create a Github repo called `XX`. Make sure to click the "Add a README file" check box.

Connect the repo to a project on your computer using `File -> New Folder from Git ...`.  Make sure to select the "Open in a new window" box. 

You need two Positon windows: this one for running the tutorial and the one you just created for writing your code and interacting with the Console.

In the new window, select `File -> New File -> Quarto Document ...`. Provide a title -- `"XX"` -- and an author (you). Render the document and save it as `analysis.qmd`.

Create a `.gitignore` file with `analysis_files` on the first line and then a blank line. Save and push.

In the Console, run:

```         
show_file(".gitignore")
```

If that fails, it is probably because you have not yet loaded `library(tutorial.helpers)` in the Console.

CP/CR.

```{r introduction-1}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

### 

<!-- XX: Insert a knowledge drop related to this project. -->
````

Feel free to copy/paste this question as-is, but replacing `XX` with whatever makes sense for your assignment. That is, you need to provide your own repo name, Quarto document title and so on. The repo name should be descriptive and also not likely to have conflicts with other repos in the students GitHub account, i.e., `golf-scores` not `project-1`. 

You do not need to use `analysis.qmd` as the name of the QMD file which the student creates. But using the same name doesn't hurt anything and is convenient since it decreases the number of things which the tutorial author needs to change.

You are, obviously, responsible for adding a knowledge drop which teaches the students something about the larger topic.

The second question in the Introduction is often:

````{verbatim}
### Exercise 2

In your QMD, put `library(tidyverse)` in a new code chunk. Render the file using `Cmd/Ctrl + Shift + K`.

Notice that the file does not look good because the code is visible and there are annoying messages. To take care of this, add `#| message: false` to remove all the messages in this `setup` chunk. Also, add the following to the YAML header to remove all code echos from the HTML:

```         
execute: 
  echo: false
```

In the Console, run:

```         
show_file("analysis.qmd", chunk = "Last")
```

CP/CR.

```{r introduction-2}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 6)
```

### 

<!-- XX: Insert a knowledge drop related to this project. -->
````

The third question generally loads the **tidyverse** library into the Console:

````{verbatim}
### Exercise 3

Place your cursor in the QMD file on the `library(tidyverse)` line. Use `Cmd/Ctrl + Enter` to execute that line.

Note that this causes `library(tidyverse)` to be copied down to the Console and then executed. 

CP/CR.

```{r introduction-3}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 8)
```

###

<!-- XX: Insert a knowledge drop related to this project. -->
````

If the target audience for the tutorial is more experienced, you can be less didactic, leaving out several of these instructions. You could also add more steps, like loading more libraries at once.

I recommend offering these explicit instructions in every tutorial. First, students need lots of practice. Second, each time you tell them to add something to the QMD, you give yourself an opportunity for a knowledge drop. The same applies when you tell students to execute, in the Console, a new addition to the QMD. 

Consider an example of creating an object in the QMD:


````{verbatim}
### Exercise 12

Create a new code chunk in your QMD. Add a code chunk option: `#| cache: true`. Copy/paste the R code for the final model into the code chunk, assigning the result to `fit_XX`. 

`Cmd/Ctrl + Shift + K`. It may take some time to render your QMD, depending on how complex your model is. But, by including `#| cache: true` you cause Quarto to cache the results of the chunk. The next time you render your QMD, as long as you have not changed the code, Quarto will just load up the saved object.

At the Console, run:

```
tutorial.helpers::show_file("XX.qmd", chunk = "Last")
```

CP/CR.

```{r courage-12}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 8)
```

### 

To confirm, `Cmd/Ctrl + Shift + K` again. It should be quick.
````

Once we have create the object in the QMD, we can execute the same code in the Console:

````{verbatim}
### Exercise 13

Place your cursor in the QMD file on the `fit_XX` line. Use `Cmd/Ctrl + Enter` to execute that line. 

At the Console, run `ls()`. CP/CR.

```{r courage-13}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 8)
```

### 

<!-- XX: Insert a knowledge drop related to this project. -->
````

Of course, you can combine these last two questions together, but you probably shouldn't. We want easier questions, not harder. Splitting up questions makes them simpler. Perhaps more importantly, more questions provide more opportunities for knowledge drops.

### Walking the path

To teach students about topic X, we first need to decide the final destination. What do we want students to be able to do on their own after completing the tutorial? For us, this will almost always be a plot, and often several plots. Having envisioned this goal, we need to create a "path" which students can use to reach that goal, first under our supervision and, second, on their own. The path will consist of several stepping stones, or stops along the way.


### Providing Answers

To ensure that students are on the right path with their code, we need to provide them with our code that is verified to be correct. Although we should not tell students to replace their code with ours at every step, if a student is lost, they should be able to refer to our code to get back on track. Also, when we are editing the tutorials, we want to easily run the code in the Console or render it. With this goal in mind, we will add R chunks to provide our code to students.

````{verbatim}
### Exercise 6

Using your favorite AI, prompt it to generate R code that ... Add the code to your QMD in a new chunk. Place your cursor on the first line of the code and run `Cmd/Ctrl + Enter`.

In the Console, run:

```         
show_file("analysis.qmd", chunk = "Last")
```

CP/CR.

```{r something-1}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 4)
```

###

```{r, echo = TRUE}

```

### 

<!-- XX: Insert a knowledge drop related to this project. -->
````

The R chunk with `echo = TRUE` allows the students to see the code we have written within it. This makes it easy for the students to copy and paste our code if they need to. 

Since `eval = TRUE` is the default argument in an R chunk, the code within will automatically be run as well. Students will be able to see any output from the code, which can be helpful if the code plots a graph. If it is inappropriate or unnecessary to include the output of the code, just set `eval = FALSE` explicitly.

In that case, the answer chunk would look like:

````{verbatim}
```{r, echo = TRUE, eval = FALSE}

```
````

### Plotting Questions

Plotting exercises are generally handled with a sequence of four questions. Prior to these, the tutorial will probably have the student practice gathering, organizing, and cleaning the data. 

The first of the three prior questions tells the student to replace the current pipe which they have in the QMD with our code. We check that they have done so with `show_file()`. The purpose of this question is to ensure that the student's data will match our data.

````{verbatim}
### Exercise 8

Before creating a plot, we need to ensure that your data matches our data. In the QMD, replace your code from the previous exercise with our code.

In the Console, run:

```         
show_file("analysis.qmd", chunk = "Last")
```

CP/CR.

```{r ai-usage-8}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 6)
```

###

<!-- XX: Insert a knowledge drop related to this project. -->
````

Note that the numbering of theses questions is arbitrary. Just run `check_current_tutorial()` to fix it.

The second question tells the student to, in the QMD, assign the result of the pipe to a new variable, often `x`. We then tell the student to `Cmd/Ctrl + Enter` this code so that the workspace includes a copy of `x`. 

````{verbatim}
### Exercise 9

Within the latest code chunk, add the option: `#| cache: true`. Assign the result of the pipe to `x`. 

`Cmd/Ctrl + Shift + K`. By including `#| cache: true` you cause Quarto to cache the results of the chunk. The next time you render your QMD, as long as you have not changed the code, Quarto will just load up the saved object.

If you have not done so already, you should add `analysis_cache` to the `.gitginore`. The content of the cache file does not belong on GitHub.

Place your cursor on the line where the pipe is assigned to `x`, run `Cmd/Ctrl + Enter`, thus ensuring that the workspace also includes a copy of `x`.

CP/CR.

```{r ai-usage-9}
question_text(NULL,
    answer(NULL, correct = TRUE),
    allow_retry = TRUE,
    try_again_button = "Edit Answer",
    incorrect = NULL,
    rows = 8)
```

###

Our code:

```{r, echo = TRUE}
x <- ...
```

<!-- XX: Insert a knowledge drop related to this project. -->
````

Note that we need `x` to be created in the QMD, not just in the Console, because later chunks will use `x` to create the plot.

The third question tells the student to type `x` in the Console, followed by "CP/CR." The purpose is both to have the student look at the tibble and also to set the stage for the actual graphics question. In defining `x`, you should probably require that the students keep only a reasonable number of variables.

````{verbatim}
### Exercise 10

Within the Console, type `x`, which we previously assigned to a pipe and ran in the Console. Hit `Enter`.

CP/CR.

```{r ai-usage-10}
question_text(NULL,
    answer(NULL, correct = TRUE),
    allow_retry = TRUE,
    try_again_button = "Edit Answer",
    incorrect = NULL,
    rows = 8)
```

###

Our code:

```{r, echo=TRUE}
x
...
```

<!-- XX: Insert a knowledge drop related to this project. -->
````

Could all four of these questions be combined into one? Probably. But spreading things has two advantages. First, it ensures that even the weaker students do not get lost. Second, it provides us with four opportunities to drop some knowledge. 

Now, we can move on to the plotting question. In the age of AI, students will have AI write code for their plot. They will do that while specifying that their data is `x` from earlier. The student will add their new code to a new code cell, and we check that they have done so with `show_file()`. The purpose of this question is to ensure that the student has generated their own code, albeit with AI help.

````{verbatim}
### Exercise 11

Ask AI to generate R code that uses `x` to plot a basic graph showing XX... Mention you want to use the data from `x` and copy/paste the `x` you ran in the Console with the resulting tibble. You only need the top 3 lines, mainly to include column names.

Within `labs()`, edit or add a proper title, subtitle, and caption. If axis labels would be useful, add them, but if unnecessary, don't bother. Don't assign the code for the plot to any variable. Put the plot code in a new code chunk. Run `Cmd/Ctrl + Shift + K` to ensure that everything works. Make your plot look nice.

In the Console, run:

```         
show_file("analysis.qmd", chunk = "Last")
```

CP/CR.

```{r ai-usage-11}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 12)
```

###

Our code:

```{r, echo=TRUE}
  ...
```

###

<!-- XX: Make sure your plotting code is good! This will take some time. You had better have a subtitle which provides the take-away message of the plot. AI sometimes gives you too much code, lots of `theme()` stuff and so on. This is no good! In most cases, we are happy with concise, straightforward code. Insert a knowledge drop related to this project. -->
````

More to come . . .


<!-- https://pkgs.rstudio.com/learnr/articles/exercises.html#exercise-setup-chunks

DK: The section after a plot is interesting. 

Our question is always a show_file().

Our answer is always an R code chunk.

One, we show our code. But our code should be good, often better than what the AI gives as default. It is clean, concise, does not include tons of junk. We don't just use the raw AI. We always ensure that there is a subtitle and caption which provides the data source. Et cetera.

 -->

---
title: Getting Started
author: David Kane
tutorial:
  id: getting-started
output:
  learnr::tutorial:
    progressive: yes
    allow_skip: yes
runtime: shiny_prerendered
description: Tutorial for introducing students to tutorials when using Positron.
---

```{r setup, include = FALSE}
library(learnr)
library(tutorial.helpers)
library(knitr)

# Needed because fs::path_home() seems to be the best thing to use to identify the core home
# directory on Windows, and it works on the Mac as well. ?fs::path_home gives lots of details
# and discussion.

library(fs)

knitr::opts_chunk$set(echo = FALSE)
knitr::opts_chunk$set(out.width = '90%')
options(tutorial.exercise.timelimit = 60, 
        tutorial.storage = "local") 
```

```{r copy-code-chunk, child = system.file("child_documents/copy_button.Rmd", package = "tutorial.helpers")}
```

<!-- Find the Positron window naming page from the Posit docs. Use that image: positron-interface.jpeg  -->

<!-- Revist the initial instructions once you can use Positron in the cloud. Perhaps add a new vignette discussing Posit Workbench. -->

<!-- Get a Windows machine and show Windows versions of all images. -->

## Introduction
### 

This tutorial is an introduction, for students using [Positron](https://positron.posit.co/), to the general use of R tutorials created with the [**learnr**](https://rstudio.github.io/learnr/) package. It is meant for beginning students who have never completed a tutorial. 

Most students come to this tutorial after having completed the set-up steps outlined [here](https://ppbds.github.io/primer/getting-started.html#summary). If you fail to do so, some of these questions will be confusing.

### 

After you started the tutorial, your Positron window looked something like this:

```{r}
include_graphics("images/run-tutorial.png")
```

Note the red square at the top of the Console menu. This indicates that the Console is "busy." You can't do anything with this Console unless you stop running the tutorial. You can complete the tutorial in the Viewer pane in the Secondary Side Bar. 

###

Or you can "pop" a tutorial up into your browser by pressing the "Open the current URL in the default browser" button --- the item in the Viewer menu just to the right of the circular arrow. It looks like a rectangle containing an arrow pointing up and to the right.

```{r}
include_graphics("images/send-tutorial-to-browser.png")
```

You can go back and forth between the Viewer tab and the browser. The tutorial is easier to read in the browser, but many students prefer to just keep everything inside of the Positron window.

###

You complete a tutorial on your computer. You download a copy of your answers to the `Downloads` folder on your computer. Then, you submit your answers to your instructor and/or keep a copy for yourself.

### 

Most tutorials begin by asking for your name and email. (Your instructor might also request that you fill in the "ID" field so that your work can be matched to a grading database. If not, leave the "ID" field blank.) Click "Continue" and then fill in your name and email. Click the **"Submit Answer"** button each time. If you make a mistake, you can click **"Try Again"** and, uh, try again.

If you are completing this tutorial as part of a class, you should use the same email address for every tutorial. It is best to use your "permanent" email address, one that you will always have access to. Do not use an email address associated with your current school or employer since you are likely to lose that email address some day.

```{r info-section, child = system.file("child_documents/info_section.Rmd", package = "tutorial.helpers")}
```

## Exercises
### 

Each tutorial includes exercises for you to complete. For non-coding questions, you write your answer and then hit the "Submit Answer" button, as you did above for your name and email address. 

### Exercise 1

Let's practice some simple math. Type `4 + 5` into the box below. Hit **"Run Code"** to see your results. Clicking **"Hint"** gives you a hint about what to type. 

```{r exercises-1, exercise = TRUE}

```

```{r exercises-1-hint-1, eval = FALSE}
4 + 5
```

### 

Once you click "Run Code," the answer `[1] 9` should appear below the code you entered. This is the same behavior you would see if you ran this code in the Console, which is where R is run.

###

There are two types of answers for tutorial questions: prose and code. With prose answers, you will use the "Submit Answer" button, as you did with your name and email. With code answers, you use the "Run Code" button, as you did for this question.

### Exercise 2

Let's use variables to perform the same calculation. Below, we have already assigned 5 to `x` and 4 to `y`. On the line below those two provided lines of code, add the two variables by typing `x + y`. Hit "Run Code" to see your results. 

```{r exercises-2, exercise = TRUE}
x <- 5
y <- 4

```

```{r exercises-2-hint-1, eval = FALSE}
x <- 5
y <- 4
x + y
```

### 

Once you click "Run Code", the answer `[1] 9` should appear below the code you entered. 

### 

If you don't "Run Code" (or "Submit Answer") your response, nothing is recorded.

### Exercise 3

We will often need to copy code from the previous exercise to continue working on a multi-step problem. To make this easier, we provide a "Copy previous code" button. Below the following box, click on the "Copy previous code" button. 

```{r exercises-3, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r exercises-3-hint-1, eval = FALSE}
x <- ...
y <- 10
x + ...
```

### 

This should paste all work from the previous exercise into the above exercise code area. Note that this will only copy code which has already been run in the previous exercise.

### 

Change the values to `x <- 6` and `y <- 10`. Hit "Run Code." You should get the answer `[1] 16`.

### 

Note that the assignment symbol, `<-`, is a less-than-sign followed by a dash. We often pronounce this symbol as "gets." So, we might say, when reading this code, that "x gets 6."

### 

Did you remember to hit the "Run Code" button above? Lots of people forget! You must hit the button for your answer to be recorded.

### Exercise 4

<!-- DK: Seems like it should be possible to restart the R session. But the various buttons don't seem to work very well, at least when you are in a tutorial. Try this again in a month. In the meantime, this radical close-Positron approach is OK. -->

Read all of the following instructions before continuing. The purpose of this exercise is to demonstrate that your tutorial answers are saved automatically.

* Quit and then restart Positron.

* Restart the tutorial. The most common approach is to run this command at the Console:

````
learnr::run_tutorial("getting-started", package = "tutorial.helpers")
````

* Copy the above code now so that you can easily paste it into the Console after you restart Positron.

Now that you have read all the instructions, go ahead and follow them.

### 

Note how all your answers to previous questions have been saved. It is not a problem to quit a tutorial. In fact, if you ignore a tutorial, it will eventually quit automatically.

### Exercise 5

Click `File -> New Window` in the Positron menu.

```{r}
include_graphics("images/new-window.png")
```

This brings up a new instance of Positron, completely independent of this one. This is the most common set up when working on tutorials. One Positron window, this one, runs the tutorial, but we don't do any work in it. We generally don't even bother to create a second Console. The second Positron window is where we do the work which the tutorial requires.

###

There are many ways to move between Positron windows. One approach is to hold down the Control key and hit the letter "W." This is usually abbreviated as `Ctrl + w`. This pulls up a list of the open Positron windows. You then select the one you want and hit `Enter`. On the Mac, you can hold down the Command key and then hit the backtick key: `` ` ``, i.e., ``Cmd + ` ``. This "cycles through" the open Positron windows. `Alt + Tab` on Windows has a similar effect, although the cycle will include all open applications, not just Positron windows.

<!-- 
DK: Not sure on the best way to have students handle this. Right now, we report the facts above and leave it at that. But we could guide them through the creation of a shortcut key which would be the same (?) for both Mac and Windows. Maybe `workbench.action.quickSwitchWindow` is the key command, and we should set that to a specific command sequence for everyone. This would involve, I think, editing/creating the `keybindings.json` file. 

{
  "key": "ctrl+alt+w",
  "command": "workbench.action.quickSwitchWindow"
}

But need to worry about how this might interact, eventually, with doing Positron in the cloud. Would need to pick a sequence which is not pre-empted by hr browser.
-->

###

In the Console in the new Positron window, type `version[['version.string']]`, hitting `return` (if using Mac) or `enter` (if using Windows). To avoid us having to type `return` or `enter` for each question we ask, we use the term `Enter` to specify hitting either the `return` or `enter` key, as appropriate.


**Do not just type the command into the answer box!** You must paste the command into the Console and then execute the command.

Copy/paste the command **and** the return value into this box.

```{r exercises-5}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

### 

We typically provide example answers (like the answer below) after you submit your own response. Don't be concerned if your submitted answer differs from the example answer we show.

```{r exercises-5-test, echo = TRUE}
version[['version.string']]
```

Your version of R may be more recent, but it should not be older. It is OK if your formatting differs from ours. Exactness does not matter. What matters is that you prove that you have executed the command as instructed.

### 

Tutorial workflow will often look like this. We tell you to run a command **in the Console**, always in your second Positron window. You run it. We then ask you to copy/paste, from the Console, both the command itself and the response returned. Because we ask you to **C**opy/**P**aste the **C**ommand/**R**esponse so often, we usually shorten this instruction to CP/CR.

### Exercise 6

The Terminal tab is next to the Console tab within the Panel portion the Positron window. Click on the word “Terminal” to access it. We use the Console tab to talk to R. We use the Terminal tab to talk directly with the computer.

```{r}
include_graphics("images/terminal-1.png")
```

###

At the Terminal, type `git --version`. Hit `Enter`. (Your computer may respond by asking if you want to install Git. You do. Go ahead and install.)

CP/CR.

```{r exercises-6}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

My answer looks like this:

````
dkane@macbook positron.tutorials % git --version
git version 2.39.3 (Apple Git-146)
dkane@macbook positron.tutorials % 
````

As long as you have some version of Git installed, you are OK. 

But many of you will get an error like:

````
command not found: git
````

Or:

````
The term 'git' is not recognized as the name of a cmdlet, function, script file, or operable program.
````

Either means that Git is not installed on your computer.

### Exercise 7

If you got an error in the previous question, follow [these instructions](https://git-scm.com/book/en/v2/Getting-Started-Installing-Git) to install Git, if it was not installed automatically. If asked, you should allow "Git BASH" to be installed. You will sometimes see "BASH" or "Bash" or "bash," but all refer to the same shell, a program which is much better than the Powershell which comes with Windows.

Next, quit Positron using the menu commands: `Positron -> Quit Positron`. This is the safest way to ensure that Positron "discovers" your new Git installation.

Start Positron again. This tutorial will have been closed when you quit Positron. Restart the tutorial with:

```
learnr::run_tutorial("getting-started", "tutorial.helpers")
```

You will need to `File -> New Window` again to have an empty Console to work in.

At the Terminal, type `git --version`. Hit `Enter`. 

CP/CR.

```{r exercises-7}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

If this works, you should be all set. If not, seek help. You need Git!

### Exercise 8

<!-- DK: Figure out what Windows does and/or looks like and then clarify these instructions. -->

The Console behaves almost identically regardless of the kind of computer you use. The Terminal, sadly, does not. Windows and Mac are very different. 

###

Every Terminal runs a “shell,” the program which serves as the interface between you and the computer.

###

On a Mac, the default shell is referred to as "[Unix-like](https://en.wikipedia.org/wiki/Unix-like)" since it acts very much like the traditional Unix command line. In Windows, the default shell is [PowerShell](https://en.wikipedia.org/wiki/PowerShell), which is an older technology. 

###

To start a new Terminal with a specific shell, press the downward arrow next to the `+` sign. 

```{r}
include_graphics("images/terminal-2.png")
```

See the red arrow above.

###

If you are using Windows, you should see an option for “Git Bash” (or maybe just “bash”). I don't, because I am not using Windows. But it doesn't matter, because I already have a Unix-like shell, `zsh`

###

If everything goes well, Git Bash should automatically be your default shell when you open the Terminal on Windows. However, if it doesn't default to Git Bash, you'll need to manually choose it. You can set it as the default option by using the "Select Default Profile" feature: 

```{r}
include_graphics("images/terminal-3.png")
```

Again, this discussion is specifically for Windows users only. They might need to click on that tool and choose the option labeled "Git Bash" or "bash." Once selected, this will make it the default terminal for future sessions.

###

Click the `+` button (see red arrow below) to start a second Terminal and see if it works!

```{r}
include_graphics("images/terminal-4.png")
```

The red rectangle highlights that we are now running two Terminals, between which we can switch back-and-forth. In fact, we can have multiple Terminals and multiple Consoles in one Positron window just by using the `+` button. Some people prefer this approach. We still recommend using a second Positron window to have an open Console for use while working on a tutorial.

###

From the Terminal, run `echo $0`. (To "run" a command at the Terminal or the Console means to type the command and then click the `Enter` key.)

CP/CR.

```{r exercises-8}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

On a Mac, I get:

````
dkane@MacBookPro ~ % echo $0 
/bin/zsh
dkane@MacBookPro ~ % 
````

This means that, in the Terminal, I am using [Z shell](https://en.wikipedia.org/wiki/Z_shell). On Windows, your answer will probably be something like `bash`, referring to [Bash](https://en.wikipedia.org/wiki/Bash_(Unix_shell)), another Unix-like shell. Don't worry about the differences between them. 

The important thing is that, first, you now have Git installed and, second, if you are a Windows user, you now have a better shell than Powershell.

If this works, you should be all set. If not, seek help. You need a decent shell in the Terminal.

### Exercise 9

Let us make the Terminal "sleep" by typing the command `sleep 5` at the Terminal. Hit `Enter`. The `sleep 5` command at the Terminal pauses execution for 5 seconds. 

CP/CR. 

```{r exercises-9}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

```
dkane@MacBookPro ~ % sleep 5
dkane@MacBookPro ~ % 
```

We can change the duration of the sleep using this format: `sleep [number]`. The number you put in is the duration (in seconds) for which the Terminal goes to sleep. 

### Exercise 10

Run `sleep 60` at the Terminal. (Again, "run" means to type the command and then hit the `Enter` key.) The command makes the Terminal go to sleep for 60 seconds. To cancel the running `sleep` command at the Terminal press `Ctrl + C` on your keyboard. Do so now, before `sleep 60` finishes.

When you press `Ctrl + C` while `sleep` is running, it immediately terminates the `sleep` command and returns you to the command prompt.

CP/CR.

```{r exercises-10}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

### 

````
dkane@MacBookPro ~ % sleep 60
^C
dkane@MacBookPro ~ % 
````

`Ctrl + C` can stop any process from the Terminal. This is often necessary.

### Exercise 11

Positron and your R installation have some default settings which are not good for new users. Run this command at the Console to change them:

```
tutorial.helpers::set_positron_settings()
```

CP/CR.

```{r exercises-11}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 12)
```

### 

The output from this function will report the changes, if any, which were made. To learn more, look up the help page by running `?tutorial.helpers::set_positron_settings` at the Console.

### 

You can generally call a function by just giving its name along with the ending parentheses, as with `set_positron_settings()`. However, it can be safer to preface the function name with the name of the package in which the function is found, separated by two colons. Since `set_positron_settings()` is in the **tutorial.helpers** package, we can call it using this "double colon" notation: `tutorial.helpers::set_positron_settings()`, as we did above.

### Exercise 12

It is useful to understand where the R process is currently executing on your computer.

In the Console, run `getwd()`. This function is **get** **w**orking **d**irectory, meaning we want to know the directory in which the R process is located.

CP/CR.

```{r exercises-12}
question_text(NULL,
    answer(NULL, correct = TRUE),
    allow_retry = TRUE,
    try_again_button = "Edit Answer",
    incorrect = NULL,
    rows = 3)
```

### 

If you are on a Mac, you answer probably looks like:

````
> getwd()
[1] "/Users/dkane"
>
````

This is especially true if you are new to Positron. If you have used Positron before, then it might start out in the last directory in which you were working. Otherwise, it defaults to your "home" directory.

###

If I were on Windows, then the home directory would be something like `C:\Users\dkane\Documents` or even  `C:\Users\dkane\OneDrive\Documents` or even one of these with `\\` instead of `\` to separate out the directories. Note that on Windows:

* The file system starts on the `C:` drive.

* The directories are separated with backward slashes, instead of the forward slashes which we see on the Mac.

* The notion of a "home" directory is tricky, both in general and when compared with a Mac. On the Mac, the "home" directory is almost always something like `/Users/dkane`. On Windows, the "home" directory generally refers to a user's `Documents` folder, which is sometimes housed on `OneDrive`.

* `OneDrive`, like many cloud-backup systems, does not play particularly well with Git, the program almost everyone uses to keep track of changes in your code. **Do not put your data science projects in a folder which is backed up with OneDrive.**

### Exercise 13

In the Console, run `fs::path_home()`. 

CP/CR.

```{r exercises-13}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

````
> fs::path_home()
/Users/dkane
````

This function returns the "path" to your home directory such that the Windows answer is more like the Mac answer. A path is simply the information which allows you to reach a location on your computer. 

More info on paths can be found [here](https://people.ischool.berkeley.edu/~kevin/unix-tutorial/section3.html).

If you are on Windows, you should have gotten a similar answer, something like:

````
C:\Users\dkane
````

Because it is Windows, the path begins with `C:` and the slashes are backwards slashes. 

The important point is that, in almost all non-pathological set-ups, you will have a home directory. It is important to know where that is.


### Exercise 14

Let's take a look at the directories in your home directory. From the Console, run `list.files(fs::path_home())`. CP/CR.

```{r exercises-14}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

### 

My answer:

```
> list.files(fs::path_home())
 [1] "AppData"      "Applications" "Desktop"      "Documents"   
 [5] "Downloads"    "Library"      "miniconda3"   "Movies"      
 [9] "Music"        "Pictures"     "Public"       "temp"        
>           
```

If you are on a Mac, then the `Desktop` directory is a handy location in which to store all your coding projects.  If you are on Windows, then the `Desktop` directory might be less convenient because it is often, but not always, backed up on OneDrive.

**Avoid doing data science in any directory which is backed up by OneDrive or any other cloud backup service.**

### Exercise 15

Whenever you start a new project, you will want to store it in a location which you can easily find. Right now, you are likely in an arbitrary location on your computer. It would be much more useful to have a location dedicated to your projects.

From the Console, run `file.exists(file.path(fs::path_home(), "Desktop"))`. 

CP/CR.

```{r exercises-15}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

### 

For me:

```
> file.exists(file.path(fs::path_home(), "Desktop"))
[1] TRUE
```

For almost all Mac users and many Windows users, this should return `TRUE` because there is a `Desktop` directory. For some Windows users, it will return `FALSE` because their `Desktop` directory is within a `OneDrive` directory which is automatically backed up to the cloud.

### Exercise 16

Let's make a folder, and call it `projects`. If `file.exists(file.path(fs::path_home(), "Desktop"))` returned `TRUE`, then run this at the Console:

```
dir.create(file.path(fs::path_home(), "Desktop", "projects"))
```

That should put your `projects` directory in your `Desktop` directory. (Note that "folder" and "directory" mean the same thing.) If you get a warning message, here or below, that this directory "already exists," then you are all set.

If `file.exists(file.path(fs::path_home(), "Desktop"))` returned `FALSE`, then run this at the Console:

```
dir.create(file.path(fs::path_home(), "projects"))
```

That will also create a `projects` directory, but in your home directory. 

CP/CR. 

```{r exercises-16}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

### 

If that command works, you are all set. (If you get a message that the directory already exists, then you are also all set.) If not, you will need to create a `projects` directory by hand. Helpful resources can be found [here](https://support.microsoft.com/en-us/office/create-a-new-folder-cbbfb6f5-59dd-4e5d-95f6-a12577952e17) for Windows, and [here](https://support.apple.com/guide/mac-help/organize-files-using-folders-mh26885/mac) for macOS.

### 

**The path to your data science projects should not include any folders with spaces (or other weird characters) in their names.** 

For example, this is a reasonable location for R projects if you are using a Mac:

```
/Users/dkane/Desktop/projects/
```

And if you are using Windows . . .

```
C:/Users/dkane/projects/
```

Again, you can place your `projects` directory where you want, as long as you keep track of where it is.

### Exercise 17

<!-- DK: Is this necessary, or will Positron by default use the last directory you did stuff from, which will be `projects`, when it gives you a choice. So, changing the default behavior is not necessary. -->

It is handy for Positron to know where your `projects` directory is so that it can offer that directory as a default when you start a new project. To do so, from the Console, run this command after replacing "PUT YOUR PROJECT PATH HERE" with your project path.

````
tutorial.helpers::set_positron_settings(positron_settings = 
   list("files.dialog.defaultPath" = "PUT YOUR PROJECT PATH HERE"))
````

So, for example, I would run:

````
tutorial.helpers::set_positron_settings(positron_settings = 
   list("files.dialog.defaultPath" = "/Users/dkane/Desktop/projects"))
````

If I were using Windows:

````
tutorial.helpers::set_positron_settings(positron_settings = 
   list("files.dialog.defaultPath" = "C:/Users/dkane/projects"))
````

But you should not run either of those! Those only work for me. You need to run the command with your `projects` path.

CP/CR.

```{r exercises-17}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

Look at the output from `tutorial.helpers::set_positron_settings()` to confirm that this worked. Don't worry if it did not.
 
### Exercise 18

Tutorials can include written responses. You have already seen several examples. Sometimes those written answers are just copies of the R commands and their results. Other times, we will ask you to write one or more sentences of prose. 

Copy and paste everything from the Exercise header above through the end of this sentence into the answer box below. That is, you are copying from this tutorial, not from the Console. Press "Submit Answer."

```{r exercises-18}
question_text(NULL,
    message = "We will sometimes provide an answer to written questions, as we do here. Other times, we won't. In the former case, we will not allow you to resubmit your answer. Don't stress! Most instructors grade tutorials on pass/fail basis, so, as long as you make an honest effort, you will do fine.",          
    answer(NULL, 
           correct = TRUE),
    allow_retry = FALSE,
    incorrect = NULL,
    rows = 6)
```

### Exercise 19

`show_file()` is a commonly-used function from the **tutorial.helpers** package. It makes it easy for you to share the contents of another file with your instructor, especially a file which you have edited. 

From the Console, run this command:

```
tutorial.helpers::show_file(file.path(R.home(), "COPYING"), end = 7)
```

CP/CR.

```{r exercises-19}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 10)
```

###

```{r exercises-19-test, include = FALSE}
tutorial.helpers::show_file(file.path(R.home(), "COPYING"), end = 7)
```

For now, don't worry about what `R.home()` and `file.path()` are doing, although you can read about them by running `?R.home` and `?file.path` at the Console. Those details are less important than seeing the intended usage of `show_file()`. 

### Exercise 20

We frequently ask that you `copy` content from an HTML file like a webpage. Don't worry if what you copy appears messy or has lost its formatting - that's perfectly fine. Our main concern is ensuring that your HTML contains roughly the correct information it's supposed to have.

###

Go to [the homepage](https://ppbds.github.io/tutorial.helpers/) of the **tutorials.package** and copy everything in the browser window from 'tutorial.helpers' in the upper left corner through the end of the `About this package` section.

CP/CR.

```{r exercises-20}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

````
tutorial.helpers
0.5.0.9003
Reference
Articles
Changelog
Search for
Tutorial Helper Functions
About this package
tutorial.helpers provides functions and RStudio Addins to help you write R tutorials, especially if you follow my advice about the best way to do so. This approach is currently used in several packages, including r4ds.tutorials and positron.tutorials.
````

Again, your answer never needs to match ours exactly. 

## Your answers
### 

At the end of every tutorial, you will save your work -- and submit it, at the discretion of your instructor. We will do this in the next topic, labeled "Download answers." Follow the instructions there. 

Make sure to answer the question about how many **minutes** you spent on the tutorial and to click "Submit Answer."

After hitting the "Download" button, you will be prompted to save a file named `getting-started_answers.html`. This file will, by default, be saved in the `Downloads` folder on your computer. 

Unless told otherwise by your instructor, you should use the default file name, which will usually be of the form `short-name-of-tutorial` + `_answers` + `.html`. 

```{r download-answers, child = system.file("child_documents/download_answers.Rmd", package = "tutorial.helpers")}
```

## Summary
### 

You now understand how tutorials function and how to download a copy of your answers. Good luck with your data science journey!

## Information
###

```{r name}
question_text(
    "Name:",
    answer(NULL, correct = TRUE),
    allow_retry = TRUE,
    try_again = "Edit Answer",
    incorrect = NULL)
```

```{r email}
question_text(
    "Email:",
    answer(NULL, correct = TRUE),
    allow_retry = TRUE,
    try_again = "Edit Answer",
    incorrect = NULL)
```

```{r ID}
question_text(
    "ID (if requested by your instructor):",
    answer(NULL, correct = TRUE),
    allow_retry = TRUE,
    try_again = "Edit Answer",
    incorrect = NULL)
```

---
title: "Instructions for Writing R Tutorials"
author: David Kane
format: html
vignette: >
  %\VignetteEngine{quarto::html}
  %\VignetteIndexEntry{Instructions for Writing R Tutorials}
  %\VignetteEncoding{UTF-8}
---

<!-- DK: Some things that work in QMD (because it uses quarto to render) do not work in RMD (because it uses R to render). -->

<!-- Expand processing_submissions section. -->

*There are no questions here. There are only simple instructions.*

*Tutorials are not challenging. They are confidence-building.*

*Create the shallowest possible learning curve.*

*Every word matters. Never waste a student's time.*

*Drop some knowledge with each exercise.*

*Use our tools, especially functions like [`check_current_tutorial()`](https://ppbds.github.io/tutorial.helpers/reference/check_current_tutorial.html) and [`make_exercise()`](https://ppbds.github.io/tutorial.helpers/reference/exercise_creation.html).*

*Do not build code line-by-line. That is the AI's job.*

## Introduction

This document describes the best way to write R tutorials using the [**learnr**](https://rstudio.github.io/learnr/) package. The most common use case is a tutorial which covers [the material in an assigned textbook](books.html), as with the [**r4ds.tutorials**](https://CRAN.R-project.org/package=r4ds.tutorials) package for [*R for Data Science (2e)*](https://r4ds.hadley.nz/) by Hadley Wickham, Mine Çetinkaya-Rundel, and Garrett Grolemund..

Instructors assign textbook readings to students. We want our students to read and, perhaps more importantly, go through the associated code, typing it in and confirming what it does. Sadly, students almost never do so. Fortunately, the [**tutorial.helpers**](https://CRAN.R-project.org/package=tutorial.helpers) package provides tools for ensuring that students type in all the assigned code.

Imagine the *shallowest* possible learning curve. Almost every student should be able to answer almost every exercise, albeit perhaps with the help of a hint. There are no hard questions. In fact, there really aren't any *questions* at all. Instead, there are *instructions*: Do one thing, then the next, and then the next.

Almost all exercises feature a knowledge drop, a bit of information, rarely more than two sentences, provided after the student has answered the question. 

Assume that you are giving the student a private lesson. You ask them a question. They give you an answer. What would you say to them next? What do you want to teach them, given that context?


We are building a "[pit of success](https://blog.codinghorror.com/falling-into-the-pit-of-success/)." Generally, students don't do the assigned reading, at least in a large class. However, they will complete required work. They will do the assigned tutorials. *Our promise: If you complete the tutorials, you will learn the material. There is simply no way not to.*


## Set Up

Tutorials themselves live in a directory within `inst/tutorials` in whichever package you are working on. We recommend that this directory name be a combination of a prefix number (which indicate the week/chapter with which a tutorial is associated and/or the order in which to do it) and a name, which is generally shorter than the full title of the tutorial although similar to it. Within each directory is an R Markdown file and, sometimes, other material like an `images` or `data` directory. The prefix number determines the order in which tutorials appear when someone gets around to creating a tutorial selector extension for Positron. By default, we name the R Markdown file `tutorial.Rmd`, but any name works as long as the file has the appropriate YAML header.

To create a new tutorial, use `File -> New File -> R Markdown...`. Choose the "From Template" option and then select "Helpers Tutorial" from the **tutorial.helpers** package. Follow the instructions.

The `id` value is important. It should be the same as the directory in which the tutorial is located. It is used for the name of the answer file which students save at the end of the tutorial. 

<!-- DK: Matching directory and id may no longer be necessary once the tutorial selector extension is built.  -->

Note that tutorials must be [R Markdown](https://rmarkdown.rstudio.com/) documents, meaning that their suffix is `.Rmd`. You can not (yet) use Quarto documents with tutorials. Fortunately, most of what you need which works in Quarto also works in R Markdown. The main difference is that code chunk options appear within the `{}`. Don't worry about this detail. 

There is a `setup` code chunk at the top of a tutorial. You must have `library(learnr)` and, if you use our tools, `library(tutorial.helpers)` in this chunk. Our template also provides useful settings for various options.

**Warning**: You must ensure that any library used in the tutorial is explicitly loaded in this setup chunk. Almost every tutorial makes use of functions from the **tidyverse** package, so be sure to load this. Unfortunately, nothing in our test suite captures the common error of using library X in the tutorial code and forgetting to load it in the setup chunk. 

We recommend always including a question which requires students to load any library used in the tutorial, other than **learnr** and **tutorial.helpers**. This is good for seveal reasons. First, students are always forgetting to load libraries. More practice helps. Second, a load-library question provides a good occasion for a knowledge drop. Third, a load-library question should include a test case code chunk which loads the library. This test will only pass if the library is loaded by us in the `setup` chunk.

If your tutorials are part of an R package, then you should ensure that **tutorial.helpers** is included under Imports and that any library loaded in a tutorial is, at least, included under Suggests.

## Structure

The beginning of every tutorial includes the `copy-code-chunk` and the `info-section` code chunks. The tutorial is then divided into different *topics* that appear as side panels. The first topic is the "Introduction" and the last is the "Summary." 

Within the topics, other than the Introduction and Summary, there are a series of *exercises* which can include writing code or writing text. At the end of the tutorial, there is a `download-answers` code chunk which provides students with instructions on how to download a copy of their answers.

The Introduction portion is two to four sentence about the main topics covered in the tutorial. Why are we here? What will students get out of giving you 60 minutes of their lives? What functions/techniques will they learn?

The Summary portion is two to four sentences which bring the lessons of the tutorial together for the student. What do they know now that they did not know before? What are the most important functions/techniques covered? This should be very similar to the Introduction. You made a promise as to what they would learn. You kept that promise.

If there are one or two other key resources about the topic of the tutorial, then those resources should be mentioned somewhere in the tutorial and also in the Summary.

Anything typed at the keyboard belongs in \`backticks\` (not "quotation marks"), except for package names, which are always **bolded**. Function names always include the parentheses: `read_csv()`, not `read_csv`. Example: the `+` sign is used to connect `ggplot()` components when using the **ggplot2** library.

Ensure that the first few questions always require students to load any libraries which are used in the tutorial. That is, look at all the libraries you load in the set up chunk. (Try not to have too many of them.) All of them, except for **learnr** and **tutorial.helpers**, merit an exercise which requires the student to type `library(package.name)`. This ensures that students get in the practice of loading libraries. And it also provides an occasion to drop some knowledge. Don't forget that all libraries you load should be included in the DESCRIPTION file --- if the tutorial is part of a package --- probably under Suggests.

## Topics

Tutorials are divided into *topics* that appear on the side panel. To create these topics, we include a double hash (##) before the text for it to show up as a side panel. This is also called the *topic title*.  Use [sentence case](https://apastyle.apa.org/style-grammar-guidelines/capitalization/sentence-case). On the line after the topic title, put three hashes. This ensures that students will see the introductory text before they see the first exercise.

The one topic which only has the double hash (##) and not a triple hash (###) on the next line is the Summary topic, since there are no exercises which follow the Summary text.

Each topic begins with a sentence or two about what this group of exercises is trying to accomplish. Example:


````{verbatim}
## Interacting with sites with `GET()`
###

In order to get data from an API, we use the **httr** package. 
The package is designed to imitate standard HTTP in R. 
Read more about HTTP [here](https://www.jmarshall.com/easy/http/).

### Exercise 1
````

The link will be formatted correctly once the tutorial is knitted. Topic introductions will sometimes have two parts: the introductory text as above and a plot which will be replicated in this portion of the tutorial. Those two parts are generally separated by a triple hash.

After the last exercise in a topic, you should put a triple hash and then give a two sentence summary about what this topic accomplished. A topic is a 10 minute transfer of knowledge from you to the student. At the beginning, you mentioned its purpose. Conclude by tying things back to that original purpose. Often, these "purposes" will be fairly trivial: You promised to go through an example of a scatter plot and, in fact, you did. And that is OK! We are not writing poetry. Not every topic leads to salvation.

One or two high quality links, specifically relevant to this topic, should be included/explained at either the beginning or end of a topic, unless the topic is very short.


## Exercises

Each topic is composed of a series of numbered exercises. 

### Flow

Each exercise should have a *flow* which requires that students hit the "Continue" button at least once. 

* Begin with a *Start* which is a sentence or two of knowledge and/or the question itself. If the length of the Start text is longer than one or two lines, then do not place the exercise code chunk in the same part. Instead, the Start includes a triple hash, thereby creating the Continue button. If the length of the text is short enough that students are willing to read it (at most two sentences), you can include the exercise code chunk in the same part.

* Most of the time there is no need for a triple hash before the exercise code chunk. 

* Do not expect students to read more than two sentences of text at a time. After two sentences, you almost always want to use a triple hash in order to create a Continue button so that students have a break. They won't read more than two sentences without a break.

* There is a danger that students will just click the Continue button until they see a question and, only then, start reading. There is little we can do about that. However, we can take advantage of students' tendency to read the sentence or two which proceeds the question fairly closely. This is a great place for teaching since students can't skip it since they don't know if it provides necessary context for answering the question.

* After the Start, come three code chunks:

  - The exercise code chunk is the location in which students will place their answers. 
  - The hint code chunk includes any hints for the students. Hint code chunks are only available for code exercises. The code chunk name in the hint code chunk is always exactly the same as the one for the exercise code chunk, except with `-hint-n` attached at the end. The `n` is replaced by the number of the hint. Almost always, there is only one hint, so the suffix is `-hint-1`. We always set `eval = FALSE` in the hint code chunk since, often, the hint will not be legal R code.
  - The test code chunk has exactly the label as the one for the exercise code chunk, except with `-test` attached at the end. Test code chunks, like hint code chunks, are only used for code exercises. It always includes the `include = FALSE` code chunk option because we never want to show the code or the results to students. Instead, the purpose of the test code chunk is to ensure that the correct answer --- that is, the code we want students to enter into the exercise code chunk --- works. 
  
* The three code chunks are always followed by a triple hash. We want a student to pause after she has submitted her answer so that she is more likely to consider the output from her submission before moving on.

* The last part of an exercise is the end, our main opportunity to drop some knowledge. 

* The last part of the topic is another knowledge drop. It is not another exercise. It is just a knowledge drop after the last exercise which tries to take a broader overview. It is often separated from that last exercise by a simple `###`. Again, this can't be more than a sentence or two. But it should be more substantive than a simple "Good job." For example, if the topic has involved creating a scatter plot, then the last exercise will be putting the final touches on that scatter plot. The last knowledge drop should be something about scatter plots in general, not a minor point about the particular scatter plot which the student has just created.

### Question types

To create the exercise headers, you use three hashes. Make sure you number your exercises -- `### Exercise 1`, `### Exercise 2` and so on. 

There are two main types of questions. First, we have normal *coding* questions. Students write code and press the Run Code button. Second, we have *text* questions which require students to either write prose or to copy/paste the results of running specific commands. Prose is needed for questions like "Explain the meaning of potential outcomes." Copy/paste situations arise when students are instructed to do something like connect to Github or edit a QMD file. We confirm that the students have completed these questions by having them issue a command like `list.files()` and then copy/pasting the command and the output. We often abbreviate those instruction using CP/CR, which stands for **c**opy and **p**aste the **c**ommand and the **r**esult.

#### Code questions

Here is an example code question:

````{verbatim}
### Exercise 2

Start your code with `cces`. Use the pipe operator `|>` to add the 
function `filter()`, selecting the  rows where `state` is equal to 
"Massachusetts". To set something equal to a value in `filter()` 
use two equal signs: `==`.

```{r filter-2, exercise = TRUE}
```

```{r filter-2-hint-1, eval = FALSE}
cces |> 
  filter(state == "...")
```

```{r filter-2-test, include = FALSE}
cces |> 
  filter(state == "Massachusetts")
```
### 

`==` is used because it is **checking** whether the value of the 
variable on the left is equal to the value on the left. See 
[here](https://stat.ethz.ch/R-manual/R-devel/library/base/html/Comparison.html) 
for discussion of other relational operators in R.
A single equation symbol, `=`, is used to set something equal to 
something else. 
````

First, the start of the exercise sets the stage. It sometimes teaches something new, connects to a previous exercise, provides a useful link, whatever. If it is long enough, it is followed by a triple hash. If not, the text continues to the instructions. Most of the time, as above, there is only the instruction, telling the student, step-by-step, what to do.

Second, the instruction requires that students write some code. Good instructions generate results when the student presses Run Code. Tutorial answers should require the smallest incremental number of characters, relative to the last question, for students to type. That is one way you know that your learning curve is shallow. If a exercise code chunk requires the students to type a lot of characters, you should split up the exercise into multiple separate exercises.

Third, any exercise which requires the copying of code from the prior exercise should place the *Copy previous code* button below the exercise code chunk.

````{verbatim}
<button onclick = "transfer_code(this)">Copy previous code</button>
````

Fourth, tutorials should be so simple that 95% of the students can answer 95% of the questions easily. One way to ensure that is to add a *hint* to almost every coding question.

Hints must always have the same code chunk name as the exercise chunk for which they are the hint, with a "-hint-n" added at the end. So, if an exercise code chunk is named "ex-1", then the hint associated with that exercise is named "ex-1-hint-1". A second hint for that same question would be named "ex-1-hint-2", and so on. 

<!-- More recent versions of learnr allow for you to use a special character --- by default I think it is underscores --- which will not generate an error, meaning that you don't nee eval = FALSE. Not sure if we care. -->

When you create a hint, always use `eval = FALSE` within the parentheses in the code chunk. This is because hints will often include "..." and other symbols which do not run as correct R code. So, we need to tell R not to run it or an error will occur during `R CMD check`. Example:

````{verbatim}
```{r ex-1-hint-1, eval = FALSE}
This is an example hint. Normally sentences like these 
would cause an error in R because it is not proper code. 
However, since we include eval = FALSE in the r-chunk, 
R ignores all errors!
```
````

**You need to wrap the text in a long hint by hand, inserting the carriage returns yourself.** R will not wrap the text automatically.

Often, hints look like this:

````{verbatim}
```{r ex-1-hint-2, eval = FALSE}
... |> 
  filter(year = ...) |> 
  ...(flights)
```
````


The "..." indicates places where the student needs to insert some code, a value or a function name. The code in hints should be formatted correctly.

Students can not see the first hint after clicking through to the next hint. So, make sure the last hint is the one you most want them to have access to, i.e., the one which provides the key information. If students can see the last hint, they should have no reason to consult any earlier hints. We rarely provide more that one hint.

Hints are [only allowed](https://forum.posit.co/t/hints-in-written-questions/108184) for coding questions, **not for text questions**.

Fifth, the third code chunk, after the exercise and hint code chunks, is the test code chunk. Place the answer --- the code which you want students to enter into the exercise code chunk --- into the test code chunk. Since the test code chunk will be evaluated when the tutorial is knitted (which also happens during testing), this guarantees that the correct answer will work for students. 

Indeed, the workflow for writing an exercise often begins by, first, entering the code which we want students to provide into the test code chunk. We then copy/paste that same code into the hint code chunk, replacing some of the functions and/or arguments with `...` as appropriate. We then ask the question which, we hope, will cause students to answer with the same code as we have in the test code chunk.

There are some instances in which we can't test code which we want students to use. The most common case is code which requires the web, generally for downloading data. In that case, we delete the test code chunk.

Sixth, we separate the code chunks from the end of the exercise by using `###`.

For simple questions which result in a display of some data, one approach is to write "You should see that the value of height in row 1 is 23." This allows the students to know that they are on the right track. Never hard-code a number. Use R to inline calculate it, even though this can be a bother.

But, much more common, is to use the end to drop some knowledge, especially about a function which was used in the answer to this exercise, or to one of the previous exercises in this topic.

Note that we provide as many exercises as possible. For example, every tutorial features a question for each package which must be loaded. We require students to type in items like `library(tidyverse)` even though they have done so many times in the past. Every exercise is another opportunity to make the learning curve as shallow as possible and to drop some knowledge. More questions are better than fewer. 

One way to measure the shallowness of the learning curve is to examine how many new characters each exercise requires for its answer. More new characters is worse. Better three exercises, each of which loads a different package, then one exercise which loads three packages at once.

#### Text questions

There are two types of text questions: 1) those that provide the students with the correct answer, after they have submitted their own answer, and, 2) those that do not provide an answer.  Examples:

````{verbatim}
### Exercise 6

Explain potential outcomes in about two sentences.

```{r definitions-6}
question_text(NULL,
    message = "This is where we place the correct answer. It will appear only after 
    students have submitted their own answers. Note that we do not need to wrap the 
    answer text by hand.",          
    answer(NULL, 
           correct = TRUE),
    allow_retry = FALSE,
    incorrect = NULL,
    rows = 6)
``` 
````

For the `message` argument, you should provide an **excellent** answer. We want to allow students to check for themselves that they got, more or less, the correct answer. Note how we set `allow_retry` to FALSE. This means that, after they see our answer, students can't modify their answer. The `rows` argument decides how many rows the empty text input will have.

Always specify (approximately) how much you want students to write. Reasonable units are: one sentence, two sentences and a paragraph. Pick one of these three unless you have a good reason not to.

For paragraph questions, you should mention specific words or phrases which the students should include in their answers. If your suggested answer includes the word "validity," for example, then tell the students to include (and define) validity as part of their answer.

However, for many written questions, we don't provide an answer, so we don't mind if students resubmit. In that case, we use 

````{verbatim}
### Exercise 7

From the Console, run `list.files()`. CP/CR.

```{r file-creation-7}
  question_text(NULL,
    answer(NULL, correct = TRUE),
    allow_retry = TRUE,
    try_again_button = "Edit Answer",
    incorrect = NULL,
    rows = 3)
```
````


This format is most commonly used for "process" questions in which we have told students to do something and then confirm that they have done it by copying/pasting the result from a command.

`tutorial.helpers::show_file()` is a handy function for confirming that students have modified text files as instructed. For example, after telling students to edit the `_quarto.yml` file, we can check that they did so with:

````{verbatim}
In the Console, run:

```
tutorial.helpers::show_file("_quarto.yml")
```

CP/CR.
````

`show_file()` provides a variety of arguments which cause it to return only selected lines rather than the entire file.

Keep in mind that `show_file()` will not be available to students in their Console by default. We can either always call it with `tutorial.helpers::show_file()`, as above, or always (and after each restart of the R Session!) have the student type `library(tutorial.helpers)` at the Console by hand.

Maybe the instructions should always use the double colon, but remind students the first time in each tutorial that they could just use `library(tutorial.helpers)`. Or, don't use the double colon and then, the first time `show_file()` is used, remind students that, when they see the "No function found" error, they need to run `library(tutorial.helpers)` at the Console.

### Tips

Each coding exercise should always spit out something. Interactivity is good! Students should always look at what their code is producing. There are some situations in which students need to make assignments and which, because of this, will result in no output when the Run Code button is pressed. But:

* Do this as little as possible. Why not just make a pipe?

* When you do this, you will need to create the permanent object yourself because student work in an exercise code chunk has no lasting effects. Doing so is annoying and error prone.

* You can have the student not only do the assignment but also, as part of this same exercise, print out the object. This works well.


Follow a coding [Style Guide](https://style.tidyverse.org/), especially spaces around operators like " = ". Use only one command per line in pipes and graphics, with proper indentation. Indent plotting commands after the call to `ggplot()`.

Do not create an object in one question and then assume that it will be available in subsequent questions. It won't be! Each question is independent of every other question. They live in separate R instances. The only exception (which we make use of) is that objects created in the initial `setup` chunk for the entire tutorial are available in all later questions, just the way that `library()` commands executed there do not need to be executed again. 

Tutorials are knitted/run from the directory in which they are located. So, if you want to read in a file from a `data/` directory, you write:


```{r, eval = FALSE}
x <- read_csv("data/myfile.csv")
```


in an R code chunk, presumably in the global `setup` chunk. But, if you try to execute that line of code with `Command + Return`, it will fail because, by default, you are located in the main directory of **your.package** when you start working on your tutorials. Using `setwd()` will solve this problem.

````{verbatim}
> getwd()
[1] "/Users/davidkane/Desktop/projects/r4ds.tutorials"
> setwd("inst/tutorials/031-data-files/")
> getwd()
[1] "/Users/davidkane/Desktop/projects/r4ds.tutorials/inst/tutorials/031-data-files"
> 
````

`Command + Return` will now work because your R session is "located" in the same location as that from which the tutorial will be run when it is knitted.


### Pipes

The most common type of code questions involve the step-by-step process of building a pipe, the final output of which is a nice looking graphic.

You want to first show the graphic that you will create by the end of the topic. You show it once at the start of the topic and once before the last exercise, as a reminder of what the graphic should look like so students do not need to scroll all the way back up. 

You should put the code for the graphic in the `setup` code chunk. Save the code to an object. The name of the object should have a "_p" suffix, where the "p" stands for "plot." This way, you only have to put the object name in the code chunk at the end of the topic rather than copying the code.

You then build up the graphic, line by line, over a series of exercises, providing hints along the way. 


### Knowledge Drops

The most difficult part of tutorial creation is writing the "knowledge drops," the snippets of wisdom (and the associated links) which are used at the end of each exercise. These generally come in two categories: details about R functions/packages/websites and background information about the substative data science problem at hand. 

Do not expect this to be easy! Good knowledge drops are hard. Make them short. Students will not read more than a sentence or two.

Perhaps the best place for a knowledge drop, especially for written questions, is at the start of the exercise. That is, instead of just asking the question immediately, provide a sentence or two of knowledge drops, even if this information is not really needed to answer the question. Students tend to read those sentences closely since they might be relevant to the question they need to answer.

Rhetorical questions (almost) always work poorly for knowledge drops.

A knowledge drop should not be a road sign. Example: "In the next section we will explore the data further." Don't waste time telling students what you expect to do next, or what you have just completed doing. Teach them something real!

#### Advice for Knowledge Drops

* Not sure what knowledge to drop? Look up the help page for one of the functions used in the topic. Ideally, this will be the function which was used in this exercise, but it can be one used a few lines above. The help page will have two areas of interest:
  * Arguments for the function, some of which will not have been used in the topic but which are still worth mentioning. A good knowledge drop will mention this argument and describe when it might be useful.
  * Related functions. For example, use the "Useful filter functions" section of the `?dplyr::filter` help page. All of these make for good knowledge drops! Another example is the "See also" section of the `?arrange` help page. We can't possibly use every R function in an exercise, but we can at least describe some of them in knowledge drops.
  
* Use high quality links. 
  * The second edition of [*R for Data Science*](https://r4ds.hadley.nz/) is amazing. Link to it as often as you can.
  * Anything associated with [Posit](https://posit.co/) is high quality. Examples include sites associated with [Tidyverse](https://tidyverse.org/) and [Tidymodels](https://www.tidymodels.org/).

* Avoid links to low quality sites like `https://rdrr.io/` , `https://www.rdocumentation.org/`, and `https://www.datamentor.io/`. Any site which uses ads should be avoided.

* Links to function documentation are useful. Indeed, we should link at least once to the documentation for every function which we use. Two categories:
  * Links to the help pages for base function are tricky. It is not clear to me which location is the best to use for this purpose. My suggestion is to use links involving the `stat.ethz.ch` server, which is the home base (I think) for the core R developers. Consider examples for [`cut()`](https://stat.ethz.ch/R-manual/R-devel/library/base/html/cut.html) and [`print()`](https://stat.ethz.ch/R-manual/R-devel/library/base/html/print.html).
  * Links to functions from the *Tidyverse* should use the docs maintained by Posit. Consider these examples for [`arrange()`](https://dplyr.tidyverse.org/reference/arrange.html) and [`aes()`](https://tidyverse.github.io/ggplot2-docs/reference/aes.html). 
  
* One common approach is like this:

````
See `?readr::locale` for [details](https://readr.tidyverse.org/articles/locales.html).
````

Note how we concisely provide both the command which brings up the help page and a link to the help page itself. 


### Functions for Working with Tutorials

The **tutorial.helpers** package provides a collection of functions which facilitate the creation and testing of tutorial packages. Read about them in the “[Functions for Working with Tutorial](https://ppbds.github.io/tutorial.helpers/articles/instructions.html)” vignette. Three of the vignettes create new tutorial exercises. A [fourth](https://ppbds.github.io/tutorial.helpers/articles/testing.html) demonstrates how to test the tutorials in your package. Highly recommended!


## Inputs

In addition to `tutorial.Rmd`, a tutorial will often use other inputs. The two most common locations for storing these inputs are `data` and `images` directories at the same level as the `tutorial.Rmd` file. Any file in `data` or `images` will be available at run time. (Note that the directories must have these names. Something like `my_data` will not work.) 

### Data

If you need for an R object to be accessible in an exercise code chunk, create it in the initial global `setup` code chunk at the top of the tutorial.

<!-- Objects created in a code chunk which is then referenced by using the exercise.setup code chunk option in any exercise which needs access to these objects. -->

Be wary of code which downloads data from the web. This won't work if the student does not have an internet connection when she creates the tutorial. Instead, save the code which downloaded the data and then place that object in an RDS file in the `data` directory. Here is an example from the "Wrangling Census data with Tidyverse tools" tutorial from the [**tidycensus.tutorials**](https://ppbds.github.io/tidycensus.tutorials/) package.

```
median_age <- get_acs(geography = "county",
                      variables = "B01002_001",
                      year = 2020)
write_rds(median_age, "data/median_age.rds")

median_age <- read_rds("data/median_age.rds")
````

The first two commands download data and save it to an RDS file in the `data` directory. 

This code assumes that you are located in the same directory as the `tutorial.Rmd` file. You only run those commands once, and then you comment them out because you don't want them re-run each time the tutorial is created. The `read_rds()` call is never commented out because we always need the `median_age` object.

When designing tutorials which use objects like `median_age`, we generally write two exercise code chunks. The first has the student run the same code as that which we used to create the object ourselves. This won't work if the student is not connected to the web but, with luck, in that case they will get a sensible error message. The second question informs the students that we have, behind the scenes, assigned the result of the function to an R object. We then ask the student to just print out that object. We don't have them do the assignment themselves, not least because we don't like questions which don't generate any output. 

We use a similar approach with models which can take awhile to fit. Example:

````
fit_gauss <- brm(formula = biden ~ 1,
                data = poll_data,
                refresh = 0,
                silent = 2,
                seed = 9)
write_rds(fit_gauss, "data/fit_gauss.rds")

fit_gauss <- read_rds("data/fit_gauss.rds")
````

Again, this code only works if you are in the tutorial directory, not in the higher directory of the R project itself. Also, the first two commands are commented out, unless you are running them by hand to create the object.

What happens if the data is too large? See the "Arrow" tutorial in the [**r4ds.tutorials**](https://ppbds.github.io/r4ds.tutorials/) for an example. First, we generally switch away from code exercises and use written exercises. Students run the required commands and then copy/paste the command/response. Big downloads don't work well in exercise code chunks. Second, we create small versions of this big data in the global `setup` chunk. This allows us to create test code chunks for most of the exercises which follow. These tests will run much more quickly with this smaller data. Also, for any package on CRAN, we need to keep the overall size of the package as small as possible.

There are two main uses for files in `data`. First, they can be used at "compile time" (when the `tutorial.Rmd` is knit to HTML) for making plots or doing anything else. Second, and more importantly, they are available to students in the exercise code blocks during "run time" (when students are doing the tutorial).

### Images

To add images to a tutorial, first make a directory called `images` in the folder that contains `tutorial.Rmd`. Store all images for that tutorial there. You can work with those files in all the usual ways.

Use `include_graphics()` to add the image into the document. Include this code in its own chunk, in the place where you want the image to appear in the tutorial. 

````{verbatim}
```{r}
include_graphics("images/example.png")
```
````

`include_graphics()` is part of the `knitr` package, so you need `library(knitr)` in the setup code chunk. Note that you do not need to name these code chunks.

Because students will complete the tutorials using screens of very different widths, it is a good idea to put `knitr::opts_chunk$set(out.width = '90%')` in your `setup` code chunk. In this way, images will appear at a sensible size regardless of whether students are using a phone screen or a big monitor.

### Complex text

You sometimes want to include "complex" text in a tutorial. This is most common when trying to teach students how to use R code chunks and other strings which R markdown wants to process in certain ways. You can sometimes get away by placing such text in environments surrounded by three, or even four, backticks. The plain three or four backtick environments create a plain code block and content inside is displayed as-is. No syntax highlighting or execution is applied, and they show example code or text exactly as typed. Four backtick environments allow three backticks to be written within the block. This works often, but not always. Comments characters like `#` are especially problematic. We also use the [**parsermd**](https://rundel.github.io/parsermd/) package behind the scenes. It does not work as well as one might like.

Further complexity arises because tutorials are R markdown whereas all our other work is in Quarto. In Quarto documents (like this vignette), you can use an environment started by four backticks along with `{verbatim}` and then ended with four more backticks. This won't work in R markdown.

The `<pre><code>` trick solves this problem. You add this before and after a code chunk to a tutorial.

````{verbatim}
<pre><code>```{r}
1 + 1
```</code></pre>
````

If you want to display an R code chunk nested inside a four backticks environment, you can replace the four backticks with `<pre><code>` like this

````{verbatim}
<pre><code>
> tutorial.helpers::show_file("quarto-1.qmd")
---
title: "Quarto 1"
author: David Kane
format: html
execute: 
  echo: false
---
<pre><code>```{r}
suppressPackageStartupMessages(library(tidyverse))
library(palmerpenguins)
```</code></pre>
>
</code></pre>
````

The `<pre><code>` serves the same purpose as the four backticks but it displays R code chunks verbatim. 

## Processing submissions

Instructors have different needs and priorities when it comes to processing student answers. We provide the functions `submissions_summary()` and `submissions_answers()` to help them. 

## Checks 

The simplest way to test the `tutorial.Rmd` with which you are working on is to run `rmarkdown::render()` on that file. This function takes your R Markdown document and transforms it into a single HTML file that you can open in a browser to see how your tutorial will appear to users. 

You can do this by running:

```{r, eval = FALSE}
rmarkdown::render("file_path")
```

This assumes that you are located in the main directory of **your.package**, as you normally would be. While this approach may not catch every possible error, it will identify many common issues and runs very quickly, making it an efficient way to test your tutorial. Remember to replace `file_path` with whatever directory name is appropriate for your specific tutorial. If you're using Positron as your IDE, it's important *not* to use the `Cmd/Ctrl + Shift + K` shortcut, as this will execute `quarto render` instead of the R Markdown rendering process you actually need, since tutorials are R Markdown documents rather than Quarto documents.

### Test before submitting a pull request

Once you are done editing a tutorial, you need to make sure it works before you submit a pull request to the package maintainer. There are several ways to do checks:

1. Type `rmarkdown::render("file_path")` in the console. This will give you a file with the same file name and a `.html` extension. Right click the file and select `Open in Browser`. This open up a webpage and you can check the tutorial for some errors. It is smart to use `rmarkdown::render("file_path")` regularly since it identify syntax errors quickly. A good practice is to render after completing each exercise. 

2. Do a full test, which means running `devtools::check()`. This is the equivalent of `R CMD check` which validates package structure, documentation, tests, and compliance with CRAN standards. A good result would be `0 errors ✓ | 0 warnings ✓ `.

### What to do if `devtools::check()` fails

1. Read the error message carefully. You always want to see `0 errors ✓ | 0 warnings ✓` under the `R CMD check results`. If not, there is a problem. The error message will often provide a clue as to where in your code the error occurred.

2. If that error message is not detailed enough, go to the `your.package.rcheck` folder, which should be located in the same directory as `your.package` on your computer. This is a folder created by the `R CMD check` process, and it will be automatically deleted if the check process succeeds. If the process fails, the `your.package.rcheck` folder stays around so that you can examine it. The key file is `testthat.Rout.fail`, which should be in the `tests` directory. It has more details on what went wrong.

The most common source of errors is something wrong with the hint code chunks, which are not evaluated when you just Run Document. Make sure the `eval = FALSE` argument is set in the code chunk for all hints. 

### Try out the tutorial like a student

Test your tutorial from a student's perspective after completing package checks by following this local testing workflow. You can do this without the hustle of needing to push changes to the main repository and re-download the package. First, temporarily disable renv by running `renv::deactivate()` to prevent installation conflicts that can cause issues with `devtools::install()`. Next, install your local package using `devtools::install()` and when prompted about updating packages, choose to update everything to ensure all dependencies are current. Once installation is complete, run your tutorial with `learnr::run_tutorial("tutorial_name", "your_package_name")`, replacing the placeholder names with your actual tutorial and package names. This allows you to test the tutorial in a clean environment similar to what students will experience, checking that all code chunks run without errors, instructions are clear, and there's no dependence on objects from your development environment. Finally, don't forget to reactivate renv by running `renv::activate()` to return to your managed environment.

### Difficult bugs

* The most common issue is that an author will use a package like **ggthemes** in their tutorial but then forget to include `library(ggthemes)` in the `setup` code chunk. This will not cause any error in checking because **ggthemes** is not actually used in the tutorial itself --- unless it is included in a test code chunk, as it should be --- so the tutorial knits without a problem. But then the student tries to type `library(ggthemes)` as an answer to an exercise and gets an error because the package has not been installed. 

* Note that `R CMD check` does not seem to catch cases in which you `library()` a package in a tutorial but that package is not in DESCRIPTION. But such a discrepancy will cause an error on Github Actions because, there, you only have access to packages that have been installed as part of that test.

* Be careful of the way that Github is sloppy in how it deals with capitalization changes, especially when you change the name of a file. For example, you might first commit a file named `Rproj.png`. Later, you decide to change all file names for images to all lower case. So, you change the name of the file to `rproj.png`. Commit and push. Everything is great, right? No! Even if Github shows you the new file name, it might still have that file as `Rproj.png` internally, with the capital "R." This will cause errors when your run your checks on Github: 

````
Error: Cannot find the file(s): "images/rproj.png"
````

But the file is there! You can see it! The tests work on your local machine. The easiest solution is to delete the file (and commit that change). And then change the name of the file to something else and use it.

* `R CMD check` will test that all tutorials have the default code chunks exactly as they are in the Helper Tutorial template. So, use the template. If either the "Information" or "Download answers" chunks are missing, `R CMD check` will return something like "Missing a component part from file /path/to/your/Rmd/file.Rmd".

---
title: "XX: Title"
author: "XX: Your Name"
tutorial:
  id: "XX: all lowercase id is the directory in which this tutorial is located"
output: 
  learnr::tutorial:
    df_print: default
    progressive: true
    allow_skip: true
runtime: shiny_prerendered
description: "XX ADD DESCRIPTION (a single sentence)"
---

```{r setup, include = FALSE}
library(learnr)
library(tutorial.helpers)

# XX: Add other packages you use. Such packages should also be added
# to the student QMD. 

knitr::opts_chunk$set(echo = FALSE)
options(tutorial.exercise.timelimit = 60, 
        tutorial.storage = "local") 
```

```{r info-section, child = system.file("child_documents/info_section.Rmd", package = "tutorial.helpers")}
```

<!-- XX: CHECKLIST BEFORE STARTING: -->
<!-- * If you need to use the copy code button, you must load the relevant child document. -->
<!-- * Use check_current_tutorial() and make_exercise("no-answer"). Very helpful! -->
<!-- * Edit yaml at the top of this file -->
<!-- * Save the file as "tutorial.Rmd" in the correct directory under inst/tutorials/xx/, where xx is the same as the tutorial id. -->
<!-- * Load any necessary libraries for the tutorial in the first code chunk -->
<!-- * Anytime you have a student `Cmd/Ctrl + Enter`, you ask for CP/CR and then provide your code as the test code chunk -->
<!-- * Delete this and the other commented instructions below. -->
<!-- Search for, read/replace, and then delete all `XX:` entries -->

## Introduction
###

<!-- XX: Two to four sentence about the main packages/functions covered in this tutorial. -->

### Exercise 1

Create a Github repo called `XX`. Set the "Add a README file" toggle box to "On."

Connect the repo to a project on your computer using `File -> New Folder from Git ...`.  Make sure to select the "Open in a new window" box. 

You need two Positon windows: this one for running the tutorial and the one you just created for writing your code and interacting with the Console.

In the new window, select `File -> New File -> Quarto Document ...`. Provide a title -- `"XX"` -- and an author (you). Save the file as `analysis.qmd`. Render the document. 

Create a `.gitignore` file with `analysis_files` on the first line and then a blank line. Save and push.

In the Console, run:

```         
show_file(".gitignore")
```

If that fails, it is probably because you have not yet loaded `library(tutorial.helpers)` in the Console.

CP/CR.

```{r introduction-1}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

### 

<!-- XX: Insert a knowledge drop related to this project. -->

<!-- XX: By default, the next two questions just mention tidyverse. If there is more than one required library, you can just add those libraries to these steps, adjusting the grammar as needed. Or, you might have separate questions for each library, thereby providing more room for knowledge drops. -->

### Exercise 2

In your QMD, put `library(tidyverse)` in a new code chunk. Render the file using `Cmd/Ctrl + Shift + K`.

Notice that the file does not look good because the code is visible and there are annoying messages. To take care of this, add `#| message: false` to remove all the messages in this `setup` chunk. Also, add the following to the YAML header to remove all code echoes from the HTML:

```         
execute: 
  echo: false
```

Using `Cmd/Ctrl + Shift + K`, render the file again. Only the title and author should appear in the HTML.

In the Console, run:

```         
show_file("analysis.qmd", chunk = "Last")
```

CP/CR.

```{r introduction-2}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 6)
```

### 

<!-- XX: Insert a knowledge drop related to this project. -->

### Exercise 3

Place your cursor in the QMD file on the `library(tidyverse)` line. Use `Cmd/Ctrl + Enter` to execute that line.

Note how this command causes `library(tidyverse)` to be copied down to the Console and then executed. 

CP/CR.

```{r introduction-3}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 8)
```

###

<!-- XX: Insert a knowledge drop related to this project. -->

### Exercise 4

<!-- XX: Delete this question if you do not make use of a `data` directory in this tutorial. Note the extra spaces after the command. These are needed to ensure separate lines for each command when students copy/paste. -->

From the Console, run these three commands:

`getwd()`  
`dir.create("data")`  
`list.files()`  

This will create a `data` directory in your project. This is a good place to store any data that you are working with.

CP/CR.

```{r introduction-4}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

You answer should look something like this, although your path will be different.

````
> getwd()
[1] "/Users/dkane/Desktop/projects/xx"
> dir.create("data")
> list.files()
 [1] "analysis_files" "analysis.html"  "analysis.qmd"   "data"           "README.md"  
>
````

<!-- XX: You will probably have one or two Sections, in between the Introduction and Summary. We provide on example below. Before editing it, you may want to copy/paste it to create the skeleton for a second section. Any tutorial which both uses a lot of AI and is supposed to take an hour or so will probably only have two sections.  -->

## XX: First section (use sentence case)
###

<!-- XX: Mention the packages/functions which you plan on covering in this Section. Not everything mentioned here is used in the Introduction/Summary, but everything in Introduction/Summary is referenced in one of these Section intro parts, the space before Exercise 1 in each section. -->

### Exercise 1

<!-- XX: In this question, "XX" is the full name of the file, like "cars.xls". -->

We begin by downloading XX directly from GitHub to the `data` directory using `download.file()`. 

In the Console, run:

```         
download.file(
  "https://github.com/PPBDS/ai.tutorials/raw/refs/heads/main/inst/tutorials/r4ds-2/data/cheeses.xlsx",
  destfile = "data/XX"
)
```

CP/CR.

```{r cheese-1}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 8)
```


### Exercise 2

<!-- XX: The meat of a Section generally involves asking AI to create a pipe which accomplishes some goal. The end of a Section always finishes up with a plot. These four questions set up and then guide the student to creating that plot. If you want the student to mimic a plot, you can place it in the `images` subdirectory and then use knitr::include_graphics("images/plot.png") to show it to students. -->

### Exercise 3

Before creating a plot, we need to ensure that your data matches our data. In the QMD, replace your code from the previous exercise with our code.

In the Console, run:

```         
show_file("analysis.qmd", chunk = "Last")
```

CP/CR.

```{r xx-first-section-use-sentence-case-3}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 6)
```

###

<!-- XX: Insert a knowledge drop related to this project. -->

### Exercise 4

Within the latest code chunk, add the option: `#| cache: true`. Assign the result of the pipe to `x`. 

`Cmd/Ctrl + Shift + K`. By including `#| cache: true` you cause Quarto to cache the results of the chunk. The next time you render your QMD, as long as you have not changed the code, Quarto will just load up the saved object.

If you have not done so already, you should add `analysis_cache` to the `.gitginore`. The content of the cache file does not belong on GitHub.

Place your cursor on the line where the pipe is assigned to `x`, run `Cmd/Ctrl + Enter`, thus ensuring that the workspace also includes a copy of `x`.

CP/CR.

```{r xx-first-section-use-sentence-case-4}
question_text(NULL,
    answer(NULL, correct = TRUE),
    allow_retry = TRUE,
    try_again_button = "Edit Answer",
    incorrect = NULL,
    rows = 8)
```

###

Our code:

```{r, echo = TRUE}
# x <- ... where the ellipsis are replaced with the code which creates x.
```

###

<!-- XX: Insert a knowledge drop related to this project. -->

### Exercise 5

Within the Console, type `x`, which we previously assigned to a pipe and ran in the Console. Hit `Enter`.

CP/CR.

```{r xx-first-section-use-sentence-case-5}
question_text(NULL,
    answer(NULL, correct = TRUE),
    allow_retry = TRUE,
    try_again_button = "Edit Answer",
    incorrect = NULL,
    rows = 8)
```

###


Our code:

```{r, echo=TRUE}
# x
```

###

<!-- XX: Insert a knowledge drop related to this project. -->

### Exercise 6

Ask AI to generate R code that uses `x` to plot a basic graph showing XX ... Mention you want to use the data from `x` and copy/paste the `x` you ran in the Console with the resulting tibble. You only need the top 3 lines, mainly to include column names.

Within `labs()`, edit or add a proper title, subtitle, and caption. If axis labels would be useful, add them, but if unnecessary, don't bother. Don't assign the code for the plot to any variable. Put the plot code in a new code chunk. Run `Cmd/Ctrl + Shift + K` to ensure that everything works. Make your plot look nice.

In the Console, run:

```         
show_file("analysis.qmd", chunk = "Last")
```

CP/CR.

```{r xx-first-section-use-sentence-case-6}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 12)
```

###

Our code:

```{r, echo=TRUE}
# XX: Make sure your plotting code is good! This will take some time. You had better have a subtitle which provides the take-away message of the plot. AI sometimes gives you too much code, lots of `theme()` stuff and so on. This is no good! In most cases, we are happy with concise, straightforward code. 
```

###

<!-- XX: Insert a knowledge drop related to this project. -->

## Summary
###

<!-- XX: The exact same two to four sentences about the main packages/functions used in the Introduction, but written here in the past tense. You made a promise and you kept it.  -->

### Exercise 1

`Cmd/Ctrl + Shift + K` to ensure that everything works.  The resulting HTML page should be attractive, showing clean versions of your plot(s).

At the Console, run:

```
tutorial.helpers::show_file("analysis.qmd")
```

CP/CR.

```{r summary-1}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 30)
```

### 

<!-- XX: Insert a knowledge drop related to this project. -->

### Exercise 2

Publish your rendered QMD to GitHub Pages. In the Terminal --- not the Console! --- run:

````
quarto publish gh-pages analysis.qmd
````

Copy/paste the resulting URL below.

```{r summary-2}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 1)
```

### 

<!-- XX: Insert a knowledge drop related to this project. -->

### Exercise 3

Commit and push all your files. Copy/paste the URL to your Github repo.

```{r summary-3}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

### 

<!-- XX: The tutorial is now over. Add any necessary acknowledgements and/or provide a link to further high quality readings, ideally readings which you mentioned in at least one knowledge drop above. -->

```{r download-answers, child = system.file("child_documents/download_answers.Rmd", package = "tutorial.helpers")}
```

---
title: "Tutorials for Books"
author: David Kane
format: html
vignette: >
  %\VignetteEngine{quarto::html}
  %\VignetteIndexEntry{Tutorials for Books}
  %\VignetteEncoding{UTF-8}
---


This vignette assumes that you have already read [our advice](instructions.html) about writing good tutorials.

A common use case for this package is to write a collection of tutorials for a book. Consider the [**r4ds.tutorials**](https://ppbds.github.io/r4ds.tutorials/) package which is a companion to [*R for Data Science (2e)*](https://r4ds.hadley.nz/) by Hadley Wickham, Mine Çetinkaya-Rundel, and Garrett Grolemund.

Instructors like to assign books with code. Ideally, we want our students to read the book, working through the included code, line-by-line. Sadly, very few students are so disciplined. In fact, in a large class, a majority of the students won't even read the book.

The **tutorial.helpers** package makes it easy to create a *structured* tutorial in which students type in (almost) every command which the book demonstrates, along with other commands which, in your judgment, are helpful. 

Consider some idiosyncratic advice for book-based tutorials using the example of [*R for Data Science (2e)*](https://r4ds.hadley.nz/) and [**r4ds.tutorials**](https://ppbds.github.io/r4ds.tutorials/).

* Do one or two of the tutorials before you start working on your own. 

* Make your Introduction look like this:

````
This tutorial covers [Chapter 20: Spreadsheets](https://r4ds.hadley.nz/spreadsheets.html) 
from [*R for Data Science (2e)*](https://r4ds.hadley.nz/) by Hadley Wickham, 
Mine Çetinkaya-Rundel, and Garrett Grolemund. You will learn how to get data from Excel 
spreadsheets using  `[read_excel()](https://readxl.tidyverse.org/reference/read_excel.html)` 
from the [**readxl**](https://readxl.tidyverse.org/) package and Google sheets using
`[read_sheet()](https://googlesheets4.tidyverse.org/reference/range_read.html)` from 
the [**googlesheets4**](https://googlesheets4.tidyverse.org/) package.
````

  - The first sentence provides a link to your chapter as well as to R4DS.
  - We highlight some of the most important packages and functions used in the tutorial.
  - We don't go overboard. You can't mention every package or every function used in the tutorial.


* Make your Summary look like this:

````
This tutorial covered [Chapter 20: Spreadsheets](https://r4ds.hadley.nz/spreadsheets.html) 
from [*R for Data Science (2e)*](https://r4ds.hadley.nz/) by Hadley Wickham, 
Mine Çetinkaya-Rundel, and Garrett Grolemund. You have learned how to get data from Excel
spreadsheets using  `[read_excel()](https://readxl.tidyverse.org/reference/read_excel.html)` 
from the [**readxl**](https://readxl.tidyverse.org/) package and Google sheets using
`[read_sheet()](https://googlesheets4.tidyverse.org/reference/range_read.html)` from 
the [**googlesheets4**](https://googlesheets4.tidyverse.org/) package.

Read “[Data Organization in Spreadsheets](https://doi.org/10.1080/00031305.2017.1375989)” 
by Karl Broman and Kara Woo for great advice about organizing your data using spreadsheets.
````

  - The first paragraph is identical to the Introduction, except that "covers" is replaced with "covered" and "will learn" with "have learned." We began the tutorial with a promise about what students would learn. One hopes that we kept that promise.
  - The second paragraph gives one or two pointers about the best material which a student might look into if she is interested in learning more about the broad topic of the tutorial. These pointers were also mentioned as knowledge drops earlier in the tutorial.
  - These pointers will often (always?) be items which were mentioned in the book itself. Those authors, presumably, have good taste, otherwise you would not have selected their book in the first place.

* Always have a separate code exercise for each library you load and each data set you use. A code exercise like "Load the **tidyverse** package with the `library()` command." is, obviously, not difficult for students. But it forces them to practice loading libraries, which many students forget, and it provides us with an opportunity to drop some knowledge.

* Regularly require them to look up the help page for a function, proving that they have done so by copy/pasting a portion of the help page, which means that these will be written exercises with no answer instead of code exercises. Students need to get in the practice of using help.

* You want students to have to type in at least 90% of the code which is used in the book. One approach is to first, set up a bunch of empty code exercises. Then, go through the chapter, copying each snippet of example code into the test code chunk of an exercise. (If the book is sourced freely, you can also copy/paste a knowledge drop associated with each code snippet.) Then, go back and write the exercises such that the answers are the code snippets from the book. The final step is to copy the code to the hint code chunk, and then edit out at least a portion of that code.

* In most books, the authors will include more than one new thing in each code example. They will add two or three lines to a pipe or pass in two or three arguments to a function. We never want to go that fast. Spread out such code snippets into two or three separate exercises, each of which makes the smallest possible change. We are building the shallowest possible learning curve. 

* Recall the distinction between books which have a permissive license, meaning that we can copy/paste text at our own discretion, and those which do not. For the latter, you can not copy/paste text. But you can express, in your own words, the key points made in each chapter. In either case, your knowledge drops should cover the most important things for students to know, in your opinion, among those topics covered in the chapter.

* A common annoyance is the creation of temporary objects. Consider this code snippet from *R for Data Science*.

````
x <- c("$1,234", "USD 3,513", "59%")
parse_number(x)
#> [1] 1234 3513   59
````

* It is unclear how to create `x` in your tutorial. You could have a code exercise in which the students create `x` themselves. But that is annoying because it doesn't really teach students anything. It is too close to busy work. Moreover, `x` won't exist for the next question, which requires the use of `parse_number()`. You will need to create `x` yourself, either in the global setup chunk or in a setup chunk for this exercise. The global setup approach is preferred, but does not work if there are multiple different `x` objects used in this chapter.

* I recommend that tutorial authors include the "busy work" in the exercise code chunk when writing the tutorial, followed by a blank line, so that students have an obvious place to type in `parse_number(x)`. 

````
x <- c("$1,234", "USD 3,513", "59%")

````

* This avoids the complication of creating/maintaining a bunch of temporary variables. It also means less busy work for students.

---
title: Introduction to R
author: David Kane
tutorial:
  id: introduction-to-r
output:
  learnr::tutorial:
    df_print: default
    progressive: yes
    allow_skip: yes
runtime: shiny_prerendered
description: Tutorial for introducing students to the R language.
---

```{r setup, include = FALSE}
library(learnr)
library(tutorial.helpers)
library(knitr)
library(tidyverse)

# The df_print: default above is the incantation which shows tibbles normally,
# without the fancy formatting associated with the pillar package.

knitr::opts_chunk$set(echo = FALSE)
knitr::opts_chunk$set(out.width = '90%')
options(tutorial.exercise.timelimit = 60, 
        tutorial.storage = "local") 
```

```{r copy-code-chunk, child = system.file("child_documents/copy_button.Rmd", package = "tutorial.helpers")}
```

```{r info-section, child = system.file("child_documents/info_section.Rmd", package = "tutorial.helpers")}
```


## Introduction
### 

This tutorial introduces you to the [R language](https://www.r-project.org/). Our approach is inspired by [*R for Data Science (2e)*](https://r4ds.hadley.nz/) by Hadley Wickham, Mine Çetinkaya-Rundel, and Garrett Grolemund. You will learn how to work with data sets using the **[tidyverse](https://tidyverse.tidyverse.org/)** meta-package. You will learn how to direct the result of one function to another using the pipe -- `|>` --- and how to make a plot using the `ggplot()` function.

### 

This tutorial assumes that you have already completed the "Getting Started" tutorial in the [**tutorial.helpers**](https://ppbds.github.io/tutorial.helpers/) package. If you haven't, do so now. It is quick!

### 

From the main Positron menu, start a new window with `File -> New Window`. This new window is the location in which you will do all the work for the tutorial. The current window, the one in which you are reading these words, is just used to run this tutorial. 

## Working with data
### 

Learn how to explore a data set using functions like `summary()`, `glimpse()`, and `slice_sample()`. 

### Exercise 1

Before you start doing data science, you must load the packages you are going to use. Use the function `library()` to load the **tidyverse** package. Click "Run Code." The check mark which appears next to "Exercise 1" above indicates that you have submitted your answer. It doesn't verify that you have answered the question correctly.

```{r working-with-data-1, exercise = TRUE}

```

```{r working-with-data-1-hint-1, eval = FALSE}
library(...)
```

```{r working-with-data-1-test, include = FALSE}
library(tidyverse)
```

### 

"Library" and "package" mean the same thing in R. We have different words for historical reasons. However, only the `library()` command will load a package/library, giving us access to the functions and data which it contains.

### Exercise 2

In this tutorial, you will sometimes enter code into the exercise blocks, as you did above. But we will also ask you to run code in the Console. (You will do this in the other Positron window, since the Console in this window is currently busy running this tutorial.) Example:

In the Console, run `library(tidyverse)`.

With Console questions, we will usually ask you to **C**opy/**P**aste the **C**ommand/**R**esponse into an answer block, like the one below. We usually shorten those instructions as CP/CR. Do that now.


```{r working-with-data-2}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 15)
```

###

Your answer should look like:

```
> library(tidyverse)
── Attaching core tidyverse packages ─────────────────────────────────────────────────────────────────────────────────────────── tidyverse 2.0.0 ──
✔ dplyr     1.1.4     ✔ readr     2.1.5
✔ forcats   1.0.0     ✔ stringr   1.5.1
✔ ggplot2   3.5.2     ✔ tibble    3.3.0
✔ lubridate 1.9.4     ✔ tidyr     1.3.1
✔ purrr     1.1.0     
── Conflicts ───────────────────────────────────────────────────────────────────────────────────────────────────────────── tidyverse_conflicts() ──
✖ dplyr::filter() masks stats::filter()
✖ dplyr::lag()    masks stats::lag()
ℹ Use the conflicted package to force all conflicts to become errors
>
```

Your answer never needs to match ours perfectly. Our goal is just to ensure that you are actually following the instructions.

### Exercise 3

Data frames, also referred to as "tibbles," are spreadsheet-type data sets. 

In the Console, run `diamonds`.

CP/CR.

```{r working-with-data-3}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 15)
```

###

```{r working-with-data-3-test, echo = TRUE}
diamonds
```

Whenever we show outputs like this after a question, then we are showing our answer to the previous question, even if we do not label it as such. 

### Exercise 4

In the Console, run `summary()` on `diamonds`.

CP/CR.

```{r working-with-data-4}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```{r working-with-data-4-test, echo = TRUE}
summary(diamonds)
```

###

This function provides a quick statistics overview of each variable in the data set. In some cases, as here, the tutorial displays the same object differently from what you were able to copy/paste. And that is OK! Your answer does not need to match our answer.

### Exercise 5

In the Console, run `slice_sample()` on `diamonds`. This selects a random row from the data set.

CP/CR.

```{r working-with-data-5}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 7)
```

###

```{r working-with-data-5-test, echo = TRUE}
slice_sample(diamonds)
```

###

Your answer will differ from this answer because of the inherent randomness in functions like `slice_sample()`.

### Exercise 6

In the Console, hit the Up Arrow to retrieve the previous command. Edit it to add the argument `n = 4` to `slice_sample(diamonds)`. This will return 10 random rows from the `diamonds` data set.

CP/CR.

```{r working-with-data-6}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 8)
```

###

```{r working-with-data-6-test, echo = TRUE}
slice_sample(diamonds, n = 4)
```

###

Editing code directly in the Console quickly becomes annoying. See the **[positron.tutorials](https://ppbds.github.io/positron.tutorials/)** package for tutorials about using Positron to write and organize your code.
 

### Exercise 7

In the Console, run `print()` on `diamonds`. This returns the same result as typing `diamonds`.

CP/CR.

```{r working-with-data-7}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```{r working-with-data-7-test, echo = TRUE}
print(diamonds)
```

###

You can choose how many rows to display by using the `n` argument in the `print()` function, and how many columns to display by using the `width` argument.

### Exercise 8

In the Console, run `print()` on `diamonds` with the argument `n = 3`. This returns the first 3 rows of the `diamonds` data set.

CP/CR.

```{r working-with-data-8}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```{r working-with-data-8-test, echo = TRUE}
print(diamonds, n = 3)
```

###

`print()`, by default, gives the top of the tibble, so your answer should match our answer. `slice_sample()`, on the other hand, picks random rows to return. But, in both cases, the result is a tibble. 

A central organizing principal of the *[Tidyverse](https://www.tidyverse.org/)* is that most functions take a tibble as their first and return a tibble. This allows us to "chain" commands together, one after the other.

### Exercise 9

In the Console, run `?diamonds`.

This will look up the help page for the `diamonds` tibble from the **ggplot2** package, which is one of the core packages in the *[Tidyverse](https://www.tidyverse.org/)*. The help page will appear on the right-side of your Positron window, in the Secondary Activity Bar, which you might need to activate in order to see. 

Copy/paste the Description section of the help page below.

```{r working-with-data-9}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```
### 

You can find help about an entire package with `help(package = "ggplot2")`. It is confusing, but unavoidable, that package names are sometimes unquoted, as in `library(ggplot2)`, and sometimes quoted, as in `help(package = "ggplot2")`. If one does not work, try the other.

### Exercise 10

In the Console, run `glimpse()` on `diamonds`.  CP/CR.

```{r working-with-data-10}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

<!-- We cheat a bit so that the display looks better, adding width = 60 and not echo'ing the code. -->

```{r working-with-data-10-test}
glimpse(diamonds, width = 60)
```

### 

`glimpse()` displays columns running down the page and the data running across across. Note how the "type" of each variable is listed next to the variable name. For example, `price` is listed as `<int>`, meaning that it is an integer variable. To learn more about the `glimpse()` function, run `?glimpse`.

`view()` is another useful function, but, because it is interactive, we should not use it within a tutorial. 

### Exercise 11

In the Console, run `sqrt(144)`.

CP/CR.

```{r working-with-data-11}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```{r working-with-data-11-test, echo = TRUE}
sqrt(144)
```

###

The square root function is one of many built-in functions in R. Most return their result, which R then, by default, prints out.

### Exercise 12

In the Console, run `x <- sqrt(144)`.

CP/CR.

```{r working-with-data-12}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```{r working-with-data-12-test, echo = TRUE}
x <- sqrt(144)
```

###

The symbol `<-` is the assignment operator. In this case, we are *assigning* the value of `sqrt(144)` to the variable `x`. Nothing is printed out because of that assignment.

Also, you can see `x` in the "Variables" tab under the "Session" pane in the Secondary Activity Bar on the right-hand side of the Positron window.


### Exercise 13

In the Console, run `x`.

CP/CR.

```{r working-with-data-13}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```{r working-with-data-13-test, echo = TRUE}
x
```

###

Now that `x` has been defined in the Console, it is available for your use. Above, we just print it out. But we could also use it in other calculations, i.e., `x + 5`.


## Pipes and plots
### 

Although the *Tidyverse* includes hundreds of commands for data manipulation, the most important are `filter()`, `select()`, `arrange()`, `mutate()`, and `summarize()`. 

### Exercise 1

Let's warm up by examining the `gss_cat` tibble from the **forcats** package. Since **forcats** is a core **tidyverse** package, you have already loaded it. Type `gss_cat` and hit "Run Code."

Instead of using the Console, we will be doing the exercises in this section using excercise blocks.

```{r pipes-and-plots-1, exercise = TRUE}

```

```{r pipes-and-plots-1-hint-1, eval = FALSE}
...
```

```{r pipes-and-plots-1-test, include = FALSE}
gss_cat
```

### 

As the help page notes, `gss_cat` is a "sample of categorical variables from the [General Social Survey](https://gss.norc.org/)."

### Exercise 2

Run `summary()` on `gss_cat`. 

```{r pipes-and-plots-2, exercise = TRUE}

```

```{r pipes-and-plots-2-hint-1, eval = FALSE}
summary(...)
```

```{r pipes-and-plots-2-test, include = FALSE}
summary(gss_cat)
```

### 

Note that there are missing values in some columns. The word `NA` stands for "Not Available" and is used to represent missing data in R.

### Exercise 3

Pipe `gss_cat` to `drop_na()`. This function removes rows with missing values.  The pipe symbol --- `->` --- allows us to chain R commands together, one after the other, with each one connected to the next with the pipe symbol. In this case, we want:

```
gss_cat |> 
  drop_na()
```

```{r pipes-and-plots-3, exercise = TRUE}

```

```{r pipes-and-plots-3-hint-1, eval = FALSE}
... |> 
  drop_na()
```

```{r pipes-and-plots-3-test, include = FALSE}
gss_cat |> 
  drop_na()
```

### 

Note the number of rows in the tibble after `drop_na()`. Since `drop_na()` removes rows with missing values, the number of rows in the tibble will be less than the original number of rows.

We could achieve the same result by running `drop_na(gss_cat)`. The symbol `|>` just "pipes" `gss_cat` into `drop_na()` as its first argument.

### Exercise 4

Pipe `gss_cat` to `filter()`.  Within `filter()`, use the argument `year == 2014`. 

```{r pipes-and-plots-4, exercise = TRUE}

```

```{r pipes-and-plots-4-hint-1, eval = FALSE}
gss_cat |> 
  ...(year == 2014)
```

```{r pipes-and-plots-4-test, include = FALSE}
gss_cat |> 
  filter(year == 2014)
```

### 

This workflow --- in which we pipe a tibble to a function, which then outputs another tibble, which we can then pipe to another function, and so on --- is very common in R programming.

The resulting tibble has the same number of columns as `gss_cat` because `filter()` only affects the rows. But there are many fewer rows.

### Exercise 5

Continue the code and pipe with `select()`, using the argument `age, marital, race, relig, tvhours`. Note that you do not need to retype the code from the last exercise. You can just click the "Copy Code" button.

```{r pipes-and-plots-5, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r pipes-and-plots-5-hint-1, eval = FALSE}
... |> 
  select(age, ..., race, ..., tvhours)
```

```{r pipes-and-plots-5-test, include = FALSE}
gss_cat |> 
  filter(year == 2014) |> 
  select(age, marital, race, relig, tvhours)
```

### 

Note how the Hint only gives the most recent line of the pipe. Because `select()` does not affect the rows, we have the same number as after `filter()`. But we only have 5 columns now, consistent with what we told `select()` to do.

### Exercise 6

Copy previous code. Continue the pipe with `summary()`

```{r pipes-and-plots-6, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r pipes-and-plots-6-hint-1, eval = FALSE}
... |> 
  summary()
```

```{r pipes-and-plots-6-test, include = FALSE}
gss_cat |> 
  filter(year == 2014) |> 
  select(age, marital, race, relig, tvhours) |>
  summary()
```

### 

Note that there are missing values in the `tvhours` column. Let's remove them.

### Exercise 7

Copy previous code. Replace the `summary()` with `drop_na()`.

```{r pipes-and-plots-7, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r pipes-and-plots-7-hint-1, eval = FALSE}
... |> 
  drop_na()
```

```{r pipes-and-plots-7-test, include = FALSE}
gss_cat |> 
  filter(year == 2014) |> 
  select(age, marital, race, relig, tvhours) |>
  drop_na()
```

### 

The number of rows has decreased because we removed rows with missing values. `drop_na()` removes all rows which have a missing value for any of the variables. If we wanted to just remove the rows which are missing `tvhours`, we would use `drop_na(tvhours)`.

### Exercise 8

Continue the pipe with `arrange()`, using `tvhours` as the argument.

```{r pipes-and-plots-8, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r pipes-and-plots-8-hint-1, eval = FALSE}
... |> 
  arrange(...)
```

```{r pipes-and-plots-8-test, include = FALSE}
gss_cat |> 
  filter(year == 2014) |> 
  select(age, marital, race, relig, tvhours) |>
  drop_na() |>
  arrange(tvhours)
```

### 

The `arrange()` function sorts the rows of a tibble. By default, it sorts in ascending order.

### Exercise 9

Copy the previous code. Put `desc()` around `tvhours` to sort in descending order.

```{r pipes-and-plots-9, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r pipes-and-plots-9-hint-1, eval = FALSE}
... |> 
  arrange(desc(...))
```

```{r pipes-and-plots-9-test, include = FALSE}
gss_cat |> 
  filter(year == 2014) |> 
  select(age, marital, race, relig, tvhours) |>
  drop_na() |>
  arrange(desc(tvhours))
```

### 

Got to respect someone who watches TV 24 hours a day!

### Exercise 10

Let's make a plot. Copy the previous code, and pipe to `ggplot()`. Set `aes(x = age, y = tvhours)`.

```{r pipes-and-plots-10, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r pipes-and-plots-10-hint-1, eval = FALSE}
... |> 
  ggplot(aes(x = ..., y = ...))
```

```{r pipes-and-plots-10-test, include = FALSE}
gss_cat |> 
  filter(year == 2014) |> 
  select(age, marital, race, relig, tvhours) |>
  drop_na() |>
  arrange(desc(tvhours)) |>
  ggplot(aes(x = age, y = tvhours))
```

### 

This will return a plain graph as we have not mapped any data to specific "aesthetics" yet.

### Exercise 11

Add another layer with `geom_jitter()` using the `+` sign. Plotting code in the **ggplot2** package uses `+`, not `|>`, to connect different commands together. This difference comes from the fact that **ggplot2** was written 10+ years before the pipe was invented.

```{r pipes-and-plots-11, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r pipes-and-plots-11-hint-1, eval = FALSE}
... + 
  geom_jitter()
```

```{r pipes-and-plots-11-test, include = FALSE}
gss_cat |> 
  filter(year == 2014) |> 
  select(age, marital, race, relig, tvhours) |>
  drop_na() |>
  arrange(desc(tvhours)) |>
  ggplot(aes(x = age, y = tvhours)) + 
    geom_jitter()
```

### 

This is a scatterplot of `age` versus `tvhours`. The x-axis is age, and the y-axis is the number of hours of TV watched per day.

### Exercise 12

Finally, add a title, subtitle, labels for x and y axes using `labs()`. The subtitle should be the one sentence of information about the graph with which you would hope a reader walks away. What is the most important fact demonstrated in the graphic? 

Consider this example graph:

```{r}
gss_cat |> 
  filter(year == 2014) |> 
  select(age, marital, race, relig, tvhours) |>
  drop_na() |>
  arrange(desc(tvhours)) |>
  ggplot(aes(x = age, y = tvhours)) + 
  geom_jitter() + 
  labs(title = "TV Hours Watched by Age", 
       subtitle = "Got to respect someone who watches TV 24 hours a day!", 
       x = "Age", 
       y = "TV Hours")
```

You can make yours look like ours, if you like.

```{r pipes-and-plots-12, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r pipes-and-plots-12-hint-1, eval = FALSE}
... + 
  labs(title = "...", 
       subtitle = "...", 
       x = "...", 
       y = "...")
```

```{r pipes-and-plots-12-test, include = FALSE}
gss_cat |> 
  filter(year == 2014) |> 
  select(age, marital, race, relig, tvhours) |>
  drop_na() |>
  arrange(desc(tvhours)) |>
  ggplot(aes(x = age, y = tvhours)) + 
  geom_jitter() + 
  labs(title = "TV Hours Watched by Age", 
       subtitle = "Got to respect someone who watches TV 24 hours a day!", 
       x = "Age", 
       y = "TV Hours")
```

### 

Note that the code in the exercise block is not saved. If you want to save the code, you can copy/paste it into an R script file.

## Generative AI
###

Gneerative AI --- tools like [ChatGPT](https://chat.openai.com/), [Grok](https://x.ai/), [Claude](https://claude.ai/), [DeepSeek](https://www.deepseek.com/en) and so on
--- are the future, of data science and everything else. The more you use these tools, the better off you will be. Unfortunately, the tools are changing so much that it is hard for a tutorial like this to stay up-to-date. This section provides some general advice and practice exercises.

### Exercise 1

Using any AI you like, ask it to write a one-sentence summary about the R programming language. Copy the answer below.

```{r generative-ai-1}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

Our answer:

```
R is a free, open-source programming language and software environment designed specifically for statistical computing, data analysis, and creating data visualizations.
```

If you do not want to pay for an AI service, then you will probably need to have free accounts with several different services. That way, if one service cuts you off for the day, you can switch to another.

### Exercise 2

Type `diamonds` in the Console. Hit `Enter`. Copy/paste the command and the first few lines of the tibble.

```{r generative-ai-2}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 7)
```

###

```{r generative-ai-2-test, echo = TRUE}
diamonds
```

When working with AI, you often need to tell it about the data set. The easiest way to do that is often to just copy/paste over the first few lines. That shows the AI what the variable names and types are, which is the key information it needs for creating plots and pipes.

### Exercise 3

Copy/paste the top of the `diamonds` tibble into your AI interface and ask it to create a pipe which calculates the average value of `carat` for all combinations of `cut` and `color`. Run the provided code in the Console. If it fails, show the AI the error and ask for better code.

CP/CR.

```{r generative-ai-3}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

Claude gave us this answer:

```{r generative-ai-3-test, echo = TRUE}
diamonds %>%
  group_by(cut, color) %>%
  summarise(avg_carat = mean(carat), .groups = "drop")
```

###

But we don't like this answer! First, it uses `%>%`, an older version of the pipe symbol, instead of `|>`. Second, it uses `group_by()`, which necessitates the addition of the confusing `.groups` argument. We would have preferred:

```{r}
diamonds |>
  summarise(avg_carat = mean(carat), .by = c(cut, color))
```

Using AI is good! But intelligent use --- use in which you understand what the AI has done and try to improve/clarify its answer --- is even better.

### Exercise 4

Ask AI to create a beautiful plot using `diamonds` and the **tidyverse** package. Run the provided code in the Console. If it fails, show the AI the error and ask for better code.


```{r generative-ai-4}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 15)
```

###

DeepSeek provides this code:

```{r generative-ai-4-test, echo = TRUE}
library(tidyverse)

# Create a sophisticated diamond price analysis plot
diamonds %>%
  sample_frac(0.3) %>% # Sample 30% for better performance
  mutate(price_per_carat = price / carat) %>%
  ggplot(aes(x = carat, y = price, color = cut)) +
  geom_point(alpha = 0.6, size = 1.5, position = position_jitter(width = 0.02)) +
  geom_smooth(formula = 'y ~ x', method = "loess", se = FALSE, linewidth = 1) +
  scale_y_continuous(labels = scales::dollar_format()) +
  scale_color_manual(
    values = c("Ideal" = "#E69F00", 
               "Premium" = "#56B4E9", 
               "Very Good" = "#009E73", 
               "Good" = "#F0E442", 
               "Fair" = "#0072B2")
  ) +
  facet_wrap(~color, labeller = labeller(color = ~paste("Color", .))) +
  labs(
    title = "Diamond Price Analysis: Carat vs Price by Cut Quality and Color",
    subtitle = "Relationship between diamond characteristics and market pricing",
    x = "Carat Weight",
    y = "Price (USD)",
    color = "Cut Quality",
    caption = "Data: ggplot2 diamonds dataset | Visualization: Tidyverse"
  ) +
  theme_minimal(base_size = 12) +
  theme(
    plot.title = element_text(face = "bold", size = 16, hjust = 0.5),
    plot.subtitle = element_text(size = 12, hjust = 0.5, color = "gray40"),
    legend.position = "bottom",
    strip.background = element_rect(fill = "gray95", color = NA),
    strip.text = element_text(face = "bold"),
    panel.grid.minor = element_blank(),
    plot.caption = element_text(color = "gray50", size = 9),
    legend.key.size = unit(0.8, "lines")
  )
```

###

Note:

* DeepSeek provided the `library(tidyverse)` even though that is not necessary. We have already loaded the **tidyverse** package. Loading it again is sloppy. It is up to you to keep track of which libraries are loaded.

* DeepSeek first provided code that used `library(ggtheme)`. This is a nice package which we have not loaded. But we also haven't installed it yet. Again, AI is good at generating code that will work if all its dependencies have been installed. You need to keep track. It is also useful to tell the AI which packages to use, and which to not use.

* AI tends to generate too much code. For example, our question does not require or suggest the use of `theme()`, much less all the specific settings which DeepSeek has provided. So, we often tell the AI to be "concise."

* AI will often make mistakes. The first version which DeepSeek provided used `size` instead of `linewidth` as an argument to `geom_smooth()` because that was the correct usage two years ago. DeepSeek also forgot the `formula = 'y ~ x'` argument. It is your job to read, understand and fix any messages/warnings/errors generated by your code.

The key is practice. Use AI every day!

## Summary
### 

This tutorial introduced you to the R language. Our approach was inspired by [*R for Data Science (2e)*](https://r4ds.hadley.nz/) by Hadley Wickham, Mine Çetinkaya-Rundel, and Garrett Grolemund. You learned how to work with data sets using the **[tidyverse](https://tidyverse.tidyverse.org/)** meta-package. You learned how to direct the result of one function to another using the pipe -- `|>` --- and how to make a plot using the `ggplot()` function.

```{r download-answers, child = system.file("child_documents/download_answers.Rmd", package = "tutorial.helpers")}
```

---
title: "Enable Students to Download Their Tutorial Answers"
author: David Kane
format: html
vignette: >
  %\VignetteEngine{quarto::html}
  %\VignetteIndexEntry{Enable Students to Download Their Tutorial Answers}
  %\VignetteEncoding{UTF-8}
---


The most common use of the **tutorial.helpers** package is to allow students to download their answers after they complete a tutorial. To enable this functionality, you need to, first, include `library(tutorial.helpers)` within the initial setup R code chunk and, second, add this line at the end of the tutorial:

````{verbatim}
```{r download-answers, child = system.file("child_documents/download_answers.Rmd", package = "tutorial.helpers")}
```
````

When students complete the tutorial, they will first be asked how long the tutorial took them to complete, in minutes. They will then be presented with the option to download a copy of their answers, in html format, to their Downloads folder. By default, the name of the downloaded file will be the `id` of the tutorial with `_answers.html` appended. 

So, for example, a tutorial with an `id` equal to `getting-started` will generate a file called `getting-started_answers.html`.

The package provides functions like `submissions_summary()` and `submissions_answers()` to work with a collection of students answers. 

---
title: Getting Started
author: David Kane
tutorial:
  id: getting-started
output:
  learnr::tutorial:
    progressive: yes
    allow_skip: yes
runtime: shiny_prerendered
description: Tutorial for introducing students to tutorials when using Visual Studio Code via GitHub Codespaces.
---

```{r setup, include = FALSE}
library(learnr)
library(tutorial.helpers)
library(knitr)

knitr::opts_chunk$set(echo = FALSE)
knitr::opts_chunk$set(out.width = '90%')
options(tutorial.exercise.timelimit = 60,
        tutorial.storage = "local")
```

<!-- Get rid of the images which use dark mode. They are unreadable! -->

## Introduction
###

This tutorial, intended for beginners, teaches students how to use [VS Code](https://code.visualstudio.com/) within [GitHub Codespaces](https://github.com/features/codespaces) to run R tutorials created with the [**learnr**](https://rstudio.github.io/learnr/) package. We also introduce students to the use of the bash Terminal, the R Terminal, and [GitHub Copilot](https://github.com/copilot), an AI coding assistant.

Most students begin this tutorial after completing the setup steps outlined [here](https://ppbds.github.io/primer/getting-started.html). If you have not done that yet, do so now.

###

After starting the tutorial, your Codespace looks something like this:

```{r}
include_graphics("images/gs-1.png")
```

###

Complete the tutorial in your Codespace. At the end, you will press a button to download a copy of your answers to the `Downloads` folder on your computer. Downloading only saves the file on your computer. If your instructor wants a copy, they will tell you how to send it to them.

###

Most tutorials begin by asking for your name and email. Your instructor might also ask you to fill in the "ID" field so that your work can be matched to a grading database. If not, leave the "ID" field blank. Click "Continue," then enter your name and email. Click the **"Submit Answer"** button each time. If you make a mistake, click **"Try Again"** and submit a new answer.

If you are completing this tutorial as part of a class, use the same email address for every tutorial.

```{r info-section, child = system.file("child_documents/info_section.Rmd", package = "tutorial.helpers")}
```

## Bash Terminal
###

Each tutorial includes exercises for you to complete. Most exercises ask you to run a command in a Terminal and then copy the result into the answer box. After writing your answer, click the "Submit Answer" button, as you did above for your name and email address.

###

The **bash** Terminal is a text-based interface in which you type commands to tell your Codespace what to do. The pattern is always the same: you send a command, the Terminal executes it, and it gives you something back. In this section you will practice the most common bash commands. (In the next section you will see that the R Terminal works exactly the same way.)

###

To do the exercises, you need a bash Terminal in which to type commands. Press the plus button (`+`) on the right side of the Panel to open one.


```{r}
include_graphics("images/gs-2.png")
```

### Exercise 1

In the bash Terminal, type `whoami` and then press `return` (on Mac) or `enter` (on Windows). To "run" a command means to type it and then press this key. To avoid repeating both terms in every question, we use `Enter` to mean either the `return` or `enter` key, as appropriate.

###

The other Mac/Windows issue to be aware of is the Command key (Mac) versus Control key (Windows). 

Many shortcut key commands use the Command key --- abbreviated to `Cmd` --- or the Control key --- abbreviated to `Ctrl` --- to perform various actions, depending on the operating system you are using. We will use `Cmd/Ctrl` to mean either key, as appropriate.

###

**Do not just type the command into the answer box!** You must run the command in the bash Terminal and then copy the result.

**C**opy and **P**aste both the **C**ommand *and* the **R**esponse from the Terminal into the box below. We will abbreviate this instruction in the future as **CP**/**CR**.

```{r bash-terminal-1}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

We typically provide example answers, like the one below, after you submit your own response. Do not be concerned if your answer differs from the example we show.

````
codespace-starter $ whoami
rstudio
codespace-starter $
````

###

`whoami` prints the username of the account you are signed in as. Your Codespace is a complete computer running in the cloud, and you interact with it as the `rstudio` user. It is OK if your formatting differs from ours. What matters is that you show that you executed the command as instructed.

### Exercise 2

In the bash Terminal, run `echo $0`. (Again, "run" means to type the command and then press `Enter`.)

CP/CR.

```{r bash-terminal-2}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

````
codespace-starter $ echo $0
/bin/bash
codespace-starter $
````

###

A "shell" is the program that reads and interprets the commands you type in the Terminal. `echo $0` prints the name of the shell you are using: `/bin/bash`, the [Bash shell](https://en.wikipedia.org/wiki/Bash_(Unix_shell)). Every command you run in this Terminal --- starting with the `whoami` you just ran --- is interpreted by bash.

### Exercise 3

In the bash Terminal, run `pwd`.

CP/CR.

```{r bash-terminal-3}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

````
codespace-starter $ pwd
/workspaces/codespace-starter
codespace-starter $
````

###

`pwd` stands for **p**rint **w**orking **d**irectory --- the folder the Terminal is currently "in." GitHub Codespaces place your repository under `/workspaces/`, so that is where you begin. Almost every command you run acts on this directory unless you tell it otherwise.

### Exercise 4

In the bash Terminal, run `ls`.

CP/CR.

```{r bash-terminal-4}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

````
codespace-starter $ ls
codespace-starter $
````

###

`ls` **l**i**s**ts the files in the working directory. Notice that nothing was printed! That is because, by default, `ls` hides any file or folder whose name begins with a dot (`.`). The directory is not actually empty, as we will see next.

### Exercise 5

In the bash Terminal, run `ls -a`.

CP/CR.

```{r bash-terminal-5}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

````
codespace-starter $ ls -a
.  ..  .devcontainer  .git
codespace-starter $
````

###

The `-a` flag (for "all") tells `ls` to reveal the hidden entries. `.` is the current directory and `..` is the directory just above it. `.git` stores the project's version history, and `.devcontainer` holds the configuration for your Codespace. A "flag" like `-a` is an option that changes how a command behaves.

### Exercise 6

In the bash Terminal, run `ls .devcontainer`.

CP/CR.

```{r bash-terminal-6}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

````
codespace-starter $ ls .devcontainer
AGENTS.md  STUDENT_WORKFLOW.md  connect-repo.sh  devcontainer.json  welcome.sh
codespace-starter $
````

###

Giving `ls` a directory name lists what is inside that directory. The `.devcontainer` folder tells GitHub Codespaces how to build your workshop --- which version of R, Python, and Quarto to install, and which AI tools to include --- so that every student gets an identical setup.

### Exercise 7

In the bash Terminal, run `head .devcontainer/STUDENT_WORKFLOW.md`.

CP/CR.

```{r bash-terminal-7}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 12)
```

###

````
codespace-starter $ head .devcontainer/STUDENT_WORKFLOW.md
# Student workflow

This Codespace is your **workshop** — it already has R, Python (with the
usual data-science libraries), Quarto, the AI assistants, and all the right
settings. But this `codespace-starter` repo isn't *yours*, so you won't save
your work here. Instead, you'll make your **own repo** and work in it. The
tools travel with the Codespace, so your repo needs no setup of its own.

## 1. Create or connect to your repo (one command)

codespace-starter $
````

###

`head` prints the first ten lines of a file, a quick way to peek at its contents without opening it. `STUDENT_WORKFLOW.md` explains the most important habit for this course: because `codespace-starter` is not your own repo, you create your own with `.devcontainer/connect-repo.sh` and do your work there.

### Exercise 8

Let us make the Terminal "sleep" by running `sleep 5` in the bash Terminal. The `sleep 5` command pauses execution for 5 seconds.

CP/CR.

```{r bash-terminal-8}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

````
codespace-starter $ sleep 5
codespace-starter $
````

###

Many commands do their work silently and print nothing --- success is simply the return of the prompt. We can change the duration with `sleep [number]`, where the number is how many seconds the Terminal pauses.

### Exercise 9

Run `sleep 60` in the bash Terminal. This command makes the Terminal sleep for 60 seconds. To cancel the running `sleep` command, press `Ctrl + C` on your keyboard. Do so now, before `sleep 60` finishes.

CP/CR.

```{r bash-terminal-9}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

````
codespace-starter $ sleep 60
^C
codespace-starter $
````

###

`Ctrl + C` interrupts whatever process is currently running in the Terminal and returns you to the prompt; the `^C` marks where you pressed it. This is one of the most useful things to know: when a command hangs or you change your mind, `Ctrl + C` stops it. You will use it again in the next exercise to stop the tutorial itself.

### Exercise 10

Read all of the following instructions before continuing. The purpose of this exercise is to demonstrate that your tutorial answers are saved automatically.

* Return to your Codespace and click the Terminal tab titled "R Tutorial" in the Panel.

```{r}
include_graphics("images/gs-3.png")
```

* Stop the tutorial by pressing `Ctrl + C` in the terminal. This is useful because it can interrupt any terminal process. A new line should appear where you can enter another command.

* To restart the tutorial, press the triangle button next to the "Getting Start" tutorial via the R Tutorials VS Code extension.

* Finally, open the tutorial.

Now that you have read all the instructions, go ahead and follow them.

###

Notice that all of your answers to previous questions have been saved. It is not a problem to quit a tutorial. In fact, if you leave a tutorial idle, it will eventually quit automatically.

## R Terminal
###

All the terminology associated with VS Code is confusing. The "Panel" is the bottom of the VS Code window. The are several "views" associated with the Panel, including "Output," "Ports," and "Terminal." Within the Terminal view, you can have multiple "terminals" running. Each terminal has an associated "tab" on the right side of the panel which you can use to select that terminal. Different terminals serve different purposes.
###

So far, we have learn about the "bash Terminal,"" which lets you talk to the computer. The "R Terminal" lets you talk to R, the programming language you will use for data science. The two work the same way: you type a command, press `Enter`, and the Terminal executes the command and gives you something back. The only real difference is the language --- bash commands like `ls` and `pwd` on the one hand, R commands like `sqrt(16)` on the other. Each exercise in this section asks you to run a one-line command and then CP/CR.

Create a new R Terminal by pressing the down arrow next to the plus button on the right side of the Panel Toolbar and selecting "New R Terminal."


```{r}
include_graphics("images/gs-4.png")
```

###

You may need to click on the new  "R Interactive" tab at the bottom of the list of terminals in order to bring it to the top of the Panel.

### Exercise 1

In the R Terminal, run `4 + 5`.

CP/CR.

```{r r-terminal-1}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

````
R 4.5.3> 4 + 5
[1] 9
R 4.5.3>
````

###

At its simplest, R is a calculator. Just as the bash Terminal took `whoami` and handed back your username, the R Terminal takes `4 + 5` and hands back the result. You send a command; the Terminal sends something back. The `[1]` at the start of the output is a counter telling you that the line begins with the first element of the result --- useful when a result has many values spread across several lines.

### Exercise 2

In the R Terminal, run `sqrt(16)`.

CP/CR.

```{r r-terminal-2}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

````
R 4.5.3> sqrt(16)
[1] 4
R 4.5.3>
````

###

`sqrt()` is a **function**. You "call" a function by writing its name followed by parentheses, placing any input --- here, `16` --- inside them. That input is called an **argument**. Almost everything you do in R involves calling functions on arguments.

### Exercise 3

In the R Terminal, run `getwd()`. This stands for **get** **w**orking **d**irectory, meaning that we want to know the directory in which the R process is currently located.

CP/CR.

```{r r-terminal-3}
question_text(NULL,
    answer(NULL, correct = TRUE),
    allow_retry = TRUE,
    try_again_button = "Edit Answer",
    incorrect = NULL,
    rows = 3)
```

###

````
R 4.5.3> getwd()
[1] "/workspaces/codespace-starter"
R 4.5.3>
````

###

`getwd()` is the R equivalent of the bash command `pwd` you ran earlier. Here it shows that R is operating inside your Codespace at `/workspaces/codespace-starter`. Note that `getwd()` takes no arguments, but you still need the parentheses to call it.

### Exercise 4

In the R Terminal, run `R.version.string`.

CP/CR.

```{r r-terminal-4}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

````
R 4.5.3> R.version.string
[1] "R version 4.5.3 (2026-03-11)"
R 4.5.3>
````

###

Your version of R may differ from the example above --- that's fine, as long as it's reasonably recent. `R.version.string` is not a function call --- it has no parentheses --- but a built-in variable whose value R simply prints.

### Exercise 5

`show_file()` is a commonly used function from the **tutorial.helpers** package. It makes it easy for you to share the contents of another file with your instructor, especially a file that you have edited.

You can generally call a function by giving its name followed by parentheses, as with `getwd()`. However, it can be safer to prefix the function name with the name of the package in which the function is found, separated by two colons. Since `show_file()` is in the **tutorial.helpers** package, we can call it using this "double-colon" notation: `tutorial.helpers::show_file()`.

In the R Terminal, run this command:

```
tutorial.helpers::show_file(file.path(R.home(), "COPYING"), end = 7)
```

CP/CR.

```{r r-terminal-5}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 10)
```

###

```{r r-terminal-5-test, include = FALSE}
tutorial.helpers::show_file(file.path(R.home(), "COPYING"), end = 7)
```

For now, don't worry about what `R.home()` and `file.path()` are doing, although you can read about them by running `?R.home` and `?file.path` in the R Terminal. Those details are less important than seeing the intended usage of `show_file()`.

### Exercise 6

Tutorials can include written responses. You have already seen several examples. Sometimes those written answers are simply copies of commands and their results. Other times, we will ask you to write one or more sentences of prose.

Copy and paste everything from the exercise header above, beginning with "Exercise 6," through the end of this sentence into the answer box below. In other words, you are copying from this tutorial, not from the Terminal. Press "Submit Answer."

```{r r-terminal-6}
question_text(NULL,
    message = "We will sometimes provide an answer to written questions, as we do here. Other times, we won't. In the former case, we will not allow you to resubmit your answer. Don't stress! Most instructors grade tutorials on pass/fail basis, so, as long as you make an honest effort, you will do fine.",
    answer(NULL,
           correct = TRUE),
    allow_retry = FALSE,
    incorrect = NULL,
    rows = 6)
```

### Exercise 7

We frequently ask you to `copy` content from an HTML file such as a webpage. Do not worry if what you copy appears messy or loses its formatting; that is perfectly fine. Our main concern is that what you paste contains roughly the information it is supposed to contain.

###

Go to [the homepage](https://ppbds.github.io/tutorial.helpers/) of the **tutorial.helpers** package and copy everything in the browser window from "tutorial.helpers" in the upper-left corner through the end of the `About this package` section.

CP/CR.

```{r r-terminal-7}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

````
tutorial.helpers
0.6.2.9008
Reference
Articles
Changelog
Search for
Tutorial Helper Functions
About this package
tutorial.helpers provides functions to help you write R tutorials, especially if you follow my advice about the best way to do so.
````

Again, your answer never needs to match ours exactly.

## GitHub Copilot
###

Your Codespace includes **GitHub Copilot**, an AI coding assistant that can answer questions, explain code, create and edit files, and run commands for you. 

Copilot Chat usually opens automatically when your Codespace starts. If it is not open, click the menu button in the top-left corner, marked with a red circle below. Select **View**, and then click **Chat**, marked with a red rectangle. You can also open Chat by pressing `Cmd/Ctrl + Alt + I`.

```{r}
include_graphics("images/cs-open-copilot-chat.png")
```

When the Copilot Chat view opens, select **Agent** from the mode menu near the chat box. Agent mode lets Copilot work with files and Terminals in your Codespace. If you are asked to sign in or authorize GitHub Copilot, follow the instructions on the screen. Keep the Chat view open throughout this section.

### Exercise 1

The bottom of the Chat window looks like this:

```{r}
include_graphics("images/gs-5.png")
```

In the Copilot chat box, type the following message and press `Enter`. Replace `<your-name>` with your actual name.

```
Hello! My name is <your-name>
```

Copy Copilot's response into the box below. This is the only exercise in which we ask you to copy what the AI says back. Usually, we will confirm its work by checking the file which it has edited.

```{r ai-assistant-1}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

###

```
Hello David! How can I help you today?
```

###

GitHub Copilot works like ChatGPT. You "talk" to it in plain English, and it responds. Your exact response will differ from ours --- AI does not give usually the same answer twice --- and that is fine.

### Exercise 2

Copilot can do much more than chat. In Agent mode, which is the default, it can create and edit files for you. Send this message:

```
Make a file called presidents.txt which includes the names of the first 5 US Presidents.
```

Copilot will create the file in your working directory. Copilot may open the new file in the Editor and display a **Keep** button. Click **Keep** to accept the file.

```{r}
include_graphics("images/cs-copilot-keep-editor.png")
```

The **Keep** button may also appear in a smaller change summary near the Copilot chat box.

```{r, out.width = '45%'}
include_graphics("images/cs-copilot-keep-summary.png")
```

From the R Terminal, run:

```
tutorial.helpers::show_file("presidents.txt")
```

CP/CR.

```{r ai-assistant-2}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 8)
```

###

````
George Washington
John Adams
Thomas Jefferson
James Madison
James Monroe
````

###

You can also see the new file without using the Terminal. Click the Explorer icon at the top of the Activity Bar. The Explorer lists every file in your working directory, and `presidents.txt` should now appear there. Click on it to open the file in the Editor and see its contents directly. The Explorer is useful whenever you want to inspect a file that AI created or edited.

The Explorer icon is marked with a red square below.

```{r}
include_graphics("images/gs-6.png")
```

### Exercise 3

We are done with the presidents example. Now ask Copilot to create a very small R program. Send this message:

```
Create an R script called script.R containing only this line: 2 + 2. Do not run the script.
```

If Copilot asks you to keep or accept the file, do so. Then run this command in the R Terminal:

```
tutorial.helpers::show_file("script.R")
```

CP/CR.

```{r ai-assistant-3}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

````
2 + 2
````

###

An R script is a plain-text file whose name ends in `.R`. Creating or opening a script does not run it. Running the script means asking R to execute the instructions saved inside it.

### Exercise 4

In the Explorer, click `script.R` to open it in the Editor. Click the triangular **Run Source** button in the upper-right corner of the Editor. Look in the R Terminal for the generated `source()` command, the line of code, and its result.

The Run Source button is marked with a red square below.

<!-- DK: Replace this nonsense. -->

<div style="position: relative; width: 90%; margin: 1em auto;">
  <img src="images/cs-run-source.png" alt="VS Code Editor with script.R open" style="display: block; width: 100%; height: auto;">
  <div aria-hidden="true" style="position: absolute; left: 93.1%; top: 0.5%; width: 1.9%; height: 4.5%; border: 4px solid red; box-sizing: border-box;"></div>
</div>

CP/CR.

```{r ai-assistant-4}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 10)
```

###

````
R 4.5.3> source("/workspaces/codespace-starter/script.R", encoding = "UTF-8", echo = TRUE)

> 2 + 2
[1] 4
R 4.5.3>
````

###

The Run Source button saves the open file and sends a `source()` command with its full path to the active R Terminal. The setting `echo = TRUE` results in each instruction being displayed before R executes it.

### Exercise 5

Make sure `script.R` is still open. Place your cursor on the line `2 + 2` and press `Cmd/Ctrl + Enter`. This runs the current line rather than sourcing the entire file.

CP/CR.

```{r ai-assistant-5}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 7)
```

###

````
R 4.5.3> 2 + 2
[1] 4
R 4.5.3>
````

###

The Editor must have focus so VS Code knows which line to send to the active R Terminal.

### Exercise 6

Keep your cursor in the `script.R` Editor and press `Cmd/Ctrl + Shift + Enter`. This sources the entire file with echo, just like the Run Source button. CP/CR.

```{r ai-assistant-6}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 10)
```

###

````
R 4.5.3> source("/workspaces/codespace-starter/script.R", encoding = "UTF-8", echo = TRUE)

> 2 + 2
[1] 4
R 4.5.3>
````

###

The button and the keyboard shortcut both source the saved file in your current R session. Running the current line also uses that session, but it sends only the line where your cursor is located.

### Exercise 7

Now run the script outside your open R session. Press the plus button (`+`) on the right side of the Panel to open a new bash Terminal. At the bash prompt, run:

```
Rscript script.R
```

CP/CR.

```{r ai-assistant-7}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 8)
```

###

````
codespace-starter $ Rscript script.R
[1] 4
codespace-starter $
````

###

`Rscript` starts a new R process, runs the file, and closes that process when the script finishes. It does not use the R session that is already open in your other Terminal.

### Exercise 8

AI can also run files for you. Return to Copilot Chat and tell your AI:

```
Run script.R
```

Copilot will usually choose a command such as `Rscript script.R`. If it asks for permission, provide it. Copilot should report the result in the Chat view. Enter that result in the box below.

```{r ai-assistant-8}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 2)
```

###

`4`

###

You have now run the same script five ways: with the Run Source button, with two keyboard shortcuts, by typing `Rscript` in a bash Terminal, and by asking Copilot to run the script. Asking Copilot delegates the work, but Copilot still uses commands on the computer to carry it out.

### Exercise 9

So far, `script.R` calculates `2 + 2` and displays the result, but it does not give that result a name. In R, an **object** is a name used to store a value so you can use it again. The instruction `x <- 2 + 2` creates an object named `x` and stores the value `4` in it. An object exists only in the R session where it was created.

Then ask Copilot to edit and run the script:

```
Edit script.R so its only line is x <- 2 + 2. Then run it with Rscript.
```

If Copilot asks for permission, inspect and approve the command. When it finishes, return to the same R Terminal and run:

```
exists("x")
```

CP/CR.

```{r ai-assistant-9}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 4)
```

###

````
R 4.5.3> exists("x")
[1] FALSE
R 4.5.3>
````

###

Copilot used `Rscript`, which created `x` in a separate R process and then closed that process. Your open R session therefore still reports `FALSE`: it does not know about the `x` that briefly existed in the other session.

### Exercise 10

Open `script.R` in the Editor. It should now contain `x <- 2 + 2`. Keep your cursor in the Editor and press `Cmd/Ctrl + Shift + Enter` to source the file in your open R session. Then run this command in the R Terminal:

```
x
```

CP/CR.

```{r ai-assistant-10}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 10)
```

###

````
R 4.5.3> source("/workspaces/codespace-starter/script.R", encoding = "UTF-8", echo = TRUE)

> x <- 2 + 2
R 4.5.3> x
[1] 4
R 4.5.3>
````

###

This time, `x` exists because the Editor sourced `script.R` in your open R session. The five ways to run the script do not all use the same R session:

| Method | R session used |
|---|---|
| Run Source button | Your open R Terminal |
| `Cmd/Ctrl + Enter` | Your open R Terminal; current line only |
| `Cmd/Ctrl + Shift + Enter` | Your open R Terminal; entire file |
| `Rscript script.R` in bash | A new process that closes afterward |
| Ask Copilot to use `Rscript` | A new process that closes afterward |

## Your answers
###

At the end of every tutorial, you will download a copy of your answers. We will do this in the next topic, labeled "Download answers." Follow the instructions there.

Make sure to answer the question about how many **minutes** you spent on the tutorial and to click "Submit Answer."

After clicking the "Download" button, you will be prompted to save a file named `getting-started_answers.html`. By default, this file will be saved in the `Downloads` folder on your computer. This action only downloads the file; it does not send the file to anyone.

Unless your instructor tells you otherwise, you should use the default file name, which will usually be of the form `short-name-of-tutorial` + `_answers` + `.html`.

```{r your-answers-0, child = system.file("child_documents/download_answers.Rmd", package = "tutorial.helpers")}
```

## Summary
###

You now understand how tutorials work, how to use bash and R Terminals, how to ask GitHub Copilot to create and edit files, and five ways to run an R script. You also know how to download a copy of your answers. Good luck with your data science journey!

<p>
  <img src="images/neo-cs.gif" alt="Neo learns codespaces" width="320">
</p>

---
title: "Python Test Document"
---

# Introduction

This is a test document with Python chunks.

```{python}
import pandas as pd
```

Some text here.

```{python}
df = pd.DataFrame({"x": [1, 2, 3]})
df.head()
```

---
title: Data transformation
author: David Kane
tutorial:
  id: data-transformation
output:
  learnr::tutorial:
    progressive: yes
    allow_skip: yes
runtime: shiny_prerendered
description: 'Tutorial for Chapter 4: Data transformation'
---

```{r setup, include = FALSE}
library(learnr)
knitr::opts_chunk$set(echo = FALSE)
options(tutorial.exercise.timelimit = 60, 
        tutorial.storage = "local") 
```

```{r copy-code-chunk, child = system.file("child_documents/copy_button.Rmd", package = "tutorial.helpers")}
```

```{r info-section, child = system.file("child_documents/info_section.Rmd", package = "tutorial.helpers")}
```

<!-- There are plenty of exercises in this chapter. We could add them here or make another tutorial: `data-transformation-exercises`. -->

<!-- This help page --- https://dplyr.tidyverse.org/reference/dplyr_tidy_select.html --- has lots of great stuff. <Should find a way to mention lots of items from this help page throughout this tutorial, maybe in the select() questions.  -->

## Introduction
### 

This tutorial covers [Chapter 4: Data transformation](https://r4ds.hadley.nz/data-transform.html) from [*R for Data Science (2e)*](https://r4ds.hadley.nz/) by Hadley Wickham, Mine Çetinkaya-Rundel, and Garrett Grolemund. You will learn about the key functions from the [**dplyr**](https://dplyr.tidyverse.org/) package for working with data including `[filter()](https://dplyr.tidyverse.org/reference/filter.html)`,  `[arrange()](https://dplyr.tidyverse.org/reference/arrange.html)`, `[select()](https://dplyr.tidyverse.org/reference/select.html)`, `[mutate()](https://dplyr.tidyverse.org/reference/mutate.html)`, and `[summarize()](https://dplyr.tidyverse.org/reference/summarise.html)`.

The goal of this chapter is to give you an overview of all the key tools for transforming a data frame. We’ll start with functions that operate on rows and then columns of a data frame, then circle back to talk more about the pipe, an important tool that you use to combine verbs.



## Summary
### 

This tutorial covered [Chapter 4: Data transformation](https://r4ds.hadley.nz/data-transform.html) from [*R for Data Science (2e)*](https://r4ds.hadley.nz/) by Hadley Wickham, Mine Çetinkaya-Rundel, and Garrett Grolemund. You learned about the key functions from the [**dplyr**](https://dplyr.tidyverse.org/) package for working with data including [`filter()`](https://dplyr.tidyverse.org/reference/filter.html),  [`arrange()`](https://dplyr.tidyverse.org/reference/arrange.html), [`select()`](https://dplyr.tidyverse.org/reference/select.html), [`mutate()`](https://dplyr.tidyverse.org/reference/mutate.html), and [`summarize()`](https://dplyr.tidyverse.org/reference/summarise.html).


<!-- Something about janitor::clean_names()` perhaps? Maybe: https://cran.r-project.org/web/packages/janitor/vignettes/janitor.html-->

```{r download-answers, child = system.file("child_documents/download_answers.Rmd", package = "tutorial.helpers")}
```

## My section
###

### Exercise 1

### Exercise 2

### Exercise 3


```{r my-section-3}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```{r my-section-3-test, echo = TRUE}

```

###

### Exercise 4


```{r my-section-4}
question_text(NULL,
	message = "Place correct answer here.",
	answer(NULL, correct = TRUE),
	allow_retry = FALSE,
	incorrect = NULL,
	rows = 6)
```

###

### Exercise 5


```{r my-section-5, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r my-section-5-hint, eval = FALSE}

```

```{r my-section-5-test, include = FALSE}

```

###

---
title: 'Data: Webscraping'
tutorial:
  id: data-webscraping
output:
  learnr::tutorial:
    progressive: yes
    allow_skip: yes
runtime: shiny_prerendered
description: Scraping data from the web.
---

```{r setup, include = FALSE}
library(learnr)
library(tutorial.helpers)
library(rvest)

knitr::opts_chunk$set(echo = FALSE)
options(tutorial.exercise.timelimit = 600, 
        tutorial.storage = "local") 

### HTML Set up objects

raw_html_1 <- "<html>Hello World</html>"

raw_html_2 <- "<html>
<table>
  <tr>
    <td>
      Hello World
    </td>
  </tr>
</table>
</html>"

raw_html_3 <- "<html>
<table>
  <tr>
    <td>
      Hello World
    </td>
    <td>
      foo
    </td>
    <td>
      bar
    </td>
  </tr>
</table>
</html>"

td_tags <- minimal_html(raw_html_3) |> html_elements("td")

### CSS Set Up

anchor <- '<a https://ppbds.github.io/primer/index.htm" id="abcd1234" class="hello hi">link to primer</a>'

### Open Secrets Set Up

web_url <- "https://www.opensecrets.org/political-action-committees-pacs/foreign-connected-pacs/2020"

local_url <- "data/opensecrets_backup.html"

raw_data_opensecrets <- read_html(local_url) |> 
  html_element("table.DataTable-Partial") |> 
  html_table()

### Wikipedia Set Up

wiki_url <- "https://en.wikipedia.org/w/index.php?title=%22,%22Gun_violence_in_the_United_States_by_state%22,%22&direction=prev&oldid=810166167"

raw_data_wiki <- read_html(wiki_url) |>
  html_element("#mw-content-text") |>
  html_element("table.wikitable") |>
  html_table()
```

<!-- DK: I think that all objects need to be created in the initial set up chunk. This is required if a student is to use Run Tutorial from the Tutorial window. I *think* however, that this is not needed if a student chooses Run Document after opening the .Rmd for the tutorial. That is a guess! Or maybe this has something to do with us testing our tutorials from within tutorial.helpers itself. That seems less likely. -->

<!-- DK: Why did I need to move the html_raw creation into this code chunk? Why didn't it work in its own code chunk? All the other set up code chunks seem to work fine? At least they do when I do the tutorial from Run Document . . . -->

<!-- DK: You skip lines a lot. Is this good or bad? I am not sure! Does it make it more likely that students will read things? -->

<!-- DK: Explain somewhere what abbreviations like "td" stand for. -->


```{r copy-code-chunk, child = "../../child_documents/copy_button.Rmd"}
```

```{r info-section, child = "../../child_documents/info_section.Rmd"}
```

## HTML

HTML, or "HyperText Markup Language" is a markup language that is used to display webpages on browsers. You might have seen html files as a result of knitting Rmd files.

### 

### Exercise 1

In this tutorial, you will use the **rvest** package to read the HTML of websites and extract data. First, load the **rvest** package.

You may need to install the **rvest** package in RStudio's console if the code box below outputs an error. If so, use `install.packages("rvest")` in Console. Do not issue this command within the tutorial itself.

```{r html-1, exercise = TRUE}

```

```{r html-1-hint-1, eval = FALSE}
library(...)
```

### 

**[rvest](https://rvest.tidyverse.org/)** is part of the **tidyverse**, a collection of packages, but it is not part of the core **tidyverse** package so you have to load it in separately.

### Exercise 2

HTML uses many nested layers of different tags to organize structure. For example: An html file like this:

```{r echo = TRUE, eval = FALSE}
<html>
<p>
Hello World
</p>
</html>
```

creates the following:

<br>
<html>
<p  style="background-color:lightgray">
Hello World
</p>
</html>
<br>

Tags are created with an "opening" tag like `<...>` and a "closing" tag like `</...>`, similar to parentheses. Note how the closing tag includes a forward slash. This example uses the html tag `<html>`, which is the main tag that includes the entire webpage.

### 

You will be parsing the following html structure:

```{r echo = TRUE, eval = FALSE}
<html>
<p>
Hello World
</p>
</html>
```

<!-- DK: This question is awkward. It had two versions of the html, one with <p> and one without. If we want to do both, this should be split into two questions. Parse the simple Hello World first. I tried to fix this by deleting the <p> tags. Is that correct? Or should the p tags be everywhere, like they seem to be with Exercise 4? Looks like the p tag is there by default, that we don't need to include it specifically. Make that clear. -->

<!-- NWL: The <p>, <body>, and <head> tags are all supplied by the function minimal_html(), so I think it would make more sense to include <p> tags in the example since usually when webscraping they would be there. -->

Assign the provided html code into the variable `raw_html_1` for easy access.

```{r html-2, exercise = TRUE}
"<html><p>Hello World</p></html>"

```

```{r html-2-hint-1, eval = FALSE}
raw_html_1 <- "<html>Hello World</html>"

```

<!-- DK: Instead of creating permanent R objects --- which can lead to all sorts of problems --- why not just do everything as a pipe. Like:  -->

<!-- "<html>Hello World</html>"  |>  -->
<!--     minimal_html() -->

<!-- Then each example is guaranteed to work.     -->

<!-- NWL: I just think it would be easier to copy/paste with variables. -->

### 

You can now use the `raw_html_1` variable

### Exercise 3

Run `minimal_html()` on `raw_html_1`. This function takes the raw html text and creates an object to easily access information.

Later, you will read html directly from websites.


```{r html-3, exercise = TRUE}

```

```{r html-3-hint-1, eval = FALSE}
raw_html_1 |>
  minimal_html()
```

### 

**rvest** reads html data and constructs a nested list-like object to represent the usual nested structure of html.

The output is similar to the output of a list.

### Exercise 4

Copy/paste your code from above and extend it using `|>`. Use the function `html_element()` to select the first paragraph tag `<p>` inside the structure.

`html_element()` takes a string that shows the type of a tag.
For example, `<p>` can be found with `"p"`.

```{r html-4, exercise = TRUE}

```

```{r html-4-hint-1, eval = FALSE}
... |>
  html_element(...)
```

```{r html-4-hint-2, eval = FALSE}
... |>
  html_element("p")
```

### 

You can see it returned an `{html_node}` object that is the tag `<p>`

### Exercise 5

Copy/paste your pipe from above and extend it. Use the function `html_text2()`, which prints the text inside a tag.

```{r html-5, exercise = TRUE}

```

```{r html-5-hint-1, eval = FALSE}
... |>
  html_text2()
```

### 

The output should be "Hello World". Now you can access the inside of a tag.

### Exercise 6

Usually, html tag structures are nested like the following:

```{r echo = TRUE, eval = FALSE}
<html>
<table>
  <tr>
    <td>
      Hello World
    </td>
  </tr>
</table>
</html>
```

<!-- DK: Maybe show what this looks like? Maybe explain it in a sentence or two? -->

This html, for example, represents a `<table>` that has one table row `<tr>`, and one table data `<td>` in that `<tr>`.

<!-- AG: I'll be putting HTML code in here that includes the picture. This also lines up with the primer. -->

<html>
<table style="border: 1px solid black">
  <tr style="border: 1px solid black">
    <td style="border: 1px solid black">
      Hello World
    </td>
  </tr>
</table>
</html>

### 

You will parse the above html

Assign the provided html to a variable named `raw_html_2`

```{r html-6, exercise = TRUE}
"<html>
<table>
  <tr>
    <td>
      Hello World
    </td>
  </tr>
</table>
</html>"
```

```{r html-6-hint-1, eval = FALSE}
raw_html_2 <- "..."
```

### 

Now you have access to `raw_html_2`

### Exercise 7

<!-- DK: Pipes beginw ith objects. Should be raw_html_2 |> minimal_html(). -->

Start a pipe using the function `minimal_html()` and `raw_html_2` as its argument

```{r html-7, exercise = TRUE}

```

```{r html-7-hint-1, eval = FALSE}
... |>
  minimal_html()
```

### 

Now you can search for elements inside the html.

### Exercise 8

Copy/paste your code from above and extend it with `|>`. Using `html_element()`, find the first `<table>` tag.

Remember that `html_element()` takes the type of tag you want to search for as an argument.

So to search for `<table>`, you can use `"table"` as an argument.

```{r html-8, exercise = TRUE}

```

<!-- DK: Format all hints correctly: new lines after the pipe symbol. -->

```{r html-8-hint-1, eval = FALSE}
... |> 
  html_element(...)
```

```{r html-8-hint-2, eval = FALSE}
... |>
  html_element("table")
```

### 

The output like last time returns an `{html_node}` object that has the `<table>` tag.

The lines after the `<table>` show the elements inside `<table>`.

### Exercise 9

Copy/paste your pipe from above and extend it. Use the function `html_element()` again to search for `<td>` tag inside the `<table>` tag.

```{r html-9, exercise = TRUE}

```

```{r html-9-hint-1, eval = FALSE}
... |>
  html_element(...)
```

```{r html-9-hint-2, eval = FALSE}
... |>
  html_element("td")
```

### 

By extending your pipe from the resulting `<table>` tag, you are limiting your searches and functions to only inside the `<table>` tag. You can see from the results that you have selected the `<td>` tag that is inside the `<table>` tag.

### Exercise 10

Copy/paste your pipe from above and extend it. Use the `html_text2()` function to view the text inside the `<td>` tag.

```{r html-10, exercise = TRUE}

```

```{r html-10-hint-1, eval = FALSE}
... |>
  html_text2()
```

### 

Note that there is an older function called `html_text()`, but its default behavior is much less convenient than `html_text2()`. You should always use `html_text2()` unless you have a good reason not to.

### Exercise 11

Usually there are many elements with the same tag on a webpage, which means selecting only the first element with a tag using `html_element()` will not work.

### 

Consider the following html:
```{r echo = TRUE, eval = FALSE}
<html>
<table>
  <tr>
    <td>
      Hello World
    </td>
    <td>
      foo
    </td>
    <td>
      bar
    </td>
  </tr>
</table>
</html>
```

There are 3 separate `<td>` tags. If you want to select the middle `<td>` with `foo` as its text, it will be hard for you to use `html_element()`.

### 

You will parse the above html

Assign the provided html to a variable named `raw_html_3`

```{r html-11, exercise = TRUE}
"<html>
<table>
  <tr>
    <td>
      Hello World
    </td>
    <td>
      foo
    </td>
    <td>
      bar
    </td>
  </tr>
</table>
</html>"
```

```{r html-11-hint-1, eval = FALSE}
raw_html_3 <- "..."
```

### 

You now have access to `raw_html_3`

### Exercise 12

Start a pipe using `minimal_html()` and `raw_html_3` as its argument.

```{r html-12, exercise = TRUE}

```

```{r html-12-hint-1, eval = FALSE}
raw_html_3 |>
  minimal_html()
```

### 

You can now search through `raw_html_3` for elements.

### Exercise 13

Copy/paste your code from above and extend it with `|>`. You will use the function `html_elements()` (Note that it is `elements`, not `element`) to find all `<td>` tags.

`html_elements()` takes the same argument as `html_element()`. Use the tag type of `<td>`, which is `"td"`.

```{r html-13, exercise = TRUE}

```

```{r html-13-hint-1, eval = FALSE}
... |>
  html_elements("...")
```

```{r html-13-hint-2, eval = FALSE}
... |>
  html_elements("td")
```

### 

`html_elements()` is very similar to `html_element()`, only that it returns all results compared to `html_element()` first result.

The output returns a `{xml_nodeset (3)}` object, which is basically a list, and then lists all the results.

### Exercise 14

Since the result is a list, you can navigate it exactly like how you would with a list.

### 

Copy/paste your pipe from above and assign it to a variable called `td_tags`. You will use it in the next exercise.

```{r html-14, exercise = TRUE}

```

```{r html-14-hint-1, eval = FALSE}
td_tags <- raw_html_3 |>
  minimal_html() |> ...
```

### 

You now have `td_tags`, which is essentially a list object.

### Exercise 15

Index `td_tags` and extract the second object using brackets.

```{r html-15, exercise = TRUE}

```

```{r html-15-hint-1, eval = FALSE}
td_tags[[...]]
```

### 

Remember to use double brackets to extract an object from a list.

You have selected the second object in `td_tags`, which should print out an `{html_node}` object like you've seen before in earlier exercises.

### Exercise 16

Copy/paste your indexing code and pass it into `html_text2()` as an argument. This will extract the text of the `<td>` tag that you have selected.

```{r html-16, exercise = TRUE}

```

```{r html-16-hint-1, eval = FALSE}
td_tags[[...]] |> 
  html_text2()
```

<!-- DK: Note how I changed the above hint to be a pipe. Everything is a pipe. Much easier for beginning students. -->

### 

Well done! This is the basic structure of HTML. It is tags inside tags inside tags.

Now you can navigate the nested structure of html, select  tags based on position and tag type, and extract the inner text of tags.

## CSS Selectors

<!-- AG: This section is very long. I plan on changing it so that it goes into less detail on what CSS is, as that should be left to the primer. -->

CSS, or Cascading Style Sheets is a language that defines the style of webpages (colors, boxes, spacing, etc).

CSS uses different attributes of tags to find which tags should be formatted in which way.

### 

Consider the following CSS code:
```{r eval = FALSE, echo = TRUE}
p {
color: red
}
```

This code formats all the `<p>` tags red.

You can utilize this system to find the tags that you want.

### Exercise 1

A tag's attributes are shown like such in html:
``` {r echo = TRUE, eval = FALSE}

<a href="https://ppbds.github.io/primer/index.html" id="abcd1234" class="hello hi">
link to primer
</a>

```

In the above example, the anchor, or hyperlink tag `<a>` has 3 different attributes:<br>
`href` which contains the destination link<br>
`id` which contains a unique id<br>
`class` which contains the classes that this tag is a part of.

### 

Assign the html for the `<a>` tag into a variable named `anchor`

```{r css-selectors-1, exercise = TRUE}
'<a href="https://ppbds.github.io/primer/index.html" id="abcd1234" class="hello hi">link to primer</a>'

```

```{r css-selectors-1-hint-1, eval = FALSE}
anchor <- '...'
```

### 

You now have access to the `anchor` variable.

The anchor tag `<a>` is simply a link to another page.

### Exercise 2

Start a pipe with `minimal_html()` and `anchor` as its argument.

```{r css-selectors-2, exercise = TRUE}

```

```{r css-selectors-2-hint-1, eval = FALSE}
anchor |> 
  minimal_html()
```

### 

You have loaded the html and can now start to select elements inside.

### Exercise 3

You have been using `html_element()` to find elements with the type of tag (`<a>`, `<p>`, `<table>`, etc), but you can also select an element with CSS attributes.

### 

``` {r echo = TRUE, eval = FALSE}

<a href="https://ppbds.github.io/primer/index.html" id="abcd1234" class="hello hi">
link to primer
</a>

```

Copy/paste your code from above and extend it with `|>`. Then use the function `html_element()`.

Select the `<a>` tag using its `id` attribute, `"abcd1234"`.

You specify that you are filtering by `id` in `html_element()` by using `"#id_name"`.

```{r css-selectors-3, exercise = TRUE}

```

```{r css-selectors-3-hint-1, eval = FALSE}
... |>
  html_element("#abcd1234")
```

### 

Since `id` is a unique identifier, every page should only have one tag that uses that `id`, which is helpful to select specific elements.

### Exercise 4

Copy/paste your pipe from above and extend it with the function `html_text2()`. This prints out the text inside the `<a>` tag that you have selected.

```{r css-selectors-4, exercise = TRUE}

```

```{r css-selectors-4-hint-1, eval = FALSE}
... |>
  html_text2()
```

### 

The output is the text inside, which is "link to primer"

### Exercise 5

Start a new pipe using `minimal_html()` and `anchor`.

```{r css-selectors-5, exercise = TRUE}

```

### 

Now we will try to select using the `class` attribute instead of `id`.

### Exercise 6

``` {r echo = TRUE, eval = FALSE}

<a href="https://ppbds.github.io/primer/index.html" id="abcd1234" class="hello hi">
link to primer
</a>

```

Copy/paste your code from above and extend it with `|>`. Then use the function `html_element()`.

Select the `<a>` tag using its class attribute, `"hello"`. You specify that you are filtering by `class` in `html_element()` by preceding a class name with `"."`.

```{r css-selectors-6, exercise = TRUE}

```

```{r css-selectors-6-hint-1, eval = FALSE}
... |>
  html_element(".hello")
```

### 

You will notice that the `<a>` tag's full class attribute is actually `"hello hi"`. This means that this `<a>` tag belongs to both the `hello` class and the `hi` class. (You can try selecting with `".hi"` instead of `".hello"`)

Classes are not unique to one tag, so usually there are multiple tags with the same class and tags with multiple classes.

### Exercise 7

Since there can be multiple tags with the same class, it is often safer to use the `html_elements()` to return all results, so you can better find the element you want.

### 

Copy/paste your pipe from above and extend it with `html_text2()`. You can view the inner text of the `<a>` tag.

```{r css-selectors-7, exercise = TRUE}

```

```{r css-selectors-7-hint-1, eval = FALSE}
... |>
  html_text2()
```

### 

Now you can select with class and id as well as extract text from that element.

### Exercise 8

You can also filter by id, class, and type of tag at the same time by chaining them together.

Consider the following code:
```{r eval = FALSE, echo = TRUE}

some_html |>
  minimal_html() |>
  html_element("p#id_name")

```

This code selects the first `<p>` tag with `"id_name"` as its `id`

### 

Start a new pipe with `minimal_html()` and `anchor`, then extend it with `|>`. Use the `html_element()` function to select the anchor tag by specifying its tag type and `id`.

Remember that to select `id`, you have to precede the `id` value with a `#`.

The `<a>` tag's `id` is `"abcd1234"`

```{r css-selectors-8, exercise = TRUE}

```

```{r css-selectors-8-hint-1, eval = FALSE}
anchor |>
  minimal_html() |>
  html_element("a#abcd1234")
```

### 

Though it may seem redundant to be even more specific, there are always weird website html that could break the constraint of having only one unique `id`.

Adding this extra layer of specificity just makes your selection more accurate.

### Exercise 9

Copy/paste your pipe from above and extend it. Use the function `html_text2()` to see the inner text of the tag.

```{r css-selectors-9, exercise = TRUE}

```

```{r css-selectors-9-hint-1, eval = FALSE}
... |> 
  html_text2()
```

### 

You can now select by chaining tag type and id value, then extract the text.

### Exercise 10

Consider the following code:
```{r eval = FALSE, echo = TRUE}

some_html |>
  minimal_html() |>
  html_element("p.class_name")

```

This code selects the first `<p>` tag with `"class_name"` as one of its `class`

### 

Start a new pipe with `minimal_html()` and `anchor`, then extend it with `|>`. Use the `html_element()` function to select the anchor tag by specifying its tag type and `class`.

Remember that to select `class`, you have to precede the `class` value with a `.`.

The `<a>` tag's `class` attribute is `"hello hi"`

```{r css-selectors-10, exercise = TRUE}

```

```{r css-selectors-10-hint-1, eval = FALSE}
... |>
  html_element("a.hello")

# or 

... |>
  html_element("a.hi")
```

### 

Now you can select using class and tag type at the same time.

### Exercise 11

Copy/paste your pipe from above and extend it. Use the function `html_text2()` to see the inner text of the tag.

```{r css-selectors-11, exercise = TRUE}

```

```{r css-selectors-11-hint-1, eval = FALSE}
... |> 
  html_text2()
```

### 

Well done.

### Exercise 12

Finally, consider the following code:
```{r eval = FALSE, echo = TRUE}

some_html |>
  minimal_html() |>
  html_element("p#id_name.class_name")

```

This code selects the first `<p>` tag with both `"id_name"` as its `id` AND `"class_name"` as one of its `class`

### 

Start a new pipe with `minimal_html()` and `anchor`.

```{r css-selectors-12, exercise = TRUE}

```

### Exercise 13

Copy/paste your code from above and extend it with `|>`. Use the `html_element()` function to select the anchor tag by specifying its tag type, `id`, and `class`.

html of `<a>` tag:
``` {r echo = TRUE, eval = FALSE}

<a href="https://ppbds.github.io/primer/index.html" id="abcd1234" class="hello hi">
link to primer
</a>

```

Remember:
To select `class`, you have to precede the `class` value with a `.`.

To select `id`, you have to precede the `id` value with a `#`.

```{r css-selectors-13, exercise = TRUE}

```

```{r css-selectors-13-hint-1, eval = FALSE}
minimal_html(anchor) |>
  html_element("a#abcd1234.hello")

# or 

minimal_html(anchor) |>
  html_element("a#abcd1234.hi")

```

### 

Now you can select using id, class, and tag type at the same time.

### Exercise 14

Copy/paste your pipe from above and extend it. Use the function `html_text2()` to see the inner text of the tag.

```{r css-selectors-14, exercise = TRUE}

```

```{r css-selectors-14-hint-1, eval = FALSE}
... |>
  html_text2()
```

### 

Well done.

### Exercise 15

Up until now, you have been extracting only the inner text of tags. However, there are times when the data you want is in the tag's attribute.

Consider again the following tag:
``` {r echo = TRUE, eval = FALSE}

<a href="https://ppbds.github.io/primer/index.html" id="abcd1234" class="hello hi">
link to primer
</a>

```

The `href` attribute can be helpful if you are collecting urls.

### 

Start a new pipe with `minimal_html()` and `anchor`.

```{r css-selectors-15, exercise = TRUE}

```

```{r css-selectors-15-hint-1, eval = FALSE}
anchor |>
  minimal_html()
```

### 

You now loaded in the html in `anchor`.

### Exercise 16

Copy/paste your pipe from above and extend it with `|>`. Use the function `html_element()` to select the `<a>` tag with both its tag type and `id` attribute, `"abcd1234"`.

```{r css-selectors-16, exercise = TRUE}

```

```{r css-selectors-16-hint-1, eval = FALSE}
... |>
  html_element("#...")
```

```{r css-selectors-16-hint-2, eval = FALSE}
... |>
  html_element("#abcd1234")
```

### 

You have now selected the `<a>` tag.

### Exercise 17

Copy/paste your pipe from above and extend it with `html_attr()`.
This function takes as an argument the name of the attribute that you want to select in quotes. Select the `href` attribute of the `<a>` tag.

```{r css-selectors-17, exercise = TRUE}

```

```{r css-selectors-17-hint-1, eval = FALSE}
... |> 
  html_attr("...")
```

```{r css-selectors-17-hint-2, eval = FALSE}
... |> 
  html_attr("href")
```

### 

Now you can also extract data from the attributes of tags. You can try `"class"` or `"id"` since both are also attributes.

Nice work!

<!-- AG: I want to move this to after Wikipedia just because it lines up a bit more with the primer. -->

## Open Secrets table

<a href="https://www.opensecrets.org">Open Secrets</a> contains <a href="https://www.opensecrets.org/political-action-committees-pacs/foreign-connected-pacs/2020">data</a> associated with the amount and sources of money donated to Democratic and Republican parties in 2020. <a href="https://www.opensecrets.org/political-action-committees-pacs/foreign-connected-pacs/2020">Check out the page</a>. You will be reading the table underneath the map.

### 

### Exercise 1

Assign the url provided below to a variable named `web_url`.

```{r open-secrets-table-1, exercise = TRUE}
"https://www.opensecrets.org/political-action-committees-pacs/foreign-connected-pacs/2020"

```

```{r open-secrets-table-1-hint-1, eval = FALSE}
web_url <- "..."
```

### Exercise 2

Start a pipe with `web_url` and the function `read_html()`.

`read_html()` takes a link and reads the html of that webpage.

If you tried the hint for this exercise and it is giving you an error like: `Error in open.connection(x, "rb") : HTTP error 403.`, use the variable `local_url` instead for all exercises in the Open Secrets section.

```{r open-secrets-table-2, exercise = TRUE}

```

```{r open-secrets-table-2-hint-1, eval = FALSE}
web_url |> read_html()
```

### 

You now loaded the html of the webpage into R, similar to how you did with `minimal_html()` in earlier sections.

You can begin to select elements in the page.

### Exercise 3

The following html shows the `table` tag's attributes:
```{r echo = TRUE, eval = FALSE}
<table class="DataTable-Partial dataTable no-footer" data-title="Foreign Connected PACs, 2019-2020" data-paging="true" data-page-length="25">...</table>
```

### 

Copy/paste your pipe from above and extend it with `|>`. Select the table using the function `html_element()` with both the tag type and its first `class` attribute.

```{r open-secrets-table-3, exercise = TRUE}

```

```{r open-secrets-table-3-hint-1, eval = FALSE}
... |>
  html_element("tag.class_name")
```

```{r open-secrets-table-3-hint-2, eval = FALSE}
... |>
  html_element("table.DataTable-Partial")
```

### 

You have selected the desired table.

You can see that it has a lot of different attributes and items inside the tag. It can be overwhelming at times but if you focus on the parts you know like `id` or `class`, you will be fine.


### Exercise 4

The **rvest** package provides a great table-parsing function, `html_table()`, that allows you to load in web-scraped tables easily.

### 

Copy/paste your pipe from above and extend it with the function `html_table()`. This allows you to load in data in the table in the form of a tibble.

```{r open-secrets-table-4, exercise = TRUE}

```

```{r open-secrets-table-4-hint-1, eval = FALSE}
... |>
  html_table()
```

Your resulting tibble should look something like this:

```{r open_example, echo = FALSE}
raw_data_opensecrets
```

### 

Nice work!

Now you have a tibble web-scraped directly from OpenSecrets.

## Wikipedia table

<!-- AG: I'm going to change this to the Canadian Wikipedia since we already have a SelectorGadget video -->

Now you will try your hand in scraping data from a Wikipedia page about <a href="https://en.wikipedia.org/w/index.php?title=%22,%22Gun_violence_in_the_United_States_by_state%22,%22&direction=prev&oldid=810166167">gun violence in the United States by state</a> and plotting it.

Let's create the following plot:
```{r make-wiki_graph, echo = FALSE}
clean_data <- raw_data_wiki |>
  rename(population = 'Population (total inhabitants) (2015) [1]',
  total_deaths = 'Murders and Nonnegligent\nManslaughter(total deaths) (2015) [2]',
  death_rate = 'Murder and Nonnegligent\nManslaughter Rate(per 100,000 inhabitants) (2015)') |>
  select(population, death_rate) |>
  mutate(population = parse_number(population)) |>
  mutate(pop_tile = ntile(population, 20)) |>
  group_by(pop_tile) |>
  summarize(sum_rate = sum(death_rate),
            num_states = n()) |>
  mutate(avg_rate = sum_rate/num_states) |>
  mutate(percent_rate = avg_rate/sum(avg_rate))

wiki_graph <- ggplot(clean_data, aes(x=pop_tile, y=percent_rate)) + 
geom_col() +
geom_smooth(method="loess", formula=y~x, se=FALSE) +
scale_x_continuous(breaks=1:20) +
scale_y_continuous(labels=scales::percent_format(accuracy=1)) +
theme_classic() +
labs(title="Distribution of US Death Rate by Guns in 2015",
subtitle="Death rate of less populated states fluctuate\nmore than states with a denser population",
x = "Quantile by Population (least to most)",
y="Percent of Average Death Rate",
caption = "Wikipedia: Gun Violence in the United States (2017)")


wiki_graph
```

### Exercise 1

Assign the url provided below to a variable called `wiki_url`.

```{r wikipedia-table-1, exercise = TRUE}
"https://en.wikipedia.org/w/index.php?title=%22,%22Gun_violence_in_the_United_States_by_state%22,%22&direction=prev&oldid=810166167"

```

```{r wikipedia-table-1-hint-1, eval = FALSE}
wiki_url <- "..."
```

### 

To make sure that updates to the Wikipedia page does not affect this tutorial, the url used is specifically for how the page looked in 2017.

### Exercise 2

Use the function `read_html()` and `wiki_url` as argument to read the html of the Wikipedia webpage.


```{r wikipedia-table-2, exercise = TRUE}

```

```{r wikipedia-table-2-hint-1, eval = FALSE}
wiki_url |> 
  read_html()
```

### 

You now loaded the html of the webpage into R, similar to how you did with `minimal_html()` in earlier sections.

You can begin to select elements in the page.

### Exercise 3

The following html is the section that contains the main content of the page:

```{r echo = TRUE, eval = FALSE}
<div id="mw-content-text" class="mw-body-content mw-content-ltr" lang="en" dir="ltr">...</div>
```

### 

Copy/paste your pipe from above and extend it with `|>`. Select the main content element with the function `html_element()` using its `id` attribute.

Remember to put `#` before `mw-content-text` to specify that you are searching by `id`.

```{r wikipedia-table-3, exercise = TRUE}

```

```{r wikipedia-table-3-hint-1, eval = FALSE}
... |> 
  html_element("#...")
```

```{r wikipedia-table-3-hint-2, eval = FALSE}
... |> 
  html_element("#mw-content-text")
```

<!-- DK: Explain why we need the "#" here, when, before we just did "p" and the like. Is that because mw-content-text is an id and not a tag? -->

### 

You have selected the main content section of the page.

### Exercise 4

The following is the tag for the table inside the section selected in the previous exercise:
```{r echo = TRUE, eval = FALSE}
<table class="wikitable sortable jquery-tablesorter">...</table>
```

### 

Copy/paste your pipe from above and extend it with the function `html_element()`. Select the table with both its tag type AND one of its `class`.

```{r wikipedia-table-4, exercise = TRUE}

```

```{r wikipedia-table-4-hint-1, eval = FALSE}
... |>
  html_element("tag_type.class_name")
```

```{r wikipedia-table-4-hint-2, eval = FALSE}
... |>
  html_element("table.wikitable")
```

### 

Now you have access to the data table. You can see some of the elements inside listed.

### Exercise 5

The **rvest** package provides a great table-parsing function, `html_table()`, that allows you to load in web-scraped tables easily.

### 

Copy/paste your pipe from above and extend it with the function `html_table()`. This allows you to load in data in the table in the form of a tibble

```{r wikipedia-table-5, exercise = TRUE}

```

```{r wikipedia-table-5-hint-1, eval = FALSE}
... |>
  html_table()
```

### 

Nice work!

Now you have a tibble webscraped directly from Wikipedia.

### Exercise 6

Copy/paste your pipe from above and assign it to a variable named `raw_data_wiki`.

```{r wikipedia-table-6, exercise = TRUE}

```

### 

Now you saved the Wikipedia data into a variable, so you do not have to repeatedly web-scrape the page every time you re-run the pipe.

### Exercise 7

Run `glimpse()` on `raw_data_wiki` to understand the structure of the data.

```{r wikipedia-table-7, exercise = TRUE}

```

```{r wikipedia-table-7-hint-1, eval = FALSE}
glimpse(...)
```

### 

`Population (total inhabitants) (2015) [1]` and `Murders and Nonnegligent\nManslaughter(total deaths) (2015) [2]` should be integer columns, obviously. However, they were read in as character columns because of the commas in the numbers on the Wikipedia page.

More often than not, data that's web-scraped are unclean. Therefore, before doing any plotting, you must clean the data.

### Exercise 8

Start a pipe with `raw_data_wiki`.

First, let's rename the columns to make referencing them easier.

Using the `rename()` function, rename all the columns as such: <br><br>
`Population (total inhabitants) (2015) [1]` to `population`<br><br>
`Murders and Nonnegligent\nManslaughter(total deaths) (2015) [2]` to `total_deaths`<br><br>
`Murder and Nonnegligent\nManslaughter Rate(per 100,000 inhabitants) (2015)` to `death_rate`<br><br>

Remember that `rename()` is similar to `mutate()` in the way it is used:<br>
`rename(new_column_name = old_column_name, 
        new_col2 = old_col2, ...)`

```{r wikipedia-table-8, exercise = TRUE}

```

<!-- DK: Should we be using backticks rather than quotes for idenifying weird column names? -->

```{r wikipedia-table-8-hint-1, eval = FALSE}
raw_data_wiki |> 
  rename(... = `...`,
         ... = `...`)
```

```{r wikipedia-table-8-hint-2, eval = FALSE}
raw_data_wiki |> 
  rename(population = `Population (total inhabitants) (2015) [1]`,
         total_deaths = `...`)
```

### 

In order to select the weird column names, you have to put either \` or " around it.

\` is preferred.

Now it is easier for you to refer to the columns.

### Exercise 9

Copy/paste your pipe and extend it. Use the function `select()` to select only the columns you need, which are `population` and `death_rate`.

```{r wikipedia-table-9, exercise = TRUE}

```

```{r wikipedia-table-9-hint-1, eval = FALSE}
... |>
  select(...)
```

### 

The only column that is unclean now is `population`.

### Exercise 10

Copy/paste your pipe and extend it.

Remove all the commas in `population` and change the column into a numeric column using `parse_number()`.

Then use `mutate()` to store the results in the same column.

```{r wikipedia-table-10, exercise = TRUE}

```

```{r wikipedia-table-10-hint-1, eval = FALSE}
... |>
  mutate(population = parse_number(population))
```

### 

As you can see from the output, right below the column name, `population` is now a `<dbl>`, or double column, which is numeric.

### Exercise 11

Now you can begin dividing the data into quantiles.

### 

Copy/paste your pipe from above and extend it. Use the `ntile()` function to split `population` into 20 quantiles.

Then use `mutate()` to store the results in a new column called `pop_tile`.

Remember that `ntile()` takes as arguments the data that it is splitting and the number of groups to split into.

```{r wikipedia-table-11, exercise = TRUE}

```

```{r wikipedia-table-11-hint-1, eval = FALSE}
... |>
  mutate(pop_tile = ntile(...))
```

```{r wikipedia-table-11-hint-2, eval = FALSE}
... |>
  mutate(pop_tile = ntile(population, 20))
```

### 

You can see the new column in the output, `pop_tile`, which has values between 1 and 20. This refers to which quantile that row is in.

### Exercise 12

Copy/paste your pipe from above and extend it. Group your data by `pop_tile`.

```{r wikipedia-table-12, exercise = TRUE}

```

```{r wikipedia-table-12-hint-1, eval = FALSE}
.. |>
  group_by(...)
```

### 

The output may look like a normal table, but the grouping has already been done.

### Exercise 13

Copy/paste your pipe from above and extend it. Summarize your data with 2 columns:
Number of States in quantile as `num_states`
Sum of death rate in quantile as `sum_rate`

```{r wikipedia-table-13, exercise = TRUE}

```

```{r wikipedia-table-13-hint-1, eval = FALSE}
... |>
  summarize(...)
```

```{r wikipedia-table-13-hint-2, eval = FALSE}
... |>
  summarize(num_states = n(),
            sum_rate = sum(...))
```

### 

The resulting tibble is your summarized data. You can see from the `num_states` column that some of the quantiles have more states than others.

### Exercise 14

In order to counter the imbalanced number of states, you will average the `sum_rate` of each quantile.

### 

Copy/paste your pipe from above and extend it. Create a new column using `mutate()` called `avg_rate`, which is the average rate in quantile per state.

```{r wikipedia-table-14, exercise = TRUE}

```

```{r wikipedia-table-14-hint-1, eval = FALSE}
... |>
  mutate(avg_rate=...)
```

```{r wikipedia-table-14-hint-2, eval = FALSE}
... |>
  mutate(avg_rate = sum_rate/num_states)
```

### 

Now you can see the new column called `avg_rate`.

### Exercise 15

Now let's calculate the percentage of each quantile's average rate over all quantiles' average rates. 

### 

Copy/paste your pipe from above and extend it. Create a new column using `mutate()` called `percent_rate`, which is the percentage of each quantile's average rate over the sum of all quantiles'.

```{r wikipedia-table-15, exercise = TRUE}

```

```{r wikipedia-table-15-hint-1, eval = FALSE}
... |>
  mutate(percent_rate = .../sum(...))
```

```{r wikipedia-table-15-hint-2, eval = FALSE}
... |>
  mutate(percent_rate = avg_rate/sum(avg_rate))
```

### 

Now you finally have everything you need for plotting!

### Exercise 16

Copy/paste your pipe from above and assign it to a variable named `clean_data`.

```{r wikipedia-table-16, echo = FALSE}
clean_data <- raw_data_wiki |>
  rename(population = 'Population (total inhabitants) (2015) [1]',
  total_deaths = 'Murders and Nonnegligent\nManslaughter(total deaths) (2015) [2]',
  death_rate = 'Murder and Nonnegligent\nManslaughter Rate(per 100,000 inhabitants) (2015)') |>
  select(population, death_rate) |>
  mutate(population = parse_number(population)) |>
  mutate(pop_tile = ntile(population, 20)) |>
  group_by(pop_tile) |>
  summarize(sum_rate = sum(death_rate),
            num_states = n()) |>
  mutate(avg_rate = sum_rate/num_states) |>
  mutate(percent_rate = avg_rate/sum(avg_rate))
```

```{r wiki16, exercise = TRUE}

```

### 

Now you have access to `clean_data`.

### Exercise 17

Start a plot with `ggplot()`. Use `clean_data` as data and map `pop_tile` to the x-axis and `percent_rate` to the y-axis. Add a `geom_col()` to create a bar graph.

```{r wikipedia-table-17, exercise = TRUE}

```

```{r wikipedia-table-17-hint-1, eval = FALSE}
ggplot(data = ...,
       mapping = aes(...)) +
  geom_col()
```

### 

You can now see the distribution.

### Exercise 18

Copy/paste you code from above and add another layer with `geom_smooth()`. Use `"loess"` as its method, `y~x` as its formula, and set `se` to `FALSE`. This adds a kind of trend line that visualizes the data better.

```{r wikipedia-table-18, exercise = TRUE}

```

```{r wikipedia-table-18-hint-1, eval = FALSE}
... +
  geom_smooth(method = ...,
              formula = ...,
              se = ...)
```

### 

The data visualization itself is basically done, but you'll notice that the x-axis isn't chunked properly and the y-axis should be in percentage.

### Exercise 19

Copy/paste your code from above and add another layer using the function `scale_x_continuous()`. Assign to its `breaks` argument `1:20` to break the x-axis into 1 through 20.

```{r wikipedia-table-19, exercise = TRUE}

```

```{r wikipedia-table-19-hint-1, eval = FALSE}
... +
  scale_x_continuous(breaks = ...)
```

```{r wikipedia-table-19-hint-2, eval = FALSE}
... +
  scale_x_continuous(breaks = 1:20)
```

### 

Now the x-axis makes sense.

### Exercise 20

Next let's format the y-axis into percentages.

### 

Copy/paste your code from above and add another layer with `scale_y_continuous()`. Assign to its `labels` argument `scales::percent_format(accuracy=1)`, which changes the y-axis into percentage format.

```{r wikipedia-table-20, exercise = TRUE}

```

```{r wikipedia-table-20-hint-1, eval = FALSE}
... +
  scale_y_continuous(labels = ...)
```

```{r wikipedia-table-20-hint-2, eval = FALSE}
... +
  scale_y_continuous(labels = scales::percent_format(accuracy=1))
```

### 

Now both the axes are labeled accordingly.

### Exercise 21

Copy/paste your code from above and add a `theme_classic()` layer to the plot. Then add the appropriate labels with `labs()`.

```{r wikipedia-table-21, exercise = TRUE}

```

```{r wikipedia-table-21-hint-1, eval = FALSE}
... +
  labs(title = "...",
       subtitle = "...",
       x = "...",
       y = "...",
       caption = "...")
```

Reminder: this is how your plot should look like.
```{r show-wiki_graph, echo = FALSE}
wiki_graph
```

### 

Well done!

### (Optional) Exercise 22

A part of web-scraping that is not covered here is actually finding the element you want to scrape from the website's html.

Below is a page explaining how to install and use SelectorGadget, which helps you find the html tag that you want:

<a href="https://rvest.tidyverse.org/articles/selectorgadget.html">Link to SelectorGadget page</a>



<!-- This should be a 60 to 90 minutes worth of exercises devoted to webscraping, with a big focus on the rvest package. Also need to teach some CSS and Selector Gadget. Resources: -->

<!-- If students would just read this paper closely and try out the associated commands -->

<!-- https://github.com/mdogucu/web-scrape -->

<!-- then a tutorial would not be necessary. But they won't! And, even if most of them would, teachers still need a way of knowing that they did. Hence, tutorials. Can we write something that, in essence, forces students to type out everything in that article and understand what it is doing? That is the goal! -->

<!-- https://github.com/rstudio-education/dsbox/tree/master/inst/tutorials/05-moneyinpolitics -->
<!-- https://minecr.shinyapps.io/dsbox-05-moneyinpolitics/ -->


<!-- There is no need to reinvest the wheel. All this code is CC0, so we can just copy/paste it, including the links to images which appear on the web. But how do we ensure that students do things? What questions do we ask? -->

<!-- https://rvest.tidyverse.org/ -->
<!-- https://github.com/tidyverse/rvest/blob/master/vignettes/rvest.Rmd -->
<!-- https://github.com/tidyverse/rvest/blob/master/vignettes/articles/selectorgadget.Rmd -->

<!-- Maybe the tutorial starts the same way as this chapter, with a concrete example of what we are trying to do. Note that we just used this prose in our chapter. Again, it is all CC0. -->

<!-- https://rafalab.github.io/dsbook/web-scraping.html -->

<!-- Reminders: -->

<!-- 1) This tutorial is just about raw webscraping, not about working with APIs or JSON. Other tutorials handle that. -->

<!-- 2) Key structure is to, first, begin with sections which highlights the key issues and key commands. Then, have 5 or so concrete examples, each of which works with an html file which we have a permanent copy of. (But how do we show how that file looks in a browser? I am confused about that.) Nothing wrong with all the examples being from Wikipedia and referencing specific versions, even if they are old. -->

<!-- 3) Plan is to write this tutorial and then, afterwards, update the Chapter to be consistent with it. -->

<!-- 4) Use html_text2(). Background: https://www.tidyverse.org/blog/2021/03/rvest-1-0-0/ -->

 

```{r download-answers, child = "../../child_documents/download_answers.Rmd"}
```

---
title: "Test Document"
author: "Test Author"
date: "2024-01-01"
---


# Introduction

This is a test R Markdown document.

```{r setup}
library(tidyverse)
```

Some text here.

```{r analysis}
x <- 1:10
mean(x)
```

---
title: "Test Document"
---

```{r}
library(tidyverse)
```

```{r}
penguins
<div data-pagedtable="false">
  <script data-pagedtable-source type="application/json">
{"columns":[{"label":["species"],"name":[1],"type":["fct"]}],"data":[["Adelie"]]}
  </script>
</div>
```

---
title: A legal tutorial
author: David Kane
tutorial:
  id: a-legal-tutorial
output:
  learnr::tutorial:
    progressive: yes
    allow_skip: yes
runtime: shiny_prerendered
description: 'Tutorial for Chapter 4: Data transformation'
---

```{r setup, include = FALSE}
library(learnr)
library(tutorial.helpers)
knitr::opts_chunk$set(echo = FALSE)
options(tutorial.exercise.timelimit = 60, 
        tutorial.storage = "local") 
```

```{r copy-code-chunk, child = system.file("child_documents/copy_button.Rmd", package = "tutorial.helpers")}
```

```{r info-section, child = system.file("child_documents/info_section.Rmd", package = "tutorial.helpers")}
```

## Reading/Writing Excel
### 

[**readxl**](https://readxl.tidyverse.org/) and [**wrirexl**](https://docs.ropensci.org/writexl/) are the two key packages for working with Excel spreadsheets.

### Exercise 1

Use the `library()` command to load the **readxl** package.


```{r readingwriting-excel-1, exercise = TRUE}

```

```{r readingwriting-excel-1-hint-1, eval = FALSE}
library(readxl)
```

### 

“[Data Organization in Spreadsheets](https://doi.org/10.1080/00031305.2017.1375989)” by Karl Broman and Kara Woo provides lots of great advice for organizing your data using spreadsheets.

### Exercise 2

Look up the help page for the package by issuing this command: "?`readxl-package`". Copy/paste the Description field from the help page.  

```{r readingwriting-excel-2}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 3)
```

### 

Almost all packages have a help page which can be accessed via `packagename-package`. Note that you need to use backticks (\` \`) around this phrase after the `?` because otherwise the Console will try to interpret the dash as a minus sign, thereby sending you to the wrong page.

### Exercise 3

Run `read_excel()` with `"data/students.xlsx"` as the argument to `path`.


```{r readingwriting-excel-3, exercise = TRUE}

```

```{r readingwriting-excel-3-hint-1, eval = FALSE}
read_excel(path = "...")
```

### 

`read_excel()` functions similarly to `read_csv()` and the other data import functions you have seen before. The **readxl** package also includes `read_xls()` (for files in xls format) and `read_xlsx()` (for files in xlsx format). `read_excel()` can read files with both xls and xlsx format. It guesses the file type based on the input.


```{r download-answers, child = system.file("child_documents/download_answers.Rmd", package = "tutorial.helpers")}
```

---
title: "format_test_1"
author: "Nuo Wen Lei"
date: "7/27/2021"
output: html_document
---

This is test 1 for the addin that formats code chunk labels. 

## Section 1
###

### Exercise 1


Purpose of this initial code chunk is to ensure format_tutorial() does not act on it. A code chunk without either "exercise," "eval," or "include" should not be changed.

```{r}

```



```{r section-1-1, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r section-1-1-hint-1, eval = FALSE}

```

```{r section-1-1-test, include=FALSE}

```


###

### Exercise 2


```{r section-1-2, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r section-1-2-hint-1, eval = FALSE}

```

Testing to see if an unlabeled code chunk at the end also is unchanged, as it should be.

```{r}

```


###

### Exercise 3


```{r section-1-3}
question_text(NULL,
	message = "answer here",
	answer(NULL,
	correct = TRUE),
	allow_retry = FALSE,
	incorrect = NULL,
	rows = 6)
```

###


### Exercise 4


```{r section-1-4, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r section-1-4-hint-1, eval = FALSE}

```

###

A code chunk with the label setup should not be changed:

```{r setup, include = FALSE}
library(learnr)
``

None of the default tutorial helpers should be changed:

```{r copy-code-chunk, child = system.file("child_documents/copy_button.Rmd", package = "tutorial.helpers")}
```

```{r info-section, child = system.file("child_documents/info_section.Rmd", package = "tutorial.helpers")}
```

```{r download-answers, child = system.file("child_documents/download_answers.Rmd", package = "tutorial.helpers")}

## Section 2
###

### Exercise 1


```{r section-2-1, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r section-2-1-hint-1, eval = FALSE}

```

###

### Exercise 2


```{r section-2-2}
question_text(
	"prompt here",
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Text for button",
	incorrect = NULL)
```

###

### Exercise 3

Ignore anything within a four backtick environment, including headers and code chunks.

````
### Exercise 6

```{r exercises-1, exercise = TRUE}

```
````

Also ignore data printouts:

````
> head(diamonds)
# A tibble: 6 × 10
  carat cut       color clarity depth table price     x     y     z
  <dbl> <ord>     <ord> <ord>   <dbl> <dbl> <int> <dbl> <dbl> <dbl>
1  0.23 Ideal     E     SI2      61.5    55   326  3.95  3.98  2.43
2  0.21 Premium   E     SI1      59.8    61   326  3.89  3.84  2.31
3  0.23 Good      E     VS1      56.9    65   327  4.05  4.07  2.31
4  0.29 Premium   I     VS2      62.4    58   334  4.2   4.23  2.63
5  0.31 Good      J     SI2      63.3    58   335  4.34  4.35  2.75
6  0.24 Very Good J     VVS2     62.8    57   336  3.94  3.96  2.48
>
````



```{r section-2-3, exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r section-2-3-hint-1, eval = FALSE}

```

###

### Exercise 4

Don't change a chunk with the "file" option.

```{r file = "images/chunk-example.txt", echo = TRUE, eval = FALSE}
```

Should work even if there are no labels, even if the comma is not included. The option tells us what kind of chunk it is, and therefor, how to label it.

```{r section-2-4, , exercise = TRUE}

```

<button onclick = "transfer_code(this)">Copy previous code</button>

```{r section-2-4-test, include = FALSE}

```

```{r section-2-4-hint-1, eval = FALSE}

```

###

## Exercises 
### 

Weird case which used to produce exercises--1 instead of exercise-1.

### Exercise 1

```{r exercises-1, exercise = TRUE}

```

## Hask marks 
### 

An example to conform that hashmarks within four backtick environments do not mess up the reformatting. This test confirms that the exercise section number and the code chunk label get changed.

````
penguins

# A tibble: 344 × 8
   species island    bill_length_mm bill_depth_mm flipper_length_mm body_mass_g
   <fct>   <fct>              <dbl>         <dbl>             <int>       <int>
 1 Adelie  Torgersen           39.1          18.7               181        3750
 2 Adelie  Torgersen           39.5          17.4               186        3800
 3 Adelie  Torgersen           40.3          18                 195        3250
 4 Adelie  Torgersen           NA            NA                  NA          NA
 5 Adelie  Torgersen           36.7          19.3               193        3450
 6 Adelie  Torgersen           39.3          20.6               190        3650
 7 Adelie  Torgersen           38.9          17.8               181        3625
 8 Adelie  Torgersen           39.2          19.6               195        4675
 9 Adelie  Torgersen           34.1          18.1               193        3475
10 Adelie  Torgersen           42            20.2               190        4250
# ℹ 334 more rows
# ℹ 2 more variables: sex <fct>, year <int>
````

### Exercise 4

```{r zz, exercise = TRUE}

```

## Four ticks problem
### 

### Exercise 2

Test to make sure that this four backtick environment does not cause trouble. Not sure what made this trouble in previous versions!

````

````  

```{r zz, exercise = TRUE}

```

## GitHub Pages Output Problem
### 

Problem is that the Terminal output from GitHub Pages publishing command causes problems.

````
dkane@macbook project-1 % quarto publish gh-pages quarto-1.qmd
? Publish site to https://davidkane9.github.io/project-1/ using gh-pages? (Y/n) › Yes
````

The first time you publish, Quarto will confirm the location, which is just the standard GitHub Pages address, using your user name and the repo name.

````
Saved working directory and index state WIP on main: 71e6d06 initial version
Switched to a new branch 'gh-pages'
[gh-pages (root-commit) b8263f3] Initializing gh-pages branch
remote: 
remote: Create a pull request for 'gh-pages' on GitHub by visiting:        
remote:      https://github.com/davidkane9/project-1/pull/new/gh-pages        
remote: 
To https://github.com/davidkane9/project-1.git
 * [new branch]      HEAD -> gh-pages
 ````

 Quarto interacts with your GitHub repo directly, creating a new branch, `gh-pages`, which will be used to construct the website. 

````
[✓] Deploying gh-pages branch to website (this may take a few minutes)
[✓] Published to https://davidkane9.github.io/project-1/

dkane@macbook project-1 % 
```` 


### Exercise 2

 

```{r zz, exercise = TRUE}

```

## Space   Test Case 1
###

This test ensures that (mistaken) extra spaces in a section name are replaced by a single dash in code chunk names rather than multiple dashes.

### Exercise 2

```{r zz, exercise = TRUE}

```

### Exercise 4

Had some problems with the new version of "no-answer" type questions. In particular, format_tutorial() had trouble with the test code chunk, perhaps because of the combination of -test in the code chunk name and `echo = TRUE`. The current problem is that the code chunk name in the test code block is not updated by check_format_tutorial().

```{r space---test-case-1-5}
question_text(NULL,
	answer(NULL, correct = TRUE),
	allow_retry = TRUE,
	try_again_button = "Edit Answer",
	incorrect = NULL,
	rows = 5)
```

###

```{r space---test-case-1-5-test, echo = TRUE}

```

###

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gather_submissions.R
\name{gather_submissions}
\alias{gather_submissions}
\title{Gather Submissions}
\usage{
gather_submissions(path, title, keep_loc = NULL, verbose = FALSE)
}
\arguments{
\item{path}{The path to the local directory containing the HTML/XML files, or a Google Drive folder URL.
If it's a Google Drive URL, the function will download the entire folder to a temporary directory.}

\item{title}{A character vector of patterns to match against the file names.
Each pattern is processed separately and results are combined.}

\item{keep_loc}{A character string specifying where to save downloaded files (only for Google Drive URLs).
If NULL (default), files are downloaded to a temporary directory and deleted after processing.
If specified, files are downloaded to this location and kept.}

\item{verbose}{A logical value (TRUE or FALSE) specifying verbosity level.
If TRUE, reports files that are removed during processing.}
}
\value{
A named list of tibbles, where each tibble contains the data from one HTML/XML file
that matches any of the specified patterns and has valid table structure.
}
\description{
This function finds and reads HTML/XML files from a local directory
or Google Drive folder that match specified patterns. It extracts tables from the
files and returns a list of tibbles containing the submission data.
}
\details{
Google Drive allows for more than one file with the exact same name.
If you download files manually ("by hand"), you will get both files but with
one of them automatically renamed by your browser. However, if you use the
Google Drive functionality in this function, the second file will overwrite
the first, potentially resulting in data loss.
}
\examples{
# Find submissions from local directory
path <- system.file("extdata", "answers_html", package = "tutorial.helpers")

tibble_list <- gather_submissions(path = path, title = "stop", verbose = TRUE)

\dontrun{
# Find submissions from Google Drive folder (temporary download)
drive_url <- "https://drive.google.com/drive/folders/10do12t0fZsfrIrKePxwjpH8IqBNVO86N"
tibble_list <- gather_submissions(
  path = drive_url, 
  title = c("positron")
)

# Find submissions from Google Drive folder (keep files)
tibble_list <- gather_submissions(
  path = drive_url, 
  title = c("introduction"),
  keep_loc = "temp_file"
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format_tutorial.R
\name{format_tutorial}
\alias{format_tutorial}
\title{Format RMarkdown tutorial code chunks}
\usage{
format_tutorial(file_path)
}
\arguments{
\item{file_path}{Character string. Path to the R Markdown file to process.}
}
\value{
Character string containing the formatted R Markdown content.
}
\description{
This function processes an R Markdown tutorial file to standardize code chunk labels
based on section names and exercise numbers. It also renumbers exercises sequentially
within each section and fixes spacing in topic headers.
}
\details{
The function applies the following formatting rules:
\itemize{
\item Topic headers (# headers) have their spacing standardized
\item Exercises are renumbered sequentially within each section
\item Code chunks are relabeled according to the pattern: section-name-exercise-number
\item Chunks with \code{eval = FALSE} receive a \code{-hint-N} suffix
\item Chunks with \code{include = FALSE} receive a \code{-test} suffix
\item Chunks with label "setup" are not modified
\item Chunks with the "file" option are not modified
\item Unlabeled chunks without key options are not modified
\item All formatted chunks preserve their original options
\item Content between quadruple backticks (\verb{ }) is preserved untouched
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/show_file.R
\name{strip_pagedtable_html}
\alias{strip_pagedtable_html}
\title{Remove embedded pagedtable HTML artifacts from a vector of lines}
\usage{
strip_pagedtable_html(lines)
}
\arguments{
\item{lines}{A character vector of lines (as returned by readLines()).}
}
\value{
A character vector with any pagedtable HTML blocks removed.
}
\description{
VS Code's interactive chunk execution can cache a chunk's rendered output
directly back into the .qmd/.Rmd source file. When the chunk's output is a
tibble, that cached output is a pagedtable HTML widget (a \verb{<div data-pagedtable="false">} block wrapping a \verb{<script data-pagedtable-source>} JSON blob). Because this cached output can end up
inside the same fenced region as the code itself, show_file()'s chunk
extraction can otherwise return this HTML noise instead of clean source
code. This helper strips any such blocks out of a character vector of
lines before they are printed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_membership.R
\name{check_membership}
\alias{check_membership}
\title{Check Membership in List of Tibbles}
\usage{
check_membership(tibble_list, key_var, membership, verbose = FALSE)
}
\arguments{
\item{tibble_list}{A named list of tibbles, each containing an "id" column and an "answer" column}

\item{key_var}{A character string specifying the key variable to check}

\item{membership}{A character vector of allowed values for the key variable}

\item{verbose}{Logical indicating whether to report removed items (default: FALSE)}
}
\value{
A list of tibbles where the key variable exists and its value is in the membership list
}
\description{
This function filters a list of tibbles based on whether a key variable's value
is among the membership values. It first uses check_key_vars() to ensure the
key variable exists, then checks membership. Useful for keeping only specific
students or participants.
}
\examples{
# Create sample data with student emails
path <- system.file("extdata", "answers_html", package = "tutorial.helpers")

tibble_list <- gather_submissions(path, "stop")

result <- check_membership(tibble_list,
                           key_var = "email",
                           membership = c("bluebird.jack.xu@gmail.com",
                                          "hassan.alisoni007@gmail.com"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/determine_code_chunk_name.R
\name{determine_code_chunk_name}
\alias{determine_code_chunk_name}
\title{Determine the code chunk name of a new exercise in a tutorial.}
\usage{
determine_code_chunk_name(file_path = NULL)
}
\arguments{
\item{file_path}{Character string of the file path to the tutorial}
}
\value{
The section id of the exercise based on its section
}
\description{
Determine the code chunk name of a new exercise in a tutorial.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/determine_exercise_number.R
\name{determine_exercise_number}
\alias{determine_exercise_number}
\title{Finds the number of the next exercise in a tutorial}
\usage{
determine_exercise_number(file_path = NULL)
}
\arguments{
\item{file_path}{Character string of the file path to the tutorial}
}
\value{
The next exercise number based on the file argument or the active document.
}
\description{
Finds the number of the next exercise in a tutorial
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submission_functions.R
\docType{data}
\name{submission_server}
\alias{submission_server}
\alias{submission_ui}
\title{Tutorial submission functions}
\format{
An object of class \code{shiny.tag} of length 3.
}
\usage{
submission_server()

submission_ui
}
\value{
No return value; called for side effects in a Shiny/learnr session.

An object of class shiny.tag
}
\description{
Provides the core Shiny server and UI hooks for collecting and downloading student answers
from a learnr tutorial. \code{submission_server()} should be called in an Rmd code chunk with
\code{context="server"}.

This function was modified from Colin Rundel's learnrhash package
(https://github.com/rundel/learnrhash).

UI block to include a download button and simple instructions for students.
}
\details{
The server function uses a Shiny downloadHandler to let students download their answers.
All main logic must be wrapped in \code{local()} with \code{parent.frame()} to ensure access to the live
learnr session and objects created in the parent environment.

The \code{session} object (created by Shiny) is only available inside the \code{downloadHandler$content}
function, so any test-case extraction or answer writing must happen there.

For reference: the \code{file} argument in \code{content} is a temporary file path created by Shiny,
and your handler's job is to write the downloadable file there.

If you want to generate test fixtures, insert \code{browser()} inside the \code{content} function,
then use functions like \code{get_submissions_from_learnr_session(session)} at the prompt.

See also: https://mastering-shiny.org/action-transfer.html#downloading-reports
}
\examples{
if(interactive()){
  submission_server()
}

if(interactive()){
  submission_ui
}

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submissions_answers.R
\name{submissions_answers}
\alias{submissions_answers}
\title{Extract Answers from Submissions with Filtering}
\usage{
submissions_answers(
  path,
  title,
  key_var = NULL,
  membership = NULL,
  vars,
  keep_file_name = NULL,
  verbose = FALSE
)
}
\arguments{
\item{path}{The path to the local directory or Google Drive folder URL containing submissions}

\item{title}{A character vector of patterns to match against file names (passed to gather_submissions)}

\item{key_var}{A character string specifying the key variable to check for membership (e.g., "email").
If NULL (default), no membership filtering is applied.}

\item{membership}{A character vector of allowed values for the key variable, or "*" to include all submissions.
If NULL (default), no membership filtering is applied. Ignored if key_var is NULL.}

\item{vars}{A character vector of variables/questions to extract, or "*" to extract all available variables}

\item{keep_file_name}{How to handle file names: NULL (don't include), "All" (full name),
"Space" (up to first space), "Underscore" (up to first underscore)}

\item{verbose}{A logical value (TRUE or FALSE) specifying verbosity level.
If TRUE, reports files that are removed during processing.}
}
\value{
A tibble with one row per valid submission, columns for each variable,
and optionally a "source" column
}
\description{
This function gathers submissions matching a title pattern, filters them by membership,
and extracts specified variables, returning a tibble with one row per valid submission
and one column per variable.
}
\examples{
# Extract specific variables from submissions matching title pattern
path <- system.file("extdata", "answers_html", package = "tutorial.helpers")

result <- submissions_answers(
  path = path,
  title = c("stop"), 
  key_var = "email",
  membership = c("bluebird.jack.xu@gmail.com", "abdul.hannan20008@gmail.com"),
  vars = c("name", "email", "introduction-1"),
  verbose = TRUE
)

# Extract all variables from submissions
result_all <- submissions_answers(
  path = path,
  title = c("stop"), 
  key_var = "email",
  membership = c("bluebird.jack.xu@gmail.com", "abdul.hannan20008@gmail.com"),
  vars = "*",
  verbose = TRUE
)

\dontrun{
drive_url <- "https://drive.google.com/drive/folders/10do12t0fZsfrIrKePxwjpH8IqBNVO86N"
x <- submissions_answers(
  path = drive_url, 
  title = c("introduction"),
  key_var = "email",
  membership = c("fmehmud325@gmail.com"),
  vars = c("email", "name", "what-you-will-learn-15")
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.check_quarto}
\alias{.check_quarto}
\title{Package build dependencies}
\usage{
.check_quarto()
}
\description{
This function ensures renv detects quarto as a dependency
while keeping it in Suggests rather than Imports. It is deliberately
never called: its mere presence (referencing the quarto namespace) is
what makes renv record quarto. Do not delete it.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_current_tutorial.R
\name{check_current_tutorial}
\alias{check_current_tutorial}
\title{Check current tutorial}
\usage{
check_current_tutorial()
}
\value{
No return value, called for the side effect of rewriting the
tutorial file in the active editor window.
}
\description{
A function which formats the exercise numbers and code chunk
labels correctly. \code{check_current_tutorial()} reads in the tutorial
which is in the active editor window in Positron. It determines what
the number of each exercise should be and fixes any mistakes. It
ensures that all code chunk labels are the correct function of the
section title and exercise number.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knit_tutorials.R
\name{knit_tutorials}
\alias{knit_tutorials}
\title{Knit a set of tutorials}
\usage{
knit_tutorials(tutorial_paths)
}
\arguments{
\item{tutorial_paths}{Character vector of the paths to the tutorials to be
knitted.}
}
\value{
No return value, called for side effects.
}
\description{
We define "testing" a tutorial as (successfully) running
\code{render()} on it. This function renders all the tutorials provided in
\code{tutorial_paths}. There is no check to see if the rendered file looks OK.
If a tutorial fails to render, then an error will be generated which will
propagate to the caller.
}
\examples{
\dontrun{
  knit_tutorials(tutorial_paths = return_tutorial_paths("tutorial.helpers"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/match_questions.R
\name{match_questions}
\alias{match_questions}
\title{Match Questions by Pattern}
\usage{
match_questions(x, pattern, ignore.case = TRUE)
}
\arguments{
\item{x}{Either a file path to an HTML file or a tibble with 'id' and 'answer'/'data' columns}

\item{pattern}{A character string to search for in the answers}

\item{ignore.case}{Logical; should the search be case-insensitive? (default: TRUE)}
}
\value{
A character vector of question IDs where the answer contains the pattern
}
\description{
This function takes a single HTML file or tibble and finds all questions/answers
that contain a specified pattern. It returns the question IDs (from the 'id' column)
for rows where the answer contains the pattern.
}
\examples{
# Search in a tibble
path <- system.file("extdata", "answers_html", package = "tutorial.helpers")

tibble <- gather_submissions(path, title = "stop")[[1]]

result <- match_questions(tibble, "http")

\dontrun{
# Search in an HTML file
question_ids <- match_questions("path/to/submission.html", "temperance")
# Returns: c("temperance-16", "temperance-19")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/return_tutorial_paths.R
\name{return_tutorial_paths}
\alias{return_tutorial_paths}
\title{Return all the paths to the tutorials in a package}
\usage{
return_tutorial_paths(package)
}
\arguments{
\item{package}{Character string of the package name to be tested.}
}
\value{
Character vector of the full paths to all installed tutorials in
\code{package}. Returns character(0) if no tutorials are found or if the
package doesn't exist.
}
\description{
Takes a package name and returns a character vector of all the
paths to tutorials in the installed package. This function looks for all
R Markdown files (.Rmd) in the inst/tutorials/ subdirectories of the
specified package. It uses learnr::available_tutorials() to identify
tutorial directories, with a fallback to directory scanning if that fails.
}
\details{
The function first checks if the package is installed and has a
tutorials directory. It then attempts to use learnr::available_tutorials()
to get the official list of tutorial directories. If that fails (e.g., if
the package doesn't properly register its tutorials with learnr), it falls
back to scanning all subdirectories under inst/tutorials/. Finally, it
collects all .Rmd files from these directories.

Returns an empty character vector if the package has no tutorials or
doesn't exist, rather than throwing an error.
}
\examples{
\dontrun{
  # Get all learnr tutorial paths
  return_tutorial_paths('learnr')
  
  # Get tutorial paths from your own package
  return_tutorial_paths('tutorial.helpers')
  
  # Returns empty vector for packages without tutorials
  return_tutorial_paths('base')
}
  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_answers.R
\name{write_answers}
\alias{write_answers}
\title{Write tutorial answers to file}
\usage{
write_answers(file, obj)
}
\arguments{
\item{file}{Output file path (should end in .html).}

\item{obj}{Either a Shiny session object (from learnr) or a list of submissions
(as returned by get_submissions_from_learnr_session()).}
}
\value{
No return value, called for the side effect of writing an HTML
file of submitted answers to \code{file}.
}
\description{
Take a tutorial session (or a submission list), extract all submitted answers,
and write out an HTML file with those answers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/show_file.R
\name{show_file}
\alias{show_file}
\title{Display the contents of a text file that match a pattern}
\usage{
show_file(path, start = 1, end = NULL, pattern = NULL, chunk = "None")
}
\arguments{
\item{path}{A character vector representing the path to the text file.}

\item{start}{An integer specifying the starting row number (inclusive) to consider. Default is 1.
If negative, it represents the number of lines to print from the end of the file.
If 0, prints the entire file.}

\item{end}{An integer specifying the ending row number (inclusive) to consider. Default is the last row.}

\item{pattern}{A regular expression pattern to match against each row. Default is NULL (no pattern matching).
Applied to the whole-file (\code{start == 0}), last-lines (\code{start < 0}), and row-range cases;
ignored when \code{chunk} is "All", "Last", or "YAML".}

\item{chunk}{A character string indicating what content to extract.
Possible values are "None" (default - no chunk processing),
"All" (print all code chunks), "Last" (print only the last code chunk),
or "YAML" (extract the YAML header without delimiters). Code chunks of
any language are recognized (e.g. \code{r}, \code{python}, \code{bash}), not just R.}
}
\value{
The function prints the contents of the specified range of rows that match the pattern (if provided),
the code lines within R code chunks (if chunk is "All" or "Last"),
or the YAML header content (if chunk is "YAML") to the console. If no rows match the pattern,
nothing is printed. If start is negative, the function prints the last abs(start) lines, ignoring
missing lines at the end of the file. If start is 0, the function prints the entire file.
}
\description{
This function reads the contents of a text file and either prints the specified range of rows
that match a given regular expression pattern, prints the code lines within code chunks,
or extracts the YAML header. If start is a negative number, it prints the last abs(start) lines,
ignoring missing lines at the end of the file. If start is 0, it prints the entire file.
}
\details{
The arguments are resolved in a fixed order of precedence: \code{chunk = "YAML"}
is handled first, then \code{start == 0} (whole file), then \code{start < 0} (last
abs(start) lines), then \code{chunk \%in\% c("All", "Last")} (code chunks), and
finally the \code{start}/\code{end} row range. The \code{pattern} filter is applied within
the whole-file, last-lines, and row-range cases, but is ignored when \code{chunk}
selects code chunks or the YAML header.
}
\examples{
\dontrun{
# Display all rows of a text file
show_file("path/to/your/file.txt")

# Display the entire file
show_file("path/to/your/file.txt", start = 0)

# Display rows 5 to 10 of a text file
show_file("path/to/your/file.txt", start = 5, end = 10)

# Display all rows of a text file that contain the word "example"
show_file("path/to/your/file.txt", pattern = "example")

# Print all code lines within R code chunks
show_file("path/to/your/file.txt", chunk = "All")

# Print only the last R code chunk
show_file("path/to/your/file.txt", chunk = "Last")

# Extract the YAML header
show_file("path/to/your/file.Rmd", chunk = "YAML")

# Display the last 5 lines of a text file, ignoring missing lines at the end
show_file("path/to/your/file.txt", start = -5)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/open_github_pages.R
\name{open_github_pages}
\alias{open_github_pages}
\title{Open Multiple GitHub Pages in Browser Tabs}
\usage{
open_github_pages(
  urls,
  url_var = NULL,
  label_var = NULL,
  delay_seconds = 0.5,
  browser = "default",
  verbose = FALSE
)
}
\arguments{
\item{urls}{Either a character vector of URLs to open, OR a tibble/data.frame
containing submission data with a URL column}

\item{url_var}{Character string specifying the column name containing URLs.
Required when urls is a tibble/data.frame. Ignored when urls is a character vector.}

\item{label_var}{Character string specifying the column name to use for
identifying each submission in verbose output (e.g., "name", "email").
Only used when urls is a tibble/data.frame.}

\item{delay_seconds}{Numeric value specifying delay between opening each URL
(default is 0.5 seconds to allow browser to process each request)}

\item{browser}{Character string specifying which browser to use. Options are
"default" (system default), "chrome", "firefox", "safari", or "edge".
On Windows, also supports "msedge". Default is "default".}

\item{verbose}{Logical value (TRUE or FALSE) specifying verbosity level.
If TRUE, reports each URL as it's being opened.}
}
\value{
Invisible NULL. Function is called for its side effect of opening browser tabs.
}
\description{
This function opens multiple GitHub.io pages in browser tabs, making it easy
for teaching fellows to quickly review student webpages. It can accept either
a vector of URLs or a tibble/data.frame containing URLs.
}
\details{
The function uses the system's default method to open URLs, which typically
opens them in the default browser. Most modern browsers will open multiple
URLs as tabs in the same window when called in quick succession.

The delay between opening URLs helps ensure the browser has time to process
each request properly. You may need to adjust this delay based on your
system performance and browser behavior.
}
\examples{
\dontrun{
# Open multiple GitHub Pages from vector
student_sites <- c("https://github.com/Abdul-Hannan96/stops.git")

open_github_pages(student_sites, verbose = TRUE)

# Open from tibble/data.frame
path <- system.file("extdata", "answers_html", package = "tutorial.helpers")

result <- submissions_answers(
  path = path,
  title = c("stop"), 
  key_var = "email",
  membership = c("bluebird.jack.xu@gmail.com", "abdul.hannan20008@gmail.com"),
  vars = c("name","email","temperance-15"),
  verbose = TRUE
)

open_github_pages(result, 
                  url_var = "temperance-15",
                  verbose = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_google_drive.R
\name{download_google_drive}
\alias{download_google_drive}
\title{Download Files from a Google Drive Folder}
\usage{
download_google_drive(
  url,
  path = NULL,
  title = NULL,
  overwrite = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{url}{Character string. The Google Drive folder URL or ID. Supports standard folder URLs
(e.g., \code{"https://drive.google.com/drive/folders/FOLDER_ID"}) or direct folder IDs.}

\item{path}{Character string or NULL. The local directory path for downloads.
If NULL (default), uses the current working directory. If the directory doesn't exist, it will be created.}

\item{title}{Character vector or NULL. Patterns to match against file names for filtering.
If provided, only files whose names contain any of these patterns (case-insensitive) are downloaded.
If NULL (default), all files are downloaded.}

\item{overwrite}{Logical. If TRUE (default), overwrites existing files. If FALSE, skips files that already exist.}

\item{verbose}{Logical. If TRUE (default), prints detailed progress messages.}
}
\value{
Character string. The path to the directory where files were downloaded.
}
\description{
Downloads all files or filtered files from a public Google Drive folder to a local directory.
}
\examples{
\dontrun{
# Download all files to current directory
download_google_drive("https://drive.google.com/drive/folders/1Rgxfiw")

# Download to a specific directory with filtering
download_google_drive(
  url = "https://drive.google.com/drive/folders/1Rgxfiw",
  path = "./my_data",
  title = c("report", ".csv"),
  overwrite = FALSE
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submissions_summary.R
\name{submissions_summary}
\alias{submissions_summary}
\title{Process Submissions Summary}
\usage{
submissions_summary(
  path,
  title = ".",
  return_value = "Summary",
  vars = NULL,
  verbose = FALSE,
  keep_file_name = NULL,
  emails = NULL
)
}
\arguments{
\item{path}{The path to the local directory containing the HTML/XML files, or a Google Drive folder URL.
If it's a Google Drive URL, the function will download individual files to a temporary directory.}

\item{title}{A character vector of patterns to match against the file names (default: ".").
Each pattern is processed separately and results are combined.}

\item{return_value}{The type of value to return. Allowed values are "Summary" (default) or "All".}

\item{vars}{A character vector of key variables to extract from the "id" column (default: NULL).}

\item{verbose}{A logical value (TRUE or FALSE) specifying verbosity level.
If TRUE, reports files that are removed during processing.}

\item{keep_file_name}{Specifies whether to keep the file name in the summary tibble. Allowed values are NULL (default), "All" (keep entire file name), "Space" (keep up to first space), or "Underscore" (keep up to first underscore). Only used when \code{return_value} is "Summary".}

\item{emails}{A character vector of email addresses to filter results by, "*" to include all emails, or NULL to skip email filtering (default: NULL).}
}
\value{
If \code{return_value} is "Summary", returns a tibble with one row for each file, columns corresponding to the \code{vars},
and an additional "answers" column indicating the number of rows in each tibble.
If \code{return_value} is "All", returns a tibble with all the data combined from all the files.
}
\description{
This function processes submissions from a local directory or Google Drive folder containing HTML/XML files.
It extracts tables from the files, filters them based on a pattern and key variables,
and returns either a summary tibble or a combined tibble with all the data.
}
\examples{
# Process submissions from local directory
path <- system.file("extdata", "answers_html", package = "tutorial.helpers")

result <- submissions_summary(path = path,
                             vars = "email",
                             title = "stop")

\dontrun{
drive_url <- "https://drive.google.com/drive/folders/10do12t0fZsfrIrKePxwjpH8IqBNVO86N"
x <- submissions_summary(
  path = drive_url, 
  title = c("positron"),
  vars = c("email", "name")
)


}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_key_vars.R
\name{check_key_vars}
\alias{check_key_vars}
\title{Check Key Variables in List of Tibbles}
\usage{
check_key_vars(tibble_list, key_vars, verbose = FALSE)
}
\arguments{
\item{tibble_list}{A named list of tibbles, each containing an "id" column with question identifiers}

\item{key_vars}{A character vector of key variables to check for}

\item{verbose}{A logical value (TRUE or FALSE) specifying verbosity level.
If TRUE, reports tibbles that are removed and why.}
}
\value{
A list of tibbles that contain all required key variables
}
\description{
This function checks if specified key variables are present in each tibble's "id" column
and returns only tibbles that contain all required key variables.
}
\examples{
# Create sample data
path <- system.file("extdata", "answers_html", package = "tutorial.helpers")

tibble_list <- gather_submissions(path, "stop")

result <- check_key_vars(tibble_list,
                         key_vars = c("name", "email"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_exercise.R
\name{make_exercise}
\alias{make_exercise}
\title{Add a tutorial code exercise or question to the active document}
\usage{
make_exercise(type = "no-answer", file_path = NULL)
}
\arguments{
\item{type}{Character of question type. Must be one of "code", "no-answer",
or "yes-answer". Abbreviations such as "no", "yes", and "co" are allowed.}

\item{file_path}{Character path to a file. If NULL, the RStudio active
document is used, which is the default behavior. An actual file path is
used for testing.}
}
\value{
Exercise skeleton corresponding to the \code{type} argument.
}
\description{
When writing tutorials, it is handy to be able to insert the
skeleton for a new code exercise or question. Note that the
function determines the correct exercise number to use and also adds
appropriate code chunk labels, based on the exercise number and section
title.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_submissions_from_learnr_session.R
\name{get_submissions_from_learnr_session}
\alias{get_submissions_from_learnr_session}
\title{Return a list of tutorial answers}
\usage{
get_submissions_from_learnr_session(sess)
}
\arguments{
\item{sess}{session object from shiny with learnr}
}
\value{
a list which includes the exercise submissions of tutorial
}
\description{
Grabs information from the \code{learnr} session environment, not
directly from the session object itself. Since we are using the session
environment, we currently don't (?) have a way to save the environment and
hence can't test this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_tutorial_defaults.R
\name{check_tutorial_defaults}
\alias{check_tutorial_defaults}
\title{Confirm that a tutorial has the recommended components}
\usage{
check_tutorial_defaults(
  tutorial_paths,
  libraries = c("learnr", "tutorial.helpers"),
  children = c("info_section", "download_answers")
)
}
\arguments{
\item{tutorial_paths}{Character vector of the paths to the tutorials to be
examined.}

\item{libraries}{Character vector of library names that should be loaded
in the tutorial. The function looks for \code{library(name)} calls. Default
is \code{c("learnr", "tutorial.helpers")}.}

\item{children}{Character vector of child document names (without the .Rmd
extension) that should be included in the tutorial. The function looks
for these in child document inclusion chunks. Default is
\code{c("info_section", "download_answers")}.}
}
\value{
No return value, called for side effects.
}
\description{
Checks that tutorials contain required libraries and child
documents. The function looks for library() calls and child document
inclusions in the tutorial files.
}
\examples{
  # Check with default requirements
  check_tutorial_defaults(
    tutorial_paths = return_tutorial_paths("tutorial.helpers")
  )
  
  # Check for specific libraries and child documents
  check_tutorial_defaults(
    tutorial_paths = return_tutorial_paths("tutorial.helpers"),
    libraries = c("learnr", "knitr"),
    children = c("info_section")
  )

}
